#include "feldliste.h"

// Kommentar: Die Feldliste ist eine Verknpfung von doppelt verketteter Liste und einem 2 - dimensionalen Feld. Dabei gilt, dass
// immer bei einem Eintrag eines Punktes folgendes gilt: Unter seinen Koordinaten des 2 - dimensionalen Feldes wird auf die
// Position der doppelt verketteten Liste gezeigt. Und umgekehrt werden in der doppelt verketteten Liste die Koordinaten des
// Punktes - und damit der Eintrag in dem 2 - dimensionalen Feld - gespeichert.

// Kommentar: Hier erfolgt eine Prfung, ob die Feldliste leer ist.
bool FeldListe::ist_leer(void)
    {
// Kommentar: An der Stelle laenge ist die Anfangsposition der Liste, an der Stelle laenge_plus1 das Ende der Liste. Zeigt
// der Anfang der Liste auf das Ende der Liste, dann ist die Liste leer.
// Bemerkung: Diese beiden Eintrge laenge und laenge_plus1 stehen Selber noch fr keine Listenelemente.
    if(nachfolger[laenge] == laenge_plus1)
        {
        return(true);
        }
    return(false);
    }

// Kommentar: Hier wird zum Durchlaufen der Liste von einer Position der Liste zur nchsten Position gegangen.
void FeldListe::naechster(void)
    {
// Kommentar: Das Feld nchster zeigt immer auf den nchsten Eintrag, und der Index position zeigt auf den aktuelle Eintrag.
    position = nachfolger[position];
    }

// Kommentar: Hier wird geprft, ob man beim Durchlauf am Ende der Liste angekommen ist.
bool FeldListe::ist_fertig_fl(void)
    {
// Kommentar: Das letzte Element der Liste - ohne das dort ein Punkt gespeichert ist - liegt auf Position laenge_plus1. Ist die
// Position dort, dann wurde das Listenende erreicht.
    if(position == laenge_plus1)
        {
        return(true);
        }
    return(false);
    }

// Kommentar: Es werden die Koordinaten des aktuellen Listenelementes gelesen. Zuerst die Koordinate f.
int FeldListe::lesen_f(void)
    {
// Kommentar: Man gibt f an der Stelle des aktuellen Indexes position zurck.
    return(f[position]);
    }

// Kommentar: Dann wird die Koordinate g gelesen.
int FeldListe::lesen_g(void)
    {
// Kommentar: Man gibt g an der Stelle des aktuellen Indexes position zurck.
    return(g[position]);
    }

// Kommentar: Die Anfangsposition der zu lesenden Elemente wird gesetzt. Das ist nicht die Anfangsposition der Liste selber.
void FeldListe::anfangsposition(void)
    {
// Kommentar: Die Stelle laenge ist die Anfangsposition, und ihr nchster Eintrag zeigt auf das erste zu lesende Element.
    position = nachfolger[laenge];
    }

// Kommentar: Am Anfang wird die Struktur geleert.
void FeldListe::leeren(void)
    {
// Kommentar: Zuerst werden Anfang und Ende der Liste aufeinander verknpft.
    nachfolger[laenge] = laenge_plus1;
    vorgaenger[laenge_plus1] = laenge;

// Kommentar: Und dann wird das Feld gelscht. Das hier macht man nur am Anfang, weil das Lschen des Feldes auf diese Weise
// sonst schneller machbar ist ber einen Durchlauf durch die Liste.
    for(i=0; i<nn; i++)
        {
        for(j=0; j < mm; j++)
            {
            feld[i][j] = -1;
            }
        }
    }

// Kommentar: Hier wird die Feldliste komplett gefllt.
void FeldListe::fuellen(void)
    {
// Kommentar: Zuerst hat man eine leere doppelt verkettete Liste.
    nachfolger[laenge] = laenge_plus1;
    vorgaenger[laenge_plus1] = laenge;

// Kommentar: Und den Stack mit den Eintrgen der doppelt verketteten Liste leer man entsprechend.
    stack1d.reset();
// Kommentar: Und dann durchluft man alle Punkte, und trgt sie ber die Funktionen diese Klasse ein.
    for(i=0; i<nn; i++)
        {
        for(j=0; j < mm; j++)
            {
            einfuegen_fl(i,j);
            }
        }
    }

// Kommentar: Hier werden die Parameter n, m und rang bergeben, und Speicherplatz bereitgestellt, und es werden Verknpfungen
// ausgefhrt.
void FeldListe::initialisieren_fl(int _n, int _m, int _rang)
// finalisieren hier OK.
    {
// Kommentar: Die Parameter werden gespeichert.
    nn = _n;
    mm = _m;
    rang_fl = _rang;
// Komemntar: Listenlnge, und Listenlnge+2 werden gespeichert. Letzteres, damit alle Eintrge plus erstem und letztem Element
// gespeichert werden knnen.
    listenlaenge = nn*mm;
    laenge_plus2 = listenlaenge + 2;
    laenge = listenlaenge;
    laenge_plus1 = listenlaenge + 1;
    laenge_minus1 = listenlaenge - 1;
// Kommentar: Der Speicher fr f  und g - das sind die 1 - dimensionalen Felder fr die Koordinaten - werden angelegt.
    f = new int [laenge_plus2];
    g = new int [laenge_plus2];

// Kommentar: Der Speicherplatz fr die Indizes von nachfolgenden oder vorhergehenden Elementen der Listen wird bereitgestellt.
    nachfolger = new int [laenge_plus2];
    vorgaenger = new int [laenge_plus2];

// Kommentar: Das erste Element auf der Position laenge, und das letzte Element auf der Position laenge_plus1, werden aufeinander
// verknpft.
    nachfolger[laenge] = laenge_plus1;
    vorgaenger[laenge_plus1] = laenge;

// Kommentar: Und das Feld wird geleert.
    feld = new int * [nn];
    for(i=0; i<nn; i++)
        {
        feld[i] = new int [mm];
        for(j=0; j < mm; j++)
            {
            feld[i][j] = -1;
            }
        }

// Kommentar: Dann wird der Stack mit der Anzahl an Eintrgen in die Liste angelegt - um jeweils fr neue Elemente eine Position
// finden zu knnen.
    stack1d.initialisieren(laenge);

// Kommentar: Das Folgende muss nicht gemacht werden, und kann nur Debugging - Zwecken dienen.
    #ifdef FELDLISTE_MARKIEREN
    for(i=0; i<laenge_plus2; i++)
        {
        f[i] = -1;
        }
    #endif
    }

// Komemntar: Hier wird der Speicherplatz, der beim Initialisieren angelegt worden ist, wieder freigegeben.
void FeldListe::finalisieren_fl(void)
    {
// Kommentar: Man gibt den Speicherplatz des Stacks wieder frei.
    stack1d.finalisieren();

// Komemntar: Man gibt den Speicher fr die Koordinaten sowie die Listen fr Vorgnger und Nachfolger wieder frei.
    delete f;
    delete g;

    delete nachfolger;
    delete vorgaenger;

// Kommentar: Man gibt das 2 - dimensionale Feld wieder frei.
    for(i=0; i<nn; i++)
        {
        delete feld[i];
        }
    delete feld;
    }

// Kommentar: Hier wird ein Element der Liste hinzugefgt.
void FeldListe::einfuegen_fl(int ff, int gg)
    {
// Man liest eine Nummer fr einen Eintrag in die doppelt verkettete Liste.
    nummer = stack1d.lesen();
// Kommentar: Diese Nummer wird ins Feld eingetragen an der Position (ff,gg)
    feld[ff][gg] = nummer; // Damit man schnell loeschen kann
// Kommentar: Die Koordinaten an der Stelle der Nummer werden von den bergabeparametern bernommen.
    f[nummer] = ff;
    g[nummer] = gg;
// Kommentar: Nun wird der Eintrag nummer in die doppelt verkettete Liste direkt nach dem Anfangselement verknpft.
// Das Vorgngerelement hat damit die Position laenge, das Nachfolgeelement die Position merken = nachfolger(laenge).
// Und das Element nummer wird dazwischen eingetragen.
    merken = nachfolger[laenge];
    nachfolger[laenge] = nummer;
    nachfolger[nummer] = merken;
    vorgaenger[merken] = nummer;
    vorgaenger[nummer] = laenge;
    }

// Kommentar: Hier wird ein Element der Liste entfernt.
void FeldListe::entfernen_fl(int ff, int gg)
    {
// Kommentar: Durch Lesen des 2 - dimensionalen Feldes wird die Nummer in der doppelt verketteten Liste bestimmt.
    nummer = feld[ff][gg];
// Kommentar: Zeigt die Nummer auf einen leeren Eintrag, dann gibt man eine Fehlermeldung aus.
    if(nummer == -1)
        {
        cout << " Fehler in entfernen_fl ff " << ff << " gg " << gg << " rang " << rang_fl << endl;
        exit(-1);
        }
    #ifdef FELDLISTE_MARKIEREN
    f[nummer] = -1;
    #endif
// Kommentar: Nun bekommt das Feld aber einen leeren Eintrag an dieser Position - weil es diese Koordinate nicht mehr in der
// Feldliste gibt.
    feld[ff][gg] = -1;
// Kommentar: Nun wird der Vorgnger der Nummer auf den Nachfolge der Nummer verknpft, und damit das Element aus der
// doppelt verketteten Liste gelscht.
    merken = vorgaenger[nummer];
    merken2 = nachfolger[nummer];
    nachfolger[merken] = merken2;
    vorgaenger[merken2] = merken;
// Kommentar: Am Ende bekommt der Stack die Nummer zurck, damit er sie wieder bei einem Punkteeintrag zur Verfgung hat.
    stack1d.schreiben(nummer);
    }

// Kommentar: Hier wird der Wert des 2 - dimensionalen Feldes zurckgegeben, damit geprft werden kann, ob es auf der Position
// (ff,gg) einen Punkt in der Liste gibt.
int FeldListe::teste_f(int ff, int gg)
// Wird -1 sein, genau dann, wenn in (ff, gg) kein Punkt mehr ist.
// Das geht nur mit feld[ff][gg] = -1 beim loeschen.
    {
    nummer = feld[ff][gg];
    return(nummer);
    }
