#include "feldliste_1d.h"

// Kommentar: Die eindimensionale Feldliste fhrt eine doppelt verkettet Liste und ein eindimensionales Feld, um die Eintrge
// einer auf eine Koordinate bezogenen Punktemenge zu fhren. Durch die Kombination von Feld und Liste ist eine
// Hinzufgung und Entfernung eines Punktes in O(1) mglich. Dabei zeigt das 1 - dimensionale Feld auf die Liste, und
// die Liste speichert unter ihrem Listeneintrag die Koordinate.

// Kommentar: Hier wird zum Durchlaufen der Liste von einer Position der Liste zur nchsten Position gegangen.
void FeldListe1D::naechster(void)
    {
    position = nachfolger[position];
    }

// Kommentar: Hier wird geprft, ob man beim Durchlauf am Ende der Liste angekommen ist.
bool FeldListe1D::ist_fertig(void)
    {
// Kommentar: Das letzte Element der Liste - ohne das dort ein Punkt gespeichert ist - liegt auf Position laenge_plus1. Ist die
// Position dort, dann wurde das Listenende erreicht.
    if(position == laenge_plus1)
        {
        return(true);
        }
    return(false);
    }

// Kommentar: Es wird die Koordinate des aktuellen Listenelementes gelesen.
int FeldListe1D::lesen(void)
    {
    return(inhalt[position]);
    }

// Kommentar: Die Anfangsposition der zu lesenden Elemente wird gesetzt. Das ist nicht die Anfangsposition der Liste selber.
void FeldListe1D::anfangsposition(void)
    {
    position = nachfolger[laenge];
    }

// Kommentar: Bei der Vektorbildung wird in der Liste nach einem Vektor gesucht, der maximal die Lastmenge an Punkten hat,
// und andererseits mglichst diese Lnge annimmt.
void FeldListe1D::vektorbildung(int _lastmenge)
    {
// Kommentar: Der bergabeparameter Lastmenge wird gespeichert.
    lastmenge = _lastmenge;

#ifdef AKTUELLE_KOMMENTARE
    if(lastmenge==1)
        {
        cout << " LLMMNN vor koordinate zum ersten Mal bestimmen laenge " << laenge << endl;
        }
#endif

// Kommentar: Es wird der erste Eintrag in die doppelt verkettete Liste genommen. Das ist an irgendeiner Stelle in der 1 -
// dimensionalen Liste. Insbesondere ist das im Allgemeinen weder der kleinste noch grte Koordinatenwert.
    koordinate = nachfolger[laenge];
// Kommentar: Ist die Liste leer, dann mache ein return. Dann kann kein Vektor in der Liste gefunden werden.
    if(koordinate==laenge_plus1)
        {
        vektoranfang = -123456;
        return;
        }

#ifdef AKTUELLE_KOMMENTARE
    if(lastmenge==1)
        {
        cout << " LLMMNN vor koordinate bestimmen " << endl;
        }
#endif

// Kommentar: Nun wird zu dem ersten Eintrag seine Koordinate gelesen.
    koordinate = inhalt[koordinate];

#ifdef AKTUELLE_KOMMENTARE
    if(lastmenge==1)
        {
        cout << " LLMMNN nach koordinate bestimmen " << endl;
        }
#endif

// Kommentar: War die Lastmenge 1, dann besteht der Vektor nur aus der Koordinate. Das wird so zurckgegeben.
    if(lastmenge == 1)
        {
        vektoranfang = koordinate;
        vektorende = koordinate;
        return;
        }

// Kommentar: Ist noch Platz fr kleinere Koordinaten, und ist die erste kleinere Koordinate verzeichnet, dann mache vektor_hoch_runter.
// Das bewirkt eine Suche nach kleineren Koordinaten und greren Koordinaten, wobei die Suche endet, wenn die Maximallast fr den
// Vektor erreicht wurde. Mache danach return, weil es keine Schritte danach mehr gibt.
    if(koordinate>0)
        {
        if(belegt[koordinate-1]==true)
            {
            vektor_hoch_runter();
            return;
            }
        }

// Kommentar: Ist noch Platz fr grere Koordinaten, und ist die erste grere Koordinate verzeichnet, dann mache vektor_runter. Das
// bewirkt eine Suche nur nach greren Koordinaten, aber die Suche endet, wenn die Maximallast erreicht wurde. Auch dann mache return.
    if(koordinate<laenge_minus1)
        {
        if(belegt[koordinate+1]==true)
            {
            vektor_runter();
            return;
            }
        }

// Kommentar: Gibt es keine Mglichkeit, den eindimensionalen Punkt mit der Koordinate koordinate zu verlngern, dann wird der
// Vektor bestehend aus dem Punkt gespeichert.
    vektoranfang = koordinate;
    vektorende = koordinate;
    }

// Kommentar: Verlngere den Punkt mit der Koordinate koordinate hin zu greren Werten, soweit mglich, und soweit der maximale
// Lastwert das ermglicht.
void FeldListe1D::vektor_runter(void)
    {
// Kommentar: Bestimme zuerst, wie lang der Vektor aufgrund der Maximallast sein kann. k ist einen Wert darber.
    k = koordinate+lastmenge;
// Kommentar: berschreitet das die Listenlnge, dann setze k auf die Listenlnge. Dann kann der Vektor bis zur Grenze des Feldes
// verlngert werden.
    if(k>laenge)
        {
        k=laenge;
        }
// Kommentar: Dann suche bis k danach, in aufsteigender Reihenfolge, ob die Liste Eintrge an der Stelle hat (belegt[j] = true).
// Es wird hochgezhlt, bis der Punkt zu der Koordinate nicht mehr in der Liste existiert.
    for(j=koordinate+2; j<k; j++) // gendert am 25.7.15. Die belegt[koordinate+1] ist true, wenn diese Funktion aufgerufen wird.
        {
        if(belegt[j]==false)
            {
            break;
            }
        }
// Kommentar: Da hin zu greren Werten gesucht worden ist, beginnt der Vektor mit der Koordinate. Er endet mit j-1, da um 1 zu
// weit aufwrts gezhlt wurde.
    vektoranfang = koordinate;
    vektorende = j-1;
    }

// Kommentar: Verlngere den Punkt mit der Koordinate koordinate erst hin zu kleineren Werten, und danach, je nach Last, hin
// zu greren Werten, soweit mglich, und soweit der maximale Lastwert das ermglicht.
void FeldListe1D::vektor_hoch_runter(void)
    {
// Kommentar: k ist die minimal Koordinate, so dass die Beschrnkung aufgrund Lastmenge eingehalten wird. Z.B. koordinate = 10,
// lastmenge = 5. Dann geht es maximal bis zur Koordinate 6.
    k = koordinate - lastmenge + 1;
    if(k < 0)
        {
        k = 0;
        }

// Kommentar: belegt[koordinate-1] = true, aufgrund der aufrufenden Funktion. Man luft maximal bis k. Es wird solange
// gelaufen, wie der Vektor sich verlngern lt.
    for(i=koordinate-2; i>=k; i--)
        {
        if(belegt[i]==false)
            {
            break;
            }
        }
// Kommentar: Lief es durch, dann return.
    if(i == k-1)
        {
        vektoranfang = k;
        vektorende = koordinate;
        return;
        }
// Kommentar: Ist die Last noch nicht erreicht, dann laufe hin zu greren Koordinaten. Maximal geht es bis: lastmenge + (i+1) - 1.
// Das ist der letzte mgliche Eintrag. Bemerkung: (i+1) ist die Anfangskoordinate.
    k = lastmenge+i;
// Kommentar: Liegt er jenseits der Gittergrenzen, dann setze ihn auf die Gittergrenze.
    if(k>=laenge)
        {
        k=laenge_minus1;
        }
// Kommentar: Prfe nun, wie weit sich der Vektor - in diesem Sinne - verlngern lt.
    for(j=koordinate+1; j<=k; j++)
        {
        if(belegt[j]==false)
            {
            break;
            }
        }
// Kommentar: Der Vektoranfang ist dann i+1, weil die Schleife um 1 weiter runter zhlt. Und j-1, weil die Schleife um 1 weiter
// hoch zhlt.
    vektoranfang = i+1;
    vektorende = j-1;
// Kommentar: Die folgende Bedingung wird so, wie diese Funktion geschrieben ist, nicht erfllt werden.
    if(vektorende-vektoranfang+1>lastmenge)
        {
        vektorende = vektoranfang+lastmenge-1;
        }
    }

// Komemntar: Die Finalisierung dieser Klasse
void FeldListe1D::finalisieren(void)
    {
// Kommentar: koordinaten ist das eindimensionale Feld, damit schnell das zu lschende Element der doppelt verketten Liste
// gefunden wird. Das wird hier freigegeben.
    delete koordinaten;
// Kommentar: inhalt, vorgaenger und nachfolger sind die Listen der doppelt verketteten Liste. Der Inhalt gibt die Koordinate
// des Eintrags wieder. Alle drei Strukturen werden hier freigegeben.
    delete inhalt;
    delete nachfolger;
    delete vorgaenger;
// Kommentar: belegt gibt an, ob auf einer Koordinate ein Eintrag vorliegt. Das knnte man aber auch vom Feld koordinaten ableiten.
    delete belegt;
// Kommentar: Und am Ende wird der Stack finalisiert.
    stack1d.finalisieren();
    }

// Kommentar: Hier wird die Klasse initialisiert. Wichtig fr die Gren der anzulegenden Strukturen ist der bergabeparameter
// _listenlaenge.
void FeldListe1D::initialisieren(int _listenlaenge)
    {
// Der bergabeparamter, sowie kleine nderungen daran werden gespeichert.
    listenlaenge = _listenlaenge;
    laenge_plus2 = listenlaenge + 2;
    laenge = listenlaenge;
    laenge_plus1 = listenlaenge + 1;
    laenge_minus1 = listenlaenge - 1;
// Kommentar: Die folgenden Strukturen wurden bei der Finalisierung erlutert. Hier werden diese Strukturen angelegt.
//    koordinaten = new int [laenge_plus2]; // Warum nicht statt laenge_plus2 nur die laenge? Das reicht doch fr die Koordinaten aus.
    koordinaten = new int [laenge]; // nderung am 25.7.15
    inhalt = new int [laenge_plus2];
    nachfolger = new int [laenge_plus2];
    vorgaenger = new int [laenge_plus2];

    belegt = new int [laenge];
// Da die Liste anfangs leer ist, ist belegt auf false zu setzen.
    for(i=0; i<laenge; i++)
        {
        belegt[i] = false;
        }

// Kommentar: Das Anfangselement laenge, und das Endelement laenge_plus1 werden aufeinander verknpft.
    nachfolger[laenge] = laenge_plus1;
    vorgaenger[laenge_plus1] = laenge;

// Kommentar: Und am Ende wird der Stack angelegt.
    stack1d.initialisieren(laenge);
    }

// Kommntar: Es wird ein Element in die Liste eingefgt.
void FeldListe1D::einfuegen(int _koordinate)
    {
// Kommentar: Zuerst wird die neu Position in der doppelt verketteten Liste bestimmt.
    nummer = stack1d.lesen();
#ifdef AUSGABEN_FELDLISTE_1D
    cout << "Nummer in einfuegen " << nummer << " Listenlaenge " << listenlaenge << endl;
#endif

// Kommentar: Das eindimensionale Feld bekommt auf der Koordinate diese Position eingetragen.
    koordinaten [_koordinate] = nummer; // Damit man schnell loeschen kann

// Kommentar: Und die doppelt verkettete Liste bekommt die Koordinate an der Position nummer im Inhalt eingetragen.
    inhalt[nummer] = _koordinate;

// Kommentar: Nun wird die nummer in die doppelt verkettete Liste aufgenommen. Zuerst wird dafr der Nachfolger der Anfangsposition
// bestimmt. Das neue Element soll zwischen diese beiden Positionen verknpft werden.
    merken = nachfolger[laenge];

// Kommentar: Mit den folgenden 4 Operationen wird folgende doppelte Verknpfung erreicht: laenge <-> nummer <-> merken.
    nachfolger[laenge] = nummer;
    nachfolger[nummer] = merken;
    vorgaenger[merken] = nummer;
    vorgaenger[nummer] = laenge;

// Kommentar: Am Ende wird vermerkt, dass diese Koordinate nun belegt ist.
    belegt[_koordinate] = true;
    }

// Kommentar: Hier wird ein Element auf der Koordinate entfernt.
void FeldListe1D::entfernen(int _koordinate)
    {
// Kommentar: Zuerst wird die Position in der doppelt verketteten Liste fr diese Koordinate bestimmt.
    nummer = koordinaten[_koordinate];
#ifdef AUSGABEN_FELDLISTE_1D
    cout << "Nummer in entfernen " << nummer << " Listenlaenge " << listenlaenge << endl;
#endif

// Kommentar: Dann werden die vorhergehende und die nachfolgende Position der doppelt verketteten Liste bestimmt:
// merken <-> nummer <-> merken2
    merken = vorgaenger[nummer];
    merken2 = nachfolger[nummer];
// Kommentar: Dann wird folgende Verknpfung hergestellt: merken <-> merken2
    nachfolger[merken] = merken2;
    vorgaenger[merken2] = merken;
// Kommentar: Dem Stack wird die freigewordenen Nummer wieder hinzugefgt.
    stack1d.schreiben(nummer);
// Kommentar: Und die Koordinate wird als nicht belegt gekennzeichnet.
    belegt[_koordinate] = false;
    }

// Kommentar: Es werden hier mit Hilfe der Feldliste alle Vektoren in Ost - West Richtung bestimmt - also die i. Zeile der
// verzeichneten Punkte. Das macht man bei z.B. bei der Visualisierung, um die Verteilung der eingetragenen Punkte auf die
// Kerne zu bekommen. Man muss das fr alle West - Ost - Feldlisten machen.
#ifdef KURZ
int FeldListe1D::alle_vektoren_horizontal(short int * buffer, int _index, int i)
#else
int FeldListe1D::alle_vektoren_horizontal(int * buffer, int _index, int i)
#endif
    {
#ifdef AUSGABEN_FELDLISTE_1D
    cout << "Anfang alle_vektoren_horizontal " << endl;
#endif
// Kommentar: Der Index ist der Bufferindex. Der wird nicht auf 0 gesetzt, weil diese Teilstcke von Vektoren
// hintereinander gesetzt werden mssen. Man bekommt bei dieser Funktion den jeweils frs Anfgen passenden
// Index zurckgegeben.
    index = _index;
// Kommentar: Im Testfeld wird verzeichnet, welche Elemenete schon behandelt wurden. Um es spter zu resetten, muss man
// sich aber die Koordinaten merken.
    testfeld_index = 0;

// Kommentar: Nun wird die Feldliste durchlaufen.
    for(anfangsposition(); ist_fertig()==false; naechster())
        {
// Kommentar: Ein Eintrag wird gelesen.
        a = lesen();
// Kommentar: wurde dieser Index noch nicht bearbeitet ...
        if(testfeld[a]==false)
            {
// Kommentar: Dann suche, wie weit man a hin zu kleineren Koordinaten verlngern kann in Bezug auf den Zustand, dass sie
// belegt sind.
            b = vektorisieren_nach_oben(a);
// Kommentar: Dann mache man dasselbe in Richtung der greren Koordinaten.
            c = vektorisieren_nach_unten(a);
#ifdef AUSGABEN_FELDLISTE_1D
            cout << " a " << a << " b " << b << " c " << c << endl;
#endif
// Kommentar: Dann markiere man diese nicht mehr verlngerbaren Vektoren ...
            for(j=b; j<=c; j++)
                {
                schreibe_testfeld(j);
                }
// Kommentar: Und trage sie in den Buffer ein.
            buffer[index++] = i; // Die i. Zeile wird besetzt.(z.B. nrdlichste) 30.08.2014
            buffer[index++] = b; // Im Westen b, im Osten c 30.08.2014
            buffer[index++] = c;
#ifdef AUSGABEN_FELDLISTE_1D
            cout << " STU horizontal i " << i << " b " << b << " c " << c << endl;
#endif
            }
        }
// Kommentar: Am Ende wird das Testfeld zurckgesetzt.
    reset_testfeld();
#ifdef AUSGABEN_FELDLISTE_1D
    cout << "Ende alle_vektoren_horizontal " << endl;
#endif
// Kommentar: Und der Bufferindex geht als return - Wert der Funktion zurck.
    return(index);
    }

#ifdef KURZ
// Kommentar: Analog zu alle_vektoren_horizontal, wobei nur statt der Ost - West - Feldlisten
// die Nord - Sd - Feldlisten verwendet werden.
int FeldListe1D::alle_vektoren_vertikal(short int * buffer, int _index, int i)
#else
int FeldListe1D::alle_vektoren_vertikal(int * buffer, int _index, int i)
#endif
    {
    index = _index;
    testfeld_index = 0;

#ifdef AUSGABEN_FELDLISTE_1D
    cout << "Anfang alle_vektoren_vertikal " << endl;
#endif

    for(anfangsposition(); ist_fertig()==false; naechster())
        {
        a = lesen();
        if(testfeld[a]==false)
            {
            b = vektorisieren_nach_oben(a);
            c = vektorisieren_nach_unten(a);
#ifdef AUSGABEN_FELDLISTE_1D
            cout << " a " << a << " b " << b << " c " << c << endl;
#endif
            for(j=b; j<=c; j++)
                {
                schreibe_testfeld(j);
                }
            buffer[index++] = b;
            buffer[index++] = c;
            buffer[index++] = i;
#ifdef AUSGABEN_FELDLISTE_1D
            cout << " STU vertikal " << " b " << b << " c " << c << " i " << i << endl;
#endif
            }
        }
    reset_testfeld();
#ifdef AUSGABEN_FELDLISTE_1D
    cout << "Ende alle_vektoren_vertikal " << endl;
#endif
    return(index);
    }

// Kommentar: Man markiert im Testfeld die Stelle unter der Variablen eintrag.
void FeldListe1D::schreibe_testfeld(int eintrag)
    {
// Kommentar: Gab es hier schon einen Eintrag => Fehlermeldung
    if(testfeld[eintrag]!=false)
        {
        cout << "Fehler in schreibe_testfeld " << endl;
        exit(-1);
        }
// Kommentar: Markiere nun das Testfeld.
    testfeld[eintrag] = true;
// Kommentar: Und speichere den Eintrag in einer Liste ab frs resetten.
    testfeld_indexliste[testfeld_index++] = eintrag;
    }

// Kommentar: resette das Testfeld, so dass berall false steht.
void FeldListe1D::reset_testfeld(void)
    {
// Kommentar: Die gemerkten testfeld_indizes werden durchlaufen ...
    for(i=0; i<testfeld_index; i++)
        {
// Kommentar: Und remarkiere das Testfeld dort durch schreiben von false.
        testfeld[testfeld_indexliste[i]] = false;
        }
// Kommentar: Setze den Testfeld Index wieder zurck, da das Testfeld als leer beschrieben ist.
    testfeld_index = 0; // Eigentlich berflssig, wegen Zeile 165
    }

// Kommentar: Man sieht hier, wie weit ab der bergebenen Koordinate Koordinateneintrge
// ununterbrochen hin zum kleineren Existieren.
int FeldListe1D::vektorisieren_nach_oben(int koord)
    {
// Kommentar: Man durchluft die Koordinaten hin zum kleineren.
    for(j=koord-1; j>=0; j--)
        {
// Kommentar: Tritt eine Unterbrechung in der Belegung auf, dann wird die Schleife abgebrochen.
        if(belegt[j]==false)
            {
            break;
            }
        }
// Kommentar: Gebe die maximale Verlngerung nach Norden zurck.
    return(j+1);
    }

// Kommentar: Analog wie vektorisieren_nach_oben, nur eben hin zu greren statt zu kleineren Koordinaten.
int FeldListe1D::vektorisieren_nach_unten(int koord)
    {
    for(j=koord+1; j<laenge; j++)
        {
        if(belegt[j]==false)
            {
            break;
            }
        }
    return(j-1);
    }

// Kommentar: Hier wird geprft, ob eine Koordinate belegt wurde.
bool FeldListe1D::teste_existenz(int stelle)
    {
    return(belegt[stelle]);
    }

void FeldListe1D::leeren(void)
    {
    // Kommentar: Das Anfangselement laenge, und das Endelement laenge_plus1 werden aufeinander verknpft.
    nachfolger[laenge] = laenge_plus1;
    vorgaenger[laenge_plus1] = laenge;

// Kommentar: Und am Ende wird der Stack resettet.
    stack1d.reset();;
    }
