#include "identify.h"

// Kommentar: Hier werden die lokalen Informationen bestimmt - entsprechend rang und groesse. Auerdem werden fr die horizontale
// Linie und vertikale Spalte der Kerntopologie zwei Kommunikatoren angelegt. Die werden bei der Precomputation - Methode
// bentigt fr die Broadcast - Operatoren, weil diese Operatoren nicht global erfolgen, sondern nur fr die jeweilige
// Spalte und Zeile der Kerntopologie.

// Kommentar: Hier wird die Klasse initialisiert. Dabei werden schon alle bentigten Daten generiert, und optional als Datei ausgegeben.
// Die bergabeparameter sind Rang und Groesse, d.h. die Anzahl an Kernen, sowie die ID des eigenen Kerns.
void Identify::initialisierung_identify(int _rang, int _groesse)
    {
// Kommentar: Zuerst werden die bergabeparameter gespeichert.
    rang = _rang;
    groesse = _groesse;
#ifdef IDENTIFY_DA
    strcpy(text, "da_precomp");
    da.initialisieren(rang, text);
    da.oeffnen();
#endif

// Kommentar: Dann werden Anzahl der Kerne fr Spalte und Zeile sowie die eigene ID in Spalte und Zeile bestimmt.
    bestimme_groessen();
    bestimme_raenge();
#ifdef IDENTIFY_DA
    * da.aus << " Nach bestimme_raenge " << endl;
#endif
// Kommentar: Dann werden die ID's in den 4 Himmelsrichtungen bestimmt. Im Falle von Kernen am Rand bzgl. der Kerntopologie
// werden negative ID's angegeben.
    bestimme_nachbar_IDs();
#ifdef IDENTIFY_DA
    * da.aus << " Nach bestimme_nachbarIDs " << endl;
#endif
// Kommentar: Dann wird festgestellt, wo der eigene Kern in der Spalte und Zeile bzgl. der Kerntopologie liegt. Es gibt
// dabei fr Spalte und Zeile 5 Mglichkeiten: 2 Am Rand, eine in der Mitte, und dann 2 zwischen Mitte und dem jeweiligen
// Rand.
    bestimme_positionen();
#ifdef IDENTIFY_DA
    * da.aus << " Nach bestimme_positionen " << endl;
#endif
// Kommentar: Dann wird bestimmt, wieviele Kerne in der Zeile links bzw. rechts liegen, bzw. in der Spalte darber bzw. darunter liegen.
    bestimme_anzahlen();
#ifdef IDENTIFY_DA
    * da.aus << " Nach bestimme_anzahlen " << endl;
#endif
// Kommentar: Bestimmung der mittleren Kerne bzgl. der Zeilen - oder Spaltenkerntopologie.
    bestimme_mitten();
#ifdef IDENTIFY_DA
    * da.aus << " Nach bestimme_mitten " << endl;
#endif
// Kommentar: Bestimmt die Rnge der Kerne bzgl. der eigenen Spalte und Zeile in Bezug auf die Kerntopologie.
    kreiere_raenge();
#ifdef IDENTIFY_DA
    * da.aus << " Nach kreiere raenge " << endl;
    * da.aus << " In Identify rang " << rang << " groesse_g " << groesse_g << endl;
#endif
// Kommentar: Erzeugt die Kommunikatoren bzgl. der eigenen Spalte und Zeile in Bezug auf die Kerntopologie.
    kreiere_kommunikatoren();
#ifdef IDENTIFY_DA
    * da.aus << " Nach kreiere kommunikatioren " << endl;
#endif
// Kommentar: Bestimmt den Rang bzgl. der eben kreierten Kommunikatoren.
    bestimme_lokale_raenge();
    }

// Kommentar: Finalisiert die Klasse, d.h. der in der Initialisierung belegte Speicherplatz wird freigegeben.
void Identify::finalisierung_identify(void)
    {
#ifdef IDENTIFY_DA
    da.schliessen();
#endif
// Kommentar: Lscht den Speicherplatz fr die Rnge, s.o..
    delete_raenge();
// Komemntar: Lscht die Kommunikatoren.
    delete_kommunikatoren();
    }

// Kommentar: Bestimmt die Spaltenlnge (groesse_f) und Zeilenlnge (groesse_g) in Bezug auf die Kerntopologie.
void Identify::bestimme_groessen(void)
// Vielleicht besser machen. Das kurz berlegen.
    {
// Kommentar: Bei einer Gesamtkernanzahl von a^2 ist die Zeilenlnge bzw. Spaltenlnge in Bezug auf die Kerntopologie jeweils a.
    switch(groesse)
        {
        case 1:
            groesse_f = 1;
            groesse_g = 1;
            break;
        case 2:
            groesse_f = 1;
            groesse_g = 2;
        case 4:
            groesse_f = 2;
            groesse_g = 2;
            break;
        case 8:
            groesse_f = 2;
            groesse_g = 4;
            break;
        case 9:
            groesse_f = 3;
            groesse_g = 3;
            break;
        case 16:
            groesse_f = 4;
            groesse_g = 4;
            break;
        case 25:
            groesse_f = 5;
            groesse_g = 5;
            break;
        case 36:
            groesse_f = 6;
            groesse_g = 6;
            break;
        case 49:
            groesse_f = 7;
            groesse_g = 7;
            break;
        case 64:
            groesse_f = 8;
            groesse_g = 8;
            break;
        case 81:
            groesse_f = 9;
            groesse_g = 9;
            break;
        case 100:
            groesse_f = 10;
            groesse_g = 10;
            break;
        }
    }

// Kommentar: Gibt die Zeilennummer (rang_f) und die Spaltennummer (rang_g) in Bezug auf die Kerntopologie an.
void Identify::bestimme_raenge(void)
    {
// Kommentar: Es gibt groesse / groesse_g viele Zeilen. Die eigene liegt in rang / groesse_g.
    rang_f = rang / groesse_g;
// Kommentar: In der wievielten Spalte der eigene Kern liegt wird bestimmt, indem vom Rang die Position
// des ersten Kernes in der Zeile abgezogen wird.
    rang_g = rang - rang_f * groesse_g; // Auch rechnen mod groesse_g
#ifdef IDENTIFY_DA
    * da.aus << " rang_f " << rang_f << " rang_g " << rang_g << endl;
#endif
    }

// Kommentar: Es wird die ID's der 4 Nachbarkerne in Richtung Ost, Nord, Sd und West bestimmt.
// Falls dort kein Kern existiert, wird eine negative ID gespeichert.
void Identify::bestimme_nachbar_IDs(void)
    {
// Kommentar: Im Westen ist der Rang 1 niedriger.
    ID_westen = rang - 1;
    // Wre ID_westen(P_3) = P_2. Das ist kappes.

// Kommentar: Ist der Kern schon ganz westlich, wird ID_westen negativ belegt.
    if(rang_g == 0)
        {
        ID_westen = -1; // streng genommen berflssig; aber der Vollstndigkeit halber.
        }
// Kommentar: ID_osten wird analog zu ID_westen bestimmt.
    ID_osten = rang + 1;
    if(rang_g == groesse_g-1)
        {
        ID_osten = -2;
        }
// Kommentar: Im Sden ist der Rang gleich der eigene Rang + groesse_g. '+ groesse_g' geht einmal die Spalten durch,
// bis die nchste Zeile an derselben Stelle erreicht wird.
    ID_sueden = rang + groesse_g;
// Kommentar: Ist der Kern schon ganz sdlich, dann wird eine negative ID gespeichert.
    if(rang_f == groesse_f-1)
        {
        ID_sueden = -2;
        }
// Kommentar: ID_norden wird analog zu ID_sueden bestimmt.
    ID_norden = rang - groesse_g;
    if(rang_f == 0)
        {
        ID_norden = -1;
        }
    }

// Kommentar: Es wird bestimmt, an welcher der 5 Typen von Positionen - vgl. die Initialisierung - der Kern liegt. Einmal
// in Bezug auf die OW - Kerntopologie, einmal in Bezug auf die NS - Kerntopologie.
void Identify::bestimme_positionen(void)
    {
    bestimme_position_ow();
    bestimme_position_ns();
    }

// Kommentar: Es wird bestimmt, an welcher der 5 Typen von Positionen der OW - Kerntopologie sich der Kern befindet.
void Identify::bestimme_position_ow(void)
    {
// Kommentar: Zuerst wird die Mittlere Spalte bestimmt.
    groesse_g_div_2 = groesse_g >> 1;

#ifdef IDENTIFY_DA
    * da.aus << " groesse_g_div_2 in identify " << groesse_g_div_2 << endl;
#endif

// Kommentar: Ist rang_g = 0, dann ist die Position der Rand im Westen.
    if(rang_g == 0)
        {
        position_ow = RAND_WESTEN;
        return;
        }

// Kommentar: Ist rang_g = groesse, dann ist die Position der Rand im Osten.
    if(rang_g == groesse_g-1)
        {
        position_ow = RAND_OSTEN;
        return;
        }

// Kommentar: Ist rang_g = mittlere Position, dann ist die Position die Mitte in Bezug auf die OW - Kerntopologie.
    if(rang_g == groesse_g_div_2)
        {
        position_ow = MITTE_WO;
        return;
        }

// Kommentar: Ist die Position links von der Mitte, dann speichere IM_WESTEN ab.
    if(rang_g < groesse_g_div_2)
        {
        position_ow = IM_WESTEN;
        return;
        }
// Kommentar: Ansonsten liegt die eigene Position im Osten.
    position_ow = IM_OSTEN;
    }

// Kommentar: Analog zu bestimme_position_wo, nur eben fr die Nord - Sd, statt der Ost - West - Kerntopologie.
void Identify::bestimme_position_ns(void)
    {
    groesse_f_div_2 = groesse_f >> 1;

    if(rang_f == 0)
        {
        position_ns = RAND_NORDEN;
        return;
        }

    if(rang_f == groesse_f-1)
        {
        position_ns = RAND_SUEDEN;
        return;
        }

    if(rang_f == groesse_f_div_2)
        {
        position_ns = MITTE_NS;
        return;
        }
    if(rang_f < groesse_f_div_2)
        {
        position_ns = IM_NORDEN;
        return;
        }
    position_ns = IM_SUEDEN;
    }

// Kommentar: Fr die Precomputation - Methode: Je nach Fall wird angegeben, wieviele Kerne im Osten / Westen, oder im Norden / Sden liegen.
void Identify::bestimme_anzahlen(void)
    {
#ifdef IDENTIFY_DA
    * da.aus << " bestimme_anzahlen position_ow " << position_ow << endl;
#endif
// Kommentar: Je nachdem, welcher der 5 Positionen in der OW - Kerntopologie der Kern annimmt, werden anzahl_westen oder anzahl_osten bestimmt,
// oder nicht. Dabei ist anzahl_westen die Anzahl in Richtung Westen in Bezug auf die Ost - West - Kerntopologie, und anzahl_osten analog festgelegt.
    switch(position_ow)
        {
// Kommentar: Fr die Punkte ganz im Westen und Osten brauchen die Werte anzahl_westen und anzahl_osten nicht bestimmt zu werden.
        case RAND_WESTEN:
            break;
        case IM_WESTEN:
// Kommentar: Im Westen leigen rang_g viele Kerne, im Osten groesse_g - rang_g - 1 viele Kerne. Gebraucht werden ggf. aber nicht mehr beide
// Informationen, vgl. die Precomputation - Berechnung fr jeden der 5 Flle.
            anzahl_westen = rang_g; // Z.B: rang_g = 2 => 2 westliche Nachbarn
            break;
        case RAND_OSTEN:
            break;
        case IM_OSTEN:
            anzahl_osten = groesse_g - rang_g - 1; // Z.B. rang_g = groesse_g-2 => 1 Nachbar stlich
            break;
        case MITTE_WO:
            anzahl_westen = rang_g;
            anzahl_osten = groesse_g - rang_g - 1;
#ifdef IDENTIFY_DA
            * da.aus << " anzahl_westen " << anzahl_westen << " anzahl_osten " << anzahl_osten << endl;
#endif
            break;
        }

// Kommentar: Analog zur Anzahl - Bestimmung im West - Ost - Fall.
    switch(position_ns)
        {
        case RAND_NORDEN:
            break;
        case IM_NORDEN:
            anzahl_norden = rang_f; // Z.B: rang_g = 2 => 2 westliche Nachbarn
            break;
        case RAND_SUEDEN:
            break;
        case IM_SUEDEN:
            anzahl_sueden = groesse_f - rang_f - 1; // Z.B. rang_g = groesse_g-2 => 1 Nachbar stlich
            break;
        case MITTE_NS:
            anzahl_norden = rang_f;
            anzahl_sueden = groesse_f - rang_f - 1;
#ifdef IDENTIFY_DA
            * da.aus << " anzahl_norden " << anzahl_norden << " anzahl_sueden " << anzahl_sueden << endl;
#endif
            break;
        }
    }

// Kommentar: Bestimmt den Kern in der Mitte der Kerntopologiespalte bzw. Kerntopologiezeile. Fr das Broadcast, weil die
// Information vom mittleren Kern stammt.
void Identify::bestimme_mitten(void)
    {
// Kommentar: Ganz links ist der Kern rang_f * groesse_g. Addiert man groesse_g_div_2 dazu, so gelangt man in die Mitte bzgl.
// der Ost - West - Kerntopologie.
    mitte_ow = groesse_g_div_2 + rang_f * groesse_g;
// Kommentar: rang_g ist die oberste Position in der NS - Spalte der Kerntopologie. Man muss von dort groesse_f_div_2 viele Zeilen
// nach Sden gehen, um zur Mitte bzgl. der Nord - Sd - Kerntopologie zu gelangen.
    mitte_ns = groesse_f_div_2 * groesse_g + rang_g;
    }

// Kommentar: Hier werden die Rnge der Kerne angelegt, die in derselbe Zeile oder Spalte bzgl. der Kerntopologie liegen, wie der
// eigene Kern. Diese Listen werden gebraucht, um die OW - oder NS - MPI - Kommunikatoren zu generieren.
void Identify::kreiere_raenge(void)
    {
// Kommentar: Zuerst wird der Speicherplatz angelegt fr die Spalte bzw. Zeile des eigenen Kerns in der Kerntopologie.
    raenge_ow = new int [groesse_g];
    raenge_ns = new int [groesse_f];
#ifdef IDENTIFY_DA
    * da.aus << " raenge_ow ";
#endif

// Kommentar: Zuerst werden die Rnge fr die eigene Zeile in der Kerntopologie in einer aufsteigenden Liste angegeben.
// Es beginnt bei dem ganz im Westen in der f. Zeile liegenden Kern mit der ID groesse_g * rang_f, und dann wird
// 1 addiert, bis man zum stlichen Kern gelangt.
    for(i=0, hilfsvariable1 = groesse_g * rang_f; i<groesse_g; i++)
        {
        raenge_ow[i] = hilfsvariable1 + i;
#ifdef IDENTIFY_DA
        * da.aus << raenge_ow[i] << " ";
#endif
        }
#ifdef IDENTIFY_DA
    * da.aus << endl << " raenge_ns ";
#endif
// Kommentar: Zuerst werden die Rnge fr die eigene Spalte in der Kerntopologie in einer aufsteigenden Liste angegeben.
// Es beginnt bei dem ganz im Norden in der g. Zeile liegenden Kern mit der ID groesse_f, und dann wird
// groesse_g addiert, bis man zum sdlichen Kern gelangt.
    for(i=0, hilfsvariable1 = rang_g; i < groesse_f; hilfsvariable1 += groesse_g, i++)
        {
        raenge_ns[i] = hilfsvariable1;
#ifdef IDENTIFY_DA
        * da.aus << raenge_ns[i] << " ";
#endif
        }
#ifdef IDENTIFY_DA
    * da.aus << endl;
#endif
    }

// Kommentar: Hier wird der Speicherplatz fr die Rnge wieder freigegeben.
void Identify::delete_raenge(void)
    {
    delete raenge_ow;
    delete raenge_ns;
    }

// Kommentar: Hier werden die Kommunikatoren fr die Spalte und Zeile des eigenen Kerns bzgl. der Kerntopologie angelegt.
void Identify::kreiere_kommunikatoren(void)
    {
// Kommentar: Zuerst werden zwei MPI_Gruppen angelegt.
    MPI_Comm_group(MPI_COMM_WORLD, &MPI_GROUP_WORLD1);
    MPI_Comm_group(MPI_COMM_WORLD, &MPI_GROUP_WORLD2);
// Kommentar: Dann werden die Listen mit den OW - und den NS - Rngen angelegt.
    MPI_Group_incl(MPI_GROUP_WORLD1, groesse_g, raenge_ow, &gruppe_ow);
    MPI_Group_incl(MPI_GROUP_WORLD2, groesse_f, raenge_ns, &gruppe_ns);
// Kommentar: Der eigene Rang in den Gruppen wird bestimmt.
    MPI_Group_rank(gruppe_ow, &ow_rang);
    MPI_Group_rank(gruppe_ns, &ns_rang);
// Kommentar: Dann wird aus jeder Gruppe je ein Kommunikator gemacht.
    MPI_Comm_create(MPI_COMM_WORLD, gruppe_ow, &MPI_COMM_OW);
    MPI_Comm_create(MPI_COMM_WORLD, gruppe_ns, &MPI_COMM_NS);
// Kommentar: Und dann werden fr jeden der zwei Kommunikatoren der eigene Rang bzgl. des jeweiligen Kommunikators bestimmt.
    MPI_Comm_rank(MPI_COMM_OW, & ow_rang_s);
    MPI_Comm_rank(MPI_COMM_NS, & ns_rang_s);
#ifdef IDENTIFY_DA
    * da.aus << " ow_rang " << ow_rang << " ow_rang_s " << ow_rang_s;
    * da.aus << " ns_rang " << ns_rang << " ns_rang_s " << ns_rang_s << endl;
#endif
    }

// Kommentar: Hier wird der bei der Erzeugung der OW - und NS - Kommunikatoren angelegte Speicherplatz wieder freigegeben.
void Identify::delete_kommunikatoren(void)
    {
    MPI_Comm_free(&MPI_COMM_OW);
    MPI_Comm_free(&MPI_COMM_NS);
    MPI_Group_free(&MPI_GROUP_WORLD1);
    MPI_Group_free(&MPI_GROUP_WORLD2);
    MPI_Group_free(&gruppe_ow);
    MPI_Group_free(&gruppe_ns);

    }

// Kommentar: Hier wird der Rang des mittleren Kerns des OW - Kommunikators und des NS - Kommunikators bestimmt.
void Identify::bestimme_lokale_raenge(void)
    {
// Nur zur Sicherheit: lokale-mittel-raenge bestimmen. Normalerweise wrde groesse_div_2 reichen. Aber so ist es ganz sicher.
// 18.09.2014

// Kommentar: Zuerst wird Speicherplatz angelegt fr das folgende Broadcasting.
    hilfsraenge_ow = new int [groesse_g];
    hilfsraenge_ns = new int [groesse_f];

// Kommentar: Dann wird beim OW - Hilfsrang die ID des eigenen Ranges an die Position ow_rang_s geschrieben.
    hilfsraenge_ow[ow_rang_s] = rang;
// Kommentar: Dann geschehen groesse_g viele Broadcasts, alle mit derselben Struktur hilfsraenge_ow, aber mit der Wurzel (root)
// i beim i.ten Kern bzgl. des Kommunikators MPI_COMM_OW.
    for(i=0; i<groesse_g; i++)
        {
        MPI_Bcast(&(hilfsraenge_ow[i]), 1, MPI_INT, i, MPI_COMM_OW);
        }
// Kommentar: Wenn der Rang des i. Kerns mitte_ow ist, dann ist i die Position, die Quelle des Broadcastings in der Precomputation
// Berechnung ist. Diese Variable wird als lokal_rang_mitte_ow gespeichert.
    for(i=0; i<groesse_g; i++)
        {
        if(hilfsraenge_ow[i]==mitte_ow)
            {
            lokal_rang_mitte_ow = i;
            break;
            }
        }
// Kommentar: Am Ende wird die Struktur hilfsraenge_ow wieder freigegeben.
    delete hilfsraenge_ow;
// Kommentar: Und dann macht man dasselbe in Bezug auf die Nord - Sd - Kerne, die in derselben Spalte bzgl. der Kerntopologie liegen.
    hilfsraenge_ns[ns_rang_s] = rang;
    for(i=0; i<groesse_f; i++)
        {
        MPI_Bcast(&(hilfsraenge_ns[i]), 1, MPI_INT, i, MPI_COMM_NS);
        }

    for(i=0; i<groesse_f; i++)
        {
        if(hilfsraenge_ns[i]==mitte_ns)
            {
            lokal_rang_mitte_ns = i;
            break;
            }
        }
    delete hilfsraenge_ns;
#ifdef IDENTIFY_DA
    * da.aus << " lokal_rang_mitte_wo " << lokal_rang_mitte_ow;
    * da.aus << " lokal_rang_mitte_ns " << lokal_rang_mitte_ns << endl;
#endif
    }
