#include "kommunikation_boolsch.h"

// Kommentar: Diese Klasse erfllt den Zweck, bitweise in einen zu kommunizierenden Sendebuffer zu schreiben, bzw. bitweise
// aus einem festgelegten Empfangsbuffer zu lesen - um krzere Nachrichten zu verschicken.

// Kommentar: Hier werden die angelegten Strukturen freigegeben.
void KommunikationBoolsch::finalisieren_kb(void)
    {
// Kommentar: Freigabe des Speicherplatzes der Indizes zu den Buffern.
    delete sendebuffer_index;
    delete empfangsbuffer_index;

// Kommentar: Freigabe des Speicherplatzes fr die Requests - die fr den Test, ob eine Kommunikation erfolgreich abgeschlossen
// ist, bentigt werden.
    delete sende_request;
    delete empfange_request;

// Kommentar: Freigabe der Zeiger auf die Sendebuffer und Empfangsbuffer.
    delete sendebuffer;
    delete empfangsbuffer;

// Kommentar: Freigabe der die Bits setzenden bzw. lesenden Klasse bool_short_int.
    delete bool_short_int;
    }

// Kommentar: Initialisierung der Klasse Kommunikation boolsch (kb)
void KommunikationBoolsch::initialisieren_kb(int _groesse, int _rang, int * _PT_grobgitter, int _empfangsbuffer_laenge)
//void KommunikationBoolsch::initialisieren_kb(int _groesse, int _rang, int _empfangsbuffer_laenge)
    {
// Kommentar: Zwei bergabeparameter werden gespeichert.
    groesse = _groesse;
    rang = _rang;

// Kommentar: Die Indizes auf die Buffer und die Requests auf die Kommunikationen werden angelegt.
    sendebuffer_index = new int [groesse];
    empfangsbuffer_index = new int [groesse];

    sende_request = new MPI_Request[groesse];
    empfange_request = new MPI_Request[groesse];

// Weitere bergabeparameter werden gespeichert, dabei insbesondere ein Zeiger auf die Grobgitter Kerntopologie.
    PT_grobgitter = _PT_grobgitter;
    empfangsbuffer_laenge = _empfangsbuffer_laenge;

// Kommentar: Fr Sendebuffer und Empfangsbuffer werden Zeiger angelegt.
    sendebuffer = new short * [groesse];
    empfangsbuffer = new short int * [groesse];

// Kommentar: Und zu jedem anderen Kern werden bool_short_int - Klassen angelegt - zum Lesen und Schreiben von Bits in und aus
// Buffern.
    bool_short_int = new BoolShortInt [groesse];
    // Wo werden die sendebuffer - Zeiger beschrieben? In "ParallelesRechnen::initialisieren_kommunikation_ep_s".
    }

// Kommentar: Der Anfang der Sendebuffer wird geschrieben.
void KommunikationBoolsch::anfang_schreiben(void)
    {
#ifdef AUSGABEN_KOMMUNIKATION_BOOLSCH
    cout << "Anfang schreiben  rang " << rang << endl;
#endif

// Kommentar: Man durchluft die bzgl. der Grobgitter - Kerntopologie benachbarten Kerne.
    for(i=0; i<groesse; i++)
        {
#ifdef AUSGABEN_KOMMUNIKATION_BOOLSCH
        cout << "AS:PTgg rang " << rang << " i " << i << " PT " << PT_grobgitter[i] << endl;
#endif
        if(PT_grobgitter[i]>0)
            {
// Kommentar: Und setzt bool_short_int fr den Kern an den Bufferanfang.
            bool_short_int[i].anfang(sendebuffer[i], 1);
            }
        }
#ifdef AUSGABEN_KOMMUNIKATION_BOOLSCH
    cout << "ENDE Anfang schreiben  rang " << rang << endl;
#endif
    }

// Kommentar: Frs Lesen aus den bool_short_int(i) Klassen werden diese - fr alle benachbarten Kerne - auf die Anfangsposition
// gesetzt.
void KommunikationBoolsch::anfang_lesen(void)
    {
    for(i=0; i<groesse; i++)
        {
        if(PT_grobgitter[i]>0)
            {
            bool_short_int[i].anfang(empfangsbuffer[i], 1);
            }
        }
    }

// Kommentar: Hier wird die Kommunikation initiiert.
void KommunikationBoolsch::kommunizieren_starten(void)
    {
// Kommentar: Man durchluft die bzgl. der Grobgitter - Kerntopologie benachbarten Kerne.
    for(i=0; i<groesse; i++)
        {
        if(PT_grobgitter[i] > 0)
            {
#ifdef AUSGABEN_KOMMUNIKATION_BOOLSCH
            cout << " Kommunikation starten i " << i << " rang " << rang << endl;
            cout << " KBlaenge / rang " << rang << " i " << i << " laenge " << bool_short_int[i].laenge() << " s[1] " << sendebuffer[i][1] << endl;
#endif
            if(rang==i)
                {
                cout << "Fehler: Prozessortopologie in KommunikationBoolsch falsch" << endl;
                exit(-1);
                }
// Kommentar: Man initiiert das Senden. Die Lnge wird in boll_short_int(i) nachgesehen.
            MPI_Isend(sendebuffer[i], bool_short_int[i].laenge(), MPI_SHORT, i, 400, MPI_COMM_WORLD, &(sende_request[i]));
// Kommentar: Man initiiert das Empfangen. Die Lnge, die an MPI bergeben wird, ist die Bufferlnge als maximal bereitstehender
// Speicherplatz.
            MPI_Irecv(empfangsbuffer[i], empfangsbuffer_laenge, MPI_SHORT, i, 400, MPI_COMM_WORLD, &(empfange_request[i]));
            }
        }
    }

// Kommentar: Hier wird geprft, ob zu allen bzgl. der Kerntopologie benachbarten Kernen Senden und Empfang der Nachrichten erfolgt
// ist.
bool KommunikationBoolsch::kommunikation_fertig(void)
    {
// Kommentar: Durchlauf ber die Rnge der benachbarten Kerne.
    for(i=0; i<groesse; i++)
        {
        if(rang != i)
            {
            if(PT_grobgitter[i] > 0)
                {
// Kommentar: Wenn Senden oder Empfangen nicht abgeschlossen ist, dann return false.
                MPI_Test(&(sende_request[i]), &flag, &sende_status);
                if(flag==false)
                    {
                    return(false);
                    }
                MPI_Test(&(empfange_request[i]), &flag, &empfange_status);
                if(flag==false)
                    {
                    return(false);
                    }
                }
            }
        }
// Kommentar: Alles erfolgreich abgeschlossen => return true.
    return(true);
    }

// Kommentar: Hier wird mit Hilfe der Klasse bool_short_int eine boolsche Variable in den zugehrigen Buffer geschrieben.
void KommunikationBoolsch::einfuegen(bool eintrag, int farbe_zielprozessor)
    {
    bool_short_int[farbe_zielprozessor].schreiben(eintrag);
    }
