#include "lastausgleichssteuerung.h"

// Kommentar: In dieser Klasse wird der Lastausgleich in der PLB - Variante fr eine Lastausgleichsrunde koordiniert - von der
// Bestimmung der Lastwerte ber die Klasse LastMengenBestimmung bis zum Precomputation - Verfahren und der Balancing -
// Simulation mit Hilfe der Klasse LastVektorAuswertung.

// Kommentar: Hier wird die Klasse finalisiert, d.h. der allokierte Speicherplatz wird freigegeben. Zur Bedeutung der Variablen
// vergleiche man die Gren der Funktion initialisierung.
void LastAusgleichsSteuerung::finalisierung(void)
    {

#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << " Ende 15c1 rang " << rang << endl;
#endif // AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    for(i=0; i<groesse_f; i++)
        {
        delete last_alle_prozessoren[i];
        }
    delete last_alle_prozessoren;

#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << " Ende 15c2 rang " << rang << endl;
#endif // AUSGABEN_LASTAUSGLEICHSSTEUERUNG

    lastvektorauswertung_ns.finalisieren();
    lastvektorauswertung_wo.finalisieren();

#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << " Ende 15c3 rang " << rang << endl;
#endif // AUSGABEN_LASTAUSGLEICHSSTEUERUNG

    delete lastvektor_horizontal;
    delete lastvektor_vertikal;

#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << " Ende 15c4 rang " << rang << endl;
#endif // AUSGABEN_LASTAUSGLEICHSSTEUERUNG

    lastmengenbestimmung.finalisieren_normal();

#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << " Ende 15c5 rang " << rang << endl;
#endif // AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    }

// Kommentar: Die Initialisierung der Klasse findet hier statt. Wichtige Parameter werden bestimmt und gespeichert, und die bentigten
// Strukturen werden mit Speicherplatz versorgt.
void LastAusgleichsSteuerung::initialisierung(int _groesse, int _rang)
    {
// Kommentar: Die Parameter groesse und rang werden gespeichert.
    groesse = _groesse;
    rang = _rang;
// Kommentar: Die Klasse LastMengenBestimmung wird mit diesen Parametern initialisiert.
    lastmengenbestimmung.initialisieren_normal(rang, groesse);
// Kommentar: groesse_f/g und position_f/g (eigener Rang in NS - und OW - Richtung) werden von der Instanz der Klasse LastMengenBestimmung
// bernommen. Dann werden in Abhngigkeit dieser Werte weitere Parameter berechnet.
    groesse_f = lastmengenbestimmung.groesse_f;
    groesse_g = lastmengenbestimmung.groesse_g;
    groesse_f_2 = groesse_f << 1;
    groesse_g_2 = groesse_g << 1;
    groesse_f_2_p1 = groesse_f_2 + 1;
    groesse_g_2_p1 = groesse_g_2 + 1;
    position_f = lastmengenbestimmung.position_f;
    position_g = lastmengenbestimmung.position_g;
    position_f_2 = position_f << 1;
    position_g_2 = position_g << 1;
    position_f_2_p1 = position_f_2 + 1;
    position_g_2_p1 = position_g_2 + 1;


#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << " LAS::init rang " << rang << " groesse_f " << groesse_f << " groesse_g " << groesse_g << " position_f " << position_f << " position_g " << position_g << endl;
#endif


// Kommentar: Die horizontalen und vertikalen Lastvektoren werden angelegt.
    lastvektor_horizontal = new int [groesse_g];
    lastvektor_vertikal = new int [groesse_f];

// Kommentar: Eine Instanz der LastVektorAuswertung fr Nord - Sd und Ost - West werden angelegt.
    lastvektorauswertung_wo.initialisieren(groesse_g, rang);
    lastvektorauswertung_ns.initialisieren(groesse_f, rang);

// Kommentar: Eine Matrix wird angelegt, die die Lasten aller Prozessoren aufnehmen kann - fr eine Gitterebene.
    last_alle_prozessoren = new int * [groesse_f];
    for(i=0; i<groesse_f; i++)
        {
        last_alle_prozessoren[i] = new int [groesse_g_2]; // nderung 15.10.15
        }
    // Alles wurde finalisiert. lastvektorauswertung_*.initialisieren wurde gegengecheckt gegenber lva_*.fina.

// Kommentar: Lastgrenzen zur Unterdrckung des Lastausgleiches werden angelegt.
    initialisieren_epsilons();
    }

// Kommentar: Hier werden die Faktoren angelegt, unterhalb derer ein Lastausgleich nicht mehr erfolgt. Das geschieht
// aber nur, wenn man die entsprechende Option whlt.
void LastAusgleichsSteuerung::initialisieren_epsilons(void)
    {
    switch(groesse)
        {
        case 1:
            epsilon_vektor = 1;
            epsilon_matrix = 1;
            break;
        case 2:
            epsilon_vektor = 1.1;
            epsilon_matrix = 1.1;
            break;
        case 4:
            epsilon_vektor = 1.1;
            epsilon_matrix = 1.21;
            break;
        case 8:
            epsilon_vektor = 1.21;
            epsilon_matrix = 1.331;
            break;
        case 16:
            epsilon_vektor = 1.21;
            epsilon_matrix = 1.4641;
            break;
        default:
            cout << " Fehler: initialisieren_epsilons in LastAusgleichsSteuerung erweitern // rang " << rang << endl;
            break;
        }
    }

// Kommentar: Diese Klasse verwendet eine Instanz der Klasse LastMengenBestimmung, um die Lastwerte einer Ost - West -
// Kerntopologie - Linie zu bestimmen. Nur fr eine Gitterebene, d.h. ohne reversen Lastausgleich. Dabei muss der
// eigene Lastwert bergeben werden.
void LastAusgleichsSteuerung::horizontaler_vektor_ohne(int eigene_last)
    {
// Kommentar: Bringe die eigene Last hin zur Instanz der LastMengenBestimmung.
    lastmengenbestimmung.setze_eigene_last(eigene_last);
// Kommentar: Wenn die all to all Kommunikation bzgl. der Ost - West - Kerntopologie erfolgen soll:
    if(plb_kn!=PLB)
        {
// Kommentar: Dann lasse die Instanz der LastMengenBestimmung die Lastwerte sammeln.
        lastmengenbestimmung.tun(KN_HORIZONTAL_OHNE);
#ifdef WARTEN
        for(j=0; j<rang*4000000; j++)
            {
            cout << "";
            }
#endif

// Kommentar: Dann kopiere man die bestimmten Werte in den horizontalen Lastvektor.
        for(i=0; i<groesse_g; i++)
            {
            lastvektor_horizontal[i] = lastmengenbestimmung.last_horizontal[i];
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
            cout << "lh[i] " << lastvektor_horizontal[i] << " ";
#endif
            }
        }
// Kommentar: Ansonsten mache die Lastmengenbestimmung ber einen Baum
    else
        {
// Kommentar: Verwende die Instanz der Klasse LastMengenBestimmung zu diesem Zweck.
        lastmengenbestimmung.tun(PLB_HORIZONTAL_OHNE);
#ifdef WARTEN
        for(j=0; j<rang*4000000; j++)
            {
            cout << "";
            }
#endif

// Kommentar: Dann kopiere man die bestimmten Werte in den horizontalen Lastvektor.
        for(i=0; i<groesse_g; i++)
            {
            lastvektor_horizontal[i] = lastmengenbestimmung.last_horizontal[i];
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
            cout << "lh[i] " << lastvektor_horizontal[i] << " ";
#endif
            }
        }
    }

// Kommentar: Das Gegenstck zur Funktion horizontaler_vektor_ohne: Hier wird auch im Falle des nicht reversen Lastausgleichs
// die Lastberechnung von Precomputation und Balancing - Simulation durchgefhrt, und die Lasttransferwerte werden gespeichert.
void LastAusgleichsSteuerung::vollende_horizontaler_vektor_ohne(void)
    {
// Kommentar: Fhre die Lastberechnung von Precomputation und Balancing - Simulation durch.
    lastvektorauswertung_wo.ablauf(lastvektor_horizontal);
// Kommentar: Speichere die Lasttransferwerte.
    west_differenz = lastvektorauswertung_wo.west_differenz_vektor[position_g];
    ost_differenz = lastvektorauswertung_wo.ost_differenz_vektor[position_g];
    }

// Kommentar: Hier werden die Funktionen horizontaler_vektor_ohne und vollende_horizontaler_vektor_ohne kombiniert, um eine
// komplette Berechnung der Lasttransferwerte durchzufhren. bergeben wird nur die eigene Last. Und: Man hat hier die Option,
// keinen Lastausgleich durchzufhren. Da es sich aber um eine Behandlung mit Lastvektoren handelt, ist diese Option
// unkoordiniert - sie gilt nur fr die einzelne Ost - West - Lastlinie.
void LastAusgleichsSteuerung::horizontaler_vektor_ohne_alles(int _eigene_last)
    {
// Kommentar: Es wird der maximale Lsatwert sowie die Lastsumme - in Bezug auf den horitontalen Lastvektor berechnet.
    horizontaler_vektor_ohne(_eigene_last);
    for(i=0, agg_wert=0, max_wert = 0; i<groesse_g; i++)
        {
        lh_wert = lastvektor_horizontal[i];
        agg_wert = agg_wert + lh_wert;
        if(lh_wert > max_wert)
            {
            lh_wert = max_wert;
            }
        }
// Kommentar: bertrifft der Maximalwert den Durchschnittswert * epsilon_vektor, dann mache den berechne die Lasttransferwerte fr den Lastausgleich,
// und vermerke das.
    if(groesse_g * max_wert > agg_wert * epsilon_vektor)
        {
        horizontaler_vektor_machen = true;
        vollende_horizontaler_vektor_ohne();
        }
// Kommentar: Ansonsten vermerke, dass kein Lastausgleich erfolgt, und berechne gar nichts.
    else
        {
        horizontaler_vektor_machen = false;
        }
    }

// Kommentar: Analog zur Funktion horizontaler_vektor_ohne_alles, nur eben fr die vertikalen Kerntopologien. Daher: Keine weitere Kommentierung hier.
void LastAusgleichsSteuerung::vertikaler_vektor_ohne_alles(int _eigene_last)
    {
    vertikaler_vektor_ohne(_eigene_last);
    for(i=0, agg_wert=0, max_wert = 0; i<groesse_f; i++) // nderung 15.10.15
        {
        lv_wert = lastvektor_vertikal[i];
        agg_wert = agg_wert + lv_wert;
        if(lv_wert > max_wert)
            {
            lv_wert = max_wert;
            }
        }
    if(groesse_g * max_wert > agg_wert * epsilon_vektor)
        {
        vertikaler_vektor_machen = true;
        vollende_vertikaler_vektor_ohne();
        }
    else
        {
        vertikaler_vektor_machen = false;
        }
    }

// Kommentar: Analog zur Funktion horizontaler_vektor_ohne, nur eben fr die vertikalen Kerntopologien. Daher: Keine weitere Kommentierung hier.
void LastAusgleichsSteuerung::vertikaler_vektor_ohne(int eigene_last)
    {
    lastmengenbestimmung.setze_eigene_last(eigene_last);

// Kommentar: Wenn die all to all Kommunikation bzgl. der Ost - West - Kerntopologie erfolgen soll:
    if(plb_kn!=PLB)
        {
        lastmengenbestimmung.tun(KN_VERTIKAL_OHNE);
        for(i=0; i<groesse_f; i++)
            {
            lastvektor_vertikal[i] = lastmengenbestimmung.last_vertikal[i];
            }
        }
    else
        {
        lastmengenbestimmung.tun(PLB_VERTIKAL_OHNE);
        for(i=0; i<groesse_f; i++)
            {
            lastvektor_vertikal[i] = lastmengenbestimmung.last_vertikal[i];
            }
        }
    }

// Kommentar: Analog zur Funktion vollende_horizontaler_vektor_ohne, nur eben fr die vertikalen Kerntopologien. Daher: Keine weitere Kommentierung hier.
void LastAusgleichsSteuerung::vollende_vertikaler_vektor_ohne(void)
    {
    lastvektorauswertung_ns.ablauf(lastvektor_vertikal);
    nord_differenz = lastvektorauswertung_ns.west_differenz_vektor[position_f];
    sued_differenz = lastvektorauswertung_ns.ost_differenz_vektor[position_f];
    }

// Kommentar: Hier werden die Funktionen horizontaler_vektor_mit und vollende_horizontaler_vektor_mit kombiniert, um eine
// komplette Berechnung der Lasttransferwerte durchzufhren. bergeben wird nur die eigene Last, und die eigene Last oben.
// Und: Man hat hier die Option, keinen Lastausgleich durchzufhren. Da es sich aber um eine Behandlung mit Lastvektoren handelt, ist diese Option
// unkoordiniert - sie gilt nur fr die einzelne Ost-West-Lastlinie.
void LastAusgleichsSteuerung::horizontaler_vektor_mit_alles(int _eigene_last, int _eigene_last_oben)
    {
// Kommentar: Mach die Lastmengenbestimmung fr die horizontale Kerntopologie fr beide Gitterebenen.
    horizontaler_vektor_mit(_eigene_last, _eigene_last_oben);
// Kommentar: Bilde den Maximalwert sowie die Lastsumme ber die Elemente des unteren Lastvektors.
    for(i=0, agg_wert=0, max_wert = 0; i<groesse_g; i++)
        {
        lh_wert = lastvektor_horizontal[i];
        agg_wert = agg_wert + lh_wert;
        if(lh_wert > max_wert)
            {
            lh_wert = max_wert;
            }
        }

// Kommentar: Wenn der Maximalwert den Durchschnittsvektor um den Faktor epsilon_vektor bersteigt, dann fhre die kompletten
// Lastberechnungen durch - und speichere, dass ein Lastausgleich erfolgen soll.
    if(groesse_g * max_wert > agg_wert * epsilon_vektor)
        {
        horizontaler_vektor_machen = true;
        vollende_horizontaler_vektor_mit();
        }
    else
        {
// Kommentar: Ansonsten speichere nur, dass kein Lastausgleich erfolgen soll.
        horizontaler_vektor_machen = false;
        }
    }

// Kommentar: Diese Klasse verwendet eine Instanz der Klasse LastMengenBestimmung, um die Lastwerte einer Ost - West -
// Kerntopologie - Linie zu bestimmen. Hier fr zwei Gitterebenen, d.h. fr den reversen Lastausgleich. Dabei mssen beide
// eigenen Lastwerte bergeben werden.
void LastAusgleichsSteuerung::horizontaler_vektor_mit(int eigene_last, int eigene_last_oben)
    {
// Kommentar: Bringe die eigenen Lastwerte zu der Instanz der Klasse LastMengenbestimmung.
    lastmengenbestimmung.setze_eigene_last(eigene_last);
    lastmengenbestimmung.setze_eigene_last_oben(eigene_last_oben);
//    cout << " HVMR horizontaler_vektor_mit rang " << rang << " ";

// Kommentar: Wenn die all to all Kommunikation bzgl. der Ost-West-Kerntopologie erfolgen soll:
    if(plb_kn!=PLB)
        {
// Kommentar: Dann lasse die Instanz der LastMengenBestimmung die Lastwerte sammeln. Fr beide Gitterebenen.
        lastmengenbestimmung.tun(KN_HORIZONTAL_MIT);
#ifdef WARTEN
        for(j=0; j<rang*4000000; j++)
            {
            cout << "";
            }
#endif
// Kommentar: Kopiere aus der Instanz der Klasse LastMengenBestimmung die Lastwerte fr die Gitterebene in den
// Lastvektor - horizontal.
        for(i=0, j=0; i<groesse_g; i++, j=j+2)
            {
            lastvektor_horizontal[i] = lastmengenbestimmung.last_horizontal[j];
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
            cout << "lh[i] " << lastvektor_horizontal[i] << " ";
#endif
            }
        }
// Kommentar: Wenn die Kommunikation ber einen Baum bzgl. der Ost-West-Kerntopologie erfolgen soll:
    else
        {
// Kommentar: Dann lasse die Instanz der LastMengenBestimmung die Lastwerte sammeln. Fr beide Gitterebenen.
        lastmengenbestimmung.tun(PLB_HORIZONTAL_MIT);
#ifdef WARTEN
        for(j=0; j<rang*4000000; j++)
            {
            cout << "";
            }
#endif

// Kommentar: Kopiere aus der Instanz der Klasse LastMengenBestimmung die Lastwerte fr die Gitterebene in den
// Lastvektor - horizontal.
        for(i=0, j=0; i<groesse_g; i++, j=j+2)  // Habe diese Schleife am 20.09.2014 geaendert. Vermutlich OK.
                                                // Aber: Wo bleiben die Werte oben?
            {
            lastvektor_horizontal[i] = lastmengenbestimmung.last_horizontal[j];
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
            cout << "lh[i] " << lastvektor_horizontal[i] << " ";
#endif
            }
        }
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << endl;
#endif
    }

// Kommentar: Das Gegenstck zur Funktion horizontaler_vektor_mit: Hier wird auch im Falle des reversen Lastausgleichs
// die Lastberechnung von Precomputation und Balancing - Simulation durchgefhrt, und die Lasttransferwerte werden gespeichert.
// Mache das fr beide Gitterebenen.
void LastAusgleichsSteuerung::vollende_horizontaler_vektor_mit(void)
    {
// Kommentar: Fhre die Lastberechnungen fr den horizontalen Lastvektor aus. Speichere dann die Lasttransferwerte.
    lastvektorauswertung_wo.ablauf(lastvektor_horizontal);
    west_differenz = lastvektorauswertung_wo.west_differenz_vektor[position_g];
    ost_differenz = lastvektorauswertung_wo.ost_differenz_vektor[position_g];

// Kommentar: Kopiere den Lastvektor der oberen Gitterebene in den horizontalen Lastvektor.
    for(i=0, j=1; i<groesse_g; i++, j=j+2)
        {
        lastvektor_horizontal[i] = lastmengenbestimmung.last_horizontal[j];
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
        cout << "lh[i] " << lastvektor_horizontal[i] << " ";
#endif
        }
// Kommentar: Fhre die Lastberechnungen fr den horizontalen Lastvektor aus. Speichere dann die Lasttransferwerte der oberen Gitterebene.
    lastvektorauswertung_wo.ablauf(lastvektor_horizontal);
    west_differenz_oben = lastvektorauswertung_wo.west_differenz_vektor[position_g];
    ost_differenz_oben = lastvektorauswertung_wo.ost_differenz_vektor[position_g];
    }

// Kommentar: Analog zur Funktion horizontaler_vektor_mit_alles. Daher keine weitere Kommentierung.
void LastAusgleichsSteuerung::vertikaler_vektor_mit_alles(int _eigene_last, int _eigene_last_oben)
    {
    vertikaler_vektor_mit(_eigene_last, _eigene_last_oben);
    for(i=0, agg_wert=0, max_wert = 0; i<groesse_g; i++)
        {
        lv_wert = lastvektor_horizontal[i];
        agg_wert = agg_wert + lv_wert;
        if(lv_wert > max_wert)
            {
            lv_wert = max_wert;
            }
        }
    if(groesse_g * max_wert > agg_wert * epsilon_vektor)
        {
        vertikaler_vektor_machen = true;
        vollende_horizontaler_vektor_mit();
        }
    else
        {
        vertikaler_vektor_machen = false;
        }
    }

// Kommentar: Analog zur Funktion horizontaler_vektor_mit. Daher keine weitere Kommentierung.
void LastAusgleichsSteuerung::vertikaler_vektor_mit(int eigene_last, int eigene_last_oben)
    {
    lastmengenbestimmung.setze_eigene_last(eigene_last);
    lastmengenbestimmung.setze_eigene_last_oben(eigene_last_oben);
#ifdef AUSGABEN_VORTRAG
    cout << " 80868 rang " << rang << " ";
#endif
    if(plb_kn!=PLB)
        {
        lastmengenbestimmung.tun(KN_VERTIKAL_MIT);
#ifdef WARTEN
        for(j=0; j<rang*4000000; j++)
            {
            cout << "";
            }
#endif
        for(i=0, j=0; i<groesse_f; i++, j=j+2)
            {
            lastvektor_vertikal[i] = lastmengenbestimmung.last_vertikal[j];
#ifdef AUSGABEN_VORTRAG
            cout << "lv[i] " << lastvektor_vertikal[i] << " ";
#endif
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
            cout << "lv[i] " << lastvektor_vertikal[i] << " ";
#endif
            }
        }
    else
        {
        lastmengenbestimmung.tun(PLB_VERTIKAL_MIT);
#ifdef WARTEN
        for(j=0; j<rang*4000000; j++)
            {
            cout << "";
            }
#endif
        for(i=0, j=0; i<groesse_f; i++, j=j+2)
            {
            lastvektor_vertikal[i] = lastmengenbestimmung.last_vertikal[j];
#ifdef AUSGABEN_VORTRAG
            cout << "lv[i] " << lastvektor_vertikal[i] << " ";
#endif
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
            cout << "lv[i] " << lastvektor_vertikal[i] << " ";
#endif
            }
        }
    }

// Kommentar: Analog zur Funktion vollende_horizontaler_vektor_mit. Daher keine weitere Kommentierung.
void LastAusgleichsSteuerung::vollende_vertikaler_vektor_mit(void)
    {
#ifdef AUSGABEN_VORTRAG
    cout << " rang " << rang << endl;
#endif
    lastvektorauswertung_ns.ablauf(lastvektor_vertikal);
    nord_differenz = lastvektorauswertung_ns.west_differenz_vektor[position_f];
    sued_differenz = lastvektorauswertung_ns.ost_differenz_vektor[position_f];

    for(i=0, j=1; i<groesse_f; i++, j=j+2) // nderung 15.10.15
        {
        lastvektor_vertikal[i] = lastmengenbestimmung.last_vertikal[j];
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
        cout << "lv[i] " << lastvektor_vertikal[i] << " ";
#endif
        }
    lastvektorauswertung_ns.ablauf(lastvektor_vertikal);
    nord_differenz_oben = lastvektorauswertung_ns.west_differenz_vektor[position_f];
    sued_differenz_oben = lastvektorauswertung_ns.ost_differenz_vektor[position_f];
    }

//////////////////////////////////////////////////////////////////////////

// Kommentar: Hier geschieht die Lastmengenbestimmung im Fall der Matrix ohne reversen Lastausgleich.
void LastAusgleichsSteuerung::matrix_ohne(int eigene_last)
    {
// Kommentar: Die Instanz der Klasse LastMengenBestimmung bekommt die eigene Last bergeben.
    lastmengenbestimmung.setze_eigene_last(eigene_last);

// Kommentar: Falls die Bestimmung ber vollstndige Graphen vertikaler und horizontaler Vektoren geschehen soll.
    if(plb_kn!=PLB)
        {
// Kommentar: Fhre das durch fr den einfachen Lastausgleich.
        lastmengenbestimmung.tun(KNxKN_OHNE);
        }
    else
        {
// Kommentar: Ansonsten mache mit Hilfe der Instanz der Klasse LastMengenBestimmung einen Ausgleich ber einen
// assymetrischen Baum, vgl. die Dissertation. Auch ber vertikale und horizontale Vektoren. Aber
// nur fr eine Gitterebene.
        lastmengenbestimmung.tun(PLBxPLB_OHNE);
        }
// Kommentar: Bestimme, ob die resultierende Lastmatrix honreichend ausgeglichen ist, oder nicht, durch
// den Grad der Abweichung des Maximalwertes vom Durchschnittswerte.
    matrix_ausgeglichen();
    }

// Kommentar: Hierdurch wird die Funktion matrix_ohne ergnzt: Falls der Lastausgleich hinreichend gut ist, dann
// berechne ihn hier.
void LastAusgleichsSteuerung::vollende_matrix_ohne(void)
    {
// Kommentar: Ist die Last schon hinreichend ausgeglichen, dann mache nichts.
    if(ausgeglichen==true)
        {
        return;
        }

#ifdef WARTEN
    for(j=0; j<rang*4000000; j++)
        {
        cout << "";
        }
#endif
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << "lh[i] ";
#endif
// Kommentar: Bestimme den horizontalen Lastvektor als position_f. Zeile der Lastmatrix.
    for(i=0; i<groesse_g; i++)
        {
        lastvektor_horizontal[i] = lastmengenbestimmung.last_alle_prozessoren[position_f][i];
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
        cout << lastvektor_horizontal[i] << " ";
#endif
        }
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << endl;
#endif
// Kommentar: Fhre eine Lastberechnung mit Precomputation und simuliertem Balancing durch.
    lastvektorauswertung_wo.ablauf_vektorupdate(lastvektor_horizontal);

// Kommentar: Speichere die bestimmten Lasttransfermengen ab.
    west_differenz = lastvektorauswertung_wo.west_differenz_vektor[position_g];
    ost_differenz = lastvektorauswertung_wo.ost_differenz_vektor[position_g];

#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << "lh2[i] ";
#endif

// Kommentar: kopiere den Lastvektor wieder in die Lastmatrix zurck von da, wo der Vektor herkam.
    for(i=0; i<groesse_g; i++)
        {
        last_alle_prozessoren[position_f][i] = lastvektor_horizontal[i];
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
        cout << lastvektor_horizontal[i] << " ";
#endif
        }
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << endl;
#endif

// Kommentar: Mache die Lastberechnung einschlielich dem simulierten Balancing fr alle brigen Lastzeilen,
// und kopiere die Ergebnisse dieser Lastberechnung in die Matrix zurck. Damit ist die Last-Matrix um eine
// Lastausgleichsrunde weiter.
    for(i=0; i<groesse_f; i++)
        {
        if(i!=position_f)
            {
            for(j=0; j<groesse_g; j++)
                {
                lastvektor_horizontal[j] = lastmengenbestimmung.last_alle_prozessoren[i][j];
                }
            lastvektorauswertung_wo.ablauf_vektorupdate(lastvektor_horizontal);

            for(j=0; j<groesse_g; j++)
                {
                last_alle_prozessoren[i][j] = lastvektor_horizontal[j];
                }
            }
        }

// Kommentar: Gebe diese Lastmatrix optionalerweise aus.
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << "last_alle_prozessoren " << endl;

    for(i=0; i<groesse_f; i++)
        {
        for(j=0; j<groesse_g; j++)
            {
            cout << last_alle_prozessoren[i][j] << " ";
            }
        cout << endl;
        }
#endif
// Das war der horizontal - Update der Matrix

// Kommentar: Bestimme den vertikalen Lastvektor aus der upgedateten Matrix.
    for(i=0; i<groesse_f; i++)
        {
        lastvektor_vertikal[i] = last_alle_prozessoren[i][position_g];
        }

// Kommentar: Fhre eine vertikale Precomputation- und simulierte Balancing-Berechnung durch.
// ndere nicht den Lastvektor, speichere aber die Lasttransferwerte ab.
    lastvektorauswertung_ns.ablauf(lastvektor_vertikal);
    nord_differenz = lastvektorauswertung_ns.west_differenz_vektor[position_f];
    sued_differenz = lastvektorauswertung_ns.ost_differenz_vektor[position_f];
    }

//////////////////////////////////////////////////////////////////////

// Kommentar: Hier geschieht die Lastmengenbestimmung im Fall der Matrix mit reversem Lastausgleich.
void LastAusgleichsSteuerung::matrix_mit(int eigene_last, int eigene_last_oben)
    {
// Kommentar: Die Instanz der Klasse LastMengenBestimmung bekommt die eigene Last bergeben - fr beide Gitterebenen.
    lastmengenbestimmung.setze_eigene_last(eigene_last);
    lastmengenbestimmung.setze_eigene_last_oben(eigene_last_oben);

// Kommentar: Falls die Bestimmung ber vollstndige Graphen vertikaler und horizontaler Vektoren geschehen soll.
    if(plb_kn!=PLB)
        {
// Kommentar: Fhre das durch fr den doppelten Lastausgleich.
        lastmengenbestimmung.tun(KNxKN_MIT);
        }
    else
        {
// Kommentar: Ansonsten mache mit Hilfe der Instanz der Klasse LastMengenBestimmung einen Ausgleich ber einen
// assymetrischen Baum, vgl. die Dissertation. Auch ber vertikale und horizontale Vektoren. Und: Fr beide
// Gitterebenen.
        lastmengenbestimmung.tun(PLBxPLB_MIT);
        }

// Kommenntar: Prfe, ob die Last auf der unteren Gitterebene hinreichend ausgeglichen ist.
    matrix_ausgeglichen_mit();
    }

// Kommentar: Hierdurch wird die Funktion matrix_mit ergnzt: Falls der Lastausgleich hinreichend gut ist, dann
// berechne ihn hier.
void LastAusgleichsSteuerung::vollende_matrix_mit(void)
    {
// Kommentar: Bei hinreichend ausgeglichener Lastmatrix mache gar nichts.
    if(ausgeglichen==true)
        {
        return;
        }

#ifdef WARTEN
    for(j=0; j<rang*4000000; j++)
        {
        cout << "";
        }
#endif

#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << "lh[i] ";
#endif

// Kommentar: Ansonsten lade die Lastwerte der position_f. Lastmatrix - Zeile in den horizontalen Lastvektor.
    for(i=0, j=0; i<groesse_g; i++, j=j+2)
        {
        lastvektor_horizontal[i] = lastmengenbestimmung.last_alle_prozessoren[position_f][j];
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
        cout << lastvektor_horizontal[i] << " ";
#endif
        }
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << endl;
#endif

// Kommentar: Fhre eine Lastberechnung mit Precomputation und simuliertem Balancing durch.
    lastvektorauswertung_wo.ablauf_vektorupdate(lastvektor_horizontal);

// Kommentar: Speichere die bestimmten Lasttransfermengen ab.
    west_differenz = lastvektorauswertung_wo.west_differenz_vektor[position_g];
    ost_differenz = lastvektorauswertung_wo.ost_differenz_vektor[position_g];
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << "lh2[i] ";
#endif

// Kommentar: kopiere den Lastvektor wieder in die Matrix last_alle_prozessoren.
    for(i=0, j=0; i<groesse_g; i++, j = j+2)
        {
        last_alle_prozessoren[position_f][j] = lastvektor_horizontal[i];
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
        cout << lastvektor_horizontal[i] << " ";
#endif
        }
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << endl;
#endif

// Kommentar: Mache die Lastberechnung einschlielich dem simulierten Balancing fr alle brigen Lastzeilen,
// und kopiere die Ergebnisse dieser Lastberechnung in die Matrix last_alle_prozessoren. Damit ist die Last-Matrix um eine
// Lastausgleichsrunde weiter.
    for(i=0; i<groesse_f; i++)
        {
        if(i!=position_f)
            {
            for(j=0, k=0; j<groesse_g; j++, k=k+2)
                {
                lastvektor_horizontal[j] = lastmengenbestimmung.last_alle_prozessoren[i][k];
                }
            lastvektorauswertung_wo.ablauf_vektorupdate(lastvektor_horizontal);

            for(j=0, k=0; j<groesse_g; j++, k=k+2)
                {
                last_alle_prozessoren[i][k] = lastvektor_horizontal[j];
                }
            }
        }

// Kommentar: Gebe diese Lastmatrix optionalerweise aus.
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << "last_alle_prozessoren " << endl;
    for(i=0; i<groesse_f; i++)
        {
        for(j=0; j<groesse_g; j++)
            {
            cout << last_alle_prozessoren[i][j] << " ";
            }
        cout << endl;
        }
#endif
// Das war der horizontal - Update der Matrix

// Kommentar: Bestimme den vertikalen Lastvektor aus der upgedateten Matrix.
    for(i=0; i<groesse_f; i++)
        {
        lastvektor_vertikal[i] = last_alle_prozessoren[i][position_g_2];
        }
// Kommentar: Fhre eine vertikale Precomputation- und simulierte Balancing-Berechnung durch.
// ndere nicht den Lastvektor, speichere aber die Lasttransferwerte ab.
    lastvektorauswertung_ns.ablauf(lastvektor_vertikal);
    nord_differenz = lastvektorauswertung_ns.west_differenz_vektor[position_f];
    sued_differenz = lastvektorauswertung_ns.ost_differenz_vektor[position_f];
//////////////////////////////////
///// Jetzt OBEN machen //////////
//////////////////////////////////



// Kommentar: Lade die Lastwerte der position_f. Lastmatrix - Zeile in den horizontalen Lastvektor - fr die obere Gitterebene.
    for(i=0, j=1; i<groesse_g; i++, j=j+2)
        {
        lastvektor_horizontal[i] = lastmengenbestimmung.last_alle_prozessoren[position_f][j];
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
        cout << lastvektor_horizontal[i] << " ";
#endif
        }
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << endl;
#endif

// Kommentar: Fhre eine Lastberechnung mit Precomputation und simuliertem Balancing durch.
    lastvektorauswertung_wo.ablauf_vektorupdate(lastvektor_horizontal);

// Kommentar: Speichere die bestimmten Lasttransfermengen ab fr den oberen Lastvektor ab.
    west_differenz_oben = lastvektorauswertung_wo.west_differenz_vektor[position_g];
    ost_differenz_oben = lastvektorauswertung_wo.ost_differenz_vektor[position_g];

#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << "lh2[i] ";
#endif

// Kommentar: kopiere den Lastvektor wieder in die Matrix last_alle_prozessoren in Bezug auf die Positionen
// der oberen Lastmatrix.
    for(i=0, j=1; i<groesse_g; i++, j=j+2)
        {
        last_alle_prozessoren[position_f][j] = lastvektor_horizontal[i];
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
        cout << lastvektor_horizontal[i] << " ";
#endif
        }
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << endl;
#endif

// Kommentar: Mache die Lastberechnung einschlielich dem simulierten Balancing fr alle brigen Lastzeilen,
// und kopiere die Ergebnisse dieser Lastberechnung in die Matrix last_alle_prozessoren - in Bezug auf die Lsatmatrix der oberen Gitterebene.
// Damit ist die Last - Matrix um eine Lastausgleichsrunde weiter.

    for(i=0; i<groesse_f; i++)
        {
        if(i!=position_f)
            {
            for(j=0, k=1; j<groesse_g; j++, k=k+2)
                {
                lastvektor_horizontal[j] = lastmengenbestimmung.last_alle_prozessoren[i][k];
                }
            lastvektorauswertung_wo.ablauf_vektorupdate(lastvektor_horizontal);

            for(j=0, k=1; j<groesse_g; j++, k=k+2)
                {
                last_alle_prozessoren[i][k] = lastvektor_horizontal[j];
                }
            }
        }

// Kommentar: Gebe beide Lastmatrizen (unten/oben) ineinander verschachtelt optionalerweise aus.
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << "last_alle_prozessoren " << endl;

    for(i=0; i<groesse_f; i++)
        {
        for(j=0; j<groesse_g_2; j++)
            {
            cout << last_alle_prozessoren[i][j] << " ";
            }
        cout << endl;
        }
#endif
// Das war der horizontal - Update der Matrix

// Kommentar: Kopiere den vertikalen Lastvektor der oberen Gitterebene nach lastvektor_vertikal.
    for(i=0; i<groesse_f; i++)
        {
        lastvektor_vertikal[i] = last_alle_prozessoren[i][position_g_2_p1];
        }
// Kommentar: Mache Precomputation und simuliertes Balancing fr den vertikalen oberen Lastvektor.
    lastvektorauswertung_ns.ablauf(lastvektor_vertikal);
// Kommentar: Und speichere die Lastransferwerte fr Norden und Sden der oberen Gitterebene ab.
    nord_differenz_oben = lastvektorauswertung_ns.west_differenz_vektor[position_f];
    sued_differenz_oben = lastvektorauswertung_ns.ost_differenz_vektor[position_f];
    }

///////////////////////////////////////////////

// Kommentar: Prfe, ob die Lastmatrix hinreichend ausgeglichen ist.
void LastAusgleichsSteuerung::matrix_ausgeglichen(void)
// true, falls sie hinreichend ausgeglichen ist.
    {
// Kommentar: Bilde die Lastsumme und das Lastmaximum ber die ganze Lastmatrix.
    durchschnitt = 0;
    maximum = 0;
    for(i=0; i<groesse_f; i++)
        {
        for(j=0; j<groesse_g; j++)
            {
            a_ij = lastmengenbestimmung.last_alle_prozessoren[i][j];
            durchschnitt = durchschnitt + a_ij;
            if(a_ij > maximum)
                {
                maximum = a_ij;
                }
            }
        }
// Kommentar: Dividiere die Lastsumme durch die Anzahl beteiligter Kerne.
    durchschnitt = durchschnitt / groesse;
// Kommentar: durchschnitt < maximum < FAKTOR * maximum
//    if(durchschnitt > FAKTOR * maximum) gendert am 8.8.15

// Kommentar: Ist der Durchschnitt Mal Faktor (>1) grer als das Maximum, ist die Last hinreichend ausgeglichen
    if(durchschnitt * FAKTOR > maximum)
        {
        ausgeglichen = true;
        }
    else
        {
        ausgeglichen = false;
        }
    }

// Kommentar: Im Wesentlichen analog zu matrix_ausgeglichen, weil die Entscheidung dafr nur die untere Lastmatrix ist.
// Einziger Unterschied: Die Lastmatrix hat hier in x - Richtung nur jeden 2.ten Wert. Deswegen hat man hier
// andere a_ij - Werte.
void LastAusgleichsSteuerung::matrix_ausgeglichen_mit(void)
// true, falls sie hinreichend ausgeglichen ist.
    {
    durchschnitt = 0;
    maximum = 0;
    for(i=0; i<groesse_f; i++)
        {
// Kommentar: Der einzige Unterschied zu matrix_ausgeglichen ist, dass hier jeder 2. LAstwert in x-Richtung gelesen wird. Es
// sind genauso viele wie in matrix_ausgeglichen.
        for(j=0; j<groesse_g_2; j+=2)
            {
            a_ij = lastmengenbestimmung.last_alle_prozessoren[i][j];
            durchschnitt = durchschnitt + a_ij;
            if(a_ij > maximum)
                {
                maximum = a_ij;
                }
            }
        }
    durchschnitt = durchschnitt / groesse;
    if(durchschnitt * FAKTOR > maximum)
        {
        ausgeglichen = true;
        }
    else
        {
        ausgeglichen = false;
        }
    }

//////////////////////////////////////////////////////////

// Kommentar: Analog zu matrix_ohne. Daher kein weiterer Kommentar.
void LastAusgleichsSteuerung::matrix_ohne_erst_ns(int eigene_last)
    {
    lastmengenbestimmung.setze_eigene_last(eigene_last);

    if(plb_kn!=PLB)
        {
        lastmengenbestimmung.tun(KNxKN_OHNE);
        }
    else
        {
        lastmengenbestimmung.tun(PLBxPLB_OHNE);
        }
    matrix_ausgeglichen();
    }

// Kommentar: Analog zu vollende_matrix_ohne, nur dass hier erst der Nord-Sd-Lastausgleich behandelt wird, und danach dann
// der Ost-West-Lastausgleich. Daher: Keine Kommentare hierzu.
void LastAusgleichsSteuerung::vollende_matrix_ohne_erst_ns(void)
    {
    if(ausgeglichen==true)
        {
        return;
        }

#ifdef WARTEN
    for(j=0; j<rang*4000000; j++)
        {
        cout << "";
        }
#endif
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << "lh[i] ";
#endif
    for(i=0; i<groesse_f; i++)
        {
        lastvektor_vertikal[i] = lastmengenbestimmung.last_alle_prozessoren[i][position_g];
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
        cout << lastvektor_vertikal[i] << " ";
#endif
        }
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << endl;
#endif
    lastvektorauswertung_ns.ablauf_vektorupdate(lastvektor_vertikal);

    nord_differenz = lastvektorauswertung_ns.west_differenz_vektor[position_f];
    sued_differenz = lastvektorauswertung_ns.ost_differenz_vektor[position_f];

#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << "lh2[i] ";
#endif
    for(i=0; i<groesse_f; i++)
        {
        last_alle_prozessoren[i][position_g] = lastvektor_vertikal[i];
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
        cout << lastvektor_horizontal[i] << " ";
#endif
        }
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << endl;
#endif

    for(i=0; i<groesse_g; i++)
        {
        if(i!=position_g)
            {
            for(j=0; j<groesse_f; j++)
                {
                lastvektor_vertikal[j] = lastmengenbestimmung.last_alle_prozessoren[j][i];
                }
            lastvektorauswertung_ns.ablauf_vektorupdate(lastvektor_vertikal); // nderung 15.10.15

            for(j=0; j<groesse_f; j++) // nderung 15.10.15
                {
                last_alle_prozessoren[j][i] = lastvektor_vertikal[j];
                }
            }
        }
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << "last_alle_prozessoren " << endl;

    for(i=0; i<groesse_f; i++)
        {
        for(j=0; j<groesse_g; j++)
            {
            cout << last_alle_prozessoren[i][j] << " ";
            }
        cout << endl;
        }
#endif
// Das war der horizontal - Update der Matrix
    for(i=0; i<groesse_g; i++)
        {
        lastvektor_horizontal[i] = last_alle_prozessoren[position_f][i];
        }
    lastvektorauswertung_wo.ablauf(lastvektor_horizontal);
    west_differenz = lastvektorauswertung_wo.west_differenz_vektor[position_g];
    ost_differenz = lastvektorauswertung_wo.ost_differenz_vektor[position_g]; // nderung 15.10.15
    }

//////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////

// Kommentar: Analog zu matrix_mit. Daher kein weiterer Kommentar.
void LastAusgleichsSteuerung::matrix_mit_erst_ns(int eigene_last, int eigene_last_oben)
    {
    lastmengenbestimmung.setze_eigene_last(eigene_last);
    lastmengenbestimmung.setze_eigene_last_oben(eigene_last_oben);

    if(plb_kn!=PLB)
        {
        lastmengenbestimmung.tun(KNxKN_MIT);
        }
    else
        {
        lastmengenbestimmung.tun(PLBxPLB_MIT);
        }

    matrix_ausgeglichen_mit();
    }

// Kommentar: Analog zu matrix_mit, nur dass hier zuerst der Nord-Sd-Lastausgleich durchgerechnet wird,
// und danach der Ost-West-Lastausgleich. Daher: Keine weiteren Kommentare.
void LastAusgleichsSteuerung::vollende_matrix_mit_erst_ns(void)
    {
    if(ausgeglichen==true)
        {
        return;
        }

#ifdef WARTEN
    for(j=0; j<rang*4000000; j++)
        {
        cout << "";
        }
#endif

#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << "lh[i] ";
#endif
    for(i=0; i<groesse_f; i++)
        {
        lastvektor_vertikal[i] = lastmengenbestimmung.last_alle_prozessoren[i][position_g_2];
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
        cout << lastvektor_vertikal[i] << " "; // nderung 15.10.15
#endif
        }
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << endl;
#endif
    lastvektorauswertung_ns.ablauf_vektorupdate(lastvektor_vertikal);

    nord_differenz = lastvektorauswertung_ns.west_differenz_vektor[position_f];
    sued_differenz = lastvektorauswertung_ns.ost_differenz_vektor[position_f];
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << "lh2[i] ";
#endif
    for(i=0; i<groesse_f; i++)
        {
        last_alle_prozessoren[i][position_g_2] = lastvektor_vertikal[i]; // nderung 15.10.15
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
        cout << lastvektor_vertikal[i] << " ";
#endif
        }
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << endl;
#endif
    for(i=0; i<groesse_g_2; i+=2)
        {
        if(i!=position_g_2)
            {
            for(j=0; j<groesse_f; j++)
                {
                lastvektor_vertikal[j] = lastmengenbestimmung.last_alle_prozessoren[j][i];
                }
            lastvektorauswertung_ns.ablauf_vektorupdate(lastvektor_vertikal); // nderung 15.10.15

            for(j=0; j<groesse_f; j++)
                {
                last_alle_prozessoren[j][i] = lastvektor_vertikal[j];
                }
            }
        }
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << "last_alle_prozessoren " << endl;
    for(i=0; i<groesse_f; i++)
        {
        for(j=0; j<groesse_g; j++)
            {
            cout << last_alle_prozessoren[i][j] << " ";
            }
        cout << endl;
        }
#endif
// Das war der horizontal - Update der Matrix
    for(i=0, j=0; i<groesse_g_2; i+=2, j++) // gendert am 15.10.15
        {
        lastvektor_horizontal[j] = last_alle_prozessoren[position_f][i]; // gendert am 15.10.15
        }
    lastvektorauswertung_wo.ablauf(lastvektor_horizontal);
    west_differenz = lastvektorauswertung_wo.west_differenz_vektor[position_g];
    ost_differenz = lastvektorauswertung_wo.ost_differenz_vektor[position_g];    // nderung am 15.10.15
//////////////////////////////////
///// Jetzt OBEN machen //////////
//////////////////////////////////
    for(i=0; i<groesse_f; i++)
        {
        lastvektor_vertikal[i] = lastmengenbestimmung.last_alle_prozessoren[i][position_g_2_p1];
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
        cout << lastvektor_vertikal[i] << " "; // nderung 15.10.15
#endif
        }
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << endl;
#endif
    lastvektorauswertung_ns.ablauf_vektorupdate(lastvektor_vertikal);

    nord_differenz_oben = lastvektorauswertung_ns.west_differenz_vektor[position_f];
    sued_differenz_oben = lastvektorauswertung_ns.ost_differenz_vektor[position_f];
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << "lh2[i] ";
#endif
    for(i=0; i<groesse_f; i++)
        {
        last_alle_prozessoren[i][position_g_2_p1] = lastvektor_vertikal[i]; // nderung 15.10.15
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
        cout << lastvektor_vertikal[i] << " ";
#endif
        }
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << endl;
#endif
    for(i=1; i<groesse_g_2_p1; i+=2)
        {
        if(i!=position_g_2_p1)
            {
            for(j=0; j<groesse_f; j++)
                {
                lastvektor_vertikal[j] = lastmengenbestimmung.last_alle_prozessoren[j][i];
                }
            lastvektorauswertung_ns.ablauf_vektorupdate(lastvektor_vertikal);

            for(j=0; j<groesse_f; j++)
                {
                last_alle_prozessoren[j][i] = lastvektor_vertikal[j];
                }
            }
        }
#ifdef AUSGABEN_LASTAUSGLEICHSSTEUERUNG
    cout << "last_alle_prozessoren " << endl;
    for(i=0; i<groesse_f; i++)
        {
        for(j=0; j<groesse_g; j++)
            {
            cout << last_alle_prozessoren[i][j] << " ";
            }
        cout << endl;
        }
#endif
// Das war der horizontal - Update der Matrix
    for(i=1, j=0; i<groesse_g_2_p1; i+=2, j++) // gendert am 15.10.15
        {
        lastvektor_horizontal[j] = last_alle_prozessoren[position_f][i]; // gendert am 15.10.15
        }
    lastvektorauswertung_wo.ablauf(lastvektor_horizontal);
    west_differenz_oben = lastvektorauswertung_wo.west_differenz_vektor[position_g];
    ost_differenz_oben = lastvektorauswertung_wo.ost_differenz_vektor[position_g]; // nderung 15.10.15
    }

///////////////////////////////////////////////////////////

// Kommentar: Hier wird eine Komplette Entscheidung getroffen, ob berhaupt Last ausgeglichen wird, und wenn
// ja, ob dann zuerst NS- oder WO-Lastausgleich geschehen soll.
int LastAusgleichsSteuerung::komplette_entscheidung_matrix_ohne(int _eigene_last)
    {
// Kommentar: Mache die Lastmengenbestimmung, und berechne Maximum und Durchschnitt der Lastmatrixeintrge.
    matrix_ohne(_eigene_last);
// Kommentar: Falls die Last schon sehr ausgeglichen ist, dann mache nichts. Dabei gilt: epsilon_matrix = epsilon_vektor^2
    if(maximum < durchschnitt * epsilon_matrix)
        {
        return(KEIN_TUN);
        }
// Kommentar: Ansonsten mache erst eine Ost-West- gefolgt von eine Nord-Sd-Lastberechnung, und bestimme das Lastmaximum hinterher.
    vollende_matrix_ohne();
    matrix_ausgeglichen();
    maximum_wo = maximum;
//    durchschnitt_wo = durchschnitt;

// Kommentar: Mache dann erst eine Nord-Sd-gefolgt von eine West-Ost-Lastberechnung, und bestimme das Lastmaximum hinterher.
    vollende_matrix_ohne_erst_ns();
    matrix_ausgeglichen();
    maximum_ns = maximum;
//    durchschnitt_ns = durchschnitt;

// Kommentar: Ist die Berechnung beim 'erst NS-Lastausgleich besser'
    if(maximum_wo > maximum_ns)
    // wo ist schlechter ausgeglichen. Also wird auf jeden Fall NS ausgefhrt.
        {
// Kommentar: Dann prfe, ob man berhaupt den West-Ost-Lastausgleich machen soll. Hier liegt zugrunde, dass nach einem Lastausgleich die Last schon besser
// ausgeglichen ist als vorher, also durchschnitt * epsilon_vektor schon viel kleiner als durchschnitt * epsilon_matrix ist. Ist das Maximum zu gro, wird auch noch WO
// - nach dem NS - ausgefhrt.
        if(maximum_wo > durchschnitt * epsilon_vektor)
            {
// Kommentar: Wenn ja, dann soll erst ein Nord-Sd-Lastausgleich, und dann ein West-Ost-Lastausgleich erfolgen.
            return(ERST_NS_DANN_WO);
            }
        else
            {
// Kommentar: Sonst mache nur den Nord-Sd-Lastausgleich. Hier ist nach diesem LA die maximale Last < maximale Last-WO <= durchschnittslast * epsilon_vektor,
// d.h. der Lastausgleich ist schon nach dem NS-Lastausgleich ziemlich gut - relativ zum epsilon-Wert.
            return(NUR_NS);
            }
        }
// Kommentar: Analog, nur im Fall, dass es besser ist, erst den West-Ost-Lastausgleich auszufhren.
    else
        {
        if(maximum_ns > durchschnitt * epsilon_vektor)
            {
            return(ERST_WO_DANN_NS);
            }
        else
            {
            return(NUR_WO);
            }
        }
// Kommentar: Diese Zeile kann nicht passieren.
    return(FEHLER_LA);
    }

// Kommentar: Analog zu komplette_entscheidung_matrix_ohne, nur dass die Funktionen 'mit' statt 'ohne' aufgerufen werden.
// Auf die Lastentscheidungen hat das aber keine Auswirkung, d.h. die geschehen genau wie in der Funktion
// komplette_entscheidung_matrix_ohne.
int LastAusgleichsSteuerung::komplette_entscheidung_matrix_mit(int _eigene_last, int _eigene_last_oben)
    {
    matrix_mit(_eigene_last, _eigene_last_oben);
    if(maximum < durchschnitt * epsilon_matrix)
        {
        return(KEIN_TUN);
        }
    vollende_matrix_mit();
    matrix_ausgeglichen_mit();
    maximum_wo = maximum;
    vollende_matrix_mit_erst_ns();
    matrix_ausgeglichen_mit();
    maximum_ns = maximum;
    if(maximum_wo > maximum_ns)
    // wo ist schlechter ausgeglichen
        {
        if(maximum_wo > durchschnitt * epsilon_vektor)
            {
            return(ERST_NS_DANN_WO);
            }
        else
            {
            return(NUR_NS);
            }
        }
    else
        {
        if(maximum_ns > durchschnitt * epsilon_vektor)
            {
            return(ERST_WO_DANN_NS);
            }
        else
            {
            return(NUR_WO);
            }
        }
    return(FEHLER_LA);
    }

// Kommentar: Analog zu komplette_entscheidung_matrix_mit, nur dass hier nicht entschieden werden kann,
// nur einen Lastausgleich durchzufhren. Es wird nur entschieden, welcher Lastausgleich zuerst erfolgen soll.
// KEIN_TUN ist hier keine Option.
int LastAusgleichsSteuerung::teilweise_entscheidung_matrix_mit(int _eigene_last, int _eigene_last_oben)
    {
    matrix_mit(_eigene_last, _eigene_last_oben);
    vollende_matrix_mit();
    matrix_ausgeglichen_mit();
    maximum_wo = maximum;
    vollende_matrix_mit_erst_ns();
    matrix_ausgeglichen_mit();
    maximum_ns = maximum;
    if(maximum_wo > maximum_ns)
    // wo ist schlechter ausgeglichen
        {
            {
            return(ERST_NS_DANN_WO);
            }
        }
    else
        {
            return(ERST_WO_DANN_NS);
        }
    return(FEHLER_LA);
    }

// Kommentar: Analog zu komplette_entscheidung_matrix_ohne, nur dass hier nicht entschieden werden kann,
// nur einen Lastausgleich durchzufhren. Es wird nur entschieden, welcher Lastausgleich zuerst erfolgen soll.
// KEIN_TUN ist hier keine Option.
int LastAusgleichsSteuerung::teilweise_entscheidung_matrix_ohne(int _eigene_last)
    {
    matrix_ohne(_eigene_last);
    if(maximum < durchschnitt * epsilon_matrix)
        {
        return(KEIN_TUN);
        }
    vollende_matrix_ohne();
    matrix_ausgeglichen();
    maximum_wo = maximum;
    vollende_matrix_ohne_erst_ns();
    matrix_ausgeglichen();
    maximum_ns = maximum;
    if(maximum_wo > maximum_ns)
    // wo ist schlechter ausgeglichen
        {
            {
            return(ERST_NS_DANN_WO);
            }
        }
    else
        {
            return(ERST_WO_DANN_NS);
        }
    return(FEHLER_LA);
    }

// Kommentar: Wichtige Bemerkung: Relevant ist hier die Beobachtung, dass alle Kerne - im Matrix-Fall - dieselbe Information haben. Sie treffen daher insbesondere
// dieselben Entscheidungen. Es wird damit vermieden, dass die einzelnen Kerne unterschiedliche Entscheidungen treffen, und das Programm dann nicht mehr korrekt arbeitet,
// weil die Kerne verschiedenes tun. Beim Fall Vektor kann dementsprechend viel weniger entschieden werden.
