#include "lastvektorauswertung.h"

// Kommentar: Diese Klasse ermglicht einerseits die Berechnung davon, wieviel Last verschickt bzw. empfangen werden soll - entsprechend einem Lastvektor.
// Allerdings geht die Leistung dieser Klasse darber hinaus: Es werden nmlich die Lasttransferwerte berechnet, das sind die real verschickbaren Lastwerte,
// und der Lastvektor wird dann entsprechend gendert. Auf diese Weise knnen beliebig viele Lastausgleichsrunden bis zum vollstndigen Lastausgleich
// simuliert werden. Die Klasse GrosseLastAusgleichsSteuerung kann mit Hilfe dieser Klasse den vollstndigen Lastausgleich herbeifhren.

// Kommentar: Hier wird die Klasse initialisiert. Hier werden die entprechenden Strukturen in Vektorlnge angelegt. Dafr wird der Parameter Vektorlnge gebraucht.
void LastVektorAuswertung::initialisieren(int _vektorlaenge, int _rang)
// finalisiert korrekt
    {
// Kommentar: Der Parameter wird gespeichert.
    vektorlaenge = _vektorlaenge;
    rang = _rang;

// Kommentar: Es wird gespeichert, wieviel Last nach Westen oder Osten verschickt wird.
    west_differenz_vektor = new int [vektorlaenge];
    ost_differenz_vektor = new int [vektorlaenge];

// Kommentar: Diese Struktur wird nur in veralteten Funktionen bentigt.
    west_differenz_vektor_td = new int [vektorlaenge];
    ost_differenz_vektor_td = new int [vektorlaenge];

// Kommentar: Auch diese Struktur wird nur in veralteten Funktionen bentigt.
    west_differenz_vektor_speichern = new int [vektorlaenge];
    ost_differenz_vektor_speichern = new int [vektorlaenge];

// Kommentar: Dieser Vektor nimmt die Partialsummen auf.
    aggregierter_lastvektor = new int [vektorlaenge];

// Kommentar: Dieser Vektor nimmt die Partialsummen des idealen Lastausgleichs auf.
    vergleichswerte = new int [vektorlaenge];

// Kommentar: Auch diese Struktur wird nur in veralteten Funktionen bentigt.
    add_last = new int [vektorlaenge];

//    lastvektor = new int [vektorlaenge];

// Kommentar: Auch diese Struktur wird nur in veralteten Funktionen bentigt.
    lastvektor_speichern = new int [vektorlaenge];

// Kommentar: Hier wird gespeichert, wieviel Last ein Kern abgeben kann. Damit wird erreicht, dass er nicht nach Osten und Westen zusammen mehr verschickt, als er als
// Last zur Verfgung hat. Die Werte hierfr werden fr jede simulierte Lastausgleichsrunde wieder hierher kopiert.
    last_reservoir = new int [vektorlaenge];
    // Die Finalisierung dazu ist korrekt.
// Kommentar: Fr Precomputation
    vL_ost = new int [vektorlaenge];
    vL_west = new int [vektorlaenge];
    }

// Kommentar: Die in der Initialisierung angelegten Speicher werden hier freigegeben. Zu den einzelnen Variablen: Vergleiche die Initialisierung.
void LastVektorAuswertung::finalisieren(void)
    {
#ifdef AUSGABEN_LASTVEKTORAUSWERTUNG
    cout << " Ende 15c2i rang " << rang << endl;
#endif // AUSGABEN_LASTVEKTORAUSWERTUNG
    delete west_differenz_vektor;
    delete ost_differenz_vektor;

#ifdef AUSGABEN_LASTVEKTORAUSWERTUNG
    cout << " Ende 15c2ii rang " << rang << endl;
#endif // AUSGABEN_LASTVEKTORAUSWERTUNG
    delete west_differenz_vektor_td;
    delete ost_differenz_vektor_td;
#ifdef AUSGABEN_LASTVEKTORAUSWERTUNG
    cout << " Ende 15c2iii rang " << rang << endl;
#endif // AUSGABEN_LASTVEKTORAUSWERTUNG
    delete west_differenz_vektor_speichern;
    delete ost_differenz_vektor_speichern;

#ifdef AUSGABEN_LASTVEKTORAUSWERTUNG
    cout << " Ende 15c2iv rang " << rang << endl;
#endif // AUSGABEN_LASTVEKTORAUSWERTUNG

    delete aggregierter_lastvektor;

    delete vergleichswerte;

    delete add_last;

    delete lastvektor_speichern;

    delete last_reservoir;

    delete vL_ost;
    delete vL_west;
#ifdef AUSGABEN_LASTVEKTORAUSWERTUNG
    cout << " Ende 15c2x rang " << rang << endl;
#endif // AUSGABEN_LASTVEKTORAUSWERTUNG
    }

// Kommentar: Hier wird eine Precomputation gemacht, gefolgt von einer simulierten Balancing - Runde.
void LastVektorAuswertung::ablauf(int * _lastvektor)
// Die Transferdaten sind hier fr alle Stellen da, nicht nur die der eigenen Prozessor_ID.
    {

// Kommentar: Zuerst wird der zu bearbeitende Lastvektor gesichert.
    lastvektor = _lastvektor;

// Kommentar: Bei Vektoren der Lnge <= 2 kann die Arbeit verkrzt werden.
    if(vektorlaenge<=2)
        {
// Kommentar: Bilde die Partialsummen
        aggregiere_lastvektor();
// Kommentar: Bilde die Partialsummen eines perfekten Lastausgleichs.
        vergleichswerte_schaffen();
        for(i=0; i<vektorlaenge; i++)
            {
// Kommentar: Bei Vektoren der Lnge <= 2 kann ein Kern nicht gleichzeitig nach Westen und Osten schicken, weshalb das Last-Reservoir nicht ntig ist.
// So entsprechend die Lasttransferwerte direkt den aus der Precomputation - Phase berechneten vL - Werten. Berechnet werden die entpsrechenden
// Gren west_differenz und ost_differenz fr alle beteiligten Kerne.
            transferdaten_bestimmen_direkt(i);
// Hier stimmt noch was nicht: Die Lasttransfervektoren werden nicht berechnet. Doch, sie werden berechnet
// fr alle Positionen.
            }
        }
// Kommentar: Nun ist die Vektorlnge > 2.
    else
        {
// Was ist mit n_p=1 => vektorlaenge = 1?

// Kommentar: Bilde die Partialsummen
        aggregiere_lastvektor();
// Kommentar: Bilde die Partialsummen eines perfekten Lastausgleichs.
        vergleichswerte_schaffen();
// Kommentar: Kopiere die Lastwerte des Vektors in die last_reservoir - Variable - damit nicht nach Westen und Osten mehr verschickt wird, als mglich ist.
        last_reservoir_anlegen();
// Kommentar: Lscht die west_differenz und ost_differenz - Variablen fr alle hier beteiligten Kerne.
        alte_daten_loeschen();

// Kommentar: Berechnet den maximal mglichen Lasttransfer nach Westen und Osten, und speichert das in den Variablen west_differenz und ost_differenz - fr
// alle beteiligten Kerne.
        for(l=0; l<vektorlaenge; l++)
            {
            differenz_daten_bestimmen(l);
            }

// Kommentar: Zu Debugging - Zwecken eine Ausgabe des Lastvektors.
        lastvektor_ausgeben();
        }
    }

// Kommentar: Hier wird eine Precomputation gemacht.
void LastVektorAuswertung::precomputation(int * _lastvektor)
    {
// Kommentar: Zuerst wird der zu bearbeitende Lastvektor gesichert.
    lastvektor = _lastvektor;
// Was ist mit n_p=1 => vektorlaenge = 1?

// Kommentar: Bilde die Partialsummen
    aggregiere_lastvektor();
// Kommentar: Bilde die Partialsummen eines perfekten Lastausgleichs.
    vergleichswerte_schaffen();

// Kommentar: Berechnet den maximal mglichen Lasttransfer nach Westen und Osten, und speichert das in den Variablen west_differenz und ost_differenz - fr
// alle beteiligten Kerne.
    for(l=0; l<vektorlaenge; l++)
        {
        vL_bestimmen(l);
        }
    }

void LastVektorAuswertung::ausgabe_vL(char * text)
    {
    cout << " GLAS rang " << rang << " text " << text;
    for(i=0;i<vektorlaenge;i++)
        {
        cout << " i " << i << " vL_west " << vL_west[i] << " vL_ost " << vL_ost[i];
        }
    cout << endl;
    }

void LastVektorAuswertung::ausgabe_bal(char * text, int iteration)
    {
    cout << " GLAS rang " << rang << " bal text " << text << " Iteration " << iteration;
    for(i=0;i<vektorlaenge;i++)
        {
        cout << " i " << i << " west_diff " << west_differenz_vektor[i] << " ost_diff " << ost_differenz_vektor[i];
        cout << " last " << lastvektor[i] << " vL_west " << vL_west[i] << " vL_ost " << vL_ost[i];
        }
    cout << endl;
    }

// Kommentar: Vor jeder simulierten Runde muss der Lastvektor nach last_reservoir kopiert werden. Man beachte, dass in ablauf_vektorupdate
// der Lastvektor nach jeder Lastausgleichsrunde angepasst wird.
void LastVektorAuswertung::last_reservoir_anlegen(void)
    {
    for(i=0; i<vektorlaenge; i++)
        {
        last_reservoir[i] = lastvektor[i];
        }
    }

// Kommentar: Lscht die West und Ost - Differenz - Vektoren. Denn in der Funktion ablauf werden nur Eintrage != 0 bearbeitet, sodass alle anderen
// Eintrge vorher 0 sein mssen.
void LastVektorAuswertung::alte_daten_loeschen(void)
    {
    for(i=0; i<vektorlaenge; i++)
        {
        west_differenz_vektor[i] = 0;
        ost_differenz_vektor[i] = 0;
        }
    }

// Kommentar: Diese Funktion berechnet die Lasttransferwerte west_differenz_vektor und ost_differenz_vektor ber die Funktion ablauf, und korrigiert dann
// entsprechend dieser Werte den Lastvektor. Auf diese Weise wird eine Lastausgleichsrunde fr den Ost - West - Fall simuliert.
void LastVektorAuswertung::ablauf_vektorupdate(int * _lastvektor)
    {
    ablauf(_lastvektor);
    for(i=0; i<vektorlaenge; i++)
        {
//        lastvektor[i] = lastvektor_speichern[i] - west_differenz_vektor[i] - ost_differenz_vektor[i];
        lastvektor[i] = lastvektor[i] - west_differenz_vektor[i] - ost_differenz_vektor[i];
#ifdef AUSGABEN_LASTVEKTORAUSWERTUNG
        cout << lastvektor[i] << " ";
#endif
        }
    }

// Kommentar: Hier werden die Partialsummen - gem der Dissertation - berechnet, d.h. lw(0), lw(0)+lw(1),lw(0)+lw(1)+lw(2) usw..
// Dabei ist die Quelle dieser Summen die Variable lastvektor.
void LastVektorAuswertung::aggregiere_lastvektor(void)
// Man hat die Lasten auf der Zeile, und braucht diese nur zu aggregieren!
    {
// Kommentar: Man beginnt bei der Last des 0. Kernes.
    lastwert = lastvektor[0];
// Hm. Nicht lastwert = lastvektor[position_g]? lastvektor[0] ist doch die Last des Prozessors 0.
// Stimmt doch. Das ist ein aggregierter Lastwert.
    aggregierter_lastvektor[0] = lastwert;
    for(i=1; i<vektorlaenge; i++)
        {
// Kommentar: Dann wird die Last des i. Kernes hinzuaddiert und gespeichert.
        lastwert = lastwert + lastvektor[i];
        aggregierter_lastvektor[i] = lastwert;
        }
    }

// Kommentar: Hier wird das Balancing fr die grosse_Lastausgleichssteuerung gemacht.
void LastVektorAuswertung::balancing(void)
    {
// Kommentar: Lscht die west_differenz und ost_differenz - Variablen fr alle hier beteiligten Kerne. Deswegen brauchen im Folgenden nur Differenzen > 0 behandelt werden.
// Ansonsten wrden bei 0-Differenz die alten Differenz-Vektoren gelesen - mit Fehler.
    alte_daten_loeschen();
// Kommentar: Schritt 1: Lasttransferwerte berechnen
    for(i=0; i<vektorlaenge; i++)
        {
        last_reservoir_hilfe = lastvektor[i];
        west_differenz = vL_west[i];
        if(west_differenz > 0)
            {
// Kommentar: Man ldt das Last - Reservoir.
// Kommentar: Dieses ist eine obere Schranke fr das an die fr den Westkern zu verschickende Lastmenge. Die gewnschte zu verschickende Last ist aber
// west_differenz.
            if(west_differenz > last_reservoir_hilfe)
                {
                west_differenz = last_reservoir_hilfe;
                }
// Kommentar: Die Westdifferenz, und die Ost - Differenz des westlichen Kernes werden dementsprechend gespeichert.
            west_differenz_vektor[i] = west_differenz;
            ost_differenz_vektor[i-1] = - west_differenz;
            last_reservoir_hilfe -= west_differenz;
// Kommentar: Und das last_reservoir wird um die verschickte Lastmenge korrigiert. Man darf nicht gleichzeitig das Last - Reservoir des Westlichen Kernes
// erhhen, weil die Last zeitversetzt eintrifft, und fr den eigenen Lasttransfer deswegen noch nicht genutzt werden kann.
// Fr die nderung der Lastsituation ist nmlich die Lastwertekorrektur in ablauf_vektorupdate zustndig. Dort steigt dann die zu verschickende Lastmenge
// bei der nchsten Lastausgleichsrunde.
            }
// Kommentar: Der Rest geschieht fr den Osten, so wie fr eben fr den Westen.
        ost_differenz = vL_ost[i];
        if(ost_differenz > 0)
            {
            if(ost_differenz > last_reservoir_hilfe)
                {
                ost_differenz = last_reservoir_hilfe;
                }
            west_differenz_vektor[i+1] = -ost_differenz;
            ost_differenz_vektor[i] = ost_differenz;
            }
        }
// Kommentar: Schritt 2: Aktualisierung des Lastvektors und von vL.
    for(i=0; i<vektorlaenge; i++)
        {
        lastvektor[i] = lastvektor[i] - west_differenz_vektor[i] - ost_differenz_vektor[i];
        vL_west[i] -= west_differenz_vektor[i];
        vL_ost[i] -= ost_differenz_vektor[i];
#ifdef AUSGABEN_LASTVEKTORAUSWERTUNG
        cout << lastvektor[i] << " ";
#endif
        }
    }

// Kommentar: Hier werden die vL's gem der Dissertation berechnet. Sie entstehen durch die Differenz aus aggregiertem Lastvektor (Partialsummen) und
// den Vergleichswerten (Lastdurchschnitt * Wegknotenanzahl; das entspricht den Partialsummen beim perfekten Lastausgleich).
void LastVektorAuswertung::differenz_daten_bestimmen(int position)
    {
// Kommentar: Wenn der Kern ganz links ist, dann ist nach Westen nichts zu verschicken, oder von dort zu empfangen. Die dazu herangezogene Lastdifferenz wird
// auf 0 gesetzt.
    if(position==0)
        {
        west_differenz = 0;
        }
    else
        {

// Kommentar: Bestimme die Differenz aus den Vergleichswerten (Partialsummen des perfekten Lastausgleichs) und dem aggregierten Lastvektor (reale Partialsummen)
// Wichtig: Bei der West - Differenz nimmt man diese Variablen an der position - 1 - das entspricht dem Reziproken der Ost - Differenz an der Stelle position-1.
        west_differenz = vergleichswerte[position-1] - aggregierter_lastvektor[position-1];
// Kommentar: Es reicht aus, die Flle west_differenz > 0 und ost_differenz > 0 zu machen. Denn West - Differenz < 0 entspricht Ost - Differenz > 0, und umgekehrt,
// und fr den Fall gleich 0 ist die Funktion alte_daten_loeschen da.
        if(west_differenz > 0)
            {
// Kommentar: Man ldt das Last - Reservoir.
            last_reservoir_hilfe = last_reservoir[position];
// Kommentar: Dieses ist eine obere Schranke fr das an die fr den Westkern zu verschickende Lastmenge. Die gewnschte zu verschickende Last ist aber
// west_differenz.
            if(west_differenz > last_reservoir_hilfe)
                {
                west_differenz = last_reservoir_hilfe;
                }
// Kommentar: Die Westdifferenz, und die Ost - Differenz des westlichen Kernes werden dementsprechend gespeichert.
            west_differenz_vektor[position] = west_differenz;
            ost_differenz_vektor[position-1] = - west_differenz;
// Kommentar: Und das last_reservoir wird um die verschickte Lastmenge korrigiert. Man darf nicht gleichzeitig das Last - Reservoir des Westlichen Kernes
// erhhen, weil die Last zeitversetzt eintrifft, und fr den eigenen Lasttransfer deswegen noch nicht genutzt werden kann.
// Fr die nderung der Lastsituation ist nmlich die Lastwertekorrektur in ablauf_vektorupdate zustndig. Dort steigt dann die zu verschickende Lastmenge
// bei der nchsten Lastausgleichsrunde.
            last_reservoir[position] -= west_differenz;
            }
        }
// Kommentar: Der Rest geschieht fr den Osten, so wie fr eben fr den Westen.
    if(position==vektorlaenge-1)
        {
        ost_differenz = 0;
        }
    else
        {
        ost_differenz = - vergleichswerte[position] + aggregierter_lastvektor[position];
        if(ost_differenz > 0)
            {
            last_reservoir_hilfe = last_reservoir[position];
            if(ost_differenz > last_reservoir_hilfe)
                {
                ost_differenz = last_reservoir_hilfe;
                }
            west_differenz_vektor[position+1] = -ost_differenz;
            ost_differenz_vektor[position] = ost_differenz;
            last_reservoir[position] -= ost_differenz;
            }
        }
#ifdef AUSGABEN_LASTVEKTORAUSWERTUNG
    cout << "Position " << position << " WD " << west_differenz << " OD " << ost_differenz << endl;
#endif
    }

// Hier wird mit der Vergleichswerte und dem aggregierten Lastvektor vL berechnet.
void LastVektorAuswertung::vL_bestimmen(int position)
    {
// Kommentar: Wenn der Kern ganz links ist, dann ist nach Westen nichts zu verschicken, oder von dort zu empfangen. Die dazu herangezogene Lastdifferenz wird
// auf 0 gesetzt.
    if(position==0)
        {
        vL_west[position] = 0;
        }
    else
        {

// Kommentar: Bestimme die Differenz aus den Vergleichswerten (Partialsummen des perfekten Lastausgleichs) und dem aggregierten Lastvektor (reale Partialsummen)
// Wichtig: Bei der West - Differenz nimmt man diese Variablen an der position - 1 - das entspricht dem Reziproken der Ost - Differenz an der Stelle position-1.
        vL_west[position] = vergleichswerte[position-1] - aggregierter_lastvektor[position-1];
        }
// Kommentar: Der Rest geschieht fr den Osten, so wie fr eben fr den Westen.
    if(position==vektorlaenge-1)
        {
        vL_ost[position] = 0;
        }
    else
        {
        vL_ost[position] = - vergleichswerte[position] + aggregierter_lastvektor[position];
        }
#ifdef AUSGABEN_LASTVEKTORAUSWERTUNG
    cout << "Position " << position << " WD " << west_differenz << " OD " << ost_differenz << endl;
#endif
    }


// Kommentar: Bei 2 Kernen kann die zu verschickende Lastmenge - nach Osten und Westen - direkt von der Precomputation - Methode
// bernommen werden. Dafr sind die Differenzen der Partialsummen (aggregierter Lastvektor) und der Lastausgleichspartialsummen
// (vergleichswerte) zu bilden. Fr nach Westen auf Position position-1, nach Osten auf der Position, aber mit anderem Vorzeichen,
// weil der Lasttransfer in die umgekehrte Richtung geht.
void LastVektorAuswertung::transferdaten_bestimmen_direkt(int position)
    {
    if(position==0)
        {
        west_differenz = 0;
        }
    else
        {
        west_differenz = vergleichswerte[position-1] - aggregierter_lastvektor[position-1];

        }

    if(position==vektorlaenge-1)
        {
        ost_differenz = 0;
        }
    else
        {
        ost_differenz = - vergleichswerte[position] + aggregierter_lastvektor[position];

        }
#ifdef AUSGABEN_LASTVEKTORAUSWERTUNG
    cout << "Position " << position << " WD " << west_differenz << " OD " << ost_differenz << endl;
#endif
    west_differenz_vektor[position] = west_differenz;
    ost_differenz_vektor[position] = ost_differenz;
    }

// Kommentar: Da die Gesamtlast und die Vektorlnge bekannt sind, kann leicht berechnet werden, wieviel Last bis zu Kern j angesammelt werden wird bei einem perfekten
// Lastausgleich. Das sind die Vergleichswerte zu den Partialsummen, die hier als aggregierte Lastwerte abgespeichert werden.
void LastVektorAuswertung::vergleichswerte_schaffen(void)
    {
    for(j=0; j<vektorlaenge; j++)
        {
        vergleichswerte[j] = ((j+1)*lastwert)/vektorlaenge;
#ifdef AUSGABEN_LASTVEKTORAUSWERTUNG
        cout << "VglWert j " << j << " VW " << vergleichswerte[j] << endl;
#endif
// Das ist (i+1) * Durchschnittswert; soviel Last wre bei ausgeglichener Last bei Prozessor (i+1)
        }
    }

// Kommentar: Das hier sind Ausgabe - Funktionen, die nur dem Debugging dienen.
void LastVektorAuswertung::ausgeben(void)
    {
#ifdef AUSGABEN_LASTVEKTORAUSWERTUNG
    cout << "West - Lastaenderung ";
    for(i=0; i<vektorlaenge; i++)
        {
        cout << west_differenz_vektor[i] << " ";
        }
    cout << endl;

    cout << "Ost - Lastaenderung ";
    for(i=0; i<vektorlaenge; i++)
        {
        cout << ost_differenz_vektor[i] << " ";
        }
    cout << endl;
#endif
    }

// Kommentar: Das hier sind Ausgabe - Funktionen, die nur dem Debugging dienen.
void LastVektorAuswertung::lastvektor_ausgeben(void)
    {
#ifdef AUSGABEN_LASTVEKTORAUSWERTUNG
    cout << "Last Ausgabe: " ;
    for(i=0; i<vektorlaenge; i++)
        {
        cout << lastvektor[i] << " ";
        }
    cout << endl;
#endif
    }

// Kommentar: Das hier sind Ausgabe - Funktionen, die nur dem Debugging dienen.
void LastVektorAuswertung::ausgabe_aggregierter_lastvektor(void)
    {
    cout << " aggregierter Lastvektor rang " << rang << " ";
    for(i=0; i<vektorlaenge; i++)
        {
        cout << aggregierter_lastvektor[i] << " ";
        }
    cout << endl;
    }

// Kommentar: Das hier sind Ausgabe - Funktionen, die nur dem Debugging dienen.
void LastVektorAuswertung::ausgabe_vergleichswerte(void)
    {
    cout << " Vergleichswerte rang " << rang << " ";
    for(i=0; i<vektorlaenge; i++)
        {
        cout << vergleichswerte[i] << " ";
        }
    cout << endl;
    }

// Kommentar: Das hier sind Ausgabe - Funktionen, die nur dem Debugging dienen.
void LastVektorAuswertung::ausgabe_transferdaten(void)
    {
    cout << " Transferdaten rang " << rang;
    for(i=0; i<vektorlaenge; i++)
        {
        cout << " West " << west_differenz_vektor_td[i] << " Ost " << ost_differenz_vektor_td[i];
        }
    cout << endl;
    }
