// Kommentar: Diese Funktion main wird aufgerufen beim Programmstart. Hier werden Parameter bestimmt aufgrund der Argumente.
// Und dann wird hier die Klasse Aufrufe definiert und aufgerufen, die ihrerseits die Klasse Programm aufruft, deren Funktion
// tun den ganzen parallelen adaptiven Mehrgitterzyklus durchfhrt.

// Kommentar: Zuerst werden die Standardheaderdateien aufgerufen. Und der Header fr mpi.
#include <iostream>

#include <mpi.h>

#include <stdlib.h>

// Danach werden Header - Dateien des Programms aufgerufen. Zentral dabei ist der Header 'aufrufe.h', damit die Klasse Aufrufe
// definiert werden kann und darauf zugegriffen werden kann.

#include "gitter.h"

#include "programm.h"

#include "liste.h"

#include "auf_ab.h"

#include "lastausgleichssteuerung.h"

#include "aufrufe.h"

#include "feldliste_sortiert.h"

using namespace std;

#define NORMAL

#ifdef NORMAL

// Kommentar: Main wird aufgerufen, und es werden die Argumente bergeben.
int main(int argc, char ** argv)
{

    int systemreturn;
    double wert_pause;

// Kommentar: Hier werden die Ausgabedateien vorher gelscht. Je nach Betriebssystem lautet der Befehl dafr verschieden.
#ifndef LINUX
    systemreturn = system("del da_*");
#else
    systemreturn = system("rm da_*");
#endif
    systemreturn--;

// Kommentar: Hier wird die Liste angelegt, mit der der Grobgitterzyklus abluft - um verschiedene Zyklen zu ermglichen.
    Liste liste_auf, liste_ab, liste_oben;

// Kommentar: MPI wird gestartet, wobei die Argumente fr MPI bergeben werden mssen. MPI bezieht daraus Informationen,
// wie die Anzahl an Kernen, die verwendet wird. Die von MPI verwendeten Informationen werden anschlieend aus der
// Argumentenliste gelscht.
    MPI_Init(& argc, & argv);

// Kommentar: Die Parameter groesse - Anzahl der verwendeten Kerne - und rang - die Identitt des eigenen Kerns - werden definiert
// und mittels MPI - Routinen bestimmt.
    int groesse, rang;

    MPI_Comm_rank(MPI_COMM_WORLD, & rang);
    MPI_Comm_size(MPI_COMM_WORLD, & groesse);

// Kommentar: Durch MPI_Barrier werden die Kerne synchronisiert.
    MPI_Barrier(MPI_COMM_WORLD);

    Programm programm;

    liste_auf.initialisieren(20);
    liste_ab.initialisieren(20);
    liste_oben.initialisieren(20);

    liste_auf.einfuegen(1, EXPLIZIT);
    liste_oben.einfuegen(1, EXPLIZIT);

// Kommentar: Die Klasse auf_ab steuert den Grobgitterzyklus. Es wird die Reihenfolge angegeben, in der die fein - grob - und
// grob - fein - Schritte ausgefhrt werden.
    AufAb auf_ab;

    auf_ab.initialisieren(20);

    auf_ab.a[0] = AUF;
    auf_ab.a[1] = AUF;
    auf_ab.a[2] = AB;
    auf_ab.a[3] = AB;
    auf_ab.a[4] = AB;
// Kommentar: auf_ab.laenge gibt an, bis zu welcher Stelle von auf_ab (k < auf_ab.laenge) der auf_ab - Zyklus ausgefhrt wird.
// Bemerkung: Die AB - Schritte sind immer ein Schritt weniger, weil nach dem Lsen auf dem grbsten Gitter direkt ein AB
// Schritt erfolgt.
    auf_ab.laenge = 3;

// Die Klasse Aufrufe wird definiert, und es werden eine ganze Reihe von Werten von aufrufe gesetzt.
    Aufrufe aufrufe;

    aufrufe.plb_kn_von_main = KN;

    int i;

    aufrufe.wieviel_punktelast = 10;

    aufrufe.uebergebe_vgitter = 0; // default

    aufrufe.lastwerte_korrektur = false;

    #ifdef DAFWV
    strcpy(aufrufe.auswertung_dateiname, "auswertung");
    #endif

    aufrufe.h_kappe = -1;
    aufrufe.echtes_plb = false;

    aufrufe.ist_zusammen_aV_LA = false;
    aufrufe.oben_nicht_adaptiv = true;

    double verf_krit;

    aufrufe.verf_krit = 0.02;
    aufrufe.verf_krit = 0.022;
    aufrufe.verf_krit = 1.0;
    aufrufe.verf_krit = 2.0;
    aufrufe.verf_krit_1 = 1.0;
    aufrufe.verf_krit_2 = 1.0;
    aufrufe.t_konst = 0.25;
    aufrufe.t_konst = 1.0;

    aufrufe.balancen = true;
    aufrufe.gL = false; // Wichtig: gL initialisieren. Ansonsten kann passieren: Wenn gL bei einigen Kernen true, und bei anderen false ist,
                        // werden verschiedene Lastausgleiche aufgerufen, und das Programm blockiert.

// Kommentar: Mit dieser Funktion werden weitere Werte fr die Klasse Aufrufe gesetzt - vor der Kommandozeilenparameterbearbeitung.
// Damit man default - Werte hat, die durch die Argumentenbearbeitung abgewandelt werden knnen.
    aufrufe.anfangswerte();

    wert_pause = 3.0;

    cout << " Version 1.03 Veroeffentlichung " << endl;

// Kommentar: Die Argumente werden durchlaufen, um damit Parameter setzen zu knnen.
    for(i=0; i<argc; i++)
        {
// Kommentar: 'm' gibt die Anzahl an Gitterpunkten des feinsten Gitters in x - Richtung an.
        if(strcmp(argv[i], "m")==0)
            {
            aufrufe.m = atoi(argv[i+1]);
            }
// Kommentar: 'n' gibt die Anzahl an Gitterpunkten des feinsten Gitters in x - Richtung an.
        if(strcmp(argv[i], "n")==0)
            {
            aufrufe.n = atoi(argv[i+1]);
            }
// Kommentar: 'iterationen' gibt an, wieviele Iterationen das Programm ausfhren soll.
        if(strcmp(argv[i], "iterationen")==0)
            {
            aufrufe.iterationsgesamtzahl = atoi(argv[i+1]);
            }

// Kommentar: 'verfeinerungstyp' gibt an, ob simuliert werden soll, bzw. welche selbstadaptive Verfeinerung gewhlt werden soll.
        if(strcmp(argv[i], "verfeinerungstyp")==0)
            {
            aufrufe.verfeinerungstyp = atoi(argv[i+1]);
#ifdef SHALLOW_WATER_EQUATIONS
            if(aufrufe.verfeinerungstyp != 1)
                {
                cout << " Fehler: Bei den Shallow Water Equations muss der Verfeinerungstyp 1 sein " << endl;
                exit(-1);
                }
#endif
            switch(aufrufe.verfeinerungstyp)
                {
                case 1:
                case 3:
                case 4:
                    break;
                default:
                    cout << " Falscher Verfeinerungstyp rang " << rang << endl;
                    exit(-1);
                }
            }

// Kommentar: 'simulationsnummer' gibt an, welche Simulation unter den programmierten gewhlt werden soll.
        if(strcmp(argv[i], "simulationsnummer")==0)
            {
            aufrufe.simulationsnummer = atoi(argv[i+1]);
            }

// Kommentar: 'vgitter' beeinflut - wenn das bei der Simulation vorgesehen ist - die Gre des simulierten Verfeinerungsgitters.
        if(strcmp(argv[i], "vgitter")==0)
            {
            aufrufe.uebergebe_vgitter = atoi(argv[i+1]);
            }

// Kommentar: 'h_min' gibt die Anzahl an Verfeinerungsebenen an.
        if(strcmp(argv[i], "h_min")==0)
            {
            aufrufe.h_min = atoi(argv[i+1]);
            }

// Kommentar: 'h_max' gibt die Anzahl an Vergrberungsebenen an.
        if(strcmp(argv[i], "h_max")==0)
            {
            aufrufe.h_max = atoi(argv[i+1]);
            }

// Kommentar: 'h_kappe' gibt an, ob es eine Kappe geben soll (sonst ist h_kappe = -1), und wenn ja, auf welcher Gitterebene.
#ifndef SHALLOW_WATER_EQUATIONS
        if(strcmp(argv[i], "h_kappe")==0)
            {
            aufrufe.h_kappe = atoi(argv[i+1]);
            }
#endif  // SHALLOW_WATER_EQUATIONS Bemerkung: Fr die SWE ist die Kappe nicht definiert. Wird aufrufe.h_kappe != -1, dann werden die Ebenen > h_kappe nicht initialisiert - vgl. Programm.
        // Diese Fehlermglichkeit wird hier abgefangen.

// Kommentar: 'mu_1' gibt die Anzahl an Vorglttungen an.
        if(strcmp(argv[i], "mu_1")==0)
            {
            aufrufe.mue1 = atoi(argv[i+1]);
            }

// Kommentar: 'mu_2' gibt die Anzahl an Nachglttungen an.
        if(strcmp(argv[i], "mu_2")==0)
            {
            aufrufe.mue2 = atoi(argv[i+1]);
            }

// Kommentar: 'lwk_typ' gibt an, welche LWK - Variante gewhlt wird, wenn die LWK gemacht werden soll.
        if(strcmp(argv[i], "lwk_typ")==0)
            {
            aufrufe.lwk_typ = atoi(argv[i+1]);
            }
// Komemntar: 'mit' gibt an, dass ein reverser Lastausgleich gemacht werden soll.
        if(strcmp(argv[i], "mit")==0)
            {
            aufrufe.mit_ohne = MIT;
            }

// Kommentar: 'ohne' gibt an, dass kein reverser Lastausgleich gemacht werden soll.
        if(strcmp(argv[i], "ohne")==0)
            {
            aufrufe.mit_ohne = OHNE;
            }

// Kommentar: Im Falle der PLB - Variante gibt 'plb_kn' an, auf welche Weise die Lastinformationen der anderen Kerne beschafft
// werden soll.
        if(strcmp(argv[i], "plb_kn")==0)
            {
            aufrufe.plb_kn = atoi(argv[i+1]);
            }

// Kommentar: 'ev_typen' gibt an, welche Klasse / Klassen bei extremen Vektoren gefhrt werden soll / sollen.
        if(strcmp(argv[i], "ev_typen")==0)
            {
            aufrufe.ev_typen = atoi(argv[i+1]);
            }

// Kommentar: 't_konst' hat einen Einflu darauf, wie stark sich die Werte der Variablen f0 in der Zeit ndern.
        if(strcmp(argv[i], "t_konst")==0)
            {
            aufrufe.t_konst = atoi(argv[i+1]);
            }

// Komemntar: 'vektorisieren' gibt an, ob der Randaustausch durch Vektorisierung oder mit Hilfe der Sortierung erreicht werden soll.
        if(strcmp(argv[i], "vektorisieren")==0)
            {
            aufrufe.vektorisieren = true;//atoi(argv[i+1]);
            }

// Kommentar: Die folgenden 4 Variablen geben an, welche Visualisierungen in welchen Zeitabstnden vorgenommen werden sollen.
        if(strcmp(argv[i], "gp_ausgabe")==0)
            {
            aufrufe.gp_ausgabe = true;//atoi(argv[i+1]);
            }

        if(strcmp(argv[i], "gp_ausgabe_mod")==0)
            {
            aufrufe.gp_ausgabe_mod = atoi(argv[i+1]);
            }

        if(strcmp(argv[i], "gp_mit_io")==0)
            {
            aufrufe.gp_mit_io = true;// = atoi(argv[i+1]);
            }

        if(strcmp(argv[i], "gp_mit_aufteilung")==0)
            {
            aufrufe.gp_mit_aufteilung = true;// = atoi(argv[i+1]);
            }

// Kommentar: 'PL' gibt an, wieviel Punktlast in einen Gitterpunkt gelegt werden soll, um die geringe Last des Poisson - Operators
// anzuheben. PL = 0 schliet die Punktlast aus.
        if(strcmp(argv[i], "PL")==0)
            {
            aufrufe.wieviel_punktelast = atoi(argv[i+1]);
            }

// Kommentar: 'gL' gibt an, ob beim ersten Schritt ein Lastausgleich ber beliebig viele Lastrunden erfolgen soll.
        if(strcmp(argv[i], "gL")==0)
            {
            aufrufe.gL = true;
            }
// Kommentar: 'plb_kn_von_main' wird an plb_kn, vgl. oben, weitergeben.
        if(strcmp(argv[i], "PLB")==0)
            {
            aufrufe.plb_kn_von_main = PLB;
            }
// Kommentar: Wird 'LwK' in die Kommandozeile gesetzt, dann wird die Last Werte Korrektur durchgefhrt.
        if(strcmp(argv[i], "LwK")==0)
            {
            aufrufe.lastwerte_korrektur = true;
            }
// Kommentar: 'adapt' unterscheidet zwischen FAS und dem Mehrgitteralgorithmus im groben Bereich des Zyklus.
        if(strcmp(argv[i], "adapt")==0)
            {
            aufrufe.oben_nicht_adaptiv = false;// Macht FAS, 7.10.14
            }
// Kommentar: 'echtes-PLB' unterscheidet zwischen PLB und der PLB - Variante, vergleiche die Dissertation.
        if(strcmp(argv[i], "echtes-PLB")==0)
            {
            if(groesse>4)
                {
                cout << " Echtes PLB rang " << rang << endl;
                aufrufe.echtes_plb = true;
                }
            }
        if(strcmp(argv[i], "einfacher-LA")==0)
            {
            aufrufe.immer_oben_la = false;
            aufrufe.ev_typen = TYP_EV;
            aufrufe.mit_ohne = OHNE;
            }
        if(strcmp(argv[i], "visualisieren")==0)
            {
#ifndef SHALLOW_WATER_EQUATIONS
            aufrufe.gp_ausgabe = true;
            aufrufe.gp_ausgabe_mod = 0;
            aufrufe.gp_mit_io = true;
            aufrufe.gp_mit_aufteilung = true;
#endif
            }
        if(strcmp(argv[i], "verf-krit")==0)
            {
            verf_krit = atof(argv[i+1]);
            aufrufe.verf_krit = verf_krit;
            aufrufe.verf_krit_1 = verf_krit;
            aufrufe.verf_krit_2 = verf_krit;
            }
        if(strcmp(argv[i], "pause")==0)
            {
            wert_pause = atof(argv[i+1]);
            }
#ifdef DEBUG
// Kommentar: Im Debugging - Fall wird mit den Schlsselworten 'AuswertungDateiname', oder kurz 'AD', der Name angegeben,
// unter dem die Dateiausgaben gespeichert werden.
        if((strcmp(argv[i], "AuswertungDateiname")==0)||(strcmp(argv[i],"AD")==0))
            {
            #ifndef DAFWV
            cout << "Argument AuswertungDateiname inkorrekt, da DAFWV nicht definiert ist." << endl;
            exit(-1);
            #endif
            strcpy(aufrufe.auswertung_dateiname, argv[i+1]);
            }
#endif
// Kommentar: 'zusammen' ist ein Schlsselwort, welches angibt, ob nach der adaptiven Verfeinerung direkt der Lastausgleich gemacht
// werden soll, oder ob dann noch Mehrgitterschritte bei unausgeglichener Last gemacht werden sollen.
        if(strcmp(argv[i], "zusammen")==0)
            {
            aufrufe.ist_zusammen_aV_LA = true;
            }
        aufrufe.ist_zusammen_aV_LA = true;
// Kommentar: 'baloff' ist ein Schlsselwort, mit dem der Lastausgleich komplett abgeschaltet werden kann.
        if(strcmp(argv[i], "baloff")==0)
            {
            aufrufe.balancen = false;
            }
        }

    if(groesse==1)
        {
        aufrufe.balancen = false;
        }

    #define EINES

#ifdef SHALLOW_WATER_EQUATIONS
    cout << " Flachwassergleichungen" << endl;
#else
    cout << " Poisson Gleichung " << endl;
#endif
#ifndef SORTIERTE_FELDLISTE
    cout << " Nicht sortierte Feldliste rang " << rang << endl;
#else
    cout << " Sortierte Feldliste rang " << rang << endl;
#endif
#ifdef TEST_FC
    cout << " Farbenfeld wird geprft. rang " << rang << endl;
#endif
    #ifdef EINES

    aufrufe.wert_pause = wert_pause;

// Kommentar: aufrufe.tun bewirkt ein Starten des Programms, wobei ggf. weitere Parameter vorher gesetzt werden.
    aufrufe.tun(groesse, rang);
    #else
    aufrufe.mehrfaches_tun(groesse, rang);
    #endif

    int variable;
    variable = 10;
    variable = variable << 1;
    int v2;
    v2 = 7;
    v2 = v2 >> 1;

// Gibt Speicherplatz der Klasse aufrufe frei.
    aufrufe.finalisieren();

// Finialisiert die oben definierten Listen.
    liste_auf.finalisieren();
    liste_ab.finalisieren();
    liste_oben.finalisieren();
    auf_ab.finalisieren();

// Kommentar: Gegen Ende des Algorithmus kommt es zu weiteren Ausgaben.
    cout << "True " << true << " False " << false << " rang " << rang << endl;
    MPI_Finalize();

    return 0;
}

#endif
