#include "multi_level_randuebertragung.h"

// Kommentar: Diese Klasse dient dem Verschicken von Randinformation für alle Gitterebenen in einem Buffer - um die Kommunikation
// zu beschleunigen.

#ifdef GEMEINSAME_BUFFER_SI
// Kommentar: Hier geschieht die Initialisierung dieser Klasse - mit den wichtigsten Parametern - und ggf. den short int
// Sendebuffern und Empfangsbuffern - bei gemeinsamer Verwendung dieser Buffer.
void MultiLevelRanduebertragung::initialisierung(int _n, int _m, int _groesse, int _rang, short int ** _sendebuffer, short int ** _empfangsbuffer)
#else
void MultiLevelRanduebertragung::initialisierung(int _n, int _m, int _groesse, int _rang)
#endif
    {
// Kommentar: Die Parameter werden kopiert.
    groesse = _groesse;
    rang = _rang;
    n = _n;
    m = _m;
// Kommentar: Die Sende - und Empfangsbuffer werden angelegt - entweder durch Zeigerübergabe oder Bufferallokation. Bei der
// Bufferallokation kann man noch unterscheiden zwischen int und short int Buffern
#ifdef KURZ_RK
#ifdef GEMEINSAME_BUFFER_SI
    sendebuffer = _sendebuffer;
    empfangsbuffer = _empfangsbuffer;
#else
    sendebuffer = new short int * [groesse];
    empfangsbuffer = new short int * [groesse];
#endif
#else
    sendebuffer = new int * [groesse];
    empfangsbuffer = new int * [groesse];
#endif
// Kommentar: Die Buffergröße wird für diese Randkommunikation recht großzügig bemessen.
    buffergroesse = m*n + 100;

    for(i=0; i<groesse; i++)
        {
        if(i!=rang)
            {
#ifdef KURZ_RK
#ifndef GEMEINSAME_BUFFER_SI
            sendebuffer[i] = new short int [buffergroesse];
            empfangsbuffer[i] = new short int [buffergroesse];
#endif
#else
            sendebuffer[i] = new int [buffergroesse];
            empfangsbuffer[i] = new int [buffergroesse];
#endif
            }
        }
// Kommentar: Nach den Buffern werden die Requests für die Kommunikation - das Senden und das Empfangen - angelegt.
    sende_request = new MPI_Request[groesse];
    empfangs_request = new MPI_Request[groesse];
    }

// Kommentar: Hier kommt es zur Finalisierung dieser Klasse.
void MultiLevelRanduebertragung::finalisierung(void)
    {
// Kommentar: Zuerst werden die Requests freigegeben.
    delete sende_request;
    delete empfangs_request;

    cout << "MLR Finalisierung Groesse " << groesse << " Rang " << rang << endl;

// Kommentar: Und dann wird, falls die Buffer nicht übergeben, sondern in dieser Klasse allokiert wurden, der Speicherplatz
// wieder freigegeben.

    for(i=0; i<groesse; i++)
        {
        if(i!=rang)
            {
#ifdef KURZ_RK
#ifndef GEMEINSAME_BUFFER_SI
            delete sendebuffer[i];
            delete empfangsbuffer[i];
#endif
#else
            delete sendebuffer[i];
            delete empfangsbuffer[i];
#endif
            }
        }


#ifdef KURZ_RK
#ifdef GEMEINSAME_BUFFER_SI
#else
    delete sendebuffer;
    delete empfangsbuffer;
#endif
#else
    delete sendebuffer;
    delete empfangsbuffer;
#endif
    }

// Kommentar: Ein Multi - Level - Randbuffer wird verschickt.
void MultiLevelRanduebertragung::senden(int i)
    {
#ifdef KURZ_RK
// Kommentar: Hiermit wird der ML - Buffer abgeschickt. Die Länge wird angegeben durch die Multi - Level - Indizes, die aufgrund
// der Anzahl an Vektoren pro Gitterebene bestimmt werden kann.
    MPI_Isend(sendebuffer[i], multi_level_indizes[i], MPI_SHORT, i, 150, MPI_COMM_WORLD, &(sende_request[i]));
#else
    MPI_Isend(sendebuffer[i], multi_level_indizes[i], MPI_INT, i, 150, MPI_COMM_WORLD, &(sende_request[i]));
#endif
#ifdef AUSGABE_MLR
    cout << "Senden // rang " << rang << " // i " << i << " mli[i] " << multi_level_indizes[i] << endl;
#endif
    }

// Kommentar: Hier wird der Multi - Level - Randbuffer empfangen. Die buffergroesse charakterisiert die maximal mögliche
// Länge, die die angekommenen Buffer annehmen dürfen.
void MultiLevelRanduebertragung::empfangen(int i)
    {
#ifdef KURZ_RK
    MPI_Irecv(empfangsbuffer[i], buffergroesse, MPI_SHORT, i, 150, MPI_COMM_WORLD, &(empfangs_request[i]));
#else
    MPI_Irecv(empfangsbuffer[i], buffergroesse, MPI_INT, i, 150, MPI_COMM_WORLD, &(empfangs_request[i]));
#endif
#ifdef AUSGABE_MLR
    cout << "Empfangen // rang " << rang << " // i " << i << " // buffergroesse " << buffergroesse << endl;
#endif
    }

// Kommentar: Hier wird getestet, ob das Senden erfolgreich abgeschlossen worden ist.
bool MultiLevelRanduebertragung::test_senden(int i)
    {
    MPI_Test(&(sende_request[i]), & flag, & sende_status);
    if(flag == true)
        {
        return(true);
        }
    return(false);
    }

// Kommentar: Hier wird getestet, ob das Empfangen erfolgreich abgeschlossen worden ist.
bool MultiLevelRanduebertragung::test_empfangen(int i)
    {
    MPI_Test(&(empfangs_request[i]), & flag, & empfangs_status);
    if(flag == true)
        {
        return(true);
        }
    return(false);
    }
