#include "mv_projektionen.h"

// Kommentar: Hier werden die Vektoren, die fr eine Gitterebene gepackt wurden, auf die anderen Gitterebenen ausgeweitet.
// Vorher gibt es die Vektoren nur auf einer Ebene, nachher sind die Punkte darber und darunter fr alle Gitterebenen
// bestimmt.

// Kommentar: Diese Funktion bewirkt die Durchfhrung der Projektionen der Vektoren einer Ebene auf die anderen Ebenen.
// Dabei ermglicht die Struktur mv die Multilevel - Speicherung, wobei die bestimmten Vektoren in mv in der Ebene
// h_LA gespeichert sind.

// Kommentar: Hier werden vertikale Vektoren projeziert, was beim Ost - West - Lastausgleich der Fall ist.
void MVProjektionen::durchfuehren(MV_vertikal * mv, int h_LA)
    {
#ifdef AUSGABEN_MV_PROJEKTIONEN
    cout << "!2010" << endl;
#endif
// Kommentar: Hier geschehen die Projektionen der Vektoren auf der Ebene h_LA hin zu den feineren Gittern. Das ist der schwierigere
// Teil, weil die neuen Vektoren unterbrochen sein knnen.
    fein_projektionen(mv, h_LA);
#ifdef AUSGABEN_MV_PROJEKTIONEN
    cout << "!2011" << endl;
#endif
// Kommentar: Hier geschehen die Projektionen der Vektoren auf der Ebene h_LA hin zu den grberen Gittern. Das ist der einfachere
// Teil, weil die neuen Vektoren nicht unterbrochen sein knnen.
    grob_projektionen(mv, h_LA);
#ifdef AUSGABEN_MV_PROJEKTIONEN
    cout << "!2012" << endl;
#endif
    }

// Kommentar: Hier wird die Klasse initialisiert. Man hat die Parameter h_gesamt, um bei den grob_projektionen zu wissen, bis zu
// welchem Gitter diese erfolgen mssen. Dasselbe gilt fr h_kappe, weil fr die grberen Ebenen in Bezug auf die Ebene h_kappe
// kein Lastausgleich erfolgen braucht bzw. erfolgen sollte. Der Parameter rang dient Ausgaben, und damit Debugging - Zwecken.
void MVProjektionen::initialisieren(int _h_gesamt, int _rang, int _h_kappe)
// wird finalisiert
    {
// Kommentar: Die Parameter werden bergeben. Dabei werden n und m fr jede Gitterebene gespeichert. Auerdem werden Zeiger fr
// die Farbenfelder angelegt, damit diese - fr die Feingitterprojektionen - zugreifbar sind.
    h_gesamt = _h_gesamt;
    rang = _rang;
#ifdef KURZ_FF
    farbenfelder = new short int ** [h_gesamt];
#else
    farbenfelder = new int ** [h_gesamt];
#endif

    n = new int [h_gesamt];
    m = new int [h_gesamt];

    h_kappe = _h_kappe;
    h_gesamt_minus1 = h_gesamt - 1;
    }

// Kommentar: Der allokierte Speicher wird freigegeben.
void MVProjektionen::finalisieren(void)
    {
// Die Zeiger farbenfelder, sowie die Felder zum Speichern der Gittermae aller Gitterebenen werden freigegeben.
    delete farbenfelder;
    delete n;
    delete m;
    }

// Kommentar: Mit dieser Funktion werden die Zeiger auf die Farbenfelder und m und n fr die jeweilige Gitterebene h geschrieben.
// Wird das fr alle Ebenen gemacht, hat diese Klasse alle zur Projektionsbearbeitung ntigen Informationen.
#ifdef KURZ_FF
void MVProjektionen::schreibe_farbenfeld(int _h, short int ** _farbenfeld, int _n, int _m)
#else
void MVProjektionen::schreibe_farbenfeld(int _h, int ** _farbenfeld, int _n, int _m)
#endif
    {
    farbenfelder[_h] = _farbenfeld;
    n [_h] = _n;
    m [_h] = _m;
    }

// Kommentar: Hier werden die Projektionen hin zu den feineren Gittern durchgefhrt.
void MVProjektionen::fein_projektionen(MV_vertikal * mv, int h_LA)
    {
#ifdef AUSGABEN_MV_PROJEKTIONEN
    cout << "Fein Projektionen " << endl;
#endif

// Kommentar: Die Idee: Die Projektionen werden Gitterebene fr Gitterebene auf die immer feineren Gitter gebracht.
    for(h=h_LA; h>0; h--)
        {
#ifdef AUSGABEN_MV_PROJEKTIONEN
        cout << " h " << h << endl;
#endif

// Kommentar: Es erfolgt eine Projektion der Vektoren der Ebene h auf die Ebene h-1.
        fein_projektion(mv, h);
        }
    }

// Kommentar: Es wird die Feingitterprojektion von der Ebene h auf die Ebene h-1 durchgefhrt. Die Vektoren werden
// gelesen und geschrieben von und nach mv.
void MVProjektionen::fein_projektion(MV_vertikal * mv, int h)
    {
#ifdef AUSGABEN_MV_PROJEKTIONEN
    cout << "!20000 rang " << rang << endl;
#endif

// Kommentar: Es wird die Vektoranzahl der zu lesenden Vektoren bestimmt.
    vektoranzahl = mv->indizes[h];

#ifdef AUSGABEN_MV_PROJEKTIONEN
    cout << "!20001" << " rang " << rang << endl;
#endif

// Kommentar: Es werden die Zeiger auf die Vektoren der Ebene h gespeichert. Frs Lesen dieser Vektoren.
    f1_oben = mv->f1_speichern[h];
    f2_oben = mv->f2_speichern[h];
    g_oben = mv->g_speichern[h];

#ifdef AUSGABEN_MV_PROJEKTIONEN
    cout << "!20002" << " rang " << rang << endl;
#endif

// Kommentar: Dasselbe wird fr die feinere Gitterebene gemacht. Frs Schreiben der bestimmten Vektoren in diese Vektorliste.
    h_minus_1 = h-1;

    f1_unten = mv->f1_speichern[h_minus_1];
    f2_unten = mv->f2_speichern[h_minus_1];
    g_unten = mv->g_speichern[h_minus_1];

#ifdef AUSGABEN_MV_PROJEKTIONEN
    cout << "!20003" << " rang " << rang << endl;
#endif

// Kommentar: Es werden bentigte Parameter bestimmt.
    m_minus_1 = m[h] - 1;
    m_h_minus1 = m[h_minus_1];
    m_h_minus1_minus1 = m_h_minus1 - 1;

#ifdef AUSGABEN_MV_PROJEKTIONEN
    cout << "!20004" << " rang " << rang << endl;
#endif

// Kommentar: Es wird die Zahl geschriebener Vektoren auf 0 gesetzt.
    schreibindex = 0;

// Kommentar: Fr die Bestimmung der Unterbrechungsstellen wird das Farbenfeld der feinen Gitters auf den Zeiger farbenfeld
// verknpft.
    farbenfeld = farbenfelder[h_minus_1];
#ifdef AUSGABEN_MV_PROJEKTIONEN
    cout << "!20005 vektoranzahl " << vektoranzahl << " rang " << rang << endl;
#endif

// Kommentar: Die zu  lesenden Vektoren werden durchlaufen.
    for(i=0; i<vektoranzahl; i++)
        {
// Die Vektorkoordinaten werden gelesen.
        vvektor_f1 = f1_oben[i];
        vvektor_f2 = f2_oben[i];
        vvektor_g = g_oben[i];

// Kommentar: Die Anfangs - und Endkoordinaten des Feingittervektors werden bestimmt. Erst wird die Grobgitterkoordinate verdoppelt,
// und dann, wenn mglich, um 1 verlngert, vgl. die Dissertation - nach oben und unten.
        vvektor_f1_hier=vvektor_f1 << 1;

        vvektor_f2_hier=vvektor_f2 << 1;
        if(vvektor_f2_hier < m_h_minus1_minus1)
            {
            vvektor_f2_hier ++;
            }
// Kommentar: Der g - Koordinatenwert fr das feinere Gitter wird bestimmt.
        vvektor_g_hier =vvektor_g  << 1;

// Kommentar: Der zugehrige Vektor fr die g - Koordinate wird bestimmt. Dabei werden Unterbrechungen bei der Funktion
// vektor_schreiben bercksichtigt.
        vektor_schreiben(vvektor_f1_hier, vvektor_f2_hier, vvektor_g_hier);

// Kommentar: Falls mglich, so wird der Vektor darunter, also mit um 1 erhhter g - Koordinate -  auch gebildet.
        if(vvektor_g<m_minus_1)
                {
                vektor_schreiben(vvektor_f1_hier, vvektor_f2_hier, vvektor_g_hier+1);
                }
        }
// Kommentar: Am Ende wird die Zahl geschriebener Vektoren im Index der feineren Ebene gespeichert.
    mv->indizes[h_minus_1] = schreibindex;
    }

// Komemntar: Hier wird ein Vektor in die Vektorliste der feineren Gitterebene eingefgt.
void MVProjektionen::vektor_einfuegen(int f1, int f2, int g)
    {
#ifdef AUSGABEN_MV_PROJEKTIONEN
    cout << "Vektor einfuegen f1 " << f1 << " f2 " << f2 << " g " << g << " rang " << rang << endl;
#endif
// Kommentar: Der Vektor wird kopiert.
    f1_unten[schreibindex] = f1;
    f2_unten[schreibindex] = f2;
    g_unten[schreibindex] = g;
// Kommentar: Und der Index wird um 1 erhht.
    schreibindex++;
    }

// Kommentar: Hier werden die auf diesem Kern liegenden Vektoren im angegebenen Vektorbereich gebildet und beschrieben.
// Die Idee dabei ist die, die Unterbrechungen aufgrund fremder Gitterpunkte zu handhaben.
void MVProjektionen::vektor_schreiben(int vvektor_f1_hier, int vvektor_f2_hier, int vvektor_g_hier)
// Wie genau funktioniert das? Es geht darum, Vektoren zu bilden, solange es gleich dem Rang ist.
    {
// Kommentar: Der zustand wird so gesetzt, dass die Gitterpunkte existieren.
    if(farbenfeld[vvektor_f1_hier][vvektor_g_hier]==rang)
        {
        f1_merken = vvektor_f1_hier;
        zustand = MITTEN;
        }
    else
        {
        zustand = ANFANG;
        }
#ifdef AUSGABEN_MV_PROJEKTIONEN
    cout << "Vor der Schleife " << " rang " << rang << endl;
#endif

// Kommentar: Der Vektor wird mit seinen Koordinaten durchlaufen.
    for(k=vvektor_f1_hier+1;k<=vvektor_f2_hier;k++)
        {
#ifdef AUSGABEN_MV_PROJEKTIONEN
        cout << " k " << k << " g " << vvektor_g_hier << " ff[k][g] " << farbenfeld[k][vvektor_g_hier] << " rang " << rang << endl;
#endif

// Kommentar: Bei dem Zustand, dass die Punkte nicht existieren, wird durchlaufen, bis sie wiederexistieren.
// Bemerkung: Man kann das Farbenfeld zur Entscheidung, ob ein Punkt existiert, nehmen, weil die Feingitterpunkte stehts
// zu den Grobgitterpunkten, fr die die Punkte auf dem eigenen Kern liegen, einen Feingitterabstand von hchstens 1 haben.
// Das Feingitterfarbenfeld ist im verwendeten Bereich deswegen korrekt.
        if(zustand == ANFANG)
            {
            if(farbenfeld[k][vvektor_g_hier]!=rang)
                {
                continue;
                }
// Kommentar: Der folgende Vektor beginnt mit f1_merken = k. Der Zustand geht nach MITTEN, weil auf diesem Kern exitierende Punkte
// gescannt werden.
            f1_merken = k;
            zustand = MITTEN;
            continue;
            }
// Kommentar: Man ist im Zustand 'MITTEN', was bedeutet, das ein Vektor hier auf dem eigenen Kern liegt, der soweit verlngert
// wird, wie es mglich ist.
        if(farbenfeld[k][vvektor_g_hier]==rang)
            {
            continue;
            }
// Kommentar: Man hat nun die mgliche Lnge des zu speichernden Vektors um1 nerschritten. Demnach erfolge ein Abspeichern der
// Koordinaten des bestimmten Vektors, und springen in den Zustand, bei dem nach nicht - eigenen
// Punkten gescannt wird. Das ist der Zustand 'ANFANG'.
        vektor_einfuegen(f1_merken, k-1, vvektor_g_hier);
        zustand = ANFANG;
        continue;
        }
#ifdef AUSGABEN_MV_PROJEKTIONEN
    cout << "Schleife zu Ende " << " rang " << rang << endl;
#endif
// Kommentar: Wenn man am Ende des Vektors nach eigenen Punkten scannt, wird das hier mit einem Abspeichern des Vektors beendet.
// Andernfalls, bei zustand = ANFANG, wurde nach nicht Punkten in Bezug auf den eigenen Kern gescannt, und es ist deswegen nichts
// zu speichern.
    if(zustand == MITTEN)
        {
        vektor_einfuegen(f1_merken, vvektor_f2_hier, vvektor_g_hier);
        }
    }

// Kommentar: Hier wird eine Grobgitterprojektion der Ebenen h_LA bis zur maximal sinnvollen Ebene.
void MVProjektionen::grob_projektionen(MV_vertikal * mv, int h_LA)
    {
#ifdef AUSGABEN_MV_PROJEKTIONEN
    cout << "Grob Projektionen " << endl;
    cout << "h_LA " << h_LA << " h_gesamt " << h_gesamt << endl;
#endif
// Kommentar: Man unterscheidet, ob es eine Kappe gibt oder nicht.
    if(h_kappe==-1)
        {
// Kommentar: Keine Kappe => Fr alle Ebenen h_LA bis h_gesamt-2 werden fr eine Ebene grber die Vektoren projeziert.
        for(h=h_LA; h<h_gesamt_minus1; h++)
            {
#ifdef AUSGABEN_MV_PROJEKTIONEN
            cout << "h " << h << endl;
#endif
            grob_projektion(mv, h);
            }
#ifdef AUSGABEN_MV_PROJEKTIONEN
        cout << endl;
#endif
        }
    else
        {
// Kommentar: Liegt eine Kappe vor, werden die Koordinaten noch auf die Kappe aufgebracht, aber nicht mehr hher.
        for(h=h_LA; h<h_kappe; h++)
        // '< h_kappe', weil nur die Information bis max. der Ebene h_kappe gebraucht wird.
            {
#ifdef AUSGABEN_MV_PROJEKTIONEN
            cout << "h " << h << endl;
#endif
            grob_projektion(mv, h);
            }
#ifdef AUSGABEN_MV_PROJEKTIONEN
        cout << endl;
#endif
        }
    }

// Kommentar: Es erfolgen die Feingitterprojektionen bis zur feinsten Ebene, aber die Grobgitterprojektionen nur fr eine Ebene
// hher. Das wird in der Klasse Lastausgleich gebraucht, um Gitterpunkt dazupacken zu knnen, wo keine Feingitterpunkte darunter liegen.
void MVProjektionen::projektionen_bis_eine_ebene_hoeher(MV_vertikal * mv, int h_LA)
    {
    fein_projektionen(mv, h_LA);
    grob_projektion(mv, h_LA);
    }

// Kommentar: Hier wird eine Grobgitterprojektion von der Ebene h auf die Ebene h + 1 gemacht. Dabei wird in mv fr die Ebene
// h gelesen, und fr die Ebene h+1 geschrieben.
void MVProjektionen::grob_projektion(MV_vertikal * mv, int h)
    {
// Kommentar: Man speichert die Vektorenliste der Ebene h zum Lesen in einer gut zugreifbaren Weise ab.
    vektoranzahl = mv->indizes[h];

    f1_unten = mv->f1_speichern[h];
    f2_unten = mv->f2_speichern[h];
    g_unten = mv->g_speichern[h];

// Kommentar: Dann macht man dasselbe fr die Ebene h+1, aber zum beschreiben.
    h_plus_1 = h+1;

    f1_oben = mv->f1_speichern[h_plus_1];
    f2_oben = mv->f2_speichern[h_plus_1];
    g_oben = mv->g_speichern[h_plus_1];

// Kommentar: Dann werden die Vektoren durchlaufen. Es werden hchstens soviele Vektoren gespeichert wie gelesen, weil
// die Vektoren maximal 1 zu 1 gebildet werden: Ein Vektor der Ebene h wird zu hchstens einem Vektor der Ebene h+1.
    for(i=0, k=0; i<vektoranzahl; i++)
        {
        vvektor_f1 = f1_unten[i];
        vvektor_f2 = f2_unten[i];
        vvektor_g = g_unten[i];
// Kommentar: Ist h ungerade, wird darber nichts gespeichert.
        if((vvektor_g&1)==1)
            {
            continue;
            }
// Kommentar: Ansonsten berechne die Grobgitterkoordinaten des Vektors.
        if((vvektor_f1&1)==1)
            {
            vvektor_f1 ++;
            }
        if((vvektor_f2&1)==1)
            {
            vvektor_f2--;
            }
// Kommentar: Falls er leer ist, dann nichts speichern.
        if(vvektor_f1>vvektor_f2)
            {
            continue;
            }
// Kommentar: Ansonsten wird der Vektor gespeichert.
        f1_oben[k] = vvektor_f1 >> 1;

        f2_oben[k] = vvektor_f2 >> 1;

        g_oben[k] = vvektor_g >> 1;


        k++;
        }
// Kommentar: Am Ende wird die Zahl der geschriebenen Vektoren in mv gespeichert.
    mv->indizes[h+1] = k;
    }

// Kommentar: Die folgende Funktion gibt die in MVProjektionen gespeicherten Vektoren zu Debugging - Zwecken aus.
void MVProjektionen::ausgeben(char * text, MV_vertikal * mv)
    {
    for(h=0; h<h_gesamt;h++)
        {
        f1 = mv->f1_speichern[h];
        f2 = mv->f2_speichern[h];
        g = mv->g_speichern[h];
        vektoranzahl = mv->indizes[h];
        cout << " MVP " << text << " Ausgeben MVProjektionen Ebene " << h << " vektoranzahl " << vektoranzahl << " Rang " << rang;
        for(i=0; i<vektoranzahl; i++)
            {
// Die Vektorkoordinaten werden gelesen.
            vvektor_f1 = f1[i];
            vvektor_f2 = f2[i];
            vvektor_g = g[i];
            if((h == 1)&&(vvektor_f1<=7)&&(vvektor_f2>=7)&&(vvektor_g==10))
                {
                cout << " Code X ";
                }
            cout << " f1 " << vvektor_f1 << " f2 " << vvektor_f2 << " g " << vvektor_g;
            }
        cout << endl;
        }
    }
