#include "precomputation.h"

// Kommentar: Hier wird die Precomputation so ausgefhrt, wie sie beim PLB entsprechend der Arbeit von Herrn Bhm erfolgt.
// Einzige Unterschiede zu seiner Methode: Man kann einen reversen Lastausgleich haben, oder nicht. Und man hat einen
// Ost - West - Lastausgleich, und einen Nord - Sd - Lastausgleich. Ohne Einschrnkung wird hier nur der Ost - West
// Lastausgleich kommentiert, da der Nord - Sd - Lastausgleich analog erfolgt.

// Kommentar: Die lambda - Variablen geben an, wieviel Last verschickt werden muss, um einen vollstndigen Lastausgleich zu erhalten.
// Wobei das mehrere Lastausgleichsrunden einschliet.
Precomputation::Precomputation(void)
    {
    lambda_ost = 0;
    lambda_ost_oben = 0;
    lambda_west = 0;
    lambda_west_oben = 0;
    lambda_sued = 0;
    lambda_sued_oben = 0;
    lambda_nord = 0;
    lambda_nord_oben = 0;
    }

// Kommentar: Mit dieser Funktion wird die eigene Last - auf zwei Ebenen - gespeichert.
void Precomputation::tun_mit(int _eigene_last, int _eigene_last_oben)
// Wird nicht in einem ausgefhrt. Nur OW und NS getrennt ausfhren.
    {
    eigene_last = _eigene_last;
    eigene_last_oben = _eigene_last_oben;
    lambda = eigene_last;
    lambda_oben = eigene_last_oben;
    }

// Kommentar: Mit dieser Funktion wird die eigene Last - auf einer Ebene - gespeichert.
void Precomputation::tun_ohne(int _eigene_last)
// Wird nicht in einem ausgefhrt. Nur OW und NS getrennt ausfhren.
    {
    eigene_last = _eigene_last;
    lambda = eigene_last;
    }

// Kommentar: Hiermit wird die Precomputation - Funktion fr den Ost - West - Lastausgleich durchgefhrt. Und zwar fr zwei
// Gitterebenen. Dabei ist im Fall groesse_g = 2 getrennt zu bearbeiten, weil in dem Fall anders berechnet wird.
void Precomputation::tun_ow_mit(void)
    {
    if(groesse_g==2)
        {
        kommunikation_ow_mit_2();
        berechne_lambdas_ow_mit_2();
        return;
        }
#ifdef AUSGABE_PLB
    * da.aus << " Vor Kommunikation ow position_ow " << position_ow << endl;
#endif
    kommunikation_ow_mit();
#ifdef AUSGABE_PLB
    * da.aus << " Nach Kommunikation ow gemeinsame_last " << gemeinsame_last << endl;
#endif
    berechne_lambdas_ow_mit();
#ifdef AUSGABE_PLB
    * da.aus << " Nach berechne_lambdas_ow_mit " << endl;
#endif
    }

// Kommentar: Hiermit wird die Precomputation - Funktion fr den Ost - West - Lastausgleich durchgefhrt. Und zwar fr eine
// Gitterebene. Dabei ist im Fall groesse_g = 2 getrennt zu bearbeiten, weil in dem Fall anders berechnet wird.
void Precomputation::tun_ow_ohne(void)
    {
    if(groesse_g==2)
        {
        kommunikation_ow_ohne_2();
        berechne_lambdas_ow_ohne_2();
        return;
        }
#ifdef AUSGABE_PLB
    * da.aus << " Vor Kommunikation ow position_ow " << position_ow << endl;
#endif
    kommunikation_ow_ohne();
#ifdef AUSGABE_PLB
    * da.aus << " Nach Kommunikation ow gemeinsame_last " << gemeinsame_last << endl;
#endif
    berechne_lambdas_ow_ohne();
#ifdef AUSGABE_PLB
    * da.aus << " Nach berechne_lambdas_ow_ohne " << endl;
#endif
    }

// Kommentar: Hier wird eine Lastinformation von Osten her empfangen. Fr beide Gitterebenen. Und dabei wird diese Kommunikation
// abgewartet.
void Precomputation::empfange_osten_mit(void)
    {
// Kommentar: Zuerst wird die Empfangs - Kommunikation zum stlichen Kern initiiert.
    MPI_Irecv(empfangsbuffer_osten, 2, MPI_INT, ID_osten, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &empfange_request_osten);
// Kommentar: Dann wird gewartet, bis die Kommunikation eintrifft.
    for(flag=false;flag==false;)
        {
        MPI_Test(&empfange_request_osten, &flag, &status);
        }
// Kommentar: Dann werden die Teilsummen von Osten fr beide Gitterebenen abgespeichert.
    teilsumme_osten = empfangsbuffer_osten[0];
    teilsumme_osten_oben = empfangsbuffer_osten[1];
    }

// Kommentar: Hier wird eine Lastinformation von Osten her empfangen. Fr beide Gitterebenen. Das erfolgt analog zu empfange_osten_mit.
void Precomputation::empfange_westen_mit(void)
    {
    MPI_Irecv(empfangsbuffer_westen, 2, MPI_INT, ID_westen, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &empfange_request_westen);
    for(flag=false;flag==false;)
        {
        MPI_Test(&empfange_request_westen, &flag, &status);
        }
    teilsumme_westen = empfangsbuffer_westen[0];
    teilsumme_westen_oben = empfangsbuffer_westen[1];
    }

// Kommentar: Hier wird eine Lastinformation von Osten her empfangen - inklusive des Wartens. Fr eine Gitterebene. Die
// Vorgehensweise ist wie bei empfange_osten_mit - es wird aber nur fr eine Gitterebene kommuniziert (ein Int wird bertragen),
// und nur fr ein Int eingefgt.
void Precomputation::empfange_osten_ohne(void)
    {
    MPI_Irecv(empfangsbuffer_osten, 1, MPI_INT, ID_osten, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &empfange_request_osten);
    for(flag=false;flag==false;)
        {
        MPI_Test(&empfange_request_osten, &flag, &status);
        }
    teilsumme_osten = empfangsbuffer_osten[0];
    }

// Kommentar: Analog zu empfangen_osten_ohne - nur fr die gespiegelte Himmelsrichtung.
void Precomputation::empfange_westen_ohne(void)
    {
    MPI_Irecv(empfangsbuffer_westen, 1, MPI_INT, ID_westen, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &empfange_request_westen);
    for(flag=false;flag==false;)
        {
        MPI_Test(&empfange_request_westen, &flag, &status);
        }
    teilsumme_westen = empfangsbuffer_westen[0];
    }

// Kommunikation: Im Falle 2er Kerne in Ost-West-Richtung wird die Last des jeweils anderen Kerns bestimmt - durch wechselseitige Kommunikation. Man bestimmt dann
// auch die Gesamtlast mit.
void Precomputation::kommunikation_ow_mit_2(void)
    {
    switch(position_ow)
        {
        case RAND_WESTEN:
// Kommentar: Die eigenen Lastwerte werden gepackt, und nach Osten verschickt, wobei das Versenden der Nachricht abgewartet wird.
            sendebuffer_osten[0] = lambda;
            sendebuffer_osten[1] = lambda_oben;
            MPI_Isend(sendebuffer_osten, 2, MPI_INT, ID_osten, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &(sende_request_osten));
            for(flag=false;flag==false;)
                {
                MPI_Test(&sende_request_osten, &flag, &status);
                }
            empfange_osten_mit();
            gemeinsame_last = teilsumme_osten + lambda;
            gemeinsame_last_oben = teilsumme_osten_oben + lambda_oben;
            break;
        case RAND_OSTEN:
            sendebuffer_westen[0] = lambda;
            sendebuffer_westen[1] = lambda_oben;
            MPI_Isend(sendebuffer_westen, 2, MPI_INT, ID_westen, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &(sende_request_westen));
            for(flag=false;flag==false;)
                {
                MPI_Test(&sende_request_westen, &flag, &status);
                }
            empfange_westen_mit();
            gemeinsame_last = teilsumme_westen + lambda;
            gemeinsame_last_oben = teilsumme_westen_oben + lambda_oben;
            break;
        default:
            cout << " Fehler in Precomputation wo mit " << endl;
            exit(-1);
        }
    }

// Kommentar: Hier wird die Precomputation Kommunikation durchgefhrt - im Falle vom 2 Ebenen Lastausgleich (reverser
// Lastausgleich).
void Precomputation::kommunikation_ow_mit(void)
    {
// Kommentar: Bei 2 Prozessoren nebeneinander geht es schneller. Hineingenommen am 8.2.16

    if(groesse_g==2)
        {
        kommunikation_ow_mit_2();
        return;
        }

// Kommentar: Wir haben hier einen Baum, der von der westlichen Ecke bis zur Mitte verluft, und auch von der stlichen Ecke
// zur Mitte hin. Es gibt in diesem Baum demnach 5 Positionen: Ein westliches, oder stliches Blatt (RAND_WESTEN und RAND_OSTEN),
// die Wurzel in der Mitte (MITTE_WO), und die Inneren Knoten westlich und stlich von der Wurzel (IM_WESTEN und IM_OSTEN).
    switch(position_ow)
        {
        case RAND_WESTEN:
// Kommentar: Die eigenen Lastwerte werden gepackt, und nach Osten verschickt, wobei das Versenden der Nachricht abgewartet wird.
            sendebuffer_osten[0] = lambda;
            sendebuffer_osten[1] = lambda_oben;
            MPI_Isend(sendebuffer_osten, 2, MPI_INT, ID_osten, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &(sende_request_osten));
            for(flag=false;flag==false;)
                {
                MPI_Test(&sende_request_osten, &flag, &status);
                }
            break;
        case IM_WESTEN:
// Kommentar: Zuerst wird von Westen aus empfangen. Durch Addition des eigenen Lastwertes zu dem Wert von teilsumme wird die nach
// Osten zu verschickende Partialsumme bestimmt, und im Buffer gespeichert. Fr beide Gitterebenen. Dann wird diese Nachricht
// verschickt, und gewartet, bis sie vollendet ist.
            empfange_westen_mit();
            sendebuffer_osten[0] = teilsumme_westen + lambda;
#ifdef AUSGABE_PLB
            * da.aus << " IM_WESTEN teilsumme_westen " << teilsumme_westen << " lambda " << lambda << " rang " << rang << endl;
#endif
            sendebuffer_osten[1] = teilsumme_westen_oben + lambda_oben;
            MPI_Isend(sendebuffer_osten, 2, MPI_INT, ID_osten, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &(sende_request_osten));
            for(flag=false;flag==false;)
                {
                MPI_Test(&sende_request_osten, &flag, &status);
                }
            break;
// Kommentar: Analog zu RAND_WESTEN, nur das hier die eigenen Lasten Richtung Westen verschickt werden.
        case RAND_OSTEN:
            sendebuffer_westen[0] = lambda;
            sendebuffer_westen[1] = lambda_oben;
            MPI_Isend(sendebuffer_westen, 2, MPI_INT, ID_westen, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &(sende_request_westen));
            for(flag=false;flag==false;)
                {
                MPI_Test(&sende_request_westen, &flag, &status);
                }
            break;
// Kommentar: Analog zu IM_WESTEN, nur das die Informationen von osten empfangen, und nach Westen weitergegeben werden.
        case IM_OSTEN:
            empfange_osten_mit();
            sendebuffer_westen[0] = teilsumme_westen + lambda;
            sendebuffer_westen[1] = teilsumme_westen_oben + lambda_oben;
            MPI_Isend(sendebuffer_westen, 2, MPI_INT, ID_westen, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &(sende_request_westen));
            for(flag=false;flag==false;)
                {
                MPI_Test(&sende_request_westen, &flag, &status);
                }
            break;
        case MITTE_WO:
// Kommentar: Hier werden die Lasten von Westen unf Osten empfangen. Die einzelnen empfangenen Teillasten einer Ebene
// plus die eigene Last dieser Ebene ergibt die Gesamtlast. Das wird fr beide Gitterebenen berechnet.
            empfange_osten_mit();
            empfange_westen_mit();
#ifdef AUSGABE_PLB
            * da.aus << " MITTE_WO teilsumme_westen " << teilsumme_westen << " lambda " << lambda << " rang " << rang << endl;
            * da.aus << " MITTE_WO teilsumme_osten " << teilsumme_osten << " lambda " << lambda << " rang " << rang << endl;
            * da.aus << " groesse_g_div_2 " << groesse_g_div_2 << " rang " << rang << endl;
#endif
            gemeinsame_last_vektor[0] = teilsumme_westen + teilsumme_osten + lambda;
            gemeinsame_last_vektor[1] = teilsumme_westen_oben + teilsumme_osten_oben + lambda_oben;
        }
#ifdef AUSGABE_PLB
    * da.aus << " groesse_g_div_2 " << groesse_g_div_2 << endl;
    * da.aus << " mitte_ow " << mitte_ow << " " << endl;
#endif

// Kommentar: Die gemeinsame Last wird nun allen Kernen derselben Ost - West - Kerntopologie gebroadcastet. Man achte darauf,
// das der Kommunikation hierbei so gendert wird, dass nur die Kerne derselben Ost - West - Kerntopologielinie diesen
// Broadcast durchfhren.
    MPI_Bcast(gemeinsame_last_vektor, 2, MPI_INT, lokal_rang_mitte_ow, MPI_COMM_OW);

// Kommentar: Die gemeinsame Last wird nun entpackt.
    gemeinsame_last = gemeinsame_last_vektor[0];
    gemeinsame_last_oben = gemeinsame_last_vektor[1];
#ifdef AUSGABE_PLB
    * da.aus << " gl " << gemeinsame_last << " gl_oben " << gemeinsame_last_oben << endl;
#endif
    }

void Precomputation::kommunikation_ow_ohne_2(void)
    {
    switch(position_ow)
        {
        case RAND_WESTEN:
// Kommentar: Die eigenen Lastwerte werden gepackt, und nach Osten verschickt, wobei das Versenden der Nachricht abgewartet wird.
            sendebuffer_osten[0] = lambda;
            MPI_Isend(sendebuffer_osten, 1, MPI_INT, ID_osten, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &(sende_request_osten));
            for(flag=false;flag==false;)
                {
                MPI_Test(&sende_request_osten, &flag, &status);
                }
            empfange_osten_ohne();
            gemeinsame_last = teilsumme_osten + lambda;
            break;
        case RAND_OSTEN:
            sendebuffer_westen[0] = lambda;
            MPI_Isend(sendebuffer_westen, 1, MPI_INT, ID_westen, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &(sende_request_westen));
            for(flag=false;flag==false;)
                {
                MPI_Test(&sende_request_westen, &flag, &status);
                }
            empfange_westen_ohne();
            gemeinsame_last = teilsumme_westen + lambda;
            break;
        default:
            cout << " Fehler in Precomputation ow ohne " << endl;
            exit(-1);
        }
    }

// Kommentar: Hier wird die Precomputation Kommunikation durchgefhrt - im Falle vom ein - Ebenen Lastausgleich (kein
// reverser Lastausgleich). Es werden genau dieselben Schritte wie bei kommunikation_ow_mit durchgefhrt, nur eben nur
// fr eine Gitterebene. Die Kommunikationsbuffer sind nur halb so lang, und es fllt auch nur die Hlfte der
// Berechnungen an. Aber die Schritte sind dieselben.
void Precomputation::kommunikation_ow_ohne(void)
    {
    if(groesse_g==2) // Hineingenommen am 8.2.16
        {
        kommunikation_ow_ohne_2();
        return;
        }
    switch(position_ow)
        {
        case RAND_WESTEN:
            sendebuffer_osten[0] = lambda;
            MPI_Isend(sendebuffer_osten, 1, MPI_INT, ID_osten, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &(sende_request_osten));
            for(flag=false;flag==false;)
                {
                MPI_Test(&sende_request_osten, &flag, &status);
                }
            break;
        case IM_WESTEN:
            empfange_westen_ohne();
            sendebuffer_osten[0] = teilsumme_westen + lambda;
            MPI_Isend(sendebuffer_osten, 1, MPI_INT, ID_osten, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &(sende_request_osten));
            for(flag=false;flag==false;)
                {
                MPI_Test(&sende_request_osten, &flag, &status);
                }
            break;
        case RAND_OSTEN:
            sendebuffer_westen[0] = lambda;
            MPI_Isend(sendebuffer_westen, 1, MPI_INT, ID_westen, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &(sende_request_westen));
            for(flag=false;flag==false;)
                {
                MPI_Test(&sende_request_westen, &flag, &status);
                }
            break;
        case IM_OSTEN:
            empfange_osten_ohne();
            sendebuffer_westen[0] = teilsumme_westen + lambda;
            MPI_Isend(sendebuffer_westen, 1, MPI_INT, ID_westen, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &(sende_request_westen));
            for(flag=false;flag==false;)
                {
                MPI_Test(&sende_request_westen, &flag, &status);
                }
            break;
        case MITTE_WO:
            empfange_osten_ohne();
            empfange_westen_ohne();
#ifdef AUSGABE_PLB
            * da.aus << " MITTE_WO teilsumme_westen " << teilsumme_westen << " lambda " << lambda << " rang " << rang << endl;
            * da.aus << " MITTE_WO teilsumme_osten " << teilsumme_osten << " lambda " << lambda << " rang " << rang << endl;
#endif
            gemeinsame_last_vektor[0] = teilsumme_westen + teilsumme_osten + lambda;
        }
    MPI_Bcast(gemeinsame_last_vektor, 1, MPI_INT, lokal_rang_mitte_ow, MPI_COMM_OW);
    gemeinsame_last = gemeinsame_last_vektor[0];
    }

// Kommunikation: Nach der vL-Berechnung wird hier bestimmt, welcher Wert von Westen nach Osten verschickt wird. Entsprechend  wird
// das negative davon von Ost nach West verschickt.
void Precomputation::berechne_lambdas_ow_mit_2(void)
    {
    switch(position_ow)
        {
        case RAND_WESTEN:
            lambda_ost = lambda - (gemeinsame_last / groesse_g);
            lambda_ost_oben = lambda_oben - (gemeinsame_last_oben / groesse_g);
            break;
        case RAND_OSTEN:
            lambda_west = (gemeinsame_last / groesse_g) - teilsumme_westen;
            lambda_west_oben = (gemeinsame_last_oben / groesse_g) - teilsumme_westen_oben;
            break;
        }
    }

void Precomputation::berechne_lambdas_ow_ohne_2(void)
    {
    switch(position_ow)
        {
        case RAND_WESTEN:
            lambda_ost = lambda - (gemeinsame_last / groesse_g);
            break;
        case RAND_OSTEN:
            lambda_west = (gemeinsame_last / groesse_g) - teilsumme_westen;
            break;
        }
    }

void Precomputation::berechne_lambdas_ns_mit_2(void)
    {
    switch(position_ns)
        {
        case RAND_NORDEN:
            lambda_sued = lambda - (gemeinsame_last / groesse_f);
            lambda_sued_oben = lambda_oben - (gemeinsame_last_oben / groesse_f);
            break;
        case RAND_SUEDEN:
            lambda_nord = (gemeinsame_last / groesse_f) - teilsumme_norden;
            lambda_nord_oben = (gemeinsame_last_oben / groesse_f) - teilsumme_norden_oben;
            break;
        }
    }

void Precomputation::berechne_lambdas_ns_ohne_2(void)
    {
    switch(position_ns)
        {
        case RAND_NORDEN:
            lambda_sued = lambda - (gemeinsame_last / groesse_f);
            break;
        case RAND_SUEDEN:
            lambda_nord = (gemeinsame_last / groesse_f) - teilsumme_norden;
            break;
        }
    }

// Kommentar: Hier werden die fr einen vollstndigen Lastausgleich nach Osten und Westen zu verschicken bzw. von dort zu
// empfangenden Lastmengen bestimmt.
void Precomputation::berechne_lambdas_ow_mit(void)
    {
// Kommentar: Es treten dieselben 5 Positionen in den Ost - West - Kerntopologielinien auf wie bei 'kommunikation_ow_mit'.
    switch(position_ow)
        {
        case RAND_WESTEN:
// Kommentar: Nach Osten wird der Unterschied zwischen der eigenen Last und der durchschnittlichen Last fr einen
// Kern der Ost - West - Kerntopologie verschickt.
            lambda_ost = lambda - (gemeinsame_last / groesse_g);
            lambda_ost_oben = lambda_oben - (gemeinsame_last_oben / groesse_g);
            // Sei lambda < phi(lambdas). Dann muss Last zurckfliessen, also lambda_ost negativ sein.
            // lambda_west bleibt 0.
            break;
        case IM_WESTEN:
// Kommentar: Hier wird die Differenz aus Durchschnittslast Mal Gitterpunkte im Westen minus die Teilsummelast von Westen aus.
// Man sehe, dass die Werte Richtung West das umgekehrte Vorzeichen haben wir die lamda - Werte Richtung Osten.
            lambda_west = (gemeinsame_last * anzahl_westen) / groesse_g - teilsumme_westen;
            lambda_west_oben = (gemeinsame_last_oben * anzahl_westen) / groesse_g - teilsumme_westen_oben;
            // ist teilsumme_westen < ziellast_nachbar_westen => schicke ber Westkante Last nach Westen

// Kommentar:  Dieselbe Formel wie nach Westen, nur mit einem Gitterpunkt mehr - fr die Durchschnittslast; und die
// Teilsumme nach Westen wird um die eigene Last erhht. Und das Vorzeichen wird nach Osten gendert.
            lambda_ost = - (gemeinsame_last * (anzahl_westen+1)) / groesse_g + teilsumme_westen + lambda;
            // Ist die aggregierte Last mit Punkt kleiner als die Durchschnittslast mit Punkt, muss von Osten Last kommen,
            // d.h. lambda_ost > 0 sein.
            lambda_ost_oben = - (gemeinsame_last_oben * (anzahl_westen+1)) / groesse_g + teilsumme_westen_oben + lambda_oben;
            break;
// Kommentar: Analog zu RAND_WESTEN. Hier wird dasselbe Vorzeichen gewhlt wie bei RAND_WESTEN. Das liegt daran, dass die
// Lastsummen von den Blttern her berechnet werden, in diesem Fall von Osten her.
        case RAND_OSTEN:
            lambda_west = lambda - (gemeinsame_last / groesse_g);
            lambda_west_oben = lambda_oben - (gemeinsame_last_oben / groesse_g);
            // Wieder: Ist lambda grsser als der Durchschnitt => Last muss abfliessen, also lambda_west > 0 sein
            break;
// Kommentar: Analog zu IM_WESTEN, nur dass hier lambda_ost dasselbe Vorzeichen hat wie lambda_west bei IM_WESTEN. Das liegt
// daran, dass die Lastsummen von den Blttern her berechnet werden, in diesem Fall von Osten her.
        case IM_OSTEN:
            lambda_ost = (gemeinsame_last * anzahl_osten) / groesse_g - teilsumme_osten;
            lambda_ost_oben = (gemeinsame_last_oben * anzahl_osten) / groesse_g - teilsumme_osten_oben;
//            lambda_west = (gemeinsame_last * (anzahl_osten+1)) / groesse_g - teilsumme_osten - lambda;
            lambda_west = - (gemeinsame_last * (anzahl_osten+1)) / groesse_g + teilsumme_osten + lambda;
//            lambda_west_oben = (gemeinsame_last_oben * (anzahl_osten+1)) / groesse_g - teilsumme_osten_oben - lambda_oben;
            lambda_west_oben = - (gemeinsame_last_oben * (anzahl_osten+1)) / groesse_g + teilsumme_osten_oben + lambda_oben;
            break;
        case MITTE_WO:
// Kommentar: Man hat hier nach Westen dieselbe Formel wie fr lambda_west im Fall IM_WESTEN, und fr lambda_ost
// dieselbe Formel wie fr lambda_ost im Fall IM_OSTEN. Diese Formeln drcken sich also bis zur Mitte durch.
            lambda_west = (gemeinsame_last * anzahl_westen) / groesse_g - teilsumme_westen;
#ifdef AUSGABE_PLB
            * da.aus << " berechne lambda_west gemeinsame_last " << gemeinsame_last << " anzahl_westen " << anzahl_westen << " groesse_g " << groesse_g << " teilsumme_westen " << teilsumme_westen << " lambda_west " << lambda_west << endl;
#endif
            lambda_west_oben = (gemeinsame_last_oben * anzahl_westen) / groesse_g - teilsumme_westen_oben;
            lambda_ost = (gemeinsame_last * anzahl_osten) / groesse_g - teilsumme_osten;
            lambda_ost_oben = (gemeinsame_last_oben * anzahl_osten) / groesse_g - teilsumme_osten_oben;
            break;
        }
    }

// Kommentar: Hier werden dieselben Formeln wie bei berechne_lambdas_ow_mit verwendet, nur eben nur fr eine Gitterebene.
// Die Kommentare folgen also vllig analog.
void Precomputation::berechne_lambdas_ow_ohne(void)
    {
    switch(position_ow)
        {
        case RAND_WESTEN:
            lambda_ost = lambda - (gemeinsame_last / groesse_g);
            // Sei lambda < phi(lambdas). Dann muss Last zurckfliessen, also lambda_ost negativ sein.
            // lambda_west bleibt 0.
            break;
        case IM_WESTEN:
            lambda_west = (gemeinsame_last * anzahl_westen) / groesse_g - teilsumme_westen;
            // ist teilsumme_westen < ziellast_nachbar_westen => schicke ber Westkante Last nach Westen
            lambda_ost = - (gemeinsame_last * (anzahl_westen+1)) / groesse_g + teilsumme_westen + lambda;
            // Ist die aggregierte Last mit Punkt kleiner als die Durchschnittslast mit Punkt, muss von Osten Last kommen,
            // d.h. lambda_ost > 0 sein.
            break;
        case RAND_OSTEN:
            lambda_west = lambda - (gemeinsame_last / groesse_g);
            // Wieder: Ist lambda grsser als der Durchschnitt => Last muss abfliessen, also lambda_west > 0 sein
            break;
        case IM_OSTEN:
            lambda_ost = (gemeinsame_last * anzahl_osten) / groesse_g - teilsumme_osten;
            lambda_west = - (gemeinsame_last * (anzahl_osten+1)) / groesse_g + teilsumme_osten + lambda;
            // teilsumme_ost_groesser => ber diese Kante muss mehr verschickt werden.
            break;
        case MITTE_WO:
            lambda_west = (gemeinsame_last * anzahl_westen) / groesse_g - teilsumme_westen;
            lambda_ost = (gemeinsame_last * anzahl_osten) / groesse_g - teilsumme_osten;
            break;
        }
    }

//////////////////////////////////////////////////////////
////////////      Precomputation Nord Sued  //////////////
//////////////////////////////////////////////////////////

// Hier war zu entscheiden, alles nochmal zu machen, oder Precomputation 2 Mal zu deklarieren - wie beim Balancing
// Da der Code wenig ist, die Verknpfung mit Identify komplizierter wre, und auch mehr Variablen
// umzusetzen wren habe ich mich fr das Duplikat entschieden. Wesentlich hierbei ist die Subklasse Identify.
// Z.B. setzen groesse_g = groesse_g bzw. groesse_g = groesse_f fhrt zu Verwicklungen. Die 'Kopie' ist daher
// - im Gegensatz zum Balancing - zu bevorzugen. 13.09.2014
// Es ist einfach zu viel variabel zu halten - bis hin zu den cases der switch - Befehle. Es ginge, ist aber nicht clever.

// Kommentar: Dieser Fall ist vllig analog zum Ost - West - Fall, und wird daher nicht gesondert kommentiert.
void Precomputation::tun_ns_mit(void)
    {
    if(groesse_f==2) // Hineingenommen am 8.2.16
        {
        kommunikation_ns_mit_2();
        berechne_lambdas_ns_mit_2();
        return;
        }
#ifdef AUSGABE_PLB
    * da.aus << " Vor Kommunikation ns position_ns " << position_ns << endl;
#endif
    kommunikation_ns_mit();
#ifdef AUSGABE_PLB
    * da.aus << " Nach Kommunikation ns gemeinsame_last " << gemeinsame_last << endl;
#endif
    berechne_lambdas_ns_mit();
#ifdef AUSGABE_PLB
    * da.aus << " Nach berechne_lambdas_ns_mit " << endl;
#endif
    }

void Precomputation::tun_ns_ohne(void)
    {
    if(groesse_f==2)
        {
        kommunikation_ns_ohne_2();
        berechne_lambdas_ns_ohne_2();
        return;
        }
#ifdef AUSGABE_PLB
    * da.aus << " Vor Kommunikation ns position_ns " << position_ns << endl;
#endif
    kommunikation_ns_ohne();
#ifdef AUSGABE_PLB
    * da.aus << " Nach Kommunikation ns gemeinsame_last " << gemeinsame_last << endl;
#endif
    berechne_lambdas_ns_ohne();
#ifdef AUSGABE_PLB
    * da.aus << " Nach berechne_lambdas_ns_ohne " << endl;
#endif
    }

void Precomputation::empfange_sueden_mit(void)
    {
    MPI_Irecv(empfangsbuffer_sueden, 2, MPI_INT, ID_sueden, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &empfange_request_sueden);
    for(flag=false;flag==false;)
        {
        MPI_Test(&empfange_request_sueden, &flag, &status);
        }
    teilsumme_sueden = empfangsbuffer_sueden[0];
    teilsumme_sueden_oben = empfangsbuffer_sueden[1];
    }

void Precomputation::empfange_norden_mit(void)
    {
    MPI_Irecv(empfangsbuffer_norden, 2, MPI_INT, ID_norden, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &empfange_request_norden);
    for(flag=false;flag==false;)
        {
        MPI_Test(&empfange_request_norden, &flag, &status);
        }
    teilsumme_norden = empfangsbuffer_norden[0];
    teilsumme_norden_oben = empfangsbuffer_norden[1];
    }

void Precomputation::empfange_sueden_ohne(void)
    {
    MPI_Irecv(empfangsbuffer_sueden, 1, MPI_INT, ID_sueden, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &empfange_request_sueden);
    for(flag=false;flag==false;)
        {
        MPI_Test(&empfange_request_sueden, &flag, &status);
        }
    teilsumme_sueden = empfangsbuffer_sueden[0];
    }

void Precomputation::empfange_norden_ohne(void)
    {
    MPI_Irecv(empfangsbuffer_norden, 1, MPI_INT, ID_norden, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &empfange_request_norden);
    for(flag=false;flag==false;)
        {
        MPI_Test(&empfange_request_norden, &flag, &status);
        }
    teilsumme_norden = empfangsbuffer_norden[0];
    }

void Precomputation::kommunikation_ns_mit_2(void)
    {
    switch(position_ns)
        {
        case RAND_NORDEN:
// Kommentar: Die eigenen Lastwerte werden gepackt, und nach Osten verschickt, wobei das Versenden der Nachricht abgewartet wird.
            sendebuffer_sueden[0] = lambda;
            sendebuffer_sueden[1] = lambda_oben;
            MPI_Isend(sendebuffer_sueden, 2, MPI_INT, ID_sueden, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &(sende_request_sueden));
            for(flag=false;flag==false;)
                {
                MPI_Test(&sende_request_sueden, &flag, &status);
                }
            empfange_sueden_mit();
            gemeinsame_last = teilsumme_sueden + lambda;
            gemeinsame_last_oben = teilsumme_sueden_oben + lambda_oben;
            break;
        case RAND_SUEDEN:
            sendebuffer_norden[0] = lambda;
            sendebuffer_norden[1] = lambda_oben;
            MPI_Isend(sendebuffer_norden, 2, MPI_INT, ID_norden, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &(sende_request_norden));
            for(flag=false;flag==false;)
                {
                MPI_Test(&sende_request_norden, &flag, &status);
                }
            empfange_norden_mit();
            gemeinsame_last = teilsumme_norden + lambda;
            gemeinsame_last_oben = teilsumme_norden_oben + lambda_oben;
            break;
        default:
            cout << " Fehler in Precomputation ns mit " << endl;
            exit(-1);
        }
    }

void Precomputation::kommunikation_ns_mit(void)
    {
// Kommentar: Bei 2 Prozessoren nebeneinander geht es schneller. Hineingenommen am 8.2.16

    if(groesse_f==2)
        {
        kommunikation_ns_mit_2();
        return;
        }

    switch(position_ns)
        {
        case RAND_NORDEN:
            sendebuffer_sueden[0] = lambda;
            sendebuffer_sueden[1] = lambda_oben;
            MPI_Isend(sendebuffer_sueden, 2, MPI_INT, ID_sueden, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &(sende_request_sueden));
            for(flag=false;flag==false;)
                {
                MPI_Test(&sende_request_sueden, &flag, &status);
                }
#ifdef AUSGABE_PLB
            * da.aus << " Durch senden durch rang " << rang << endl;
#endif
            break;
        case IM_NORDEN:
            empfange_norden_mit();
            sendebuffer_sueden[0] = teilsumme_norden + lambda;
            sendebuffer_sueden[1] = teilsumme_norden_oben + lambda_oben;
            MPI_Isend(sendebuffer_sueden, 2, MPI_INT, ID_sueden, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &(sende_request_sueden));
            for(flag=false;flag==false;)
                {
                MPI_Test(&sende_request_sueden, &flag, &status);
                }
            break;
        case RAND_SUEDEN:
            sendebuffer_norden[0] = lambda;
            sendebuffer_norden[1] = lambda_oben;
            MPI_Isend(sendebuffer_norden, 2, MPI_INT, ID_norden, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &(sende_request_norden));
            for(flag=false;flag==false;)
                {
                MPI_Test(&sende_request_norden, &flag, &status);
                }
            break;
        case IM_SUEDEN:
            empfange_sueden_mit();
            sendebuffer_norden[0] = teilsumme_norden + lambda;
            sendebuffer_norden[1] = teilsumme_norden_oben + lambda_oben;
            MPI_Isend(sendebuffer_norden, 2, MPI_INT, ID_norden, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &(sende_request_norden));
            for(flag=false;flag==false;)
                {
                MPI_Test(&sende_request_norden, &flag, &status);
                }
            break;
        case MITTE_NS:
#ifdef AUSGABE_PLB
            * da.aus << " ID_norden " << ID_norden << " ID_sueden " << ID_sueden << endl;
#endif
            empfange_sueden_mit();
            empfange_norden_mit();
            gemeinsame_last_vektor[0] = teilsumme_norden + teilsumme_sueden + lambda;
            gemeinsame_last_vektor[1] = teilsumme_norden_oben + teilsumme_sueden_oben + lambda_oben;
#ifdef AUSGABE_PLB
            * da.aus << " teilsumme_norden " << teilsumme_norden << " teilsumme_sueden " << teilsumme_sueden << " lambda " << lambda << endl;
            * da.aus << " teilsumme_norden_oben " << teilsumme_norden_oben << " teilsumme_sueden_oben " << teilsumme_sueden_oben << " lambda_oben " << lambda_oben << endl;
#endif
            break;
        default:
            cout << " Falsche position_ns " << position_ns << " rang " << rang << endl;
            exit(-1);
        }
#ifdef AUSGABE_PLB
    * da.aus << " Durch kommunikation_ns_mit fast durch rang " << rang << " mitte_ns " << mitte_ns << endl;
#endif
    MPI_Bcast(gemeinsame_last_vektor, 2, MPI_INT, lokal_rang_mitte_ns, MPI_COMM_NS);
#ifdef AUSGABE_PLB
    * da.aus << " Durch kommunikation_ns_mit fast durch rang " << rang << " mitte_ns " << mitte_ns << endl;
#endif
    gemeinsame_last = gemeinsame_last_vektor[0];
    gemeinsame_last_oben = gemeinsame_last_vektor[1];
    }

void Precomputation::kommunikation_ns_ohne_2(void)
    {
    switch(position_ns)
        {
        case RAND_NORDEN:
// Kommentar: Die eigenen Lastwerte werden gepackt, und nach Osten verschickt, wobei das Versenden der Nachricht abgewartet wird.
            sendebuffer_sueden[0] = lambda;
            MPI_Isend(sendebuffer_sueden, 1, MPI_INT, ID_sueden, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &(sende_request_sueden));
            for(flag=false;flag==false;)
                {
                MPI_Test(&sende_request_sueden, &flag, &status);
                }
            empfange_sueden_ohne();
            gemeinsame_last = teilsumme_sueden + lambda;
            break;
        case RAND_SUEDEN:
            sendebuffer_norden[0] = lambda;
            MPI_Isend(sendebuffer_norden, 1, MPI_INT, ID_westen, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &(sende_request_norden));
            for(flag=false;flag==false;)
                {
                MPI_Test(&sende_request_norden, &flag, &status);
                }
            empfange_norden_ohne();
            gemeinsame_last = teilsumme_norden + lambda;
            break;
        default:
            cout << " Fehler in Precomputation ns ohne " << endl;
            exit(-1);
        }
    }

void Precomputation::kommunikation_ns_ohne(void)
    {
#ifdef AUSGABE_PLB
    * da.aus << " Start Kommunikation ns ohne " << endl;
#endif // AUSGABE_PLB

// Kommentar: Bei 2 Prozessoren nebeneinander geht es schneller. Hineingenommen am 8.2.16

    if(groesse_f==2)
        {
        kommunikation_ns_ohne_2();
        return;
        }

    switch(position_ns)
        {
        case RAND_NORDEN:
            sendebuffer_sueden[0] = lambda;
            MPI_Isend(sendebuffer_sueden, 1, MPI_INT, ID_sueden, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &(sende_request_sueden));
            for(flag=false;flag==false;)
                {
                MPI_Test(&sende_request_sueden, &flag, &status);
                }
            break;
        case IM_NORDEN:
            empfange_norden_ohne();
            sendebuffer_sueden[0] = teilsumme_norden + lambda;
#ifdef AUSGABE_PLB
            * da.aus << " IM Norden: teilsumme_norden " << teilsumme_norden << " lamdba " << lambda << endl;
#endif
            MPI_Isend(sendebuffer_sueden, 1, MPI_INT, ID_sueden, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &(sende_request_sueden));
            for(flag=false;flag==false;)
                {
                MPI_Test(&sende_request_sueden, &flag, &status);
                }
            break;
        case RAND_SUEDEN:
            sendebuffer_norden[0] = lambda;
            MPI_Isend(sendebuffer_norden, 1, MPI_INT, ID_norden, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &(sende_request_norden));
            for(flag=false;flag==false;)
                {
                MPI_Test(&sende_request_norden, &flag, &status);
                }
            break;
        case IM_SUEDEN:
            empfange_sueden_ohne();
            sendebuffer_norden[0] = teilsumme_norden + lambda;
            MPI_Isend(sendebuffer_norden, 1, MPI_INT, ID_norden, TAG_PRECOMPUTATION, MPI_COMM_WORLD, &(sende_request_norden));
            for(flag=false;flag==false;)
                {
                MPI_Test(&sende_request_norden, &flag, &status);
                }
            break;
        case MITTE_NS:
            empfange_sueden_ohne();
            empfange_norden_ohne();
#ifdef AUSGABE_PLB
            * da.aus << " MITTE_NS: teilsumme_norden " << teilsumme_norden << " teilsumme_sueden " << teilsumme_sueden << " lambda " << lambda << endl;
#endif
            gemeinsame_last_vektor[0] = teilsumme_sueden + teilsumme_norden + lambda;
            break;
        default:
            cout << " Falsche position_ns " << position_ns << " rang " << rang << endl;
            exit(-1);
        }
    MPI_Bcast(gemeinsame_last_vektor, 1, MPI_INT, lokal_rang_mitte_ns, MPI_COMM_NS);
    gemeinsame_last = gemeinsame_last_vektor[0];
    }

void Precomputation::berechne_lambdas_ns_mit(void)
    {
    switch(position_ns)
        {
        case RAND_NORDEN:
            lambda_sued = lambda - (gemeinsame_last / groesse_f);
            lambda_sued_oben = lambda_oben - (gemeinsame_last_oben / groesse_f);
            // Sei lambda < phi(lambdas). Dann muss Last zurckfliessen, also lambda_ost negativ sein.
            // lambda_west bleibt 0.
            break;
        case IM_NORDEN:
            lambda_nord = (gemeinsame_last * anzahl_norden) / groesse_f - teilsumme_norden;
            lambda_nord_oben = (gemeinsame_last_oben * anzahl_norden) / groesse_f - teilsumme_norden_oben;
            // ist teilsumme_westen < ziellast_nachbar_westen => schicke ber Westkante Last nach Westen
            lambda_sued = - (gemeinsame_last * (anzahl_norden+1)) / groesse_f + teilsumme_norden + lambda;
            // Ist die aggregierte Last mit Punkt kleiner als die Durchschnittslast mit Punkt, muss von Osten Last kommen,
            // d.h. lambda_ost > 0 sein.
            lambda_sued_oben = - (gemeinsame_last_oben * (anzahl_norden+1)) / groesse_f + teilsumme_norden_oben + lambda_oben;
            break;
        case RAND_SUEDEN:
            lambda_nord = lambda - (gemeinsame_last / groesse_f);
            lambda_nord_oben = lambda_oben - (gemeinsame_last_oben / groesse_f);
            // Wieder: Ist lambda grsser als der Durchschnitt => Last muss abfliessen, also lambda_west > 0 sein
            break;
        case IM_SUEDEN:
            lambda_sued = (gemeinsame_last * anzahl_sueden) / groesse_f - teilsumme_sueden;
            lambda_sued_oben = (gemeinsame_last_oben * anzahl_sueden) / groesse_f - teilsumme_sueden_oben;
            lambda_nord = - (gemeinsame_last * (anzahl_sueden+1)) / groesse_f + teilsumme_sueden + lambda;
            lambda_nord_oben = - (gemeinsame_last_oben * (anzahl_sueden+1)) / groesse_f + teilsumme_sueden_oben + lambda_oben;
            // So richtig: beides Mal die anzahl_sueden, weil von dort die Teilsumme stammt
            break;
        case MITTE_NS:
            lambda_nord = (gemeinsame_last * anzahl_norden) / groesse_f - teilsumme_norden;
            lambda_nord_oben = (gemeinsame_last_oben * anzahl_norden) / groesse_f - teilsumme_norden_oben;
            lambda_sued = (gemeinsame_last * anzahl_sueden) / groesse_f - teilsumme_sueden;
            lambda_sued_oben = (gemeinsame_last_oben * anzahl_sueden) / groesse_f - teilsumme_sueden_oben;
            break;
        default:
            cout << " Falsche position_ns " << position_ns << " rang " << rang << endl;
            exit(-1);
        }
    }

void Precomputation::berechne_lambdas_ns_ohne(void)
    {
    switch(position_ns)
        {
        case RAND_NORDEN:
            lambda_sued = lambda - (gemeinsame_last / groesse_f);
            // Sei lambda < phi(lambdas). Dann muss Last zurckfliessen, also lambda_ost negativ sein.
            // lambda_west bleibt 0.
            break;
        case IM_NORDEN:
            lambda_nord = (gemeinsame_last * anzahl_norden) / groesse_f - teilsumme_norden;
            // ist teilsumme_westen < ziellast_nachbar_westen => schicke ber Westkante Last nach Westen
            lambda_sued = - (gemeinsame_last * (anzahl_norden+1)) / groesse_f + teilsumme_norden + lambda;
            // Ist die aggregierte Last mit Punkt kleiner als die Durchschnittslast mit Punkt, muss von Osten Last kommen,
            // d.h. lambda_ost > 0 sein.
            break;
        case RAND_SUEDEN:
            lambda_nord = lambda - (gemeinsame_last / groesse_f);
            // Wieder: Ist lambda grsser als der Durchschnitt => Last muss abfliessen, also lambda_west > 0 sein
            break;
        case IM_SUEDEN:
            lambda_sued = (gemeinsame_last * anzahl_sueden) / groesse_f - teilsumme_sueden;
            lambda_nord = - (gemeinsame_last * (anzahl_sueden+1)) / groesse_f + teilsumme_sueden + lambda;
            break;
        case MITTE_NS:
            lambda_nord = (gemeinsame_last * anzahl_norden) / groesse_f - teilsumme_norden;
            lambda_sued = (gemeinsame_last * anzahl_sueden) / groesse_f - teilsumme_sueden;
            break;
        default:
            cout << " Falsche position_ns " << position_ns << " rang " << rang << endl;
            exit(-1);
        }
    }
