#include "punkt_feld_liste.h"

// Kommentar: In dieser Feldliste knnen Punkte aufgrund ihrer Koordinaten in O(1) gelesen und eingefgt und entfernt werden.
// Ferner kann die ganze Punktliste in O(Anzahl Eintrgen) durchlaufen, und komplett gelscht werden. Sie ist ein
// schnelleres quivalent zur Feldliste, und braucht auch weniger Speicherplatz, weil eine doppelt verkettete Liste mehr
// Speicherplatz als eine einfache Liste bentigt.

// Kommentar: Die bentigten Parameter dieser Klasse sind die Gittermae m und n.
void PunktFeldListe::initialisierung(int _m, int _n)
// Wird jetzt unten komplett finalisiert.
    {
// Kommentar: m, n, und mn werden abgespeichert.
    m = _m;
    n = _n;
    mn = m * n;

// Kommentar: Der Speicherplatz fr die Koordinaten der Liste wird angelegt.
    f = new short int [mn];
    g = new short int [mn];

// Kommentar: Das auf die Stelle der Listen f und g zeigende Feld wird als integer oder short integer Feld angelegt, und mit den
// Werten -1 angelegt. -1 steht fr kein Punkt, weil der Eintrag nicht auf ein Element der Koordinatenlisten zeigt.
#ifdef PUNKTE_FELD_LISTE_INT
    feld = new int * [n];
#else
    feld = new short int * [n];
#endif
    for(i=0; i<n; i++)
        {
#ifdef PUNKTE_FELD_LISTE_INT
        feld[i] = new int [m];
#else
        feld[i] = new short int [m];
#endif
        for(j=0; j<m; j++)
            {
            feld[i][j] = -1;
            }
        }
// Kommentar: Anzahl = 0 heisst, dass diese Liste leer ist.
    anzahl = 0;
    }

// Kommentar: Die Klasse wird finalisiert.
void PunktFeldListe::finalisierung(void)
    {
// Kommentar: Der Speicherplatz des Feldes sowie der Koordinatenliste wird freigegeben.
    for(i=0; i<n; i++)
        {
        delete feld[i];
        }
    delete feld;
    // Neu hinzu:
    delete f;
    delete g;
    }

// Kommentar: Ein Punkt wird hinzugefgt.
void PunktFeldListe::hinzu(int ff, int gg)
    {
// Kommentar: Zuerst wird geprft, dass die Gittermae nicht berschritten werden.
    if(ff < 0)
        {
        return;
        }
    if(gg < 0)
        {
        return;
        }
    if(ff >= n)
        {
        return;
        }
    if(gg >= m)
        {
        return;
        }
#ifdef AUSGABEN_PUNKT_FELDLISTE
    if((ff==8)&&(gg==3))
        {
        cout << " (PFL 8,3) " << feld[ff][gg] << " ";
        }
#endif
#ifdef AUSGABE_AALL
    if((ff==126)&&(gg==992))
        {
        cout << " AALL hinzu Anfang feld[ff][gg] " << " ff " << ff << " gg " << gg << " feld " << feld[ff][gg] << " anzahl " << anzahl << endl;
        }
#endif
// Kommentar: Falls das Feld noch keinen Eintrag hat, dann wird es auf die Position 'anzahl' fr die Koordinatenliste gesetzt,
// und es werden die Koordinaten an dieser Stelle in die Koordinatenlisten eingetragen. Die Lnge der Liste 'anzahl' wird um
// eins erhht.
    if(feld[ff][gg]==-1)
        {
#ifdef AUSGABE_AALL
        if((ff==126)&&(gg==992))
            {
            cout << " AALL hinzu Mitte feld[ff][gg] " << " ff " << ff << " gg " << gg << " feld " << feld[ff][gg] << " anzahl " << anzahl << endl;
            }
#endif
        feld[ff][gg] = anzahl;
#ifdef AUSGABE_AALL
        if((ff==126)&&(gg==992))
            {
            cout << " AALL hinzu Mitte2 feld[ff][gg] " << " ff " << ff << " gg " << gg << " feld " << feld[ff][gg] << " anzahl " << anzahl << endl;
            }
#endif
        f[anzahl] = ff;
        g[anzahl] = gg;
        anzahl ++;
        }
#ifdef AUSGABE_AALL
    if((ff==126)&&(gg==992))
        {
        cout << " AALL hinzu Ende feld[ff][gg] " << " ff " << ff << " gg " << gg << " feld " << feld[ff][gg] << " anzahl " << anzahl << endl;
        }
#endif
    }

// Kommentar: Hier wird in O(1) ein Punkt mit den Koordinaten (ff,gg) aus der Liste entfernt.
void PunktFeldListe::hinweg(int ff, int gg)
    {
/*
    if(ff < 0)
        {
        return;
        }
    if(gg < 0)
        {
        return;
        }
    if(ff >= n)
        {
        return;
        }
    if(gg >= m)
        {
        return;
        }
Die wurden schon einmal kontrolliert! Diese Kontrollen aber behalten. 8.2.16
*/

// Kommentar: Die Position in den Koordinatenlisten wird aus der Klasse Feld unter den Koordinaten (ff,gg) bestimmt.
    position = feld[ff][gg];

// Kommentar: Ist der Wert -1, existier in (ff,gg) kein Punkt, und einen nicht existierenden Punkt kann man nicht lschen.
    if(position == -1) // Falls feld[ff][gg]==-1, so existiert in (ff,gg) kein Punkt, und es ist nichts zu tun.
        {
        return;
        }
// Kommentar: Die Liste wird um eins verkrzt. Eigentlich wrde der Punkt (f(anzahl),g(anzahl)) damit gelscht.
    anzahl --;
// Kommentar: Dieser Punkt wird nun genommen.
    f_hilfe = f[anzahl];
    g_hilfe = g[anzahl];
//    feld[ff][gg] = feld[f_hilfe][g_hilfe];

// Kommentar: Das Feld an dieser Stelle wird auf die Position position gesetzt - dorthin wird nmlich der Punkt (f_hilfe,g_hilfe)
// verschoben.
    feld[f_hilfe][g_hilfe] = position;
// Kommentar: Der Punkt (f_hilfe,g_hilfe) wird auf die Stelle position in den Koordinatenlisten verschoben. Dabei wird der Punkt
// (ff,gg) berschrieben.
    f[position] = f_hilfe;
    g[position] = g_hilfe;
// Kommentar: Nun muss noch der Eintrag des Punktes (ff,gg) im Feld gelscht werden, indem das Feld an der Stelle auf -1 gesetzt
// wird.
    feld[ff][gg] = -1; // Hinzugefgt. Fehlte bisher.
    }

// Kommentar: Hier wird die gesamte Liste gelscht.
void PunktFeldListe::reset(void)
    {
// Kommentar: Zuerst wird das Feld auf allen Punkten der Liste gelscht. Es ist damit komplett auf -1.
    for(i=0; i<anzahl; i++)
        {
        feld[f[i]][g[i]] = -1;
        }
// Kommentar: Und dann wird die Liste bzgl. der Koordinaten f und g geleert.
    anzahl = 0;
    }

// Kommentar: Hier werden die Koordinaten der in (f,g) verzeichneten Punkte ausgegeben.
void PunktFeldListe::ausgabe(void)
    {
#ifdef AUSGABEN_PUNKT_FELDLISTE
    cout << "PFLa ";

    for(i=0; i<anzahl; i++)
        {
        cout << f[i] << " " << g[i] << " ";
        }
    cout << endl;
#endif
    }

// Kommentar: Hier wird das Feld, 2 dimensional, ausgegeben, in einer Datei.
void PunktFeldListe::ausgabe_feld(DateiAusgabe * da)
    {
#ifdef AUSGABEN_DATEI_NEU
    * da->aus << "PFL 2" << endl;

    for(i=0; i<n; i++)
        {
        * da->aus << i << " ";
        if(i<10)
            {
            * da->aus << " ";
            }
        for(j=0; j<m; j++)
            {
            if(feld[i][j]==-1)
                {
                * da->aus << "0" << " ";
                }
            else
                {
                * da->aus << "1" << " ";
                }
            }
        * da->aus << endl;
        }
    for(i=0; i<anzahl; i++)
        {
        * da->aus << "(" << f[i] << "," << g[i] << ")";
        }
    * da->aus << endl << " Anzahl " << anzahl << endl;
    * da->aus << "Ende PFL 2" << endl;
#endif
    }

// Kommentar: Hier wird ein boolsches Feld auf den Gitterpunkten der Liste auf false gesetzt.
void PunktFeldListe::entferne_besetzte_punkte(void)
    {
    for(i=0; i<anzahl; i++)
        {
        besetzte_punkte[f[i]][g[i]] = false;
        }
    }

// Kommentar: Hier wird ein boolsches Feld auf den Gitterpunkten der Liste auf true gesetzt.
void PunktFeldListe::fuege_besetzte_punkte_ein(void)
    {
    for(i=0; i<anzahl; i++)
        {
        besetzte_punkte[f[i]][g[i]] = true;
        }
    }

