#ifndef RANDSTRUKTUR_KOMMUNIKATION
#define RANDSTRUKTUR_KOMMUNIKATION 1

#include <mpi.h>

#include "randstukturpv.h"

#include "ausgaben.h"

#include "kleine_randstruktur_kommunikation.h"
#include "kleine_randstruktur_test.h"

class RandStrukturKommunikation :   virtual public RandStrukturPV,
                                    public KleineRandStrukturTest,
                                    public KleineRandStrukturKommunikation
    {
private:
public:
#ifdef KURZ_RK
    short int * sendebuffer;
    short int * empfangsbuffer;
#else
    int * sendebuffer;
    int * empfangsbuffer;
#endif
private:
    MPI_Request sende_request;
    MPI_Request empfangs_request;

    MPI_Request sende_request_ep;
    MPI_Request empfangs_request_ep;

    MPI_Status sende_status;
    MPI_Status empfangs_status;
    int flag;

private:
public:
#ifdef KURZ_RAND_ELEMENTE
    short int * f1_v_vektor_e, * f2_v_vektor_e, * g_v_vektor_e;

    short int * f_h_vektor_e, * g1_h_vektor_e, * g2_h_vektor_e;

    short int * f_e, * g_e;
#else
    int * f1_v_vektor_e, * f2_v_vektor_e, * g_v_vektor_e;

    int * f_h_vektor_e, * g1_h_vektor_e, * g2_h_vektor_e;

    int * f_e, * g_e;
#endif
    int anzahl_v_vektor_e;
    int anzahl_h_vektor_e;
    int punkteanzahl_e;
    int punkteanzahl_alle;
public:
    int aussenrand_bestimmen(int anzahl_aussen);
private:
        int j1, j2, k1, k2;
        int pos;
public:
    #ifdef dimensionen1
    void initialisieren_kommunikation(double ** _x0, double ** _y0);
    #endif
    #ifdef dimensionen2
    void initialisieren_kommunikation(double ** _x0, double ** _y0, double ** _x1, double ** _y1);
    #endif
    #ifdef dimensionen3
    void initialisieren_kommunikation(double ** _x0, double ** _y0, double ** _x1, double ** _y1, double ** _x2, double ** _y2);
    #endif
    #ifdef dimensionen4
    void initialisieren_kommunikation(double ** _x0, double ** _y0, double ** _x1, double ** _y1, double ** _x2, double ** _y2, double ** _x3, double ** _y3);
    #endif

    void initialisieren_kommunikation_rest(double ** _u0, double ** _gg_u0, double ** _r0, double ** _f0, double ** _Iu0, double ** _rS0, double ** _e0, double ** _Lu0, double ** _u0_hilf);

    void finalisieren_kommunikation(void);
private:
    int i, j, k, l;
public:
    void packen(void);
    void entpacken(void);
    void bereinigen(void);
    bool ist_fertig(void);
    void kommunikation_starten(void);
    void ausgeben_empfangen(void);
private:
    int a, b, c;
public:
    #ifdef dimensionen1
    double ** y0, ** x0;
    #endif
    #ifdef dimensionen2
    double ** y0, ** x0;
    double ** y1, ** x1;
    #endif
    #ifdef dimensionen3
    double ** y0, ** x0;
    double ** y1, ** x1;
    double ** y2, ** x2;
    #endif
    #ifdef dimensionen4
    double ** y0, ** x0;
    double ** y1, ** x1;
    double ** y2, ** x2;
    double ** y3, ** x3;
    #endif

    double ** u0;
    double ** r0;
    double ** f0;
    double ** Iu0;
    double ** rS0; // rechte Seite
    double ** e0;
    double ** Lu0;
    double ** u0_hilf;
    double ** gg_u0;
    double ** gg_e0;
    double ** gg_v9;
    double ** fg_u0;
    double ** fg_v6;

    double ** v1;
    double ** v2;
    double ** v3;
    double ** v4;
    double ** v5;
    double ** v6;
    double ** v7;
    double ** v8;
    double ** v9;
    double ** v10;
    double ** v11;

    short int ** ep_aktuell;
private:
public:
    double * empfange_werte_buffer;
    double * sende_werte_buffer;
public:

public:
    void werte_kommunikation_starten(void);
public:
    int schreibe_in_sendebuffer(int _sendebuffer_index);
private:
    int sendebuffer_index;
    int empfangsbuffer_index;
public:
    int lese_aus_empfangsbuffer(int _empfangsbuffer_index);
private:
    int a_mod_2, b_mod_2, c_mod_2;
    int f_rs_speichern, g_rs_speichern;
    int f_e_speichern, g_e_speichern;

public:
#ifdef KURZ_TESTFELD
    void schreibe_aussen_testfeld(bool ** testfeld);
    void gebe_testfeld_aus(bool ** testfeld);
#else
    void schreibe_aussen_testfeld(int ** testfeld);
    void gebe_testfeld_aus(int ** testfeld);
#endif
private:
    void verzoegern(void);
public:
#ifdef GEMEINSAME_BUFFER
    void initialisieren_werte(double * _sende_werte_buffer, double * _empfange_werte_buffer);
#else
    void initialisieren_werte(void);
#endif
    void finalisieren_werte(void);
public:
    void initialisieren_kommunikation_ep(short int ** _ep_aktuell);
    void packen_ep(void);
    void entpacken_ep(void);
    bool ist_fertig_ep(void);
    void kommunikation_starten_ep(void);
public:
    short int * f_aussen, * g_aussen, ** testfeld_aussen;
public:
    void bilde_kleine_randstruktur_innen(void);
    void bilde_kleine_randstruktur_aussen(void);
public:
    void packen_werte_u0(void);
    void packen_werte_u0_red(void);
    void packen_werte_u0_black(void);
    void packen_werte_gg_u0(void);
    void packen_werte_fg_u0(void);
    void packen_werte_gg_e0(void);
    void packen_werte_gg_e0_und_u0(void);
    void packen_werte_Lu0(void);
    void packen_werte_r0(void);
    void packen_werte_f0(void);
    void packen_werte_Iu0(void);
    void packen_werte_rS0(void);
    void packen_werte_e0(void);
    void packen_werte_Iu0r0(void);
    void packen_werte_u0r0(void);

    void entpacken_werte_u0(void);
    void entpacken_werte_u0_red(void);
    void entpacken_werte_u0_black(void);
    void entpacken_werte_gg_u0(void);
    void entpacken_werte_fg_u0(void);
    void entpacken_werte_gg_e0(void);
    void entpacken_werte_gg_e0_und_u0(void);
    void entpacken_werte_Lu0(void);
    void entpacken_werte_r0(void);
    void entpacken_werte_f0(void);
    void entpacken_werte_Iu0(void);
    void entpacken_werte_rS0(void);
    void entpacken_werte_e0(void);
    void entpacken_werte_Iu0r0(void);
    void entpacken_werte_u0r0(void);
    void packen_werte_u0_hilf(void);
public:
    int anzahl_punkte_innen;
    int anzahl_punkte_aussen(void);
private:
    int f_rs_i, g_rs_i;
    int f_e_i, g_e_i;
    int c_halbe, a_halbe;
    int c_doppelt, a_doppelt;
    int b_doppelt;
public:
    void erzeuge_aussen_punkte(void);
private:
    int j_durch_2;
    int f_e_i_durch_2;
    int g_e_i_durch_2;
    int f_rs_i_durch_2;
    int g_rs_i_durch_2;
public:
    void packen_werte_fg_v6(void);
    void packen_werte_gg_v9(void);
    void entpacken_werte_gg_v9(void);
    void packen_werte_v6(void);
    void entpacken_werte_v6(void);
    void entpacken_werte_fg_v6(void);
    void packen_werte_v1v2v3v4(void);
    void entpacken_werte_v1v2v3v4(void);
    void packen_werte_v5v6v7(void);
    void entpacken_werte_v5v6v7(void);
    void packen_werte_v8v9(void);
    void entpacken_werte_v8v9(void);
    void packen_werte_v4v6v11(void);
    void entpacken_werte_v4v6v11(void);
    void packen_werte_v6_red(void);
    void entpacken_werte_v6_red(void);
    void packen_werte_v6_black(void);
    void entpacken_werte_v6_black(void);
    };


#endif
