#include "rechnen.h"

#ifdef AUCH_ECHTE_RANDPUNKTE
// Kommentar: Hier wird fr die Punktmenge (_f,_g) und der Anzahl anzahl das Residuum berechnet. Am echten Rand
// ist es nicht definiert. Die u0 - Werte daneben oder darber / darunter werden weggenommen, wenn
// das Gitter seine Grenze erreicht, und der Multiplikationsfaktor fr den mittleren Wert dann um 1 gesenkt.
// Eigentlich ist das Residuum aber nur im Inneren zu berechnen. Die rechte Seite rS0 ist entweder f, oder
// entsprechend des FAS - Schemas zu berechnen.
void Rechnen::residuum_berechnen(short int * _f, short int * _g, int anzahl)
    {
// Kommentar: Die Punktmenge wird durchlaufen.
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];
// Kommentar: Der Faktor fr 4 u0 - Werte drum herum ist 4.0.
        faktor = 4.0;

// Kommentar: Falls die Gittergrenze erreicht wird, wird in der Resiuum - Formel nichts dazuaddiert (hilf_HR = 0),
// aber der faktor wird reduziert. Ansonsten wird der Wert hilf_HR = u0(HR(ff,gg)).
        if((ff==0)||(farbenfeld[ff-1][gg]<0))
            {
            hilf_nord = 0;
            faktor=faktor - 1.0;
            }
        else
            {
            hilf_nord = u0[ff-1][gg];
            }
// Kommentar: Dasselbe nun fr die anderen Himmelsrichtungen Sd,  Ost und West.
        if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
            {
            hilf_sued = 0;
            faktor=faktor - 1.0;
            }
        else
            {
            hilf_sued = u0[ff+1][gg];
            }

        if((gg==0)||(farbenfeld[ff][gg-1]<0))
            {
            hilf_west = 0;
            faktor=faktor - 1.0;
            }
        else
            {
            hilf_west = u0[ff][gg-1];
            }

        if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
            {
            hilf_ost = 0;
            faktor=faktor - 1.0;
            }
        else
            {
            hilf_ost = u0[ff][gg+1];
            }

// Kommentar: Am Ende steht dann die Residuums - Formel, die im Inneren des Gitters exakt mit der Formel
// im Buch von Prof. Joppich bereinstimmt.
        r0[ff][gg] = rS0[ff][gg] - eins_durch_h_xh_y * (-hilf_nord-hilf_sued-hilf_ost-hilf_west+faktor*u0[ff][gg]);
        }
    }
#else
void Rechnen::residuum_berechnen(short int * _f, short int * _g, int anzahl)
    {
// Kommentar: Die Punktmenge wird durchlaufen.
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

// Kommentar: Am Ende steht dann die Residuums - Formel, die im Inneren des Gitters exakt mit der Formel
// im Buch von Prof. Joppich bereinstimmt.
        r0[ff][gg] = rS0[ff][gg] - eins_durch_h_xh_y * (-u0[ff-1][gg]-u0[ff+1][gg]-u0[ff][gg+1]-u0[ff][gg-1]+4.0*u0[ff][gg]);
        }
    }
#endif

// Kommentar: bergeben wird wieder eine Punktmenge. Fr diese Punktmenge wird e0 = u0 - Iu0 auf dem
// grberen Gitter (bei 2 Ebenen) berechnet, vgl. die Formel dieser Funktion.
void Rechnen::korrektur_nach_loesen(short int * _f, short int * _g, int anzahl)
    {
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

        e0[ff][gg] = u0[ff][gg] - Iu0[ff][gg];
        }
    }

// Kommentar: bergeben wird wieder eine Punktmenge. Fr diese Punktmenge wird u0 += e0 auf dem feineren Gitter
// berechnet. Das ist die sogenannte Grobgitterkorrektur.
void Rechnen::korrektur_auf_feinem_gitter(short int * _f, short int * _g, int anzahl)
    {
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];
/*
        if(neuer_punkt->feld[ff][gg]==true)
            {
            continue;
            }
*/
        u0[ff][gg] += e0[ff][gg];
        }
    }


// Kommentar: Auch hier wird eine Punktmenge bergeben. Dafr wird u0 = Iu0 gerechnet.
void Rechnen::kopiere_Iu0_nach_u0(short int * _f, short int * _g, int anzahl)
    {
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

        u0[ff][gg] = Iu0[ff][gg];
        }
    }

void Rechnen::gebe_e0_aus(short int * _f, short int * _g, int anzahl)
    {
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

        if((3<=ff)&&(ff<=13)&&(gg>=4)&&(gg<=7)&&(t==1)&&(gitterebene==1))
            {
            cout << " ausgabe e0 ZR rang " << rang << " t " << t << " ff "  << ff << " gg " << gg << " e0 " << e0[ff][gg] << endl;
            }
        }
    }

// Kommentar: Auf der bergebenen Punktmenge wird e0 = u0 - Iu0 berechnet. Ist aber darunter ein
// neuer Punkt entstanden, dann ist Iu0 unzulssig, und e0 kann nicht nach dieser Formel berechnet werden.
// Ggf. sollte man sie auch nicht berechnen, wo kein Gitterpunkt darunter existiert.
void Rechnen::berechne_e0(short int * _f, short int * _g, int anzahl)
    {
#ifdef AUSGABEN_DATEI_101214
    da_zr->oeffnen();
    * da_zr->aus << " Berechne e0 ff,gg, neuer-punkt " << endl;
#endif
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];
#ifdef AUSGABEN_DATEI_101214
        * da_zr->aus << "(" << ff << "," << gg << ")";
#endif

        if(neuer_punkt->feld[ff+ff][gg+gg]==true)
            {
#ifdef AUSGABEN_DATEI_101214
            * da_zr->aus << "*";
#endif
            e0[ff][gg] = u0[ff][gg]; // Kommentar: So, als wre Iu0 = 0.
            continue;
            }

        e0[ff][gg] = u0[ff][gg] - Iu0[ff][gg];

#ifdef AUSGABEN_DATEI_101214
        if((ff==4)&&(gg==4))
            {
            * da_zr -> aus << " [4,4] u0 " << u0[4][4] << " Iu0 " << Iu0[4][4] << " e0 " << e0[4][4] << endl;
            }
#endif
        }
#ifdef AUSGABEN_DATEI_101214
    * da_zr -> aus << endl;
    da_zr->schliessen();
#endif
    }

#ifdef AUCH_ECHTE_RANDPUNKTE
// Kommentar: Der diskretisierte Differentialoperator wird auf die bergebene Punktmenge angewendet.
// Auch hier wird ein Faktor genommen, der an einem Randpunkt in Bezug auf die Gittermae reduziert wird.
// Im Inneren hat man die Formel Lu0(ff,gg) = 1 / h_x * 1 / h_y * (-u0(ff-1,g)-u0(ff+1,g)-u0(ff,gg-1)-u0(ff,gg+1)+4 u0(ff,gg))
// Ansonsten wird ein Term in Nord, Sd, West oder Ost entfernt, und der Faktor reduziert.
void Rechnen::berechne_Lu0(short int * _f, short int * _g, int anzahl)
    {
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];
        faktor = 4.0;

        if((ff==0)||(farbenfeld[ff-1][gg]<0))
            {
            hilf_nord = 0;
            faktor=faktor - 1.0;
            }
        else
            {
            hilf_nord = u0[ff-1][gg];
            }

        if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
            {
            hilf_sued = 0;
            faktor=faktor - 1.0;
            }
        else
            {
            hilf_sued = u0[ff+1][gg];
            }

        if((gg==0)||(farbenfeld[ff][gg-1]<0))
            {
            hilf_west = 0;
            faktor=faktor - 1.0;
            }
        else
            {
            hilf_west = u0[ff][gg-1];
            }

        if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
            {
            hilf_ost = 0;
            faktor=faktor - 1.0;
            }
        else
            {
            hilf_ost = u0[ff][gg+1];
            }

        Lu0[ff][gg] = eins_durch_h_xh_y * (-hilf_nord-hilf_sued-hilf_ost-hilf_west+faktor*u0[ff][gg]);
        }
    }
#else
void Rechnen::berechne_Lu0(short int * _f, short int * _g, int anzahl)
    {
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

        Lu0[ff][gg] = eins_durch_h_xh_y * (-u0[ff-1][gg]-u0[ff+1][gg]-u0[ff][gg+1]-u0[ff][gg-1]+4.0*u0[ff][gg]);
        }
    }
#endif

// Kommentar: Der Diskretisierungsfehler (LIu - ILu) wird auf die bergebene Punktmenge angewendet.
void Rechnen::berechne_DF0(short int * _f, short int * _g, int anzahl)
    {
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

        if(neuer_punkt->feld[ff+ff][gg+gg]==true)
            {
            continue;
            }

        DF0[ff][gg] = Lu0[ff][gg] - ILu0[ff][gg];
        }
    }

// Kommentar: Es kommt zu einer Glttung, aber nur im Inneren. Die Punktmenge wird wieder bergeben.
void Rechnen::vorglaetter(short int * _f, short int * _g, int anzahl)
    {
// Kommentar: Die Punktmenge wird durchlaufen.
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

#ifdef AUCH_ECHTE_RANDPUNKTE
// Kommentar: Falls in einer der Himmelsrichtungen Ost, Sd, West oder Nord ein echter Rand vorliegt, dann wird
// dort nicht geglttet.
        if((ff==0)||(farbenfeld[ff-1][gg]<0))
            {
            continue;
            }

        if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
            {
            continue;
            }

        if((gg==0)||(farbenfeld[ff][gg-1]<0))
            {
            continue;
            }

        if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
            {
            continue;
            }
#endif

// Kommentar: Der Gltter entspricht einer Jacoby - oder Gauss - Seidel - Glttung, je nachdem in welcher
// Iterationsstufe die u0 - Werte sind. f ist dabei ersetzt durch rS0.
        u0_hilf[ff][gg] = 0.25 * (h_xh_y * rS0[ff][gg] + u0[ff-1][gg] + u0[ff][gg-1] + u0[ff][gg+1] + u0[ff+1][gg]);
        }

#ifdef ZUSATZLAST_RECHNEN
// Kommentar: Zu jedem Glttungs - oder Lsungsschritt wird diese Gleichung mit der Anzahl
// 'wieviel_punktelast' Mal durchlaufen.
    rechnen_nur_als_punktelast(_f, _g, anzahl);
#endif
    }

// Kommentar: Analog zum Vorgltter.
void Rechnen::nachglaetter(short int * _f, short int * _g, int anzahl)
    {
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

#ifdef AUCH_ECHTE_RANDPUNKTE
        if((ff==0)||(farbenfeld[ff-1][gg]<0))
            {
            continue;
            }

        if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
            {
            continue;
            }

        if((gg==0)||(farbenfeld[ff][gg-1]<0))
            {
            continue;
            }

        if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
            {
            continue;
            }
#endif
        u0_hilf[ff][gg] = 0.25 * (h_xh_y * rS0[ff][gg] + u0[ff-1][gg] + u0[ff][gg-1] + u0[ff][gg+1] + u0[ff+1][gg]);
        }
#ifdef ZUSATZLAST_RECHNEN
// Kommentar: Zu jedem Glttungs - oder Lsungsschritt wird diese Gleichung mit der Anzahl
// 'wieviel_punktelast' Mal durchlaufen.
    rechnen_nur_als_punktelast(_f, _g, anzahl);
#endif
    }

// Kommentar: Analog dem vorglaetter, nur mit der Funktion f0 statt rS0.
void Rechnen::vorglaetter_f(short int * _f, short int * _g, int anzahl)
    {
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

#ifdef AUCH_ECHTE_RANDPUNKTE
        if((ff==0)||(farbenfeld[ff-1][gg]<0))
            {
            continue;
            }

        if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
            {
            continue;
            }

        if((gg==0)||(farbenfeld[ff][gg-1]<0))
            {
            continue;
            }

        if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
            {
            continue;
            }
#endif
        u0_hilf[ff][gg] = 0.25 * (h_xh_y * f0[ff][gg] + u0[ff-1][gg] + u0[ff][gg-1] + u0[ff][gg+1] + u0[ff+1][gg]);
        }
#ifdef ZUSATZLAST_RECHNEN
// Kommentar: Zu jedem Glttungs - oder Lsungsschritt wird diese Gleichung mit der Anzahl
// 'wieviel_punktelast' Mal durchlaufen.
    rechnen_nur_als_punktelast(_f, _g, anzahl);
#endif
    }

// Kommentar: Analog zu vorglaetter_f.
void Rechnen::nachglaetter_f(short int * _f, short int * _g, int anzahl)
    {
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

#ifdef AUCH_ECHTE_RANDPUNKTE
        if((ff==0)||(farbenfeld[ff-1][gg]<0))
            {
            continue;
            }

        if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
            {
            continue;
            }

        if((gg==0)||(farbenfeld[ff][gg-1]<0))
            {
            continue;
            }

        if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
            {
            continue;
            }
#endif
        u0_hilf[ff][gg] = 0.25 * (h_xh_y * f0[ff][gg] + u0[ff-1][gg] + u0[ff][gg-1] + u0[ff][gg+1] + u0[ff+1][gg]);
        }
#ifdef ZUSATZLAST_RECHNEN
// Kommentar: Zu jedem Glttungs - oder Lsungsschritt wird diese Gleichung mit der Anzahl
// 'wieviel_punktelast' Mal durchlaufen.
    rechnen_nur_als_punktelast(_f, _g, anzahl);
#endif
    }

// Kommentar: Analog zum vorglaetter.
void Rechnen::loeser(short int * _f, short int * _g, int anzahl)
    {
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

#ifdef AUCH_ECHTE_RANDPUNKTE
        if((ff==0)||(farbenfeld[ff-1][gg]<0))
            {
            continue;
            }

        if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
            {
            continue;
            }

        if((gg==0)||(farbenfeld[ff][gg-1]<0))
            {
            continue;
            }

        if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
            {
            continue;
            }
#endif
        u0_hilf[ff][gg] = 0.25 * (h_xh_y * rS0[ff][gg] + u0[ff-1][gg] + u0[ff][gg-1] + u0[ff][gg+1] + u0[ff+1][gg]);
        }
#ifdef ZUSATZLAST_RECHNEN
// Kommentar: Zu jedem Glttungs - oder Lsungsschritt wird diese Gleichung mit der Anzahl
// 'wieviel_punktelast' Mal durchlaufen.
    rechnen_nur_als_punktelast(_f, _g, anzahl);
#endif
    }

// Kommentar: Analog zum vorglaetter_f
void Rechnen::loeser_f(short int * _f, short int * _g, int anzahl)
    {
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

#ifdef AUCH_ECHTE_RANDPUNKTE
        if((ff==0)||(farbenfeld[ff-1][gg]<0))
            {
            continue;
            }

        if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
            {
            continue;
            }

        if((gg==0)||(farbenfeld[ff][gg-1]<0))
            {
            continue;
            }

        if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
            {
            continue;
            }
#endif
        u0_hilf[ff][gg] = 0.25 * (h_xh_y * f0[ff][gg] + u0[ff-1][gg] + u0[ff][gg-1] + u0[ff][gg+1] + u0[ff+1][gg]);
        }
#ifdef ZUSATZLAST_RECHNEN
// Kommentar: Zu jedem Glttungs - oder Lsungsschritt wird diese Gleichung mit der Anzahl
// 'wieviel_punktelast' Mal durchlaufen.
    rechnen_nur_als_punktelast(_f, _g, anzahl);
#endif
    }

// Kommentar: Beim Jacoby - Verfahren wird in die Variable u0_hilf geschrieben. Nachdem alles berechnet worden ist,
// mssen die Werte nach u0 zurckkopiert werden. Es wird dafr wieder eine Punktmenge bergeben.
void Rechnen::rueckkopieren_u0(short int * _f, short int * _g, int anzahl)
    {
// Kommentar: Mandurchlaufe die Punktmenge.
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

// Kommentar: Und rechnet dort u0 = u0_hilf
        u0[ff][gg] = u0_hilf[ff][gg];
        }
    }

// Kommentar: Eine Funktion zu Debugging - Zwecken: Wieviel Randpunkte entsprechend den Gittermaen hat die
// bergebene Punktmenge? Diese Zahl wird hier bestimmt.
int Rechnen::wieviel_mal_rand(short int * _f, short int * _g, int anzahl)
    {
    wieviel_mal_rand_zahl = 0;

    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

        if(ff==0)
            {
            wieviel_mal_rand_zahl++;
            continue;
            }

        if(ff==n_minus1)
            {
            wieviel_mal_rand_zahl++;
            continue;
            }

        if(gg==0)
            {
            wieviel_mal_rand_zahl++;
            continue;
            }

        if(gg==m_minus1)
            {
            wieviel_mal_rand_zahl++;
            continue;
            }
        }
    return(wieviel_mal_rand_zahl);
    }

// Kommentar: Hier wird die Punktmenge mit einer Zusatzlast belegt.
void Rechnen::rechnen_nur_als_punktelast(short int * _f, short int * _g, int anzahl)
    {
// Kommentar: gitterebene = -10 ist unmglich => es wird rechnen_nur_als_punktelastII aufgerufen.
    if(gitterebene == -10)
        {
        rechnen_nur_als_punktelastI(_f, _g, punkteanzahl);
        }
    else
        {
        rechnen_nur_als_punktelastII(_f, _g, anzahl);
        }
    }

// Kommentar: Diese Funktion hat den Zweck, die Zeitkosten einer vorgegebenen Punktmenge zu bestimmen.
// (f_arbeit,g_arbeit) (i) wurde in der Klasse Programm definiert - genau wie die punkteanzahl.
// Das sollte dem Entwickler nur Laufzeiteffekte klarmachen. Fr Messungen ist das herausgenommen,
// denn gitterebene = -10 ist unmglich.
void Rechnen::rechnen_nur_als_punktelastI(short int * _f, short int * _g, int anzahl)
    {
#ifdef AKTUELLE_WERTE
    cout << " Rechnen nur als Punktelast I rang " << rang << endl;
#endif
// Kommentar: wieviel_punktelast setzt das Gewicht fest, dass in die Gitterpunkte gelegt wird.
    for(k=0; k<wieviel_punktelast; k++)
        {
// Kommentar: Man durchluft f_arbeit(i) und g_arbeit(i).
        for(i=0; i<anzahl; i++)
            {
            ff = f_arbeit[i];
            gg = g_arbeit[i];

#ifdef AUCH_ECHTE_RANDPUNKTE
// Kommentar: Falls ein echter Randpunkt vorliegt, dann gehe zum nchsten Punkt.
            if((ff==0)||(farbenfeld[ff-1][gg]<0))
                {
                y0[ff][gg] = arbeit[ff][gg];
                continue;
                }

            if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
                {
                y0[ff][gg] = arbeit[ff][gg];
                continue;
                }

            if((gg==0)||(farbenfeld[ff][gg-1]<0))
                {
                y0[ff][gg] = arbeit[ff][gg];
                continue;
                }

            if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
                {
                y0[ff][gg] = arbeit[ff][gg];
                continue;
                }
#endif

// Kommentar: Mache einen Rechenschritt.
            y0[ff][gg] = 0.25 * (h_xh_y * f0[ff][gg] + arbeit[ff-1][gg] + arbeit[ff][gg-1] + arbeit[ff][gg+1] + arbeit[ff+1][gg]);
            }

// Kommentar: Und simuliere das zurckkopieren.
        for(i=0; i<anzahl; i++)
            {
            ff = f_arbeit[i];
            gg = g_arbeit[i];

            y0[ff][gg] = arbeit[ff][gg];
            }

        }
    }

// Kommentar: Analog zu rechnen_nur_als_punktelastI, nur mit rechnen auf der bergebenen Punktmenge, statt auf
// f_arbeit und g_arbeit.
void Rechnen::rechnen_nur_als_punktelastII(short int * _f, short int * _g, int anzahl)
    {
#ifdef AKTUELLE_WERTE
    cout << " Rechnen nur als Punktelast II rang " << rang << endl;
#endif
    timer_messen_punktlast.starten();
// Kommentar: wieviel_punktelast setzt das Gewicht fest, dass in die Gitterpunkte gelegt wird.
    for(k=0; k<wieviel_punktelast; k++)
        {
// Kommentar: Man durchluft die bergebene Punktmenge
        for(i=0; i<anzahl; i++)
            {
            ff = _f[i];
            gg = _g[i];

#ifdef AUCH_ECHTE_RANDPUNKTE
// Kommentar: Bei echten Randpunkten geht man ber zum nchsten Punkt.
            if((ff==0)||(farbenfeld[ff-1][gg]<0))
                {
                continue;
                }

            if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
                {
                continue;
                }

            if((gg==0)||(farbenfeld[ff][gg-1]<0))
                {
                continue;
                }

            if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
                {
                continue;
                }
#endif

// Kommentar: Man rechnnet einen Glttungsschritt bzw.. Lsungsschritt.
            y0[ff][gg] = 0.25 * (h_xh_y * f0[ff][gg] + arbeit[ff-1][gg] + arbeit[ff][gg-1] + arbeit[ff][gg+1] + arbeit[ff+1][gg]);
            zaehlen_rechnungsschritte++;
            }

// Kommentar: Man verwendet die Zeit zum zurckkopieren - im Jacoby - Fall.
        for(i=0; i<anzahl; i++)
            {
            ff = _f[i];
            gg = _g[i];

            y0[ff][gg] = arbeit[ff][gg];
            }

        }
    timer_messen_punktlast.stoppen();
    }


// Kommentar: Hier wird die grobe Seite so berechnet, wie es im Buch von Prof. Joppich steht.
// Das Einzige, was anders ist, ist die Behandlung an echten Rndern.
void Rechnen::berechne_grobe_rechte_seite(short int * _f, short int * _g, int anzahl)
// L_H I_h^H u_h + r_H
    {
// Kommentar: Die bergebene Punktmenge wird durchlaufen.
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

// Kommentar: Fr berschreitung der maximalen Gittergrenzen

#ifdef AUCH_ECHTE_RANDPUNKTE
// Kommentar: Falls der Punkt am echten Rand ist => nichts tun.
        if((ff==0)||(farbenfeld[ff-1][gg]<0))
            {
            continue;
            }
        else
            {
            }

        if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
            {
            continue;
            }
        else
            {
            }

        if((gg==0)||(farbenfeld[ff][gg-1]<0))
            {
            continue;
            }
        else
            {
            }

        if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
            {
            continue;
            }
        else
            {
            }
#endif
        f2 = ff << 1;
        g2 = gg << 1;

// Kommentar: Falls ein Gitterpunkt darunter liegt ...
        if((feld_darunter[f2][g2]>=0)&&
            (farbenfeld_darunter[f2+1][g2]>=0)&&
            (farbenfeld_darunter[f2][g2+1]>=0)&&
            (farbenfeld_darunter[f2-1][g2]>=0)&&
            (farbenfeld_darunter[f2][g2-1]>=0))
            {

// Kommentar: Berechne die Rechte Seite nach der Formel 'L_H I_h^H u_h + r_H'.
            rS0[ff][gg] = eins_durch_h_xh_y*(4.0*Iu0[ff][gg] - Iu0[ff][gg+1] - Iu0[ff][gg-1] - Iu0[ff-1][gg] - Iu0[ff+1][gg]) + r0[ff][gg];
#ifdef AKTUELLE_KOMMENTARE_NEU
            if((gitterebene==1)&&(ff==16)&&(gg==20))
                {
                cout << " FGH rS0 1 " << rS0[ff][gg] << " rang " << rang;
                cout << " r0() " << r0[ff][gg] << " Iu0 () " << Iu0[ff][gg];
                cout << " (,+1) " << Iu0[ff][gg+1] << " (0,-1) "  << Iu0[ff][gg-1];
                cout << " (+1,) " << Iu0[ff+1][gg] << " (-1,) " << Iu0[ff-1][gg] << " 1/h_x*h_y " << eins_durch_h_xh_y << endl;
                }
#endif
            }
        else
            {
// Kommentar: Liegt kein Punkt darunter, dann bernehme den Wert von f0.
            rS0[ff][gg] = f0[ff][gg];

// Kommentar: Alles weitere sind Hilfsausgaben frs Debugging.
#ifdef AKTUELLE_KOMMENTARE_NEU
            if((gitterebene==3)&&(ff==2)&&(gg==2))
                {
                cout << " FGH rS0 2 " << rS0[ff][gg] << " rang " << rang << endl;
                }
#endif
            }
        }
    }

// Kommentar: Es wird die grobe rechte Seite berechnet, wie es im Buch von Prof. Joppich steht. Dabei wird von einem vollen Gitter darunter ausgegangen, d.h.
// man braucht im Gegensatz zur Funktion 'berechne_grobe_rechte_seite' nicht das Farbenfeld darunter prfen.
void Rechnen::berechne_grobe_rechte_seite_volles_gitter_darunter(short int * _f, short int * _g, int anzahl)
// L_H I_h^H u_h + r_H
    {
// Kommentar: Durchlauf der bergebenen Punktmenge.
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

#ifdef AUCH_ECHTE_RANDPUNKTE
        if((ff==0)||(farbenfeld[ff-1][gg]<0))
            {
            continue;
            }
        if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
            {
            continue;
            }
        if((gg==0)||(farbenfeld[ff][gg-1]<0))
            {
            continue;
            }

        if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
            {
            continue;
            }
#endif

        f2 = ff << 1;
        g2 = gg << 1;

            {
            rS0[ff][gg] = eins_durch_h_xh_y*(faktor*Iu0[ff][gg] - Iu0[ff][gg+1] - Iu0[ff][gg-1] - Iu0[ff-1][gg] - Iu0[ff+1][gg]) + r0[ff][gg];
#ifdef AKTUELLE_KOMMENTARE_NEU
            if((gitterebene==3)&&(ff==2)&&(gg==2)&&(m>3)&&(n>3))
                {
                cout << " FGH rS0 2 " << rS0[ff][gg] << " rang " << rang;
                cout << " r0() " << r0[ff][gg] << " Iu0 () " << Iu0[ff][gg];
                cout << " (,+1) " << Iu0[ff][gg+1] << " (0,-1) "  << Iu0[ff][gg-1];
                cout << " (+1,) " << Iu0[ff+1][gg] << " (-1,) " << Iu0[ff-1][gg] << " faktor " << faktor << endl;
                }
#endif
            }
        }
    }

//////////////////     Rechnen red

// Kommentar: Man macht den Vorgltter fr alle Punkte mit (ff+gg)&1 = 0.
void Rechnen::vorglaetter_red(short int * _f, short int * _g, int anzahl)
    {
// Kommentar: Man durchluft die bergebene Punktmenge
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

// Kommentar: Nur fr die Punkte mit (ff+gg)&1 = 0 wird gerechnet.
        if(((ff+gg)&1)==1)
            {
            continue;
            }

#ifdef AUCH_ECHTE_RANDPUNKTE
// Kommentar: Bei echten Randpunkten wird nicht geglttet.
        if((ff==0)||(farbenfeld[ff-1][gg]<0))
            {
            // Bei diesem Befehl bin ich mir nicht ganz sicher. 29.08.2014
            continue;
            }

        if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
            {
            continue;
            }

        if((gg==0)||(farbenfeld[ff][gg-1]<0))
            {
            continue;
            }

        if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
            {
            continue;
            }
#endif

 // Kommentar: Die Gauss - Seidel - Glttungsformel wird angewendet.
        u0[ff][gg] = 0.25 * (h_xh_y * rS0[ff][gg] + u0[ff-1][gg] + u0[ff][gg-1] + u0[ff][gg+1] + u0[ff+1][gg]);
#ifdef AKTUELLE_KOMMENTARE_NEU
        if((ff==3)&&(gg==3)&&(gitterebene == 2))
            {
            cout << " QRS (3,3) u0 " << u0[ff][gg] << " rS0 " << rS0[ff][gg] << " u0(-1,) " << u0[ff-1][gg];
            cout << " u0(,-1) " << u0[ff][gg-1] << " u0(1,) " << u0[ff+1][gg] << " u0(,1) " << u0[ff][gg+1] << endl;
            }
#endif
        }

#ifdef ZUSATZLAST_RECHNEN
// Kommentar: Und die Zusatzlast wird behandelt.
//    rechnen_nur_als_punktelast(_f, _g, anzahl);
    rechnen_nur_als_punktelast_red(_f, _g, anzahl); // nderung 27.10.15
#endif
    }

// Kommentar: Entspricht dem vorglaetter_red.
void Rechnen::nachglaetter_red(short int * _f, short int * _g, int anzahl)
    {
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

        if(((ff+gg)&1)==1)
            {
            continue;
            }

#ifdef AUCH_ECHTE_RANDPUNKTE
        if((ff==0)||(farbenfeld[ff-1][gg]<0))
            {
            continue;
            }

        if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
            {
            continue;
            }

        if((gg==0)||(farbenfeld[ff][gg-1]<0))
            {
            continue;
            }

        if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
            {
            continue;
            }
#endif
        u0[ff][gg] = 0.25 * (h_xh_y * rS0[ff][gg] + u0[ff-1][gg] + u0[ff][gg-1] + u0[ff][gg+1] + u0[ff+1][gg]);
#ifdef AKTUELLE_KOMMENTARE_NEU
        if((gitterebene==0)&&(ff==32)&&(gg==36))
            {
            cout << " VVV nachglaetter_red u0(0,0) " << u0[ff][gg] << " rS0 " << rS0[ff][gg] << " u0(-1,0) " << u0[ff-1][gg] << " u0(0,-1) " << u0[ff][gg-1] << " u0(0,+1) " << u0[ff][gg+1];
            cout << " u0(+1,0) " << u0[ff+1][gg] << endl;
            }
#endif
        }
#ifdef ZUSATZLAST_RECHNEN
    rechnen_nur_als_punktelast_red(_f, _g, anzahl); // nderung 27.10.15
#endif
    }

// Kommentar: Entspricht der Funktion vorglaetter_red, bis auf die Formel, wo f0 statt rS0 genommen wird.
void Rechnen::vorglaetter_f_red(short int * _f, short int * _g, int anzahl)
    {
// Kommentar: Man durchlaufe die Punktmenge.
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

        if(((ff+gg)&1)==1)
            {
            continue;
            }

#ifdef AUCH_ECHTE_RANDPUNKTE
// Kommentar: Bei echten Randpunkten glttet man nicht.
        if((ff==0)||(farbenfeld[ff-1][gg]<0))
            {
            continue;
            }

        if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
            {
            continue;
            }

        if((gg==0)||(farbenfeld[ff][gg-1]<0))
            {
            continue;
            }

        if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
            {
            continue;
            }
#endif

// Kommentar: Die Formel zur Glttung, aber mit f0 statt rS0.
        u0[ff][gg] = 0.25 * (h_xh_y * f0[ff][gg] + u0[ff-1][gg] + u0[ff][gg-1] + u0[ff][gg+1] + u0[ff+1][gg]);
        }
#ifdef ZUSATZLAST_RECHNEN
    rechnen_nur_als_punktelast_red(_f, _g, anzahl); // nderung 27.10.15
#endif
    }

// Kommentar: Analog zu vorglaetter_f_red.
void Rechnen::nachglaetter_f_red(short int * _f, short int * _g, int anzahl)
    {
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

        if(((ff+gg)&1)==1)
            {
            continue;
            }

#ifdef AUCH_ECHTE_RANDPUNKTE
// Kommentar: Bei echten Randpunkten glttet man nicht.
        if((ff==0)||(farbenfeld[ff-1][gg]<0))
            {
            continue;
            }

        if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
            {
            continue;
            }

        if((gg==0)||(farbenfeld[ff][gg-1]<0))
            {
            continue;
            }

        if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
            {
            continue;
            }
#endif
        u0[ff][gg] = 0.25 * (h_xh_y * f0[ff][gg] + u0[ff-1][gg] + u0[ff][gg-1] + u0[ff][gg+1] + u0[ff+1][gg]);
        }
#ifdef ZUSATZLAST_RECHNEN
    rechnen_nur_als_punktelast_red(_f, _g, anzahl); // nderung 27.10.15
#endif
    }

// Kommentar: Analog zu vorglaetter_red, auch die Kommentare.
void Rechnen::loeser_red(short int * _f, short int * _g, int anzahl)
    {
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

        if(((ff+gg)&1)==1)
            {
            continue;
            }

#ifdef AUCH_ECHTE_RANDPUNKTE
        if((ff==0)||(farbenfeld[ff-1][gg]<0))
            {
            continue;
            }

        if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
            {
            continue;
            }

        if((gg==0)||(farbenfeld[ff][gg-1]<0))
            {
            continue;
            }

        if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
            {
            continue;
            }
#endif
        u0[ff][gg] = 0.25 * (h_xh_y * rS0[ff][gg] + u0[ff-1][gg] + u0[ff][gg-1] + u0[ff][gg+1] + u0[ff+1][gg]);
        }
#ifdef ZUSATZLAST_RECHNEN
    rechnen_nur_als_punktelast_red(_f, _g, anzahl); // nderung 27.10.15
#endif
    }

// Kommentar: Genau wie vorglaetter_f_red. Also auch dieselben Kommentare.
void Rechnen::loeser_f_red(short int * _f, short int * _g, int anzahl)
    {
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

        if(((ff+gg)&1)==1)
            {
            continue;
            }

#ifdef AUCH_ECHTE_RANDPUNKTE
        if((ff==0)||(farbenfeld[ff-1][gg]<0))
            {
            continue;
            }

        if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
            {
            continue;
            }

        if((gg==0)||(farbenfeld[ff][gg-1]<0))
            {
            continue;
            }

        if(gg==m_minus1)
            {
            u0[ff][gg] = f0[ff][gg];
            continue;
            }
#endif
        u0[ff][gg] = 0.25 * (h_xh_y * f0[ff][gg] + u0[ff-1][gg] + u0[ff][gg-1] + u0[ff][gg+1] + u0[ff+1][gg]);
        }
#ifdef ZUSATZLAST_RECHNEN
    rechnen_nur_als_punktelast_red(_f, _g, anzahl); // nderung 27.10.15
#endif
    }

//////////////////     Rechnen black

// Kommentar: Analog zu vorglaetter_red. Der Unterschied wird kommentiert.
void Rechnen::vorglaetter_black(short int * _f, short int * _g, int anzahl)
    {
#ifdef TIMER_INFORMATIONEN
            strcpy(zeichenkette,"Vor der FeldGlttung");
            zeitausgabe(zeichenkette);
#endif
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

// Kommentar: Diese Zeile weicht ab. Hier werden nur die Punkte mit ungerader Koordinatensumme gerechnet.
        if(((ff+gg)&1)==0)
            {
            continue;
            }

#ifdef AUCH_ECHTE_RANDPUNKTE
        if((ff==0)||(farbenfeld[ff-1][gg]<0))
            {
            continue;
            }

        if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
            {
            continue;
            }

        if((gg==0)||(farbenfeld[ff][gg-1]<0))
            {
            continue;
            }

        if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
            {
            continue;
            }
#endif
        u0[ff][gg] = 0.25 * (h_xh_y * rS0[ff][gg] + u0[ff-1][gg] + u0[ff][gg-1] + u0[ff][gg+1] + u0[ff+1][gg]);

#ifdef AKTUELLE_KOMMENTARE_NEU
        if((ff==4)&&(gg==3)&&(gitterebene == 2))
            {
            cout << " QRS (4,3) u0 " << u0[ff][gg] << " rS0 " << rS0[ff][gg] << " u0(-1,) " << u0[ff-1][gg];
            cout << " u0(,-1) " << u0[ff][gg-1] << " u0(1,) " << u0[ff+1][gg] << " u0(,1) " << u0[ff][gg+1] << endl;
            }
#endif
        }

#ifdef TIMER_INFORMATIONEN
            strcpy(zeichenkette,"Vor der Punktlast");
            zeitausgabe(zeichenkette);
#endif
#ifdef ZUSATZLAST_RECHNEN
    rechnen_nur_als_punktelast_black(_f, _g, anzahl); // nderung 27.10.15
#endif
#ifdef TIMER_INFORMATIONEN
            strcpy(zeichenkette,"Nach der Punktlast");
            zeitausgabe(zeichenkette);
#endif
    }

// Kommentar: Das ist dasselbe wie vorglaetter_black, also dieselben Kommentare.
void Rechnen::nachglaetter_black(short int * _f, short int * _g, int anzahl)
    {
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

        if(((ff+gg)&1)==0)
            {
            continue;
            }

#ifdef AUCH_ECHTE_RANDPUNKTE
        if((ff==0)||(farbenfeld[ff-1][gg]<0))
            {
            continue;
            }

        if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
            {
            continue;
            }

        if((gg==0)||(farbenfeld[ff][gg-1]<0))
            {
            continue;
            }

        if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
            {
            continue;
            }
#endif
        u0[ff][gg] = 0.25 * (h_xh_y * rS0[ff][gg] + u0[ff-1][gg] + u0[ff][gg-1] + u0[ff][gg+1] + u0[ff+1][gg]);
        }
#ifdef ZUSATZLAST_RECHNEN
    rechnen_nur_als_punktelast_black(_f, _g, anzahl); // nderung 27.10.15
#endif
    }

// Kommentar: Analog zu vorglaetter_f_red. Der Unterschied wird kommentiert.
void Rechnen::vorglaetter_f_black(short int * _f, short int * _g, int anzahl)
    {
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

// Kommentar: Diese Zeile weicht ab. Hier werden nur die Punkte mit ungerader Koordinatensumme gerechnet.
        if(((ff+gg)&1)==0)
            {
            continue;
            }

#ifdef AUCH_ECHTE_RANDPUNKTE
        if((ff==0)||(farbenfeld[ff-1][gg]<0))
            {
            continue;
            }

        if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
            {
            continue;
            }

        if((gg==0)||(farbenfeld[ff][gg-1]<0))
            {
            continue;
            }

        if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
            {
            continue;
            }
#endif

        u0[ff][gg] = 0.25 * (h_xh_y * f0[ff][gg] + u0[ff-1][gg] + u0[ff][gg-1] + u0[ff][gg+1] + u0[ff+1][gg]);
        }
#ifdef ZUSATZLAST_RECHNEN
    rechnen_nur_als_punktelast_black(_f, _g, anzahl); // nderung 27.10.15
#endif
    }

// Kommentar: Die Funktion ist identisch zu vorglaetter_f_black. Also ist die Kommentierung gleich.
void Rechnen::nachglaetter_f_black(short int * _f, short int * _g, int anzahl)
    {
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

        if(((ff+gg)&1)==0)
            {
            continue;
            }

#ifdef AUCH_ECHTE_RANDPUNKTE
        if((ff==0)||(farbenfeld[ff-1][gg]<0))
            {
            continue;
            }

        if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
            {
            continue;
            }

        if((gg==0)||(farbenfeld[ff][gg-1]<0))
            {
            continue;
            }

        if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
            {
            continue;
            }
#endif
        u0[ff][gg] = 0.25 * (h_xh_y * f0[ff][gg] + u0[ff-1][gg] + u0[ff][gg-1] + u0[ff][gg+1] + u0[ff+1][gg]);
        }
#ifdef ZUSATZLAST_RECHNEN
    rechnen_nur_als_punktelast_black(_f, _g, anzahl); // nderung 27.10.15
#endif
    }

// Kommentar: Identisch wie vorglaetter_black. Daher nehme man die Kommentare dieser Funktion.
void Rechnen::loeser_black(short int * _f, short int * _g, int anzahl)
    {
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

        if(((ff+gg)&1)==0)
            {
            continue;
            }

#ifdef AUCH_ECHTE_RANDPUNKTE
        if((ff==0)||(farbenfeld[ff-1][gg]<0))
            {
            continue;
            }

        if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
            {
            continue;
            }

        if((gg==0)||(farbenfeld[ff][gg-1]<0))
            {
            continue;
            }

        if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
            {
            continue;
            }
#endif
        u0[ff][gg] = 0.25 * (h_xh_y * rS0[ff][gg] + u0[ff-1][gg] + u0[ff][gg-1] + u0[ff][gg+1] + u0[ff+1][gg]);
        }
#ifdef ZUSATZLAST_RECHNEN
    rechnen_nur_als_punktelast_black(_f, _g, anzahl); // nderung 27.10.15
#endif
    }

// Kommentar: Identisch wie vorglaetter_f_black.
void Rechnen::loeser_f_black(short int * _f, short int * _g, int anzahl)
    {
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

        if(((ff+gg)&1)==0)
            {
            continue;
            }

#ifdef AUCH_ECHTE_RANDPUNKTE
        if((ff==0)||(farbenfeld[ff-1][gg]<0))
            {
            continue;
            }

        if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
            {
            continue;
            }

        if((gg==0)||(farbenfeld[ff][gg-1]<0))
            {
            continue;
            }

        if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
            {
            continue;
            }
#endif
        u0[ff][gg] = 0.25 * (h_xh_y * f0[ff][gg] + u0[ff-1][gg] + u0[ff][gg-1] + u0[ff][gg+1] + u0[ff+1][gg]);
        }
#ifdef ZUSATZLAST_RECHNEN
    rechnen_nur_als_punktelast_black(_f, _g, anzahl); // nderung 27.10.15
#endif
    }

// Kommentar: u0 wird auf der bergebenen Punktmenge genullt. Selbsterklrend. Wird gebraucht beim normalen
// Mehrgitterschema.
void Rechnen::reset_u0(short int * _f, short int * _g, int anzahl)
    {
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

        u0[ff][gg] = 0;
        }
    }

// Kommentar: e0 wird auf der bergebenen Punktmenge durch u0 berschrieben.
void Rechnen::kopiere_u0_e0(short int * _f, short int * _g, int anzahl)
    {
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

        e0[ff][gg] = u0[ff][gg];
        }
    }

// Kommentar: Beim FAS - Schema kann der Diskretisierungsfehler als Differenz von rS0 und If0 berechnet werden.
// Das geht allerdings nur fr eine Gitterebene, vgl. die Dissertation. Nach einer Idee von Herrn Joppich.
// Allerdings ist es vom Autor umgerechnet.
void Rechnen::berechne_tau(short int * _f, short int * _g, int anzahl)
    {
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

        tau[ff][gg] = rS0[ff][gg] - If0[ff][gg];
#ifdef AUSGABE_GHIJ
        cout << " GHIJ tau ff " << ff << " gg " << gg << " tau " << tau[ff][gg] << " rS0 " << rS0[ff][gg] << " If0 " << If0[ff][gg] << " r0 " << r0[ff][gg] << " t " << t << endl;
#endif
        }
    }

// Kommentar: Dieses Verfeinerungskriterium ist vom Autor der Dissertation entwickelt worden.
void Rechnen::berechne_tau12(short int * _f, short int * _g, int anzahl)
    {
// Kommentar: Die bergebene Punktmenge wird durchlaufen.
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

// Kommentar: Das Kriterium kann nur fr gemeinsame Punkte berechnet werden.
        if((ff&1)==1)
            {
            continue;
            }
        if((gg&1)==1)
            {
            continue;
            }

// Kommentar: An den maximalen Grenzen des Gitters kann es nicht berechnet werden. Aber: Punkte, die gemeinsame Punkte sind, und nicht am echten Rand liegen, haben immer
// einen Mindestabstand 2 zur maximalen Grenze. Daher ist die folgende 'if'-Abfrage berflssig.
#ifdef UNMOEGLICH
        if((ff<2)||(gg<2)||(ff>=f_minus1)||(gg>=g_minus1))
            {
            continue;
            }
#endif

// Kommentar: Der Diskretisierungsfehler - horizontal und vertikal - wird bestimmt durch die Abweichung eines
// kubischen Polynoms von den Werten in (ff,gg).
// Bemerkung: Da (ff,gg) ein gemeinsamer Punkt ist, der kein echter Randpunkt ist, existieren die Punkte (ff-2,gg), (ff+2,gg), (ff,gg-2) und (ff,gg+2). Denn gemeinsame Punkte knnen sich nicht
// im Abstand 1 vom echten Rand befinden.
        tau_1[ff][gg] = zwei_drittel*(u0[ff-1][gg]+u0[ff+1][gg]) + ein_sechstel*(u0[ff-2][gg]+u0[ff+2][gg]) - u0[ff][gg];
        tau_2[ff][gg] = zwei_drittel*(u0[ff][gg-1]+u0[ff][gg+1]) + ein_sechstel*(u0[ff][gg-2]+u0[ff][gg+2]) - u0[ff][gg];
        }
    }

// Kommentar: Es geht hierbei um eine zeitliche Simulation der Implementierung von Herrn Hess der Shallow Water Equations. Hier geschieht
// ein Teil einer Abwandlung seiner Funktion 'solve_pre'.
void Rechnen::loesen_vorher_teil1(short int * _f, short int * _g, int anzahl)
    {
// Kommentar: Die Punktmenge wird durchlaufen. Bei der Implementierung von Herrn Hess der SWE gibt es Unterschiede in den
// Punktmengen in Bezug auf den Rand.
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

// Kommentar: Die folgenden Rechnungen werden durchgefhrt:

// Bereich1
        var1 = V1_fg_1;
        var2 = V2_fg_1;
        var3 = V3_fg_1;
// Bereich2
        var4 = V1_fg_2;
        var5 = V1_fg_3;
        var6 = V1_fg_4;

        var7 = V2_fg_2;
        var8 = V2_fg_3;

        var9 = V3_fg_2;
        var10 = V3_fg_3;

        var11 = V4_fg_2;
        var12 = V4_fg_3;
// Bereich3
        v5[ff][gg] = 0.12* (var1 + var4 * var5 + konstante2 * var6 * var9);
        v5[ff][gg] -= 0.13* (var1 + var4 * var6 + konstante2 * var6 * var10);
        v5[ff][gg] -= var1;

        v6[ff][gg] = 0.162* (var2 + var4 * var7 + konstante2 * var6 * var9);
        v6[ff][gg] -= 0.163* (var2 + var4 * var8 + konstante2 * var5 * var11);
        v6[ff][gg] -= var2;

        v7[ff][gg] = 0.172* (var3 + var5 * var6 + konstante2 * var4 * var12);
        v7[ff][gg] -= var1;
        }
    }

// Kommentar: Es geht hierbei um eine zeitliche Simulation der Implementierung von Herrn Hess der Shallow Water Equations. Hier geschieht
// ein Teil einer Abwandlung seiner Funktion 'solve_pre'.
void Rechnen::loesen_vorher_teil2(short int * _f, short int * _g, int anzahl)
    {
// Kommentar: Die Punktmenge wird durchlaufen. Bei der Implementierung von Herrn Hess der SWE gibt es Unterschiede in den
// Punktmengen in Bezug auf den Rand.
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

// Kommentar: Die folgenden Rechnungen werden durchgefhrt:
        var1 = V5_fg_1;
        var2 = V6_fg_1;

        v8[ff][gg] = 0.8 * var1*V5_fg_2;
        v8[ff][gg] -= 0.8 * var2*V5_fg_3;
        v8[ff][gg] -= var1*konstante1*konstante2+ konstante3 * konstante4;
        v9[ff][gg] = 0.9 * var1 * V6_fg_2;
        v9[ff][gg] -= 0.9 * var2 * V6_fg_3;
        v9[ff][gg] -= var2 * konstante2*konstante3 + konstante4 * konstante1;
        v10[ff][gg] = 1.01 * var1*V7_fg_2;
        v10[ff][gg] -= 1.01 * var1*V7_fg_3;
        }
    }

// Kommentar: Es geht hierbei um eine zeitliche Simulation der Implementierung von Herrn Hess der Shallow Water Equations. Hier geschieht
// ein Teil einer Abwandlung seiner Funktion 'solve_pre'.
void Rechnen::loesen_vorher_teil3(short int * _f, short int * _g, int anzahl)
    {
// Kommentar: Die Punktmenge wird durchlaufen. Bei der Implementierung von Herrn Hess der SWE gibt es Unterschiede in den
// Punktmengen in Bezug auf den Rand.
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

// Kommentar: Die folgenden Rechnungen werden durchgefhrt:
        v8[ff][gg] -= 0.8 * V7_fg_2 + v4[ff+1][gg] - v4[ff-1][gg];
        v9[ff][gg] = 0.9 * V7_fg_3 + v4[ff][gg+1] - v4[ff][gg-1];
        v10[ff][gg] -= 1.01 * v1[ff][gg]*V5_fg_4;
        v10[ff][gg] -= 1.01 * v1[ff][gg]*V6_fg_4;
        }
    }

// Kommentar: Es geht hierbei um eine zeitliche Simulation der Implementierung von Herrn Hess der Shallow Water Equations. Hier geschieht
// ein Teil einer Abwandlung seiner Funktion 'solve_pre'.
void Rechnen::loesen_vorher_teil4(short int * _f, short int * _g, int anzahl)
    {
// Kommentar: Die Punktmenge wird durchlaufen. Bei der Implementierung von Herrn Hess der SWE gibt es Unterschiede in den
// Punktmengen in Bezug auf den Rand.
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

// Kommentar: Die folgenden Rechnungen werden durchgefhrt:
        var6 = (v3[ff-1][gg] + v6[ff-1][gg])*konstante1 + (v3[ff][gg] + v6[ff][gg])*konstante2 + (v3[ff+1][gg] + v6[ff+1][gg])*konstante3;
        var6 -= (v3[ff][gg-1] + v6[ff][gg-1])*konstante1 + (v3[ff][gg] + v6[ff][gg])*konstante2 + (v3[ff][gg+1] + v6[ff][gg+1])*konstante3 + v3[ff][gg]+v6[ff][gg];
        var6 -= v2[ff-1][gg] + v2[ff+1][gg] * konstante1 * konstante3 * vf[ff] * vf[ff];
        var6 -= v3[ff][gg-1] + v3[ff][gg+1] * konstante2 * konstante4 * vf[ff] * vf[ff];
//        var6 -= var5[ff][gg] * var1[ff][gg];
        var6 -= var5 * var1;
        }
    }

// Kommentar: Es geht hierbei um eine zeitliche Simulation der Implementierung von Herrn Hess der Shallow Water Equations. Hier geschieht
// ein Teil einer Abwandlung seiner Funktion 'solve_pre'.
void Rechnen::loesen_vorher_teil5(short int * _f, short int * _g, int anzahl)
    {
// Kommentar: Die Punktmenge wird durchlaufen. Bei der Implementierung von Herrn Hess der SWE gibt es Unterschiede in den
// Punktmengen in Bezug auf den Rand.
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

// Kommentar: Die folgenden Rechnungen werden durchgefhrt:
        var6 -= v8[ff-1][gg] + v8[ff+1][gg] * konstante1 * konstante2 * vf[ff] * vf[ff];
        var6 -= v9[ff][gg-1] + v9[ff][gg+1] * konstante3 * konstante4 * vf[ff] * vf[ff];
        }
    }

// Kommentar: Es geht hierbei um eine zeitliche Simulation der Implementierung von Herrn Hess der Shallow Water Equations. Hier geschieht
// ein Teil einer Abwandlung seiner Funktion 'solve_pre'.
void Rechnen::loesen_vorher_teil6(short int * _f, short int * _g, int anzahl)
    {
// Kommentar: Die Punktmenge wird durchlaufen. Bei der Implementierung von Herrn Hess der SWE gibt es Unterschiede in den
// Punktmengen in Bezug auf den Rand.
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

// Kommentar: Die folgenden Rechnungen werden durchgefhrt:
        var10 -= v1[ff-1][gg] + v1[ff+1][gg] * konstante1 * konstante4 * vf[ff];
        var10 -= v2[ff][gg-1] + v2[ff][gg+1] * konstante2 * konstante3 * vf[ff];
        }
    }

// Kommentar: Es geht hierbei um eine zeitliche Simulation der Implementierung von Herrn Hess der Shallow Water Equations. Hier geschieht
// ein Teil einer Abwandlung seiner Funktion 'solve_post'.
void Rechnen::loesen_nachher_teil1(short int * _f, short int * _g, int anzahl)
    {
// Kommentar: Die Punktmenge wird durchlaufen. Bei der Implementierung von Herrn Hess der SWE gibt es Unterschiede in den
// Punktmengen in Bezug auf den Rand.
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

// Kommentar: Die folgenden Rechnungen werden durchgefhrt:
        v4[ff][gg] += (v4[ff-1][gg] - v6[ff-1][gg] + v11[ff-1][gg])*konstante1 - (v4[ff+1][gg] - v6[ff+1][gg] + v11[ff+1][gg])*konstante2;
        v5[ff][gg] += (v4[ff][gg-1] - v6[ff][gg-1] + v11[ff][gg-1])*konstante3 - (v4[ff][gg+1] - v6[ff][gg+1] + v11[ff][gg+1])*konstante4;
        }
    }

// Kommentar: Es geht hierbei um eine zeitliche Simulation der Implementierung von Herrn Hess der Shallow Water Equations. Hier geschieht
// ein Teil einer Abwandlung seiner Funktion 'solve_post'.
void Rechnen::loesen_nachher_teil2(short int * _f, short int * _g, int anzahl)
    {
// Kommentar: Die Punktmenge wird durchlaufen. Bei der Implementierung von Herrn Hess der SWE gibt es Unterschiede in den
// Punktmengen in Bezug auf den Rand.
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

// Kommentar: laden ist ein 2 - dimensionales Feld. Die Werte dieses Feldes werden zu v1, v2 und bv3 hinzuaddiert.
        v1[ff][gg] += laden[ff][gg];
        v2[ff][gg] += laden[ff][gg];
        v3[ff][gg] += laden[ff][gg];
        }
    }

// Kommentar: Diese Funktion wird nicht aufgerufen, und daher auch nicht kommentiert.
void Rechnen::glaetten_swe(short int * _f, short int * _g, int anzahl)
    {
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

#ifdef AUCH_ECHTE_RANDPUNKTE
        if((ff==0)||(farbenfeld[ff-1][gg]<0))
            {
            continue;
            }

        if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
            {
            continue;
            }

        if((gg==0)||(farbenfeld[ff][gg-1]<0))
            {
            continue;
            }

        if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
            {
            continue;
            }
#endif
        v6[ff][gg] = (v6[ff-1][gg]+v6[ff+1][gg]+v6[ff][gg-1]+v6[ff][gg+1]+vf_2[2*ff-1]+vf_2[2*ff]+vf_2[2*ff+1]+v7[ff][gg]+v8[ff][gg])/(konstante2+konstante3);
        }
    }

// Kommentar: Diese Funktion macht eine Black-Glttung der von mir genderten Version der Implementierung von Herrn Hess der Shallow Water Equations.
void Rechnen::glaetten_swe_black(short int * _f, short int * _g, int anzahl)
    {
// Kommentar: Ein Durchlauf durch die bergebene Punktmenge.
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

// Kommentar: Werte fr Red - Punkte werden nicht berechnet. Die Koodinatensumme muss ungerade sein.
        if(((ff+gg)&1)==0)
            {
            continue;
            }

#ifdef AUCH_ECHTE_RANDPUNKTE
// Kommentar: Fr echte Randpunkte wird nicht geglttet.
        if((ff==0)||(farbenfeld[ff-1][gg]<0))
            {
            continue;
            }

        if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
            {
            continue;
            }

        if((gg==0)||(farbenfeld[ff][gg-1]<0))
            {
            continue;
            }

        if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
            {
            continue;
            }
#endif

// Kommentar: Der Glttungsoperator wird aufgerufen. Er ist von mir von dem Glttungsoperator der Implementierung von Herrn Hess der Shallow Water Equations abgewandelt.
        v6[ff][gg] = (v6[ff-1][gg]+v6[ff+1][gg]+v6[ff][gg-1]+v6[ff][gg+1]+vf_2[2*ff-1]+vf_2[2*ff]+vf_2[2*ff+1]+v7[ff][gg]+v8[ff][gg])/(konstante2+konstante3);
#ifdef AKTUELLE_KOMMENTARE_NEU
        if((ff==4)&&(gg==3)&&(gitterebene == 2))
            {
            cout << " QRS (4,3) u0 " << u0[ff][gg] << " rS0 " << rS0[ff][gg] << " u0(-1,) " << u0[ff-1][gg];
            cout << " u0(,-1) " << u0[ff][gg-1] << " u0(1,) " << u0[ff+1][gg] << " u0(,1) " << u0[ff][gg+1] << endl;
            }
#endif
        }
    }

// Kommentar: Analog zu glaetten_swe_black. Es werden hier nur red-Punkte berechnet.
void Rechnen::glaetten_swe_red(short int * _f, short int * _g, int anzahl)
    {
    for(i=0; i<anzahl; i++)
        {
        ff = _f[i];
        gg = _g[i];

// Kommentar: Die Koordinatensumme muss gerade sein. Damit fr keine Black Punkte gerechnet wird.
        if(((ff+gg)&1)==1)
            {
            continue;
            }

#ifdef AUCH_ECHTE_RANDPUNKTE
        if((ff==0)||(farbenfeld[ff-1][gg]<0))
            {
            continue;
            }

        if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
            {
            continue;
            }

        if((gg==0)||(farbenfeld[ff][gg-1]<0))
            {
            continue;
            }

        if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
            {
            continue;
            }
#endif
// Kommentar: Der Glttungsoperator wird aufgerufen. Er ist von mir von dem Glttungsoperator der Implementierung von Herrn Hess der Shallow Water Equations abgewandelt.
        v6[ff][gg] = (v6[ff-1][gg]+v6[ff+1][gg]+v6[ff][gg-1]+v6[ff][gg+1]+vf_2[2*ff-1]+vf_2[2*ff]+vf_2[2*ff+1]+v7[ff][gg]+v8[ff][gg])/(konstante2+konstante3);
#ifdef AKTUELLE_KOMMENTARE_NEU
        if((ff==4)&&(gg==3)&&(gitterebene == 2))
            {
            cout << " QRS (4,3) u0 " << u0[ff][gg] << " rS0 " << rS0[ff][gg] << " u0(-1,) " << u0[ff-1][gg];
            cout << " u0(,-1) " << u0[ff][gg-1] << " u0(1,) " << u0[ff+1][gg] << " u0(,1) " << u0[ff][gg+1] << endl;
            }
#endif
        }
    }

// Kommentar: Die Punktlast im Falle, dass fr red Punkte eine zustzlich Last gebraucht wird.
void Rechnen::rechnen_nur_als_punktelast_red(short int * _f, short int * _g, int anzahl)
    {
#ifdef AKTUELLE_WERTE
    cout << " Rechnen nur als Punktelast II rang " << rang << endl;
#endif
    timer_messen_punktlast.starten();
// Kommentar: wieviel_punktelast setzt das Gewicht fest, dass in die Gitterpunkte gelegt wird.
    for(k=0; k<wieviel_punktelast; k++)
        {
// Kommentar: Man durchluft die bergebene Punktmenge
        for(i=0; i<anzahl; i++)
            {
            ff = _f[i];
            gg = _g[i];

            if(((ff+gg)&1)==1)
                {
                continue;
                }
#ifdef AUCH_ECHTE_RANDPUNKTE
// Kommentar: Bei echten Randpunkten geht man ber zum nchsten Punkt.
            if((ff==0)||(farbenfeld[ff-1][gg]<0))
                {
                continue;
                }

            if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
                {
                continue;
                }

            if((gg==0)||(farbenfeld[ff][gg-1]<0))
                {
                continue;
                }

            if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
                {
                continue;
                }
#endif

// Kommentar: Man rechnnet einen Glttungsschritt bzw.. Lsungsschritt.
            y0[ff][gg] = 0.25 * (h_xh_y * f0[ff][gg] + arbeit[ff-1][gg] + arbeit[ff][gg-1] + arbeit[ff][gg+1] + arbeit[ff+1][gg]);
            zaehlen_rechnungsschritte++;
            }

// Kommentar: Man verwendet die Zeit zum zurckkopieren - im Jacoby - Fall.
        for(i=0; i<anzahl; i++)
            {
            ff = _f[i];
            gg = _g[i];

            if(((ff+gg)&1)==1)
                {
                continue;
                }

            y0[ff][gg] = arbeit[ff][gg];
            }

        }
    timer_messen_punktlast.stoppen();
    }

// Kommentar: Die Punktlast im Falle, dass fr black Punkte eine zustzlich Last gebraucht wird.
void Rechnen::rechnen_nur_als_punktelast_black(short int * _f, short int * _g, int anzahl)
    {
#ifdef AKTUELLE_WERTE
    cout << " Rechnen nur als Punktelast II rang " << rang << endl;
#endif
    timer_messen_punktlast.starten();
// Kommentar: wieviel_punktelast setzt das Gewicht fest, dass in die Gitterpunkte gelegt wird.
    for(k=0; k<wieviel_punktelast; k++)
        {
// Kommentar: Man durchluft die bergebene Punktmenge
        for(i=0; i<anzahl; i++)
            {
            ff = _f[i];
            gg = _g[i];

            if(((ff+gg)&1)==0)
                {
                continue;
                }

#ifdef AUCH_ECHTE_RANDPUNKTE
// Kommentar: Bei echten Randpunkten geht man ber zum nchsten Punkt.
            if((ff==0)||(farbenfeld[ff-1][gg]<0))
                {
                continue;
                }

            if((ff==n_minus1)||(farbenfeld[ff+1][gg]<0))
                {
                continue;
                }

            if((gg==0)||(farbenfeld[ff][gg-1]<0))
                {
                continue;
                }

            if((gg==m_minus1)||(farbenfeld[ff][gg+1]<0))
                {
                continue;
                }
#endif

// Kommentar: Man rechnnet einen Glttungsschritt bzw.. Lsungsschritt.
            y0[ff][gg] = 0.25 * (h_xh_y * f0[ff][gg] + arbeit[ff-1][gg] + arbeit[ff][gg-1] + arbeit[ff][gg+1] + arbeit[ff+1][gg]);
            zaehlen_rechnungsschritte++;
            }

// Kommentar: Man verwendet die Zeit zum zurckkopieren - im Jacoby - Fall.
        for(i=0; i<anzahl; i++)
            {
            ff = _f[i];
            gg = _g[i];

            if(((ff+gg)&1)==0)
                {
                continue;
                }
            y0[ff][gg] = arbeit[ff][gg];
            }

        }
    timer_messen_punktlast.stoppen();
    }
