#include "uebertrage_gitterwerte.h"

// Was macht diese Klasse? Ist sie Bestandteil der adaptiven Verfeinerung, oder ein einzelner Teil
// in der Klasse Programm?
// Welche Gitterwerte werden bertragen?
// Ich vermute, es geht hierbei um die Kappe. Das sieht wirklich so aus. Es ist in
// Koordinaten_und_Werte_sammeln, und das ist Teil von grober_Zyklus.

// Kommentar: Diese Klasse dient dem bertragen von Gitterwerten bei der Kappe: Es wird 'gesammelt' von allen anderen Kernen auf den
// mit der ID 0. Und umgekehrt wird von dem Kern mit der ID 0 'zerstreut' auf die anderen Kerne. Dabei sind die Punktmengen in beiden
// Flle dieselben. Es kann aber zwischen verschiedenen Variablen bertragen werden.
// Es werden also Werte gepackt und entpackt, sowohl beim Sammeln wie Zerstreuen. Und es werden die Zeiger auf die Variablen gesetzt,
// deren Werte gelesen oder geschrieben werden.

// Kommentar: Das ausgeben_flag wird im Konstruktor auf false gesetzt. Es geht dabei nur um Debugging.
UebertrageGitterwerte::UebertrageGitterwerte(void)
    {
    ausgeben_flag = false;
    }

// Kommentar: Durch Aufrufen dieser Funktion ausgeben wird das Debugging Flag auf true gesetzt - mit der Folge einer Ausgabe frs
// Debugging.
void UebertrageGitterwerte::ausgeben(void)
    {
    ausgeben_flag = true;
    }

// Kommentar: Die gitterwerte1 usw. werden gar nicht gebraucht. Diese Funktion wird nicht benutzt.
void UebertrageGitterwerte::schreibe_x_i(double ** _gitterwerte1, double ** _gitterwerte2, double ** _gitterwerte3, double ** _gitterwerte4)
    {
    gitterwerte1 = _gitterwerte1;
    gitterwerte2 = _gitterwerte2;
    gitterwerte3 = _gitterwerte3;
    gitterwerte4 = _gitterwerte4;
    }

// Kommentar: Hier wird in die Variable gitterwerte_sammeln1 die Variable geschrieben, bzgl. der das Sammeln erfolgen soll. Diese
// wird durch die Funktion schreibe_sammeln_i hier abgespreichert.
void UebertrageGitterwerte::schreibe_sammeln_i(double ** _gitterwerte_sammeln1, double ** _gitterwerte_sammeln2, double ** _gitterwerte_sammeln3, double ** _gitterwerte_sammeln4)
    {
    gitterwerte_sammeln1 = _gitterwerte_sammeln1;
    gitterwerte_sammeln2 = _gitterwerte_sammeln2;
    gitterwerte_sammeln3 = _gitterwerte_sammeln3;
    gitterwerte_sammeln4 = _gitterwerte_sammeln4;
    }

// Kommentar: Hier wird in die Variable gitterwerte_zerstreuen1 die Variable geschrieben, bzgl. der das Zerstreuen erfolgen soll. Diese
// wird durch die Funktion schreibe_zerstreuen_i hier abgespreichert.
void UebertrageGitterwerte::schreibe_zerstreuen_i(double ** _gitterwerte_zerstreuen1, double ** _gitterwerte_zerstreuen2, double ** _gitterwerte_zerstreuen3, double ** _gitterwerte_zerstreuen4)
    {
    gitterwerte_zerstreuen1 = _gitterwerte_zerstreuen1;
    gitterwerte_zerstreuen2 = _gitterwerte_zerstreuen2;
    gitterwerte_zerstreuen3 = _gitterwerte_zerstreuen3;
    gitterwerte_zerstreuen4 = _gitterwerte_zerstreuen4;
    }

// Kommentar: Hier wird der von dieser Klasse bentigte Speicherplatz freigegeben. Am Meisten Speicherplatz wird dabei fr die
// Buffer bereitgestellt, die verschickt werden - also frs Senden und Empfangen.
void UebertrageGitterwerte::finalisieren(void)
    {
// Kommentar: Die ganze Klasse ist assymetrisch bzgl. des Kernes 0 zu den anderen Kernen. Deshalb muss auch das Belegen und
// Freigeben von Speicherplatz assymetrisch erfolgen.
// Hat der Kern den Rang 0, so werden Buffer zu allen anderen Kernen freigegeben.
    if(rang==0)
        {
        for(i=1; i<groesse; i++)
            {
            delete koordinatenbuffer[i];
            }
        delete koordinatenbuffer;
        }
// Kommentar: Hat der Kern einen anderen Rang, so wird nur der eine Buffer wieder freigegeben.
    else
        {
        delete ein_koordinatenbuffer;
        }

// Kommentar: Hier wird gemerkt, wie lang die einzelnen Buffer sind. Einen Buffer dafr anzulegen ist im Prinzip nur relevant
// fr den Kern 0.
    delete werteindex_merken;

// Kommentar: Hier werden die Requests freigegeben. Die werden bentigt, um prfen zu knnen, ob die Kommunikation
// erfolgreich abgeschlossen worden ist.
    delete empfangs_requests;
    delete empfangs_requests2;
    delete sende_requests;
// Kommentar: Die Flags nehmen normalerweise die Information auf, ob eine Kommunikation abgeschlossen worden ist.
    delete flags;
// Kommentar: Die hier freigegebenen Flags stammen aus der Funktion 'empfangen0', damit schon entpackt werden kann,
// auch wenn noch nicht alle Nachrichten angekommen sind.
    delete uebertrage_gitterwerte_flag;
    }

#ifdef dimensionen1
// Kommentar: Hier wird die Klasse UebertrageGitterwerte initialisiert, wobei die wichtigsten Parameter bergeben werden und
// der bentigte Speicherplatz angelegt wird.
void UebertrageGitterwerte::initialisieren(int _n, int _m, int _groesse, int _rang, int _gitterebene, bool _oben_nicht_adaptiv, double ** _gitterwerte1)
#endif
#ifdef dimensionen2
void UebertrageGitterwerte::initialisieren(int _n, int _m, int _groesse, int _rang, int _gitterebene, bool _oben_nicht_adaptiv, double ** _gitterwerte1, double ** _gitterwerte2)
#endif
#ifdef dimensionen3
void UebertrageGitterwerte::initialisieren(int _n, int _m, int _groesse, int _rang, int _gitterebene, bool _oben_nicht_adaptiv, double ** _gitterwerte1, double ** _gitterwerte2, double ** _gitterwerte3)
#endif
#ifdef dimensionen4
void UebertrageGitterwerte::initialisieren(int _n, int _m, int _groesse, int _rang, int _gitterebene, bool _oben_nicht_adaptiv, double ** _gitterwerte1, double ** _gitterwerte2, double ** _gitterwerte3, double ** _gitterwerte4)
#endif
// Das Finalisieren dieser Funktion ist komplett gecheckt.
    {
// Zuerst werden die wichtigsten Parameter bergeben: n, m, groesse und rang. Auch die Gitterebene (h_kappe) wird mit bergeben.
    n = _n;
    m = _m;
    groesse = _groesse;
    rang = _rang;
    gitterebene = _gitterebene;
    oben_nicht_adaptiv = _oben_nicht_adaptiv;

// Kommentar: Die Flags werden angelegt. Damit geprft werden kann, welche Nachrichten bermittelt worden sind.
    flags = new int [groesse];

// Kommentar: Sie erst einmal auf false zu setzen, ist richtig. Allerdings muss das fr alle beteiligten Kerne
// - und das sind alle Kerne - zu Beginn ggf. wiederholt werden. Allerdings von Prozessor 0 in Bezug auf groesse - 1
// viele Kerne, und fr die anderen Kerne nur zu dem Kern 0.
    for(i=0; i<groesse; i++)
        {
        flags[i] = false;
        }

    #ifdef dimensionen1
// Kommentar: Dieser Parameter ist nicht von belang.
    gitterwerte1 = _gitterwerte1;
    #endif
    #ifdef dimensionen2
    gitterwerte1 = _gitterwerte1;
    gitterwerte2 = _gitterwerte2;
    #endif
    #ifdef dimensionen3
    gitterwerte1 = _gitterwerte1;
    gitterwerte2 = _gitterwerte2;
    gitterwerte3 = _gitterwerte3;
    #endif
    #ifdef dimensionen4
    gitterwerte1 = _gitterwerte1;
    gitterwerte2 = _gitterwerte2;
    gitterwerte3 = _gitterwerte3;
    gitterwerte4 = _gitterwerte4;
    #endif

// Kommentar: Mit Hilfe dieser Parameter wird die Gittergre auf der Ebene 'gitterebene' berechnet.
    n_hilfe = n;
    m_hilfe = m;

    for(i=0; i<gitterebene; i++)
        {
        n_hilfe = (n_hilfe+1)/2;
        m_hilfe = (m_hilfe+1)/2;
        }

// Kommentar: Zuerst kommt die Buffergre fr die Koordinaten. Da Vektoren kommuniziert werden - und Vektoren haben 3 Koordinaten - und die
// Bufferlnge mit verschickt wird, ist die Lnge dieses Buffers wie angegeben.
    buffergroesse = 3*n_hilfe*m_hilfe + 2;
    #ifdef dimensionen1
// Kommentar: Die zweite Buffergre verzeichnet, wieviele Werte bermittelt werden sollen. Je nach Anzahl der bermittelten
// Variablen muss man entsprechend viele Werte im Maximalfall bertragen.
    buffergroesse2 = n_hilfe * m_hilfe;
    if(oben_nicht_adaptiv==false)
        {
        buffergroesse2_sammeln = buffergroesse2 << 1;
        }
    else
        {
        buffergroesse2_sammeln = buffergroesse2;
        }
    #endif
    #ifdef dimensionen2
    buffergroesse2 = 2 * n_hilfe * m_hilfe;
    #endif
    #ifdef dimensionen3
    buffergroesse2 = 3 * n_hilfe * m_hilfe;
    #endif
    #ifdef dimensionen4
    buffergroesse2 = 4 * n_hilfe * m_hilfe;
    #endif

// Kommentar: Beim Kern mit der ID 0 werden die Koordinatenbuffer zu allen anderen Kernen angelegt. Die Wertebuffer werden in initialisieren2
// angelegt. Die Koordinaten werden von allen Kernen zu dem Kern 0 nur abgesendet, und vom Kern 0 nur empfangen. Die Kommunikation geht nicht
// in die andere Richtung.
    if(rang==0)
        {
#ifdef KURZ
        koordinatenbuffer = new short int * [groesse];
#else
        koordinatenbuffer = new int * [groesse];
#endif

        for(i=1; i<groesse; i++)
            {
#ifdef KURZ
            koordinatenbuffer[i] = new short int [buffergroesse];
#else
            koordinatenbuffer[i] = new int [buffergroesse];
#endif
            }
        }
// Kommentar: Die anderen Kerne legen hingegeben nur einen Koordinatenbuffer an - nmlich bzgl. den Koordinaten, die sie an den Kern
// 0 verschicken.
    else
        {
#ifdef KURZ
        ein_koordinatenbuffer = new short int [buffergroesse];
#else
        ein_koordinatenbuffer = new int [buffergroesse];
#endif
        }

// Kommentar: werteindex_merken verzeichnet, wie lange der zu verschickende Wertebuffer ist.
    werteindex_merken = new int [groesse];

// Kommentar: Die Requests dienen dem Zweck, zu prfen, ob die jeweilige Kommunikation abgeschlossen ist. Sie werden hier angelegt, wobei
// der Parameter Gre bei der Allokation eigentlich nur relevant ist fr Kern 0, weil nur dieser Kern zu mehr als einem anderen Kern
// kommuniziert.
    empfangs_requests = new MPI_Request[groesse];
    empfangs_requests2 = new MPI_Request[groesse];
    sende_requests = new MPI_Request[groesse];
// Kommentar: Die hier mit Speicherplatz Flags stammt aus der Funktion 'empfangen0', damit schon entpackt werden kann,
// auch wenn noch nicht alle Nachrichten angekommen sind.
    uebertrage_gitterwerte_flag = new int [groesse];
    }

#ifdef GEMEINSAME_BUFFER
// Kommentar: Hier werden die Wertebuffer angelegt - fr den Rang 0 die Buffer zu allen anderen Kernen, und fr die Kerne mit Rang != 0
// zum Kern mit dem Rang = 0. Also fr den Kern 0 sind es groesse - 1 viele Buffer, und fr die anderen Kerne ist es ein Buffer.
// Im Falle der Verwendung gemeinsamer Buffer wird der Speicherplatz bergeben, und andernfalls ...
void UebertrageGitterwerte::initialisieren2(double ** buffer)
    {
    if(rang==0)
        {
        wertebuffer = buffer;
        }
    else
        {
        ein_wertebuffer = buffer[0];
        }
    }
#else
// Kommentar: ... werden die Buffer der Lnge buffergroesse2 fr den Kern 0 groesse - 1 fach angelegt, und fr die anderen Kerne
// werden sie einfach angelegt.
void UebertrageGitterwerte::initialisieren2(void)
// Das Finialisieren dazu ist auch gecheckt.
    {
    if(rang == 0)
        {
        wertebuffer = new double * [groesse];

        for(i=1; i<groesse; i++)
            {
            wertebuffer[i] = new double [buffergroesse2];
            }
        }
    else
        {
        ein_wertebuffer = new double [buffergroesse2];
        }
    }
#endif

// Kommentar: Hier werden die Wertebuffer vom Speicherplatz befreit.
void UebertrageGitterwerte::finalisieren2(void)
    {
#ifdef GEMEINSAME_BUFFER
// Kommentar: Im Falle gemeinsamer Buffer ist hier nichts zu tun.
#else
// Kommentar: Ansonsten werden im Falle von Kern 0 die groesse - 1 Buffer befreit - zusammen mit dem Zeiger auf diese Buffer. Und
// die anderen Kerne entfernen den einen angelegten Buffer.
    if(rang==0)
        {
        for(i=1; i<groesse; i++)
            {
            delete wertebuffer[i];
            }
        delete wertebuffer;
        }
    else
        {
        delete ein_wertebuffer;
        }
#endif
    }

// Kommentar: Mit dieser Funktion werden die Koordinaten verschickt - von den Kernen mit Rang != 0.
void UebertrageGitterwerte::sende_koordinaten_ungleich_0(void)
    {
#ifdef KURZ
// Kommentar: Das hier ist der Sendebefehl. Der Sendebuffer ist ein_koordinatenbuffer, usw..
    MPI_Isend(ein_koordinatenbuffer, ein_koordinatenbuffer[0], MPI_SHORT, 0, 120, MPI_COMM_WORLD, & sende_request);
#else
    MPI_Isend(ein_koordinatenbuffer, ein_koordinatenbuffer[0], MPI_INT, 0, 120, MPI_COMM_WORLD, & sende_request);
#endif
    }

// Kommentar: Mit dieser Funktion empfngt der Kern mit der ID 0 die Nachrichten der anderen Kerne bzgl. der Koordinaten
// auf denen dieser ihre Gitterpunkte haben.
void UebertrageGitterwerte::empfange_koordinaten_gleich_0(void)
    {
// Kommentar: Fr alle anderen Kerne ...
    for(i=1; i<groesse; i++)
        {
#ifdef KURZ
// Kommentar: ... empfange eine Nachricht in dem koordinatenbuffer[i].
        MPI_Irecv(koordinatenbuffer[i], buffergroesse, MPI_SHORT, i, 120, MPI_COMM_WORLD, & (empfangs_requests[i]));
#else
        MPI_Irecv(koordinatenbuffer[i], buffergroesse, MPI_INT, i, 120, MPI_COMM_WORLD, & (empfangs_requests[i]));
#endif
        }
    }

// Kommentar: Beim 'zerstreuen' sendet der Kern mit der ID 0 die dafr zu bermittelnden Gitterwerte an die anderen Kerne.
void UebertrageGitterwerte::sende_werte_gleich_0(void)
    {
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
    cout << "IL" << endl;
#endif
// Kommentar: Fr alle anderen Kerne ...
    for(i=1; i<groesse; i++)
        {
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
        cout << " i " << i << " wm[i] " << werteindex_merken[i];
#endif
// Kommentar: ... verschicke die Werte des i.ten Wertebuffers an den Kern mit dem Rang i.
        MPI_Isend(wertebuffer[i], werteindex_merken[i], MPI_DOUBLE, i, 122, MPI_COMM_WORLD, & (sende_requests[i]));
        }
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
    cout << endl;
#endif
    }

// Kommentar: Hier werden die Koordinaten der Gitterpunkte ausgegeben mit cout. Der benutzte Buffer dabei ist ein_koordantenbuffer,
// d.h. er wird verwendet von den Kernen mit Rang != 0. Das dient nur dem Debugging.
void UebertrageGitterwerte::ausgabe_koordinaten(void)
    {
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
    cout << " In bertrage Gitterwerte ausgabe_koordinaten . Rang " << rang << endl;
#endif
// Kommentar: Die Lnge des Buffers wird gelesen, sowie, ob Nord - Sd oder Ost - West Vektoren verschickt worden sind.
    hoechstindex = ein_koordinatenbuffer[0];
    wons = ein_koordinatenbuffer[1];
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
    cout << " In bertrage Gitterwerte ausgabe_koordinaten . Rang " << rang << " hoechstindex " << hoechstindex << " wons " << wons << endl;
    if(wons == NORD_SUED)
        {
        cout << "Nord Sd " << endl;
        }
    else
        {
        cout << "Ost West " << endl;
        }
#endif
// Kommentar: Bei Nord Sd Vektoren ...
    if(wons == NORD_SUED)
        {
        for(i=2; i<hoechstindex;)
            {
            a = ein_koordinatenbuffer[i++];
            b = ein_koordinatenbuffer[i++];
            c = ein_koordinatenbuffer[i++];
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
// Kommentar: ... werden ebensolche ausgegeben.
            cout << " f1 " << a << " f2 " << b << " g " << c << " rang " << rang << endl;
#endif
            }
        }
// Kommentar: Ansonsten werden Ost - West - Vektoren ausgegeben.
    else
        {
        for(i=2; i<hoechstindex;)
            {
            a = ein_koordinatenbuffer[i++];
            b = ein_koordinatenbuffer[i++];
            c = ein_koordinatenbuffer[i++];
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
            cout << " f " << a << " g1 " << b << " g2 " << c << " rang " << rang << endl;
#endif
            }
        }
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
    cout << "Ende ausgabe bermitteln Gitterwerte / rang " << rang << endl;
#endif
    }

// Kommentar: Diese Funktion dient dem Schreiben von Werten in den Wertebuffer - zu spteren bermittlung. Das erfolgt bei Kernen
// mit Rang != 0, und beim Sammeln. Bei Zerstreuen werden bei diesen Kernen die Werte ausgepackt.
void UebertrageGitterwerte::wertebuffer_bestimmen(void)
// Dient dem Sammeln
    {
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
    cout << " Wertebuffer bestimmen 0 / rang " << rang << endl;
    for(i=0; i<5; i++)
        {
        cout << " i " << i << " ein_koord[i] " << ein_koordinatenbuffer[i];
        }
    cout << endl;
#endif
// Kommentar: Grundlage der zu packenden Werte ist der Buffer ein_koordinatenbuffer. Zuerst word gelesen, was fr Vektoren gelesen
// werden, und wieviele es sind.
    hoechstindex = ein_koordinatenbuffer[0];
    werteindex = 0;
    wons = ein_koordinatenbuffer[1];

#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
    cout << " Wertebuffer bestimmen 1 / rang " << rang << " laufindex " << hoechstindex << " wons " << wons << endl;

    if(ausgeben_flag==true)
        {
        cout << "UU" << endl;
        }

#endif

// Kommentar: Im ersten Fall werden Nord - Sd Vektoren gelesen.
    if(wons == NORD_SUED) // NORD_SUED
        {
// Kommentar: Nun werden die Koordinaten aus dem Koordinatenbuffer gelesen.
        for(i=2; i<hoechstindex;)
            {
            a = ein_koordinatenbuffer[i++];
            b = ein_koordinatenbuffer[i++];
            c = ein_koordinatenbuffer[i++];
            // f1 - a, f2 - b, g - c

#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
            cout << " f1 " << a << " f2 " << b << " g " << c << " Rang " << rang << endl;
#endif

// Kommentar: Die Gitterpunktkoordinaten des jeweiligen Vektor's werden durchlaufen.
            for(j=a; j<=b;)
                {
                #ifdef dimensionen1
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
                cout << gitterwerte1[j][c] << " ";
#endif
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
                cout << " RST sammeln f " << j << " g " << c << " wert " << gitterwerte_sammeln1[j][c] << " rang " << rang << endl;
#endif

// Kommentar: Und je nach Dimensionsanzahl werden fr jede Dimension aus der Variablen 'gitterwerte_sammeln1', ggf. aus der Variablen
// 'gitterwerte_sammeln2' usw. - je nach Anzahl der Dimensionen - gelesen und in den Wertebuffer 'ein_wertebuffer' geschrieben.
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln1[j++][c];
                #endif
                #ifdef dimensionen2
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln1[j][c];
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln2[j++][c];
                #endif
                #ifdef dimensionen3
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln1[j][c];
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln2[j][c];
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln3[j++][c];
                #endif
                #ifdef dimensionen4
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln1[j][c];
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln2[j][c];
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln3[j][c];
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln4[j++][c];
                #endif
                }
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
            cout << endl;
#endif
            }
        }
// Kommentar: Im alternativen Fall werden Ost - West Vektoren bearbeitet. Alles weitere dazu erfolgt analog dem Nord - Sd - Fall -
// nur das Ost - West Vektoren durchlaufen werden.
    else
        {
        for(i=2; i<hoechstindex;)
            {
            a = ein_koordinatenbuffer[i++];
            b = ein_koordinatenbuffer[i++];
            c = ein_koordinatenbuffer[i++];
            // f - a, g1 - b, g2 - c
            for(j=b; j<=c;)
                {
                #ifdef dimensionen1
                cout << gitterwerte1[a][j] << " ";
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln1[a][j++];
                #endif
                #ifdef dimensionen2
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln1[a][j];
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln2[a][j++];
                #endif
                #ifdef dimensionen3
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln1[a][j];
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln2[a][j];
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln3[a][j++];
                #endif
                #ifdef dimensionen4
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln1[a][j];
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln2[a][j];
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln3[a][j];
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln4[a][j++];
                #endif
                }
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
            cout << endl;
#endif
            }
        }
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
    cout << " Ende Wertebuffer bestimmen / rang " << rang << " laufindex " << hoechstindex << " wons " << wons << endl;
#endif
    }

// Kommentar: Diese Funktion ist fast analog zu wertebuffer_bestimmen. Der einzige Unterschied besteht darin, dass
// im Fall 'einer Dimension' doch zwei Variablen bertragen werden.
void UebertrageGitterwerte::wertebuffer_bestimmen_adapt(void)
// Dient dem Sammeln
    {
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
    cout << " Wertebuffer bestimmen 0 / rang " << rang << endl;
    for(i=0; i<5; i++)
        {
        cout << " i " << i << " ein_koord[i] " << ein_koordinatenbuffer[i];
        }
    cout << endl;
#endif
    hoechstindex = ein_koordinatenbuffer[0];
    werteindex = 0;
    wons = ein_koordinatenbuffer[1];

#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
    cout << " Wertebuffer bestimmen 1 / rang " << rang << " laufindex " << hoechstindex << " wons " << wons << endl;

    if(ausgeben_flag==true)
        {
        cout << "UU" << endl;
        }

#endif
    if(wons == NORD_SUED) // NORD_SUED
        {
        for(i=2; i<hoechstindex;)
            {
            a = ein_koordinatenbuffer[i++];
            b = ein_koordinatenbuffer[i++];
            c = ein_koordinatenbuffer[i++];
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
            cout << " f1 " << a << " f2 " << b << " g " << c << " Rang " << rang << endl;
#endif
            // f1 - a, f2 - b, g - c
            for(j=a; j<=b;)
                {
                #ifdef dimensionen1
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
                cout << gitterwerte1[j][c] << " ";
#endif
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
                cout << " RST sammeln f " << j << " g " << c << " wert " << gitterwerte_sammeln1[j][c] << " rang " << rang << endl;
#endif
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln1[j][c];
// Kommentar: Hier ist der Unterschied zu wertebuffer_bestimmen.
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln2[j++][c];
                #endif
                #ifdef dimensionen2
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln1[j][c];
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln2[j++][c];
                #endif
                #ifdef dimensionen3
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln1[j][c];
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln2[j][c];
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln3[j++][c];
                #endif
                #ifdef dimensionen4
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln1[j][c];
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln2[j][c];
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln3[j][c];
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln4[j++][c];
                #endif
                }
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
            cout << endl;
#endif
            }
        }
    else
        {
        for(i=2; i<hoechstindex;)
            {
            a = ein_koordinatenbuffer[i++];
            b = ein_koordinatenbuffer[i++];
            c = ein_koordinatenbuffer[i++];
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
            cout << " f " << a << " g1 " << b << " g2 " << c << " Rang " << rang << endl;
#endif
            // f - a, g1 - b, g2 - c
            for(j=b; j<=c;)
                {
                #ifdef dimensionen1
                cout << gitterwerte1[a][j] << " ";
// Kommentar: Hier ist der Unterschied zu wertebuffer_bestimmen.
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln1[a][j];
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln2[a][j++];
                #endif
                #ifdef dimensionen2
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln1[a][j];
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln2[a][j++];
                #endif
                #ifdef dimensionen3
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln1[a][j];
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln2[a][j];
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln3[a][j++];
                #endif
                #ifdef dimensionen4
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln1[a][j];
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln2[a][j];
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln3[a][j];
                ein_wertebuffer[werteindex++] = gitterwerte_sammeln4[a][j++];
                #endif
                }
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
            cout << endl;
#endif
            }
        }
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
    cout << " Ende Wertebuffer bestimmen / rang " << rang << " laufindex " << hoechstindex << " wons " << wons << endl;
#endif
    }

// Kommentar: Hier werden die Werte verschickt von den Rang != 0 Kernen, natrlich nachdem die Wertebuffer gepackt wurden von diesen
// Kernen. Empfnger der Wertebuffer ist der Kern 0. Da diese Funktion im wesentlichen aus der einen Zeile besteht, wo das Senden
// des Buffers ein_wertebuffer initiiert wird, erfolgen keine weiteren Kommentare. Diese Funktion gehrt zum Sammeln.
void UebertrageGitterwerte::sende_werte_ungleich_0(void)
    {
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
    cout << "Anfang SWU0 / werteindex " << werteindex << " bg2 " << buffergroesse2 << endl;
    for(i=0; i<werteindex; i++)
        {
        cout << ein_wertebuffer[i] << " ";
        }
    cout  << endl;
#endif
    MPI_Isend(ein_wertebuffer, werteindex, MPI_DOUBLE, 0, 121, MPI_COMM_WORLD, & sende_request2);
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
    cout << "Ende SWU0 buffergroesse 2 " << buffergroesse2 << endl;
#endif
    }

// Kommentar: Auch diese Funktion gehrt zum Sammeln, nur dass es der Kern 0 ist, der diese Aktion macht. Er empfngt die Wertebuffer
// der anderen Kerne.
void UebertrageGitterwerte::empfange_werte_gleich_0(void)
    {
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
    cout << " Code C1 buffergroesse2 " << buffergroesse2 << " buffergroesse2_sammeln " << buffergroesse2_sammeln << endl;
#endif // AUSGABEN_UEBERTRAGE_GITTERWERTE
// Kommentar: Die ID's der anderen Kerne werden durchlaufen.
    for(i=1; i<groesse; i++)
        {
// Kommentar: Und fr die Werte der anderen Kerne wird eine Empfangskommunikation initiiert.
        MPI_Irecv(wertebuffer[i], buffergroesse2_sammeln, MPI_DOUBLE, i, 121, MPI_COMM_WORLD, & (empfangs_requests2[i])); // nderung 10.11.15
        }
    }

// Kommentar: Das hier ist eine Ausgabe  - Funktion zu Debugging - Zwecken.
void UebertrageGitterwerte::sehe_empfange_werte_gleich_0(void)
    {
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
    cout << "Anfang SEWG0 " << endl;
    for(i=1; i<groesse; i++)
        {
        for(j=0; j<30; j++)
            {
            cout << wertebuffer[i][j] << " ";
            }
        cout << endl;
        }
    cout << "Ende SEWG0 " << endl;
#endif
    }

// Kommentar: Es wird hier durch den Kern 0 geprft, ob der Empfang der Koordinaten der anderen Kerne vollendet ist - zu allen Kernen.
// Falls ja, dann wird true zurckgegeben, ansonsten false.
bool UebertrageGitterwerte::teste_empfange_koordinaten_gleich_0(void)
    {
// Kommentar: Die ID's der Kerne mit Rang != 0 werden durchlaufen.
    for(i=1; i<groesse; i++)
        {
// Kommentar: Ist die Nachricht zu Kern i nicht angekommen, dann gebe false zurck.
        MPI_Test(&(empfangs_requests[i]), & flag, & empfangs_status);
        if(flag == false)
            {
            return(false);
            }
        }
// Kommentar: Sind alle flags true - weil kein false zurckgegeben worden ist - so gebe true zurck.
    return(true);
    }

// Kommentar: Es wird hier durch den Kern 0 geprft, ob das Senden der Werte zu den anderen Kernen vollendet ist - zu allen Kernen.
// Falls ja, dann wird true zurckgegeben, ansonsten false.
bool UebertrageGitterwerte::teste_sende_koordinaten_gleich_0(void)
    {
// Kommentar: Die ID's der Kerne mit Rang != 0 werden durchlaufen.
    for(i=1; i<groesse; i++)
        {
// Kommentar: Ist die Nachricht zu Kern i nicht abgeschickt, dann gebe false zurck.
        MPI_Test(&(sende_requests[i]), & flag, & sende_status);
        if(flag == false)
            {
            return(false);
            }
        }
// Kommentar: Sind alle flags true - weil kein false zurckgegeben worden ist - so gebe true zurck.
    return(true);
    }


// Kommentar: Diese Funktion wird nicht verwendet. Ihr Aufruf ist entweder bei nicht definierter Prprozessorvariablen KUWS2_ALT,
// oder auskommentiert.
void UebertrageGitterwerte::ausgabe_empfang(void)
    {
    for(i=1; i<groesse; i++)
        {
        if(flags[i]==false)
            {
            MPI_Test(&(empfangs_requests2[i]), & flag, & empfangs_status);
            if(flag == true)
                {
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
                cout << "FLAG i " << i << " flag " << flag << endl;
#endif
                flags[i] = true;
                }
            }
        }
    }

// Kommentar: Analog zu 'teste_empfange_koordinaten_gleich_0', nur dass statt der Prfung, ob die Koordinatenbuffer bzgl. des Sammelns
// angekommen sind, geprft wird, ob die Wertebufer bzgl. des Sammelns angekommen sind. Die Funktion hat deshalb nur den kleinen
// Unterschied, dass statt empfangs_requests die empfangs_requests2 verwendet werden. Alles andere ist gleich.
bool UebertrageGitterwerte::teste_empfange_werte_gleich_0(void)
    {
    for(i=1; i<groesse; i++)
        {
        MPI_Test(&(empfangs_requests2[i]), & flag, & empfangs_status);
        if(flag == false)
            {
            return(false);
            }
        }
    return(true);
    }

// Kommentar: Das ist wieder eine Funktion, die fr die Kern 0 ausgefhrt wird. Nachdem er die Werte aller anderen Kerne empfangen hat,
// trgt er diese Werte jedes Kerns in seiner Ebene ein.
void UebertrageGitterwerte::trage_werte_ein_gleich_0(void)
    {
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
    cout << " trage werte ein gleich 0. groesse " << groesse << endl;
#endif

// Kommentar: Durchlauf die ID's aller Kerne.
    for(k=1; k<groesse; k++)
        {
        cout << " Trage werte einmal ein gleich 0 zu Kern " << k << endl;
// Kommentar: Trage die Werte nach den Informationen eines Kernes ein. Eingetragen wird das fr den Kern 0.
        trage_werte_einmal_ein_gleich_0(k);
        }
    }

// Kommentar: Hier werden fr den Kern 0 die Werte des Wertebuffers des Kerns auswahl_rang - entsprechend den Koordinaten,
// die dieser Kern geschickt hat, entpackt.
void UebertrageGitterwerte::trage_werte_einmal_ein_gleich_0(int auswahl_rang)
    {
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
    cout << " Falls ona = false stimmt das nicht mit keinen Nullen rang " << rang << endl;
#endif
// Dient dem Sammeln
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
    cout << " twe " << auswahl_rang << endl;
    if(ausgeben_flag==true)
        {
        cout << " WW Anfang trage werte einmal ein " << endl;
        }
#endif

// Kommentar: Koordianten - und Wertebuffer werden bestimmt. Dazu die Vektorenanzahl, und, ob es Nord - Sd oder Ost - West
// Vektoren sind.
    koord = koordinatenbuffer[auswahl_rang];
    werte = wertebuffer[auswahl_rang];
    hoechstindex = koord[0];
    werteindex = 0;
    wons = koord[1];

// Komemntar: Nun wird stellvertretend fr beide Flle der Nord - Sd - Fall besprochen.
    if(wons == NORD_SUED)
        {
// Kommentar: Die Koordinaten der Vektoren werden durchlaufen.
        for(i=2; i<hoechstindex;)
            {
            a = koord[i++];
            b = koord[i++];
            c = koord[i++];
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
            cout << " XYZ abc a " << a << " b " << b << " c " << c << endl;
#endif
            // f1 - a, f2 - b, g - c
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
            cout << " f1 " << a << " f2 " << b << " g " << c << " ";
#endif

// Kommentar: Die Gitterpunktkoordinaten des ausgewhlten Vektors werden durchlaufen.
            for(j=a; j<=b;)
                {
                #ifdef dimensionen1
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
                cout << " werteindex " << werteindex << " wert ";
                cout << werte[werteindex] << " " << endl;
#endif
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
                cout << " XYZ trage werte ein f " << j << " g " << c << " von rang " << auswahl_rang << " werte " << werte[werteindex] << endl;
#endif

// Kommentar: In die Variablen gitterwerte_sammeln1 werden die Werte des Wertebuffers geschrieben. Bei Nord - Sd Vektoren wird
// die erste Koordinate durchlaufen (zwichen a und b), und die zweite festgehalten (Koordinatenwert c).
                gitterwerte_sammeln1[j++][c] = werte[werteindex++];
                #endif
                #ifdef dimensionen2
                gitterwerte_sammeln1[j][c] = werte[werteindex++];
                gitterwerte_sammeln2[j++][c] = werte[werteindex++];
                #endif
                #ifdef dimensionen3
                gitterwerte_sammeln1[j][c] = werte[werteindex++];
                gitterwerte_sammeln2[j][c] = werte[werteindex++];
                gitterwerte_sammeln3[j++][c] = werte[werteindex++];
                #endif
                #ifdef dimensionen4
                gitterwerte_sammeln1[j][c] = werte[werteindex++];
                gitterwerte_sammeln2[j][c] = werte[werteindex++];
                gitterwerte_sammeln3[j][c] = werte[werteindex++];
                gitterwerte_sammeln4[j++][c] = werte[werteindex++];
                #endif
                }
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
            cout << endl;
#endif
            }
        }
// Kommentar: Dieser Fall behandelt die Ost - West Vektoren, und das geschieht analog.
    else
        {
        for(i=2; i<hoechstindex;)
            {
            a = koord[i++];
            b = koord[i++];
            c = koord[i++];

            // f - a, g1 - b, g2 - c
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
            cout << " f " << a << " g1 " << b << " g2 " << c << " ";
#endif
            for(j=b; j<=c;)
                {
                #ifdef dimensionen1
                cout << werte[werteindex] << " ";
                gitterwerte_sammeln1[a][j++] = werte[werteindex++];
                #endif
                #ifdef dimensionen2
                gitterwerte_sammeln1[a][j] = werte[werteindex++];
                gitterwerte_sammeln2[a][j++] = werte[werteindex++];
                #endif
                #ifdef dimensionen3
                gitterwerte_sammeln1[a][j] = werte[werteindex++];
                gitterwerte_sammeln2[a][j]= werte[werteindex++];
                gitterwerte_sammeln3[a][j++] = werte[werteindex++];
                #endif
                #ifdef dimensionen4
                gitterwerte_sammeln1[a][j] = werte[werteindex++];
                gitterwerte_sammeln2[a][j] = werte[werteindex++];
                gitterwerte_sammeln3[a][j] = werte[werteindex++];
                gitterwerte_sammeln4[a][j++] = werte[werteindex++];
                #endif
                }
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
            cout << endl;
#endif
            }
        }
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
    cout << " Ende trage werte einmal ein " << endl;
    cout << " ENDEtwee " << endl;
    cout << " In AUG: dieses / rang " << rang << endl;
#endif
    }

// Kommentar: Das hier ist wieder eine Funktion fr den Kern mit dem Rang 0. Es geht ums Zerstreuen. Es sollen alle zu verschickenden
// Wertebuffer gepackt werden.
void UebertrageGitterwerte::alle_wertebuffer_bestimmen(void)
    {
// Kommentar: Dies geschieht, indem man die Rnge aller anderen Kerne durchluft, und zu jedem Rand den Wertebuffer schreibt.
    for(i=1; i<groesse; i++)
        {
        einen_wertebuffer_bestimmen(i);
        }
    }

// Kommentar: Diese Funktion ist fr den Kern 0 da. Sie beschreibt den Wertebuffer zu einem anderen Kern - beim Zerstreuen.
void UebertrageGitterwerte::einen_wertebuffer_bestimmen(int auswahl_rang)
    {
// Dient dem Zerstreuen

// Kommentar: Der Koordinaten - und Wertebuffer zu dem ausgewhlten Rang werden geladen. Dazu die Bufferlnge, und die Ausrichtung
// - Ost - West oder Nord - Sd.
    koord = koordinatenbuffer[auswahl_rang];
    werte = wertebuffer[auswahl_rang];
    hoechstindex = koord[0];
    werteindex = 0;
    wons = koord[1];

// Kommentar: Behandelt wird hier der Nord - Sd - Fall.
    if(wons == NORD_SUED) // NORD_SUED
        {
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
        cout << " EWB 0 / rang " << rang << " auswahl_rang " << auswahl_rang << endl;
#endif

// Kommentar: Die Nord - Sd - Vektoren werden durchlaufen.
        for(l=2; l<hoechstindex;)
            {
            a = koord[l++];
            b = koord[l++];
            c = koord[l++];
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
            cout << " f1 " << a << " f2 " << b << " g " << c << endl;
#endif
            // f1 - a, f2 - b, g - c

// Kommentar: Dann werden die Gitterpunktkoordinaten des Nord - Sd - Vektors durchlaufen.
            for(j=a; j<=b;)
                {
                #ifdef dimensionen1
                cout << gitterwerte1[j][c] << " ";

// Kommentar: Und die Werte werden in den Wertebuffer geschrieben. Sie stammen von der Variablen gitterwerte_zerstreuen1,
// die fr das Zerstreuen - und das passiert bei dieser Funktion - zustndig ist. Bei NS - Vektoren wird die erste Koordinate
// durchlaufen (zwischen a und b), und die zweite wird fixiert (Wert c).
                werte[werteindex++] = gitterwerte_zerstreuen1[j++][c];
                #endif
                #ifdef dimensionen2
                werte[werteindex++] = gitterwerte_zerstreuen1[j][c];
                werte[werteindex++] = gitterwerte_zerstreuen2[j++][c];
                #endif
                #ifdef dimensionen3
                werte[werteindex++] = gitterwerte_zerstreuen1[j][c];
                werte[werteindex++] = gitterwerte_zerstreuen2[j][c];
                werte[werteindex++] = gitterwerte_zerstreuen3[j++][c];
                #endif
                #ifdef dimensionen4
                werte[werteindex++] = gitterwerte_zerstreuen1[j][c];
                werte[werteindex++] = gitterwerte_zerstreuen2[j][c];
                werte[werteindex++] = gitterwerte_zerstreuen3[j][c];
                werte[werteindex++] = gitterwerte_zerstreuen4[j++][c];
                #endif
                }
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
            cout << endl;
#endif
            }
        }
// Kommentar: Der Ost - West - Fall geschieht analog.
    else
        {
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
        cout << " EWB 1 / rang " << rang << " auswahl_rang " << auswahl_rang << endl;
#endif
        for(l=2; l<hoechstindex;)
            {
            a = koord[l++];
            b = koord[l++];
            c = koord[l++];
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
            cout << " f " << a << " g1 " << b << " g2 " << c << endl;
#endif
            // f - a, g_1 - b, g_2 - c
            for(j=b; j<=c;)
                {
                #ifdef dimensionen1
                cout << gitterwerte1[a][j] << " ";
                werte[werteindex++] = gitterwerte_zerstreuen1[a][j++];
                #endif
                #ifdef dimensionen2
                werte[werteindex++] = gitterwerte_zerstreuen1[a][j];
                werte[werteindex++] = gitterwerte_zerstreuen2[a][j++];
                #endif
                #ifdef dimensionen3
                werte[werteindex++] = gitterwerte_zerstreuen1[a][j];
                werte[werteindex++] = gitterwerte_zerstreuen2[a][j];
                werte[werteindex++] = gitterwerte_zerstreuen3[a][j++];
                #endif
                #ifdef dimensionen4
                werte[werteindex++] = gitterwerte_zerstreuen1[a][j];
                werte[werteindex++] = gitterwerte_zerstreuen2[a][j];
                werte[werteindex++] = gitterwerte_zerstreuen3[a][j];
                werte[werteindex++] = gitterwerte_zerstreuen4[a][j++];
                #endif
                }
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
            cout << endl;
#endif
            }
        }
// Kommentar: Am wird der Werteindex, der auf den nchsten Buffereintrag zeigt, gespeichert.
    werteindex_merken[auswahl_rang] = werteindex;
    }

// Kommentar: Die Position ist diese: Es geht um einen Kern vom Rand != 0 beum Zerstreuen. Er empfngt von Kern 0 die Werte,
// und bei dieser Funktion schreibt er die empfangenen Werte in die ausgewhlte Variable.
void UebertrageGitterwerte::wertebuffer_eintragen(void)
// Dient dem Zerstreuen
    {
// Kommentar: Es wird die Lnge des Koordinatenbuffer gelesen sowie der Typ von Vektor.
    hoechstindex = ein_koordinatenbuffer[0];
    werteindex = 0;
    wons = ein_koordinatenbuffer[1];

// Kommentar: Hier werden die Nord - Sd - Vektoren bearbeitet.
    if(wons == NORD_SUED) // NORD_SUED
        {
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
        cout << " WBE wons = 0 / rang " << rang << endl;
#endif

// Kommentar: Die Vektorkoordinaten werden gelesen.
        for(i=2; i<hoechstindex;)
            {
            a = ein_koordinatenbuffer[i++];
            b = ein_koordinatenbuffer[i++];
            c = ein_koordinatenbuffer[i++];
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
            cout << " f1 " << a << " f2 " << b << " g " << c << endl;
#endif
            // f_1 - a, f_2 - b, g - c

// Kommentar: Die Gitterpunktkoordinaten werden durchlaufen.
            for(j=a; j<=b;)
                {
                #ifdef dimensionen1
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
                cout << ein_wertebuffer[werteindex] << " ";
#endif

// Kommentar: Die Werte des Buffers werden in die verknpfte Variable gitterwerte_zerstreuen1 an den jeweiligen Gitterkoordinaten
// geschrieben. Bei Nord - Sd Vektoren variiert die erste Koordinate (y).
                gitterwerte_zerstreuen1[j++][c] = ein_wertebuffer[werteindex++];
                #endif
                #ifdef dimensionen2
                gitterwerte_zerstreuen1[j][c] = ein_wertebuffer[werteindex++];
                gitterwerte_zerstreuen2[j++][c] = ein_wertebuffer[werteindex++];
                #endif
                #ifdef dimensionen3
                gitterwerte_zerstreuen1[j][c] = ein_wertebuffer[werteindex++];
                gitterwerte_zerstreuen2[j][c] = ein_wertebuffer[werteindex++];
                gitterwerte_zerstreuen3[j++][c] = ein_wertebuffer[werteindex++];
                #endif
                #ifdef dimensionen4
                gitterwerte_zerstreuen1[j][c] = ein_wertebuffer[werteindex++];
                gitterwerte_zerstreuen2[j][c] = ein_wertebuffer[werteindex++];
                gitterwerte_zerstreuen3[j][c] = ein_wertebuffer[werteindex++];
                gitterwerte_zerstreuen4[j++][c] = ein_wertebuffer[werteindex++];
                #endif
                }
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
            cout << endl;
#endif
            }
        }
// Kommentar: Der Fall mit den Ost -  West Vektoren erfolgt in derselben Weise - mit variierender zweiten Koordinate.
    else
        {
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
        cout << " WBE wons = 1 / rang " << rang << endl;
#endif
        for(i=2; i<hoechstindex;)
            {
            a = ein_koordinatenbuffer[i++];
            b = ein_koordinatenbuffer[i++];
            c = ein_koordinatenbuffer[i++];
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
            cout << " f " << a << " g1 " << b << " g2 " << c << endl;
#endif
            // f - a, g_1 - b, g_2 - c
            for(j=b; j<=c;)
                {
                #ifdef dimensionen1
                cout << ein_wertebuffer[werteindex] << " ";
                gitterwerte_zerstreuen1[a][j++] = ein_wertebuffer[werteindex++];
                #endif
                #ifdef dimensionen2
                gitterwerte_zerstreuen1[a][j] = ein_wertebuffer[werteindex++];
                gitterwerte_zerstreuen2[a][j++] = ein_wertebuffer[werteindex++];
                #endif
                #ifdef dimensionen3
                gitterwerte_zerstreuen1[a][j] = ein_wertebuffer[werteindex++];
                gitterwerte_zerstreuen2[a][j]= ein_wertebuffer[werteindex++];
                gitterwerte_zerstreuen3[a][j++] = ein_wertebuffer[werteindex++];
                #endif
                #ifdef dimensionen4
                gitterwerte_zerstreuen1[a][j] = ein_wertebuffer[werteindex++];
                gitterwerte_zerstreuen2[a][j] = ein_wertebuffer[werteindex++];
                gitterwerte_zerstreuen3[a][j] = ein_wertebuffer[werteindex++];
                gitterwerte_zerstreuen4[a][j++] = ein_wertebuffer[werteindex++];
                #endif
                }
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
            cout << endl;
#endif
            }
        }
    }

// Kommentar: Hier empfangen die Kerne mit Rang != 0 die Werte beim Zerstreuen. Diese Kommunikation wird hier initiiert, und
// nach erhalt der Nachrichten in Form von Werten knnen die dann entpackt werden.
void UebertrageGitterwerte::empfange_werte_ungleich_0(void)
    {
    MPI_Irecv(ein_wertebuffer, buffergroesse2, MPI_DOUBLE, 0, 122, MPI_COMM_WORLD, & empfangs_request);
    }

// Kommentar: Hier wird auf einem Kern mit ID != 0 geprft, ob die Werte gem 'empfange_werte_ungleich_0' angekommen sind.
// Wenn ja, dann return true, sonst return false.
bool UebertrageGitterwerte::teste_empfange_werte_ungleich_0(void)
    {
    MPI_Test(&(empfangs_request), & flag, & empfangs_status);
    return(flag);
    }

// Kommentar: Hier wird getesetet, ob - fr Kerne mit Rand != 0 - das Senden von Koordinaten und Werten abgeschlossen ist.
bool UebertrageGitterwerte::teste_ungleich_0(void)
    {
// Kommentar: Wenn das Senden der Koordinaten nicht abgeschlossen ist => return false.
    MPI_Test(&(sende_request), & flag, & empfangs_status);
    if(flag==false)
        {
        return(false);
        }
// Kommentar: Wenn das Senden der Werte nicht abgeschlossen ist => return false.
    MPI_Test(&(sende_request2), & flag, & empfangs_status);
    if(flag==false)
        {
        return(false);
        }
// Kommentar: Beide Kommunikationen sind abgeschlossen => return true.
    return(true);
    }

// Kommentar: Hier wir die Variable u0 (auf der Ebene der Kappe, wie alles hier) - kopiert. Wo wird sie benutzt?
void UebertrageGitterwerte::setze_u0(double ** _u0)
    {
    u0 = _u0;
    }

// Kommentar: Analog zu trage_werte_ein_gleich_0, nur dass hier die Funktion 'trage_werte_einmal_ein_gleich_0_ohne_nullen'
// aufgerufen wird. Kern 0 trgt die von den anderen Kernen verschickten Werte ein.
void UebertrageGitterwerte::trage_werte_ein_gleich_0_ohne_nullen(void)
    {
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
    cout << " trage werte ein gleich 0. groesse " << groesse << endl;
#endif

// Kommentar: Es werden die Rnge der anderen Kerne durchlaufen.
    for(k=1; k<groesse; k++)
        {
// Kommentar: Und fr jeden empfangenen Koordinaten - und Wertebuffer werden die Werte entpackt.
        trage_werte_einmal_ein_gleich_0_ohne_nullen(k);
        }
    }

// Kommentar: Fr jeden empfangenen Koordinaten - und Wertebuffer werden die Werte entpackt. Dabei werden in diesem
// Fall die zwei Variablen 'gitterwerte_sammeln1' und 'gitterwerte_sammeln2' beschrieben. Ansonsten analog zu
// 'trage_werte_einmal_ein_gleich_0', d.h. man vergleiche die Kommentare von dort.
void UebertrageGitterwerte::trage_werte_einmal_ein_gleich_0_ohne_nullen(int auswahl_rang)
    {
    cout << " Falls ona = false stimmt das mit Nullen rang " << rang << endl;
// Dient dem Sammeln
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
    cout << " twe " << auswahl_rang << endl;
    if(ausgeben_flag==true)
        {
        cout << " WW Anfang trage werte einmal ein " << endl;
        }
#endif
    koord = koordinatenbuffer[auswahl_rang];
    werte = wertebuffer[auswahl_rang];
    hoechstindex = koord[0];
    werteindex = 0;
    wons = koord[1];
    if(wons == NORD_SUED)
        {
        for(i=2; i<hoechstindex;)
            {
            a = koord[i++];
            b = koord[i++];
            c = koord[i++];
            cout << " XYZ abc a " << a << " b " << b << " c " << c << endl;
            // f1 - a, f2 - b, g - c
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
            cout << " f1 " << a << " f2 " << b << " g " << c << " ";
#endif
            for(j=a; j<=b;)
                {
                #ifdef dimensionen1
                cout << werte[werteindex] << " ";
                cout << " XYZ trage werte ein f " << j << " g " << c << " von rang " << auswahl_rang << " werte " << werte[werteindex] << endl;
                gitterwerte_sammeln1[j][c] = werte[werteindex++];
                gitterwerte_sammeln2[j++][c] = werte[werteindex++];
                #endif
                #ifdef dimensionen2
                gitterwerte_sammeln1[j][c] = werte[werteindex++];
                gitterwerte_sammeln2[j++][c] = werte[werteindex++];
                #endif
                #ifdef dimensionen3
                gitterwerte_sammeln1[j][c] = werte[werteindex++];
                gitterwerte_sammeln2[j][c] = werte[werteindex++];
                gitterwerte_sammeln3[j++][c] = werte[werteindex++];
                #endif
                #ifdef dimensionen4
                gitterwerte_sammeln1[j][c] = werte[werteindex++];
                gitterwerte_sammeln2[j][c] = werte[werteindex++];
                gitterwerte_sammeln3[j][c] = werte[werteindex++];
                gitterwerte_sammeln4[j++][c] = werte[werteindex++];
                #endif
                }
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
            cout << endl;
#endif
            }
        }
    else
        {
        for(i=2; i<hoechstindex;)
            {
            a = koord[i++];
            b = koord[i++];
            c = koord[i++];

            // f - a, g1 - b, g2 - c
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
            cout << " f " << a << " g1 " << b << " g2 " << c << " ";
#endif
            for(j=b; j<=c;)
                {
                #ifdef dimensionen1
                cout << werte[werteindex] << " ";
                gitterwerte_sammeln1[a][j] = werte[werteindex++];
                gitterwerte_sammeln2[a][j++] = werte[werteindex++];
                #endif
                #ifdef dimensionen2
                gitterwerte_sammeln1[a][j] = werte[werteindex++];
                gitterwerte_sammeln2[a][j++] = werte[werteindex++];
                #endif
                #ifdef dimensionen3
                gitterwerte_sammeln1[a][j] = werte[werteindex++];
                gitterwerte_sammeln2[a][j]= werte[werteindex++];
                gitterwerte_sammeln3[a][j++] = werte[werteindex++];
                #endif
                #ifdef dimensionen4
                gitterwerte_sammeln1[a][j] = werte[werteindex++];
                gitterwerte_sammeln2[a][j] = werte[werteindex++];
                gitterwerte_sammeln3[a][j] = werte[werteindex++];
                gitterwerte_sammeln4[a][j++] = werte[werteindex++];
                #endif
                }
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
            cout << endl;
#endif
            }
        }
#ifdef AUSGABEN_UEBERTRAGE_GITTERWERTE
    cout << " Ende trage werte einmal ein " << endl;
    cout << " ENDEtwee " << endl;
    cout << " In AUG: dieses / rang " << rang << endl;
#endif
    }

// Kommentar: Hier wird etwas wichtiges gemacht: Kern 0 stellt bei dieser Kommunikation einen Flaschenhals da.
// Das betrifft insbesondere den Empfang von Koordinaten und Werten. Mit dieser Funktion hier wird erreicht, dass nachdem die
// Koordinaten und Werte angekommen sind, schon entpackt werden kann, auch wenn noch nicht zu allen Kernen diese
// Informationen angekommen sind.
void UebertrageGitterwerte::empfangen0(void)
// Mit dieser Funktion wird bereits entpacken, whrend andere Pakete noch unterwegs sind.
    {
// Kommentar: Zuerst wird das Feld uebertrage_gitterwerte_flag mit false initiiert.
    for(i_flag=1; i_flag<groesse; i_flag++)
        {
        uebertrage_gitterwerte_flag[i_flag] = false;
        }
// Kommentar: Dann wird die Anzahl an true - flags auf 1 gesetzt, und die Schleife abgearbeitet, bis alle Flags true sind.
    for(anzahl_flag_true = 1; anzahl_flag_true < groesse;)
        {
// Kommentar: Man geht alle flags durch.
        for(i_flag=1; i_flag<groesse; i_flag++)
            {
// Kommentar: Ist flag i false, dann gibt es dort noch was zu tun.
            if(uebertrage_gitterwerte_flag[i_flag] == false)
                {
// Kommentar: Man testet, ob die Koordinaten und Werte dafr angekommen sind.
                MPI_Test(&(empfangs_requests[i_flag]), & flag, & empfangs_status);
                if(flag == true)
                    {
                    MPI_Test(&(empfangs_requests2[i_flag]), & flag, & empfangs_status);
                    if(flag == true)
                        {
// Kommentar: Falls ja, dann wird das uebertrage_gitterwerte_flag dort true, die Werte werden entpackt, und die Anzahl an true
// Flags wird um eins erhht.
                        anzahl_flag_true++;
                        uebertrage_gitterwerte_flag[i_flag] = true;
                        trage_werte_einmal_ein_gleich_0(i_flag);
                        }
                    }
                }
            }
        }
    }
