#include "werte_sammeln.h"
#include "ausgaben.h"

// Kommentar: Diese Klasse dient der Ausgabe - Klasse DateiAusgabefuerWertevergleich (kurz DAFWV). Denn die Werte der parallelen
// Ausfhrung sind verteilt, und mssen fr die Dateiausgabe beim Kern 0 gesammelt werden - es sei denn, dass
// nur ein Kern arbeitet, oder man im 'Kappen - Bereich' arbeitet, wo auch nur ein Kern da ist.

// Kommentar: Die Initialisierungsfunktion. Die Einzelheiten - siehe die Kommentare in dieser Funktion
void WerteSammeln::initialisieren_werte_sammeln(int _n, int _m, int _h_gesamt, int _groesse, int _rang, ParallelesRechnen * _paralleles_rechnen, char * _auswertung_dateiname)
    {
// Kommentar: Die 7 bergabeparameter werden kopiert.
    n = _n;
    m = _m;
    groesse = _groesse;
    rang = _rang;
    h_gesamt = _h_gesamt;

    paralleles_rechnen = _paralleles_rechnen;
    auswertung_dateiname = _auswertung_dateiname;

#ifdef AKTUELLE_KOMMENTARE_NEU
    cout << " init_werte_sammeln n " << n << " m " << m << " groesse " << groesse << " rang " << rang << " h_gesamt " << h_gesamt;
    cout << " auswertung_dateiname " << auswertung_dateiname << endl;
#endif

// Kommentar: Fr die folgenden Variablen werden Werte ausgegeben. Daher mssen die Zeiger darauf kopiert werden.
// Bemerkung dazu: Es werden nur Zeiger kopiert. Das hier verbraucht kaum Speicherplatz.
    u0 = new double ** [h_gesamt];
    r0 = new double ** [h_gesamt];
    e0 = new double ** [h_gesamt];
    rS0 = new double ** [h_gesamt];
    Iu0 = new double ** [h_gesamt];

    for(i=0; i<h_gesamt; i++)
        {
        u0[i] = paralleles_rechnen[i].u0;
        e0[i] = paralleles_rechnen[i].e0;
        r0[i] = paralleles_rechnen[i].r0;
        rS0[i] = paralleles_rechnen[i].rS0;
        Iu0[i] = paralleles_rechnen[i].Iu0;
        }

// Kommentar: Das folgende sind die Gittergren der jeweiligen Ebene.
    n_h = new int [h_gesamt];
    m_h = new int [h_gesamt];

    for(i=0, n_hh = n, m_hh = m; i<h_gesamt; i++)
        {
        n_h[i] = n_hh;
        m_h[i] = m_hh;

        n_hh = (n_hh+1) >> 1;
        m_hh = (m_hh+1) >> 1;
        }

// Kommentar: Bei nur einem Kern
    if(groesse == 1)
        {
// Kommentar: Punkte Indikatormatrix und Wertematrix anlegen
        matrizen_anlegen();
#ifdef AKTUELLE_KOMMENTARE_NEU
        cout << " Auswertung Dateiname " << auswertung_dateiname << endl;
#endif
// Kommentar: Datei Ausgabe erffnen
        datei_ausgabe.initialisieren(rang, auswertung_dateiname);
        datei_ausgabe.oeffnen();
// Kommentar: Eine erste Auegabe der Punkte Indikatormatrix bewirken.
        t_v = -1;
// Kommentar: Und fertig mit der Initialisierung.
        return;
        }

// Kommentar: Die Buffer anlegen. Dafr zuerst die Gre bestimmen.
    buffergroesse = m * n;
    buffergroesse_doppelt = buffergroesse << 1;

    if(rang == 0)
        {
// Kommentar: Bei rang = 0 Empfangsbuffer zu allen anderen Kernen anlegen
        buffer_anlegen_empfangen();
// Kommentar: Neben den Buffern die Wertematrix und Punkteindikatormatrix anlegen - fr die Ausgabe
        matrizen_anlegen();
// Kommentar: Den Rest fr den Empfang anlegen.
        flags_anlegen(); // empfangen
        requests_anlegen(); // empfangen
        bufferlaengen = new int [groesse];
        }
    else
        {
// Kommentar: Fr Kerne != 0 die Sendebuffer anlegen
        buffer_anlegen_senden();
        }
#ifdef AKTUELLE_KOMMENTARE_NEU
    cout << " LLKKJJ Auswertung Dateiname " << auswertung_dateiname << " rang " << rang << endl;
#endif
// Kommentar: Datei Ausgabe erffnen
    datei_ausgabe.initialisieren(rang, auswertung_dateiname);
    datei_ausgabe.oeffnen();
// Kommentar: Eine erste Auegabe der Punkte Indikatormatrix bewirken.
    t_v = -1;
    }

// Den Koordinatenbuffer und den Wertebuffer fr Kerne mit ID > 0 anlegen - zwecks Empfang
void WerteSammeln::buffer_anlegen_empfangen(void)
    {
    fg_buffer_empfangen = new short int * [groesse];
    werte_buffer_empfangen = new double * [groesse];
    for(i=1; i<groesse; i++)
        {
        fg_buffer_empfangen[i] = new short int [buffergroesse_doppelt];
        werte_buffer_empfangen[i] = new double [buffergroesse];
        }
    }

// Frs Senden den Buffer fr das Verschicken an Kern 0 anlegen
void WerteSammeln::buffer_anlegen_senden(void)
    {
    fg_buffer_senden = new short int [buffergroesse_doppelt];
    werte_buffer_senden = new double [buffergroesse];
    }

// Kommentar: Legt die Werte und Punkte Indikatormatrix an.
void WerteSammeln::matrizen_anlegen(void)
    {
    werte_matrix = new double * [n];
    punkte_indikator_matrix = new bool * [n];
    for(i=0; i<n; i++)
        {
        werte_matrix[i] = new double [m];
        punkte_indikator_matrix[i] = new bool [m];
        for(j=0; j<m; j++)
            {
            werte_matrix[i][j] = 0.0;
            punkte_indikator_matrix[i][j] = false;
            }
        }
    }

// Kommentar: Den bei der Initialisierung angelegten Speicher freigeben.
void WerteSammeln::finalisieren_werte_sammeln(void)
    {
// Kommentar: Die Datei - Ausgabe beenden.
    datei_ausgabe.schliessen();

// Kommentar: Die Zeiger auf die Variablenmatrizen freigeben
    delete u0;
    delete r0;
    delete e0;
    delete rS0;
    delete Iu0;

// Kommentar: In diesem Fall Punkte Indikatormatrix und die Wertematrix befreien
    if(groesse == 1)
        {
        matrizen_befreien();
        return;
        }
    if(rang == 0)
        {
// Kommentar: In diesem Fall Punkte Indikatormatrix und die Wertematrix befreien und die Empfangsstrukturen befreien
        buffer_befreien_empfangen();
        matrizen_befreien();
        flags_befreien();
        requests_befreien();
        delete bufferlaengen;
        }
    else
        {
// Kommentar: Die Sende - Strukturen befreien
        buffer_befreien_senden();
        }
    }

// Kommentar: Die Empfangsbuffer befreien
void WerteSammeln::buffer_befreien_empfangen(void)
    {
    for(i=1; i<groesse; i++)
        {
        delete fg_buffer_empfangen[i];
        delete werte_buffer_empfangen[i];
        }
    delete fg_buffer_empfangen;
    delete werte_buffer_empfangen;
    }

// Kommentar: Die Sendebuffer befreien
void WerteSammeln::buffer_befreien_senden(void)
    {
    delete fg_buffer_senden;
    delete werte_buffer_senden;
    }

// Kommentar: Die Wertematrix und Punkte Indikatormatrix befreien
void WerteSammeln::matrizen_befreien(void)
    {
    for(i=0; i<n; i++)
        {
        delete werte_matrix[i];
        delete punkte_indikator_matrix[i];
        }
    delete werte_matrix;
    delete punkte_indikator_matrix;
    }

// Kommentar: Die Funktion tun

void WerteSammeln::tun(void)
    {
// Kommentar: Fhrt nur ein Kern den Programmcode aus, wird ein anderes tun aufgerufen, als wenn mehrere Kerne den
// Programmcode aufrufen. Im ersterem Fall muss nicht kommuniziert werden.
    if(groesse == 1)
        {
        tun_ein_kern();
        }
    else
        {
        tun_mehrere_kerne();
        }
    }

// Kommentar: Fhrt tun bei mehreren beteiligten Kernen aus
void WerteSammeln::tun_mehrere_kerne(void)
    {
// Fr den ausgebenden Kern sind die folgenden Schritte zu tun
    if(rang == 0)
        {
// Kommentar: Falls die Punkte Indikatormatrix dieses Mal mit ausgegeben werden soll:
        if(punkte_indikator_matrix_schreiben == true)
            {
// Kommentar: Beide Matrizen resetten
            matrizen_reset();
            }
        else
            {
// Kommentar: Sonst nur die Wertematrix resetten
            matrix_reset();
            }
#ifdef AKTUELLE_KOMMENTARE_NEU
        cout << " DAfWV tun mehrere Kerne 1 ebene " << ebene << " rang " << rang << endl;
#endif
// Kommunikationsflags zurcksetzen
        flags_reset();
#ifdef AKTUELLE_KOMMENTARE_NEU
        cout << " DAfWV tun mehrere Kerne 2 ebene " << ebene << " rang " << rang << endl;
#endif
// Kommentar: Die Kommunikation fr den Empfang initiieren
        kommunikation_empfangen_initiieren();
#ifdef AKTUELLE_KOMMENTARE_NEU
        cout << " DAfWV tun mehrere Kerne 3 ebene " << ebene << " rang " << rang << endl;
#endif

// Kommentar: Solange die Werteflags nicht alle true sind:
        for(;test_auf_beenden_empfangen()==false;)
            {
// Kommentar: ndere die Flags fr einkommende Nachrichten und bearbeite diese Nachrichten
            flags_aendern_und_schreiben();
            }
#ifdef AKTUELLE_KOMMENTARE_NEU
        cout << " DAfWV tun mehrere Kerne 4 ebene " << ebene << " rang " << rang << endl;
#endif
// Kommentar: Falls die Punkte Indikatormatrix geschrieben werden soll:
        if(punkte_indikator_matrix_schreiben == true)
            {
// Kommentar: Bestimme den eigenen Anteil an der Wertematrix sowie der Punkte Indikatormatrix
            tun_eigenes_beides();
            }
        else
            {
// Kommentar: Bestimme den eigenen Anteil an der Wertematrix
            tun_eigenes_eines();
            }
#ifdef AKTUELLE_KOMMENTARE_NEU
        cout << " DAfWV tun mehrere Kerne 5 ebene " << ebene << " rang " << rang << endl;
#endif
        }
// Kommentar: Fr die sendenden Kerne
    else
        {
// Kommentar: Packe Koordinaten und Werte der entsprechenden Ebene
        packen(ebene);
// Kommentar: Initiiere die sendende Kommunikation
        kommunikation_senden_initiieren();
// Kommentar: Warte, bis die Nachrichten raus sind.
        for(;test_auf_beenden_senden()==false;)
            {
            ;
            }
        }
    }

// Kommentar: Packen der sendenden Kerne
void WerteSammeln::packen(int ebene)
    {
// Kommentar: Zugriff auf paralleles rechnen, um die Koordinaten und Werte lesen zu knnen.
    paralleles_rechnen_ebene = &(paralleles_rechnen[ebene]);
// Kommentar: Reset der Buffer Indizes
    bufferindex_werte = 0;
    bufferindex_koordinaten = 0;
// Kommentar: Durchlaufe alle Gitterpunkte
    for(paralleles_rechnen_ebene->anfangsposition();paralleles_rechnen_ebene->ist_fertig_fl()==false; paralleles_rechnen_ebene->naechster())
        {
// Kommentar: Lese die Koordinaten
        f = paralleles_rechnen_ebene->lesen_f();
        g = paralleles_rechnen_ebene->lesen_g();

// Kommentar: Packe die Koordinaten
        fg_buffer_senden[bufferindex_koordinaten++] = f;
        fg_buffer_senden[bufferindex_koordinaten++] = g;
// Kommentar: Packe die Werte zu den Koordinaten
        werte_buffer_senden[bufferindex_werte++] = ausgabe_werte[f][g];
        }
#ifdef AKTUELLE_KOMMENTARE_NEU
    cout << " WS::packen rang " << rang << " laenge " << bufferindex_koordinaten << endl;
#endif
    }

// Kommentar: Iniitiere die Kommunikation on Koordinaten und Werten mit MPI_Isend
void WerteSammeln::kommunikation_senden_initiieren(void)
    {
    MPI_Isend(fg_buffer_senden, bufferindex_koordinaten, MPI_SHORT, 0, TAG_WERTE_SAMMELN, MPI_COMM_WORLD, &fg_sende_request);
    MPI_Isend(werte_buffer_senden, bufferindex_werte, MPI_DOUBLE, 0, TAG_WERTE_SAMMELN + 1, MPI_COMM_WORLD, &werte_sende_request);
    }

// Kommentar: Beende Senden wenn Koordinaten - und Wertetest abgeschlossen sind.
bool WerteSammeln::test_auf_beenden_senden(void)
    {
    MPI_Test(&fg_sende_request, &flag, & status_senden);
    if(flag == false)
        {
        return(false);
        }
    MPI_Test(&werte_sende_request, &flag, & status_senden);
    if(flag == false)
        {
        return(false);
        }
    return(true);
    }

// Kommentar: Flags fr das Empfangen von Koordinaten und Werte anlegen. Man braucht unterschiedlich Flags,
// um den Fortschritt zu sehen.
void WerteSammeln::flags_anlegen(void)
    {
    fg_empfangen_flags = new bool[groesse];
    werte_empfangen_flags = new bool[groesse];
    for(i=1; i<groesse; i++)
        {
        fg_empfangen_flags[i] = false;
        werte_empfangen_flags[i] = false;
        }
    }

// Kommentar: Befreien des Speichers fr die Flags.
void WerteSammeln::flags_befreien(void)
    {
    delete fg_empfangen_flags;
    delete werte_empfangen_flags;
    }

// Kommentar: Die Requests frs Empfangen anlegen - wird fr die Tests auf Abschlu von Kommunikation bentigt.
void WerteSammeln::requests_anlegen(void)
    {
    fg_requests_empfangen = new MPI_Request[groesse];
    werte_requests_empfangen = new MPI_Request[groesse];
    }

// Kommentar: Den Speicher fr die Empfangs - Requests freigeben.
void WerteSammeln::requests_befreien(void)
    {
    delete fg_requests_empfangen;
    delete werte_requests_empfangen;
    }

// Kommentar: Das Empfangen ist beendet, wenn alle werte_empfangen flags der Kerne mit Rang > 0 true sind.
bool WerteSammeln::test_auf_beenden_empfangen(void)
    {
    for(i=1; i<groesse; i++)
        {
        if(werte_empfangen_flags[i] == false)
            {
            return(false);
            }
        }
    return(true);
    }

// Kommentar: Die Kommunikation fr Koordinaten und Werte frs Empfangen wird mit MPI_Irecv initiiert.
void WerteSammeln::kommunikation_empfangen_initiieren(void)
    {
    for(i=1; i<groesse; i++)
        {
        MPI_Irecv(fg_buffer_empfangen[i], buffergroesse_doppelt, MPI_SHORT, i, TAG_WERTE_SAMMELN, MPI_COMM_WORLD, &(fg_requests_empfangen[i]));
        MPI_Irecv(werte_buffer_empfangen[i], buffergroesse, MPI_DOUBLE, i, TAG_WERTE_SAMMELN + 1, MPI_COMM_WORLD, &(werte_requests_empfangen[i]));
        }
    }

// Kommentar: Es wird - frs Empfangen - geprft, ob Nachrichten angekommen sind. Diese werden dann bearbeitet.
void WerteSammeln::flags_aendern_und_schreiben(void)
    {
// Kommentar: Man geht alle Kerne durch
    for(i=1; i<groesse; i++)
        {
// Kommentar: Wurde fg_empfangen bisher noch nicht festgestellt
        if(fg_empfangen_flags[i] == false)
            {
// Kommentar: Mache den Test, ob jetzt Koordinaten angekommen sind.
            MPI_Test(&(fg_requests_empfangen[i]), &flag, &status_empfangen);
// Kommentar: Falls ja, dann setzt das fg_empfangen flag auf true, und bestimme die Lnge des Koordinatenbuffers
            if(flag == true)
                {
#ifdef AKTUELLE_KOMMENTARE_NEU
                cout << " flags aendern und schreiben 1 i " << i << " rang " << rang << endl;
#endif
                fg_empfangen_flags[i] = true;
                MPI_Get_count(&status_empfangen, MPI_SHORT, &laenge);
                bufferlaengen[i] = laenge;
#ifdef AKTUELLE_KOMMENTARE_NEU
                cout << " flags aendern und schreiben 2 i " << i << " rang " << rang << " laenge " << laenge << endl;
#endif
                }
            }
        else
// Kommentar: Die Koordinaten sind angekommen
            {
            if(werte_empfangen_flags[i] == false)
                {
                MPI_Test(&(werte_requests_empfangen[i]), &flag, &status_empfangen);
// Kommentar: Sind auch die Werte angekommen
                if(flag == true)
                    {
#ifdef AKTUELLE_KOMMENTARE_NEU
                    cout << " flags aendern und schreiben 3 i " << i << " rang " << rang << endl;
#endif
// Kommentar: Vermerke, dass fr diesen Kern die Kommunikation abgeschlossen ist.
                    werte_empfangen_flags[i] = true;
// Kommentar: Entpacken Koordinaten und Werte in Wertematrix und ggf. Punkteindikatormatrix.
                    entpacken(i);
#ifdef AKTUELLE_KOMMENTARE_NEU
                    cout << " flags aendern und schreiben 4 i " << i << " rang " << rang << endl;
#endif
                    }
                }
            }
        }
    }

// Kommentar: Entpacke die Nachrichten von Kern '_kern_ID'
void WerteSammeln::entpacken(int _kern_ID)
    {
    kernID = _kern_ID;
// Kommentar: Je nach punkte_indikator_matrix_schreiben wird nur die Wertematrix, oder es werden die
// Wertematrix und die Punkte Indikatormatrix entpackt.
    if(punkte_indikator_matrix_schreiben==true)
        {
        entpacken_beides();
        }
    else
        {
        entpacken_eines();
        }
    }

// Kommentar: Resette die Empfangsflag vor dem Empfange
void WerteSammeln::flags_reset(void)
    {
    for(i=1; i<groesse; i++)
        {
        fg_empfangen_flags[i] = false;
        werte_empfangen_flags[i] = false;
        }
    }

// Kommentar: Resette die Wertematrix und die Punkte Indikatormatrix
void WerteSammeln::matrizen_reset(void)
    {
    n_hh = n_h[ebene];
    m_hh = m_h[ebene];

    for(i=0; i<n_hh; i++)
        {
        for(j=0; j<m_hh; j++)
            {
            werte_matrix[i][j] = 0.0;
            punkte_indikator_matrix[i][j] = false;
            }
        }
    }

// Kommentar: Resette die Wertematrix
void WerteSammeln::matrix_reset(void)
    {
    n_hh = n_h[ebene];
    m_hh = m_h[ebene];

    for(i=0; i<n_hh; i++)
        {
        for(j=0; j<m_hh; j++)
            {
            werte_matrix[i][j] = 0.0;
//            punkte_indikator_matrix[i][j] = false;
            }
        }
    }

// Kommentar: Entpacke die Wertematrix und die Punkteindikatormatrix
void WerteSammeln::entpacken_beides(void)
    {
// Kommentar: Bestimme die Ausmae der Matrizen - entsprechend der Ebene
    n_hh = n_h[ebene];
    m_hh = m_h[ebene];

#ifdef AKTUELLE_KOMMENTARE_NEU
    cout << " entpacken beides 2 ebene " << ebene << " kernID " << kernID << endl;
#endif

// Kommentar: Entsprechend der ID des Kerns werden die Buffer in die Variablen fg_buffer und werte_buffer gelesen.
    fg_buffer = (fg_buffer_empfangen[kernID]); // stimmt das? Ja.
    werte_buffer = (werte_buffer_empfangen[kernID]); // stimmt das? Ja.
#ifdef AKTUELLE_KOMMENTARE_NEU
    cout << " entpacken beides 3 ebene " << ebene << " kernID " << kernID << " laenge " << bufferlaengen[kernID] << endl;
#endif

    for(i=0, laenge = bufferlaengen[kernID], j=0; i<laenge;)
        {
// Kommentar: Zuerst werden die Koordinaten ausgelesen
        f = fg_buffer[i++];
        g = fg_buffer[i++];

#ifdef AKTUELLE_KOMMENTARE_NEU
        cout << " entpacken beides 3a ebene " << ebene << " kernID " << kernID << " f " << f << " g " << g << endl;
#endif
// Kommentar: Dort wird die Punkte Indikatormatrix auf true gesetzt. Der Punkt existiert auf dem Kern mit der kernID
        punkte_indikator_matrix[f][g] = true;
// Kommentar: Der Wertebuffer fr diese Koordinate wird ausgelesen, und in die Wertematrix gechrieben.
        werte_matrix[f][g] = werte_buffer[j++];
// Kommentar: Die folgenden Ausgaben tragen nicht zur Funktion des Programmes bei.
#ifdef AKTUELLE_KOMMENTARE_NEU
        if((f==25)&&(g==8))
            {
            cout << " IJK (25,8) liegt auf Prozessor " << kernID << endl;
            }
#endif
        }
#ifdef AKTUELLE_KOMMENTARE_NEU
    cout << " entpacken beides 4 ebene " << ebene << " kernID " << kernID << endl;
#endif
    }

// Kommentar: Die Wertematrix und die Punkte Indikatormatrix werden beschrieben - fr den eigenen Kern, d.h. ohne
// Kommunikation
void WerteSammeln::tun_eigenes_beides(void)
    {
// Kommentar: Zugriff auf paralleles rechnen, um die Koordinaten und Werte zu lesen
    paralleles_rechnen_ebene = &(paralleles_rechnen[ebene]);
// Kommentar: Lauf durch alle Gitterpunkte
    for(paralleles_rechnen_ebene->anfangsposition();paralleles_rechnen_ebene->ist_fertig_fl()==false; paralleles_rechnen_ebene->naechster())
        {
// Kommentar: Lesen der Koordinaten
        f = paralleles_rechnen_ebene->lesen_f();
        g = paralleles_rechnen_ebene->lesen_g();
#ifdef AUSGABEN_WS
        cout << " tun_eigenes_beides ebene " << ebene << " f " << f << " g " << g << " rang " << rang << endl;
#endif
// Kommentar: Schreiben in die Wertematrix und Punkte Indikatormatrix
        werte_matrix[f][g] = ausgabe_werte[f][g];
        punkte_indikator_matrix[f][g] = true;
        }
    }

// Kommentar: Genau wie entpacken_beides, nur dass jetzt nur die Wertematrix beschrieben wird. Es wird ausgepackt,
// und damit die Wertematrix beschrieben.
void WerteSammeln::entpacken_eines(void)
    {
// Kommentar: Bestimmung der Gittergre
    n_hh = n_h[ebene];
    m_hh = m_h[ebene];

// Kommentar: Bestimmen der Buffer zu der kernID
    fg_buffer = (fg_buffer_empfangen[kernID]); // stimmt das?
    werte_buffer = (werte_buffer_empfangen[kernID]); // stimmt das?

// Kommentar: Durchlaufen der Koordinaten
    for(i=0, laenge = bufferlaengen[kernID], j=0; i<laenge;)
        {
// Kommentar: Lesen der Koordinaten
        f = fg_buffer[i++];
        g = fg_buffer[i++];
// Kommentar: Schreiben der Werte
        werte_matrix[f][g] = werte_buffer[j++];
#ifdef AKTUELLE_KOMMENTARE_NEU
        if((f==25)&&(g==8))
            {
            cout << " IJK (25,8) liegt auf Prozessor " << kernID << endl;
            }
#endif
        }
    }

// Kommentar: Fr die eigenen Gitterpunkte werden die Ausgabewerte in die Wertematrix kopiert.
void WerteSammeln::tun_eigenes_eines(void)
    {
// Kommentar: paralleles_rechnen wird fr die Ebene aktiviert.
    paralleles_rechnen_ebene = &(paralleles_rechnen[ebene]);
// Kommentar: Man durchluft die Gitterpunktemenge
    for(paralleles_rechnen_ebene->anfangsposition();paralleles_rechnen_ebene->ist_fertig_fl()==false; paralleles_rechnen_ebene->naechster())
        {
// Kommentar: Bestimmen der Koordinaten der Gitterpunkte
        f = paralleles_rechnen_ebene->lesen_f();
        g = paralleles_rechnen_ebene->lesen_g();
// Kopieren der Ausgabewerte in die Wertematrix
        werte_matrix[f][g] = ausgabe_werte[f][g];
        }
    }

// Kommentar: Fr einen Kern wird ohne Kommunikation nur kopiert, und zwar entweder die Wertematrix, oder auch noch
// zustzlich die Punkte Indikatormatrix
void WerteSammeln::tun_ein_kern(void)
    {
// Kommentar: wre nicht notwendig. Aber: Beim Test kommt es auf die Laufzeiten nicht so strikt an.

// Kommentar: Falls die Punkte Indikatormatrix erstellt wird:
    if(punkte_indikator_matrix_schreiben == true)
        {
// Kommentar: Setzen der Matrizen in den Ausgangszustand
        matrizen_reset();
// Kommentar: Bestimmen von Wertematrix und Punkte Indikatormatrix
        tun_eigenes_beides();
        }
    else
        {
// Kommentar: Rcksetzen der Wertematrix
        matrix_reset();
// Kommentar: Kopieren in die Wertematrix
        tun_eigenes_eines();
        }
    }

// Kommentar: 'tun' auf den Ebenen, wo die Kappe wirkt.
void WerteSammeln::tun_kappe(void)
    {
// Kommentar: Fr rang!=0 Kerne ist nichts zu tun.
    if(rang!=0)
        {
//        matrix_reset();
        return;
        }
// Kommentar: Bestimmung der Gre der Matrix
    n_hh = n_h[ebene];
    m_hh = m_h[ebene];

#ifdef AKTUELLE_KOMMENTARE_NEU
//    cout << " entpacken beides 1 kappe ebene " << ebene << " kernID " << kernID << endl;
#endif

// Kommentar: Kopieren der Ausgabewerte in die Wertematrix.
    for(i=0; i<n_hh; i++)
        {
        for(j=0; j<m_hh; j++)
            {
            werte_matrix[i][j] = ausgabe_werte[i][j];
            }
        }
    }
