#include "farbenupdate.h"

// Kommentar: Der FarbenUpdate ist ein wesentlicher Schritt der parallelen adaptiven Verfeinerung. Ausgehend von
// verschiedenen Rndern sowie P+ / P- bzw. ep-ff kommt es zu einer Aktualisierung des Farbenfeldes aufgrund
// der parallelen adaptiven Verfeinerung.
// Unwissend ist ein Punkt genau dann, wenn er nicht die EP - Informationen hat, und das ist der Fall, wenn
// er kein gemeinsamer Punkt ist und er von anderer Farbe ist als die benachbarten gemeinsamen Punkte.
// Vergleiche die Funktion 'ist_unwissend'.

// Kommentar: Damit wird die PunktFeldListe der unwissenden Punkte zurckgesetzt auf den Anfangszustand.
void FarbenUpdate::reset_unwissend(void)
    {
    unwissende_punkte.reset();
    }

// Kommentar: Der fr diese Klasse bentigte Speicherplatz wird bei der Finalisierung wieder freigegeben.
void FarbenUpdate::finalisieren(void)
    {
// Kommentar: Die PunktFeldListen punktfeldliste und unwissende_punkte werden finalisiert, d.h. ihr Speicherplatz wird freigegeben.
    punktfeldliste.finalisierung();
    unwissende_punkte.finalisierung();
    }

// Kommentar: Die Klasse FarbenUpdate wird hier initialisiert. Wichtige Zeiger werden kopiert, und es wird
// Speicherplatz fr die genutzten Feldlisten bereitgestellt.
void FarbenUpdate::initialisieren(short int ** _fgff, short int ** _ggff, int ** _feingitter_snb, short int ** _ep_ff, FeldListe * _dynamischer_rand, Kommunikation4er * _kommunikation4er, int * _PT_feingitter, DateiAusgabe * _da_fu, int ** _grobgitter_feld, int ** _feingitter_feld, EP_SNBupdate * _ep_snb_update)
// wird finalisiert
    {
// Kommentar: Zuerst erfolgt ein kopieren der bentigten Zeiger, wie diverse Grobgitter - und Feingitterinformationen,
// vorneweg das Grobgitter - und Feingitterfarbenfeld - ggff und fgff. (GrobGitterFarbenFeld - ggff,
// Feingitterdarbenfeld - fgff)
    fgff = _fgff;
    ggff = _ggff;

// Kommentar: Das sind die Felder 'feld' der Feldlisten des groben und feinen Gitters.
    grobgitter_feld = _grobgitter_feld;
    feingitter_feld = _feingitter_feld;

// Kommentar: Das ist die Struktur selbstnachbarn des feinen Gitters.
    feingitter_snb = _feingitter_snb;
// Kommentar: Das ist das Feld ep_ff - als Aggregation der Informationen aus P+ und P-.
    ep_ff = _ep_ff;
// Kommentar: Der dynamische Rand ermglicht es, den Innenrand schnell zu durchlaufen, und ihn auch schnell zu aktualisieren.
    dynamischer_rand = _dynamischer_rand;
// Kommentar: Kommunikation4er ermglicht es, die zu verschickenden Informationen von Punkte - und Farbenupdate
// sowohl fr einzufgende als auch zu entfernende Punkte gemeinsam zu kommunizieren.
    kommunikation4er = _kommunikation4er;
// Kommentar: Die Topologie auf dem feinen Gitter bezglich der Eigenschaft, benachtbarte Gitterpunkte zu haben.
    PT_feingitter = _PT_feingitter;
// Kommentar: Anzahl der Kerne, sowie die eigene ID = Identittszahl.
    groesse = kommunikation4er->groesse;
    rang = kommunikation4er->rang;
// Kommentar: Die Gre des eigenen Gitters.
    m = kommunikation4er -> m;
    n = kommunikation4er -> n;
// Kommentar: Jeweils die Gre des eigenen Gitters -1 - zum schnelleren Zugriff auf diese Zahl.
    m_minus1 = m-1;
    n_minus1 = n-1;

// Kommentar: Die Klasse ep_snb_update zur Aktualisierung des Feldes SNB.
    ep_snb_update = _ep_snb_update;
    ep_snb = ep_snb_update->SNB;

// Kommentar: Die 2 PunktFeldListen: unwissende_punkte speichert die unwissenden Punkte, die besonders
// behandelt werden mssen. Die punktfeldliste selber wird nur bei schon berholten Funktionen gebraucht.
    punktfeldliste.initialisierung(m, n); // Nur beim Test! Kann ich hier Speicherplatz gewinnen? spter
    unwissende_punkte.initialisierung(m, n);
// Kommentar: da_fu ist die Datei Ausgabe der Klasse FarbenUpdate.
    da_fu = _da_fu;
// Kommentar: Setze char = -1
    farbe_short_int = -1;
    farbe_char_m1 = (char) farbe_short_int;
    }

// Kommentar: Speicherplatz von PunkteListen wird in diese Klasse ohne weitere Allokationen verwendbar.
void FarbenUpdate::initialisieren_punkte(short int * _punkte_f, short int * _punkte_g)
// braucht nicht finalisiert zu werden
    {
    punkte_f = _punkte_f;
    punkte_g = _punkte_g;
    }

///////////////////////////////////////////////////
//////////// Vervollstndigen  ////////////////////
///////////////////////////////////////////////////

// VV

// Kommentar: Diese Funktion ist definiert, und wird in der Klasse AdaptiveVerfeinerung verwendet.
// Sie hat folgenden Zweck: Aufgrund des Einfgens von Gitterpunkten gibt es einen Farbenupdate. Teil dessen ist
// eine Kommunikation, und das, was da verschickt wird, wird hier behandelt.
void FarbenUpdate::vervollstaendigen_einfuegen(void)
    {
// Kommentar: Man durchluft die ID's aller Kerne
    for(i=0; i<groesse; i++)
        {
// Kommentar: Falls zu dem Kern i auf dem groben oder feinen Gitter eine Punktnachbarschaft existiert, mache die
// folgenden Schritte.
        if(kommunikation4er->PT_feingitter_und_grobgitter[i] > 0) // 21.2.2015
//
            {
// Kommentar: Das folgende sind optionale Ausgaben zum Fehlerfinden.
#ifdef AUSGABEN_FARBENUPDATE
            if(rang==0)
                {
                cout << " ZZZY fr Kern " << i << endl;
                }
#endif
#ifdef AAHH
            if((rang==2)&&(t==7))
                {
                cout << " AAHH vervoll-einfuegen fgff " << fgff[19][7] << " i " << i << endl;
                }
#endif
#ifdef AUSGABEN_FARBENUPDATE
            cout << " ** rang " << rang << " i " << i << " 1 " << endl;
#endif
// Kommentar: Der Empfangsbuffer zum Kern i wird geladen.
            empfangsbuffer = kommunikation4er->empfangsbuffer[i];
#ifdef AUSGABEN_FARBENUPDATE
            cout << " ** rang " << rang << " i " << i << " 2 " << endl;
#endif
// Kommentar: Anfangs - und Endposition des Empfangsbuffers fr FarbenUpdate::einfuegen - Kommunikationsbefehle werden
// bestimmt.
            anfangsposition = kommunikation4er->fu_einfuegen_anfang[i];
#ifdef AUSGABEN_FARBENUPDATE
            cout << " ** rang " << rang << " i " << i << " 3 " << endl;
#endif
            endposition = kommunikation4er->fu_einfuegen_ende[i];
#ifdef AUSGABEN_FARBENUPDATE
            cout << " ** rang " << rang << " i " << i << " 4 " << " Anfangsposition " << anfangsposition << " Endposition " << endposition << endl;
#endif
// Kommentar: Der Empfangsbuffer wird fr die Indizes dazwischen durchlaufen.
            for(j=anfangsposition;j<endposition;)
                {
#ifdef AUSGABEN_FARBENUPDATE
                cout << " ** rang " << rang << " i " << i << " j " << j << " 5a " << endl;
#endif
// Kommentar: Die Koordinaten f und g sowie der zu schreibende Farbwert werden aus dem Buffer gelesen.
                f = lese_short_int(j);
#ifdef AUSGABEN_FARBENUPDATE
                cout << " ** rang " << rang << " i " << i << " j " << j << " 5b " << endl;
#endif
                j=j+2;
                g = lese_short_int(j);
#ifdef AUSGABEN_FARBENUPDATE
                cout << " ** rang " << rang << " i " << i << " j " << j << " 5c " << endl;
#endif
                j=j+2;
                farbe = empfangsbuffer[j++];
#ifdef AUSGABEN_FARBENUPDATE
                cout << " ** rang " << rang << " i " << i << " j " << j << " 5e " << " f " << f << " g " << g << " farbe " << farbe << endl;
#endif
                    {
// Kommentar: Der Farbwert wird unter den Koordinaten (f,g) in das Farbenfeld geschrieben.
                    fgff[f][g] = farbe;
                    // Das kann auch schon von einem anderen Punkt her entstanden sein.
#ifdef ZAEHLER_ALT
// Kommentar: Der ebene_ebenen_zaehler wird nicht mehr gefhrt.
                    ebene_ebenen_zaehler_h_minus1->inkrementieren(farbe);
#endif
#ifdef AUSGABEN_DATEI
                    da_eez->aus << " Vervollstaendigen (ink)  Rang " << rang << " Ziel P " << i << " f " << f << " g " << g << " m " << m << endl;
#endif
                    }
                }
            }
        }
    }

// Kommentar: Der Empfangsbuffer ist ein char - Buffer. Mittels dieser Funktion kann man short int - Werte, die
// dort geschrieben wurden, lesen.
short int FarbenUpdate::lese_short_int(int index)
    {
    merken = &(empfangsbuffer[index]);
    merken_si = (short int *) merken;
    return(merken_si[0]);
    }

///////////////////////////////////////////////////
//////////// Vervollstndigen  entfernen //////////
///////////////////////////////////////////////////

// Kommentar: Diese Funktion ist definiert, und wird in der Klasse AdaptiveVerfeinerung verwendet.
// Sie hat folgenden Zweck: Aufgrund des Entfernens von Gitterpunkten gibt es einen Farbenupdate. Teil dessen ist
// eine Kommunikation, und das, was da verschickt wird, wird hier behandelt.
void FarbenUpdate::vervollstaendigen_entfernen(void)
    {
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Rang " << rang << " vervollstaendigen_entfernen " << endl;
#endif
// Kommentar: Es wird eine Schleife zu den ID's aller Kerne durchlaufen.
    for(i=0; i<groesse; i++)
        {
// Kommentar: Falls zu dem Kern i auf dem groben oder feinen Gitter eine Punktnachbarschaft existiert, mache die
// folgenden Schritte.
        if(kommunikation4er->PT_feingitter_und_grobgitter[i] > 0) // 21.2.2015
            {
// Kommentar: Der Empfangsbuffer zum Kern i wird geladen.
            empfangsbuffer = kommunikation4er->empfangsbuffer[i];
// Kommentar: Anfangs - und Endposition des Empfangsbuffers fr FarbenUpdate::einfuegen - Kommunikationsbefehle werden
// bestimmt.
            anfangsposition = kommunikation4er->fu_entfernen_anfang[i];
            endposition = kommunikation4er->fu_entfernen_ende[i];
#ifdef AUSGABEN_FARBENUPDATE
            cout << " Rang " << rang << " vervollstaendigen_entfernen anfangsposition " << anfangsposition << " endposition " << endposition << " Kommt von Kern " << i << endl;
            cout << " Rang " << rang << " vervollstaendigen_entfernen " << " Kommt von Kern " << i;
#endif
// Kommentar: Der Empfangsbuffer wird fr die Indizes dazwischen durchlaufen.
            for(j=anfangsposition;j<endposition;)
                {
// Kommentar: Die Koordinaten f und g werden aus dem Buffer gelesen. Da entfernt wird, wird dorthin der Farbwert -1
// geschrieben, weshalb kein Farbwert zu lesen ist.
                f = lese_short_int(j);
                j=j+2;
                g = lese_short_int(j);
                j=j+2;
#ifdef AUSGABEN_FARBENUPDATE
                cout << " f " << f << " g " << g;
#endif // AUSGABEN_FARBENUPDATE
// Kommentar: Wenn eine -1 vorliegt, dann ist das kein unwissender Punkt, und er wurde schon direkt behandelt,
// und eine zweite Behandlung ist dann nicht ntig, weshalb dann die Schleife weiter durchlaufen wird.
/*
                if(unwissende_punkte.feld[f][g]==-1)
                // unwissende_punkte == -1 <=> kein Eintrag in unwissende Punkte <=> der Punkt (f,g) wurde schon behandelt
                // Die letzte quivalenz ist mir nicht klar. Doch: Er (f,g) wurde behandelt NICHT via Kommunikation.
                // Jetzt ganz klar: Wenn unwissende Punkte, dann nicht doppelt gemoppelt auch noch ein Eintrag
                // via Kommunikation!
                    {
                    continue;
                    }
*/
// Kommentar: Bei einem unwissenden Punkt fhrt diese Kommunikation zum Schreiben des Wertes -1 (entfernen) in das Farbenfeld.
                fgff[f][g] = -1;
#ifdef AUSGABEN_FARBENUPDATE
                cout << " entferne Punkt Rang " << rang << " f " << f << " g " << g << endl;
#endif
#ifdef AUSGABEN_DATEI
                da_eez->aus << " Vervollstaendigen entfernen (dek) Rang " << rang << " Ziel P " << i << " f " << f << " g " << g << " m " << m << endl;
#endif
                }
#ifdef AUSGABEN_FARBENUPDATE
            cout << endl;
#endif // AUSGABEN_FARBENUPDATE
            }
        }
    }

///////////////////////////////////////////////////////
///   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! ////
///////////////////////////////////////////////////////

// Kommentar: Ab hier beginnt der aktuelle Farbenupdate: 6.5.15

// Kommentar: Das Feingitterfarbenfeld fgff wird ausgegeben. Allerdings berall, auch dort, wo es ungltig ist.
void FarbenUpdate::fgff_ausgabe(void)
    {
#ifdef AUSGABEN_FARBENUPDATE
    for(i=0; i<m; i++)
        {
        for(j=0; j<n; j++)
            {
            cout << fgff[i][j] << " ";
            }
        cout << endl;
        }
    cout << endl;
#endif
    }

// Kommentar: Hier wird festgestellt, ob an der Stelle (f,g) ein unwissender Punkt vorliegt.
// Das ist ein Punkt, fr den auf dem jeweiligen Kern die EP - Informationen nicht vorliegen, und fr den
// deshalb auf einem anderen Kern eine Farbennderung entschieden wird.
bool FarbenUpdate::ist_unwissend(int f, int g)
// Ob der Punkt EP - Information hat, oder nicht. Hat er, dann ist er nicht unwissend, d.h. return (false);
    {
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Beginn unwissend / rang " << rang << endl;
#endif
// Kommentar: Dieser Wert gibt an, wo ein Punkt des feinen Gitters in Bezug auf das grobe Gitter liegt.
// Bei 0 liegt er auch auf dem groben Gitter, bei 1 horizontal zwischen zwei Grobgitterpunkten,
// bei 2 vertikal zwischen zwei Grobgitterpunkten, und bei 3 diagonal zwischen 4 Grobgitterpunken.
    switch(((f&1)<<1)+(g&1))
        {
// Kommentar: In diesem Fall liegt ein gemeinsamer Punkt vor - er liegt auf dem groben und feinen Gitter - und der ist
// nie 'unwissend'.
        case 0:
#ifdef AUSGABEN_FARBENUPDATE
            cout << " Ende unwissend 0 / rang " << rang << endl;
#endif
            return(false);
// Kommentar: Der Punkt liegt horizontal zwischen zwei Grobgitterpunkten, die die EP - Information haben. Also wird geprft, ob
// mindestens einer der zwei Grobgitterpunkte auf demselben Kern liegen wie dieser Punkt. Wenn ja, dann ist er wissend,
// also return(false); liegen beide auf einem anderen Kern, dann ist er unwissend, also return(true).
        case 1:
            if(g>0)
                {
                if(fgff[f][g-1] == rang)
                    {
#ifdef AUSGABEN_FARBENUPDATE
                    cout << " Ende unwissend 1a / rang " << rang << endl;
#endif
                    return(false);
                    }
                }
            if(g<m-1)
                {
                if(fgff[f][g+1] == rang)
                    {
#ifdef AUSGABEN_FARBENUPDATE
                    cout << " Ende unwissend 1b / rang " << rang << endl;
#endif
                    return(false);
                    }
                }
#ifdef AUSGABEN_FARBENUPDATE
            cout << " Ende unwissend 1c / rang " << rang << endl;
#endif
            return(true);
// Kommentar: Der Punkt liegt vertikal zwischen zwei Grobgitterpunkten, die die EP - Information haben. Also wird geprft, ob
// mindestens einer der zwei Grobgitterpunkte auf demselben Kern liegen wie dieser Punkt. Wenn ja, dann ist er wissend,
// also return(false); liegen beide auf einem anderen Kern, dann ist er unwissend, also return(true).
        case 2:
            if(f>0)
                {
                if(fgff[f-1][g] == rang)
                    {
#ifdef AUSGABEN_FARBENUPDATE
                    cout << " Ende unwissend 2a / rang " << rang << endl;
#endif
                    return(false);
                    }
                }
            if(f<n-1)
                {
                if(fgff[f+1][g] == rang)
                    {
#ifdef AUSGABEN_FARBENUPDATE
                    cout << " Ende unwissend 2b / rang " << rang << endl;
#endif
                    return(false);
                    }
                }
#ifdef AUSGABEN_FARBENUPDATE
            cout << " Ende unwissend 2c / rang " << rang << endl;
#endif
            return(true);
// Kommentar: Der Punkt liegt diagonal zwischen vier Grobgitterpunkten, die die EP - Information haben. Also wird geprft, ob
// mindestens einer der vier Grobgitterpunkte auf demselben Kern liegen wie dieser Punkt. Wenn ja, dann ist er wissend,
// also return(false); liegen alle vier auf einem anderen Kern, dann ist er unwissend, also return(true).
        case 3:
            if((f-1>0)&&(g-1>0))
                {
                if(fgff[f-1][g-1] == rang)
                    {
#ifdef AUSGABEN_FARBENUPDATE
                    cout << " Ende unwissend 3a / rang " << rang << endl;
#endif
                    return(false);
                    }
                }
            if((f-1>0)&&(g<m-1))
                {
                if(fgff[f-1][g+1] == rang)
                    {
#ifdef AUSGABEN_FARBENUPDATE
                    cout << " Ende unwissend 3b / rang " << rang << endl;
#endif
                    return(false);
                    }
                }
            if((f<n-1)&&(g-1>0))
                {
                if(fgff[f+1][g-1] == rang)
                    {
#ifdef AUSGABEN_FARBENUPDATE
                    cout << " Ende unwissend 3c / rang " << rang << endl;
#endif
                    return(false);
                    }
                }
            if((f<n-1)&&(g<m-1))
                {
                if(fgff[f+1][g+1] == rang)
                    {
#ifdef AUSGABEN_FARBENUPDATE
                    cout << " Ende unwissend 3d / rang " << rang << endl;
#endif
                    return(false);
                    }
                }
#ifdef AUSGABEN_FARBENUPDATE
            cout << " Ende unwissend 3e / rang " << rang << endl;
#endif
            return(true);
        }
    exit(-1);
    return(true);
    // Dieser Wert wird nicht angenommen.
    }

// Kommentar: Diese Funktion ist aktuell, und sie wird in der Klasse AdaptiveVerfeinerung aufgerufen. Hier wird
// berprft, welche Punkte hinzugekommen sind, und wie sich aufgrund dessen das Farbenfeld ndern muss. Und
// diese nderungen werden dann durchgefhrt.
void FarbenUpdate::einfuegen(void)
    {
// Kommentar: Das Folgende sind optionale Ausgaben.
#ifdef DEBUG
    cout << " Mache einfgen XXXY rang " << rang << endl;
#endif
#ifdef DEBUG
    cout << " FUein 2 / rang " << rang << endl;
#endif
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Feingitterfarbenfeld in einfuegen " << endl;
    fgff_ausgabe();
    cout << " Feingitterfarbenfeld in einfuegen II " << " anzahl2 " << anzahl2 << endl;
#endif

#ifdef AAHH
    test2fu("einfuegen start");
#endif

// Kommentar: Die folgende Punktmenge wird in 'RandFeldListeZuZweiZeitpunktenVerknuepfen' generiert, und von der
// Klasse AdaptiveVerfeinerung aufgerufen. Das sind die Punkte, die zu Beginn der adaptiven Verfeinerung auf dem
// Rand des Gitterbereichs des Kerns liegen, aber nach dem Einfgen der P+ - Punkte sich nicht mehr am Rand befinden.
    for(i=0; i<anzahl2; i++)
        {
#ifdef AUSGABEN_FARBENUPDATE
        cout << " f2 " << f2[i] << " g2 " << g2[i] << endl;
#endif
// Muss hier keine Abfrage ep_ff[f2[i], g2[i])==0 hinein?

// Kommentar: Es wird geprft, ob der ehemalige Randpunkt ein unwissender Punkt ist - vgl. den Kommentar zu dieser Funktion.
// Das if ist true, wenn ein wissender Punkt vorliegt. Dann macht man ein einfuegen_ein_punkt_ep_ff_0.
        if(ist_unwissend(f2[i], g2[i])==false)
        // == true ist es nur dann, wenn man einen Punkt hat, der keine EP - Information hat. 13.10.2014
        // Diese Punkte auf den alten Koordinaten mssen einen Farbupdate erhalten. 12.2.15
            {
// Kommentar: Die Variablen erweiteres_snb_0 wird nur fr stellvertretendes_einfuegen verwendet, welches
// aber herausgenommen ist.

// #ifdef SV_herein
            erweitertes_snb_0 = 0;
// #endif

// Kommentar: Diese Punkte sind Punkte die schon existierten. Hier kommt es zu direktem Einfgen, wenn Nachbarpunkte
// vom ep-ff Typ 2 sind: Weil - vom Typ 1, das gehrt zum entfernen, und Typ 0: Dann existierten Punkt und
// Nachbarpunkt schon beide, und es ist kein Farbenupdate ntig.
// Ein indirektes Einfgen unter der jeweiligen Himmelsrichtung kann dabei auch erfolgen - aber nur unter
// starken Restriktionen an eine Kommunikation.
// Dann gibt es noch die Verbindung bei inneren Ecken durch 'stellvertretung_fg_gleich_gg_HR'. (HR ist die Himmelsrichtung,
// also Nord oder West usw.)
            einfuegen_ein_punkt_ep_ff_0(f2[i], g2[i]);
// Kommentar: Das hier ist aktuell nicht definiert.
#ifdef SV_herein
            stellvertretendes_einfuegen();
#endif
            }
        else
            {
// Kommentar: Ein Eintrag hier fhrt dazu, dass eine Kommunikation nach hier in vervollstaendigen_entfernen
// ermglicht wird. Nur unwissende Punkte sollen Kommunikationsinformationen schreiben.
            unwissende_punkte.hinzu(f2[i], g2[i]);
            // Frage: Mssen sie in diesem Fall nicht kommuniziert werden? Nein. Das geschieht darber.
            }
        }
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Feingitterfarbenfeld in einfuegen III " << endl;
    fgff_ausgabe();
#endif

#ifdef AAHH
    test2fu("einfuegen nach f2,g2 ");
#endif

// Kommentar: Der akutelle Rand wird durchlaufen.
    for(dynamischer_rand->anfangsposition(); dynamischer_rand->ist_fertig_fl()==false; dynamischer_rand->naechster())
        {
// Kommentar: Die Koordinaten werden gelesen.
        f = dynamischer_rand->lesen_f();
        g = dynamischer_rand->lesen_g();
// Kommentar: Zuerst wird - wie oben - das bearbeitet, was schon da war. (ep_ff[f][g] = 0)
// Sind das nicht dieselben Punkte wie dort? Denn neu hinzugekommene Punkte haben ep_ff = 2.
// Das ist noch zu klren.
        if(ep_ff[f][g]==0)
// Kommentar: Damit wird dann auch dasselbe wie eben gemacht, vgl. also die Kommentare dort.
            {
            if(ist_unwissend(f, g)==false)
                {
                erweitertes_snb_0 = 0;
                einfuegen_ein_punkt_ep_ff_0(f, g);
                // Dort wird f=f geschrieben. Das noch aendern.
#ifdef SV_herein
                stellvertretendes_einfuegen();
#endif
                }
            else
                {
// Kommentar: Ein Eintrag hier fhrt dazu, dass eine Kommunikation nach hier in vervollstaendigen_entfernen
// ermglicht wird. Nur unwissende Punkte sollen Kommunikationsinformationen schreiben.
                unwissende_punkte.hinzu(f, g);
                }
            }
        }
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Feingitterfarbenfeld in einfuegen IV " << endl;
    fgff_ausgabe();
#endif

#ifdef AAHH
    test2fu("Nach 0 ");
#endif

// Kommentar: Wiederrum wird der aktuelle Rand durchlaufen.
    for(dynamischer_rand->anfangsposition(); dynamischer_rand->ist_fertig_fl()==false; dynamischer_rand->naechster())
        {
// Kommentar: Die Koordinaten werden gelesen.
        f = dynamischer_rand->lesen_f();
        g = dynamischer_rand->lesen_g();
        // Warum zweimal durchlaufen - nicht ganz klar. 12.2.15
// 27.02.15: Ist unwissend hier auch bringen?
            {
// Kommentar: Nun werden die neu hinzugekommenen Punkte bearbeitet.
            if(ep_ff[f][g] == 2)
                {
// Kommentar: Fr diese Punkte wird 'einfuegen_ein_punkt_ep_ff_2' aufgerufen.
                einfuegen_ein_punkt_ep_ff_2(f, g);
                // Bei (2,0) und (2,2) wird ggff eingefgt.
                // Problem: Bei (2,0) ist es nicht ntig, dass auerhalb des Gitters verfeinert wird, und dennoch wird dort
                // eingefuegt.
                }
            }
        }
#ifdef AAHH
    test2fu("Nach 2");
#endif
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Feingitterfarbenfeld in einfuegen V " << endl;
    fgff_ausgabe();
#endif
    }

// Kommentar: Diese Funktion ist aktuell, und sie wird in der Klasse AdaptiveVerfeinerung aufgerufen. Hier wird
// berprft, welche Punkte hinweggenommen wurden, und wie sich aufgrund dessen das Farbenfeld ndern muss. Und
// diese nderungen werden dann durchgefhrt.
void FarbenUpdate::entfernen(void)
    {
#ifdef AKTUELLE_WERTE
    cout << " entfernen / rang " << rang << endl;
#endif

#ifdef AUSGABEN_FARBENUPDATE
    cout << " FarbenUpdate::entfernen: Vor der Schleife rang " << rang << endl;
#endif
//    anzahl = 0;
// Kommentar: Der aktuelle Rand wird durchlaufen.
    for(dynamischer_rand->anfangsposition(); dynamischer_rand->ist_fertig_fl()==false; dynamischer_rand->naechster())
        {
// Kommentar: Die Koordinaten werden gelesen.
        f = dynamischer_rand->lesen_f();
        g = dynamischer_rand->lesen_g();

#ifdef AUSGABEN_FARBENUPDATE
        cout << " FarbenUpdate::entfernen_ein_punkt 8900 f " << f << " g " << g << endl;
#endif
#ifdef AKTUELLE_KOMMENTARE
        if(rang==4)
            {
            cout << " CCDDEE FU:entfernen f " << f << " g " << g << " 255-snb " << 255 - feingitter_snb[f][g] << endl;
            }
#endif
// Fr wissende Punkte wird 'entfernen_ein_punkt' vorgenommen, vgl. den Kommentar zu dieser Funktion.
// Bei unwissenden Punkte wird vervollstndigen_entfernen durchgefhrt.
// (Warum werden diese Punkte dann nicht als unwissend markiert? Weil diese schon beim Einfgen markiert wurden.)
        if(ist_unwissend(f, g)==false) // Versuch 24.2.2015. Ist aber logisch, da bei nicht direkt nur stellvertretend.
            {
            entfernen_ein_punkt(f, g);
            }
        }
#ifdef AUSGABEN_FARBENUPDATE
    cout << " FarbenUpdate::entfernen: Nach der Schleife rang " << rang << endl;
#endif
    }

// Kommentar: Die folgende Funktion macht direktes und indirektes Einfgen in Richtung Osten fr Punkte,
// die es vor P+ einfgen schon gab. Es werden auch die Inneren Ecken behandelt fr Osten, vgl. die Funktionen
// stellvertretung_fg_gleich_gg_HR  (HR fr Himmelsrichtung).
void FarbenUpdate::fu_punkt_farbpunkt_einfuegen_0_ost(void)
    {
// Kommentar: stlichste Punkte brauchen nicht behandelt zu werden.
    if(g==m_minus1)
        {
        return;
        }
// Kommentar: Der Paarpunkt wird fixiert.
    f_schlange = f;
    g_schlange = g+1;

// Kommentar: Direktes und indirektes einfgen wird durchgefhrt.
    direktes_einfuegen0();
    indirektes_einfuegen0_ost();

// Naechste Zeile neu 15.10.14
// Kommentar: Innere Ecken werden behandelt.
    stellvertretung_fg_gleich_gg_ost();
    }

// Kommentar: Analog zu fu_punkt_farbpunkt_einfuegen_0_ost: Nur die inneren Ecken mssen im disgonalen Fall nicht behandelt werden.
void FarbenUpdate::fu_punkt_farbpunkt_einfuegen_0_nordost(void)
    {
    if(g==m_minus1)
        {
        return;
        }
    if(f==0)
        {
        return;
        }
    f_schlange = f-1;
    g_schlange = g+1;

    direktes_einfuegen0();
#ifdef SV_herein2
    indirektes_einfuegen0_nordost();
#endif
    }

// Kommentar: Analog zu fu_punkt_farbpunkt_einfuegen_0_ost.
void FarbenUpdate::fu_punkt_farbpunkt_einfuegen_0_nord(void)
    {
    if(f==0)
        {
        return;
        }
    f_schlange = f-1;
    g_schlange = g;

    direktes_einfuegen0();
    indirektes_einfuegen0_nord();

// Naechste Zeile neu 15.10.14
    stellvertretung_fg_gleich_gg_nord();
    }

// Kommentar: Analog zu fu_punkt_farbpunkt_einfuegen_0_nordost.
void FarbenUpdate::fu_punkt_farbpunkt_einfuegen_0_suedost(void)
    {
    if(g==m_minus1)
        {
        return;
        }
    if(f==n_minus1)
        {
        return;
        }
    f_schlange = f+1;
    g_schlange = g+1;

    direktes_einfuegen0();
#ifdef SV_herein2
    indirektes_einfuegen0_suedost();
#endif
    }

// Kommentar: Analog zu fu_punkt_farbpunkt_einfuegen_0_ost.
void FarbenUpdate::fu_punkt_farbpunkt_einfuegen_0_sued(void)
    {
    if(f==n_minus1)
        {
        return;
        }
    f_schlange = f+1;
    g_schlange = g;

    direktes_einfuegen0();
    indirektes_einfuegen0_sued();

// Naechste Zeile neu 15.10.14
    stellvertretung_fg_gleich_gg_sued();
    }

// Kommentar: Analog zu fu_punkt_farbpunkt_einfuegen_0_nordost.
void FarbenUpdate::fu_punkt_farbpunkt_einfuegen_0_suedwest(void)
    {
    if(f==n_minus1)
        {
        return;
        }
    if(g==0)
        {
        return;
        }
    f_schlange = f+1;
    g_schlange = g-1;

    direktes_einfuegen0();
#ifdef SV_herein2
    indirektes_einfuegen0_suedwest();
#endif
    }

// Kommentar: Analog zu fu_punkt_farbpunkt_einfuegen_0_ost.
void FarbenUpdate::fu_punkt_farbpunkt_einfuegen_0_west(void)
    {
    if(g==0)
        {
        return;
        }
    f_schlange = f;
    g_schlange = g-1;

    direktes_einfuegen0();
    indirektes_einfuegen0_west();

// Naechste Zeile neu 15.10.14
    stellvertretung_fg_gleich_gg_west();
    }

// Kommentar: Analog zu fu_punkt_farbpunkt_einfuegen_0_nordost.
void FarbenUpdate::fu_punkt_farbpunkt_einfuegen_0_nordwest(void)
    {
    if(g==0)
        {
        return;
        }
    if(f==0)
        {
        return;
        }
    f_schlange = f-1;
    g_schlange = g-1;

    direktes_einfuegen0();
#ifdef SV_herein2
    indirektes_einfuegen0_nordwest();
#endif
    }

// Kommentar: Wie oben kommentiert, wird von einem Einfgen ausgegangen, und von schon existierenden Punkten.
// Dafr wird hier eine Funktion geschrieben, die dass bernimmt.
void FarbenUpdate::direktes_einfuegen0(void)
    {
// Kommentar: Da ep_ff[f][g] = 0 ist, und es um einfgen geht, sind ep_ff[f_schlange][g_schlange] = 0 / 1 auszunehmen:
// ' = 0' fllt weg, da in dem Fall der richtige Farbwert fr beide Punkte auf beiden Kernen schon da ist.
// ' = 1' fllt weg, da es dann nicht um einfgen, sondern um entfernen geht.
    if(ep_ff[f_schlange][g_schlange]!=2)
        {
        return;
        }
// Kommentar: Da auf (f_schlange,g_schlange) ein neuer Punkt da ist, hat der den Farbwert des groben Punktes ber
// (f_schlange/2,g_schlange/2), weil immer die 2 x 2 Blcke ((2f,2g) + {0,1}x{0,1}) unterhalb des groben Punktes
// eingefgt werden.
    fgff[f_schlange][g_schlange] = ggff[f_schlange >> 1][g_schlange >> 1];
    // Keine Kommunikation in diesem Fall. Weil der Prozessor P(f_schlange, g_schlange) im allgemeinen den Wert von direktes
    // einfuegen uebernimmt.
// Kommentar: Der Kern mit dem Punkt (f_schlange,g_schlange) weiss i.a. nichts von (f,g), weil dieser Punkt gerade
// erst existiert. Deshalb bekommt dieser Kern die Information (f,g,fgff[f][g]), um sein Farbenfeld auf (f,g) zu aktualieren.
    kommunikation4er->einfuegen_fu(f, g, fgff[f][g], fgff[f_schlange][g_schlange]);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
    strcpy(text, "9785");
    test_einfuegen(f, g, fgff[f][g], fgff[f_schlange][g_schlange], text);
#endif
#ifdef AUSGABEN_DATEI
    da_eez->aus << " direktes einfuegen 0 (ink) Rang " << rang << " Ziel P " << fgff[f_schlange][g_schlange] << " f " << f << " g " << g << " m " << m << endl;
#endif
    // Man kann ohne die Information fgff[f][g] auskommen. Dieses hier ist provisorisch, falls diese Programmierung
    // noch nicht stimmt.
    }

// Kommentar: Auch hier wird direkt eingefgt, allerdings im Fall ep_ff[f][g] = 2, d.h. fr einen neu hinzugekommenen Punkt.
void FarbenUpdate::direktes_einfuegen2(void)
    {
// Kommentar: Eigentlich ist die Abfrage ep_ff[f][g]==2 berflssig, da nur fr neue Gitterpunkte sirekte_einfuegen2 da ist.
// Am Anfang wird der Fall ep_ff[f_schlange][g_schlange] == 0 behandelt.
// Kommentar: Normalerweise bekommt der Kern in diesem Fall die Farbwerte per Kommunikation, weil auf dem anderen
// Kern direktes_einfuegen0 die Kommunikation bewirkt.
    if((ep_ff[f][g]==2)&&(ep_ff[f_schlange][g_schlange]==0))
        {
        f_schlange_div_2 = f_schlange >> 1;
        g_schlange_div_2 = g_schlange >> 1;

// Kommentar: Wenn aber die Information feld_aktuell auf dem zugeordneten Grobgitterpunkt und drum herum verschwindet,
// wird der Farbgitterwert dort gelscht (setzen = -1). Es stellt sich nur die Frage nach einem Szenario, wo dann
// trotzdem ep_ff[f_schlange][g_schlange] = 0 sein kann (vgl. die AUsgabe AAHH). Korrekt sind diese Zeilen aber.
// Zum Szenario: Wenn feld_aktuell und feld_vergangen beide false sind, tritt dieser Fall auf. Dann muss fr den neu
// hinzugekommenen Punkt fgff[f_schlange][g_schlange] auf -1 gesetzt werden!
        if((feld_aktuell[f_schlange_div_2][g_schlange_div_2] == false)&&(ep_snb[f_schlange_div_2][g_schlange_div_2]==0))
            {
            fgff[f_schlange][g_schlange] = -1;
            return;
            }
        }
// Bemerkung: Am 23.2.2015 hinzugefgt. Vgl. s.u.

// Kommentar: Jetzt wird der Fall abgearbeitet, dass auf beiden Paarpunkten der Wert 2 ist.
    if(ep_ff[f_schlange][g_schlange]!=2)
        {
        return;
        }

// Die letzten 3 Zeilen wurden am 23.2.15 hinzugefgt und sind dafr da, frher etablierte Farbenfelder zu lschen,
// wenn die Verfeinerung zurckluft.

// Kommentar: Beide Paarpunkte fhren denselben Befehl aus, nmlich das der Farbwert des anderen Punktes vom Grobgitter-
// farbenfeld abgeleitet wird. Wenn der jeweils andere Punkt neu ist, leitet sich nmlich der Feingitterfarbenwert
// vom Grobgitterfarbewert ab.
    fgff[f_schlange][g_schlange] = ggff[f_schlange >> 1][g_schlange >> 1];
#ifdef AUSGABEN_DATEI
    da_eez->aus << " direktes einfuegen 2 (ink) Rang " << rang << " Ziel P " << fgff[f_schlange][g_schlange] << " f " << f << " g " << g << " m " << m << endl;
#endif
    // Keine Kommunikation in diesem Fall.
//    cout << " AABB Ende rang " << rang << endl;
    }
// Scheinbar gleich. Bei (2,2) sowieso. Bei (2,0) aber auch, weil 0 im Allgemeinen auf ggff den richtigen Farbwert hat.
// In den anderen Fllen wird das spter durch Kommunikation berschrieben.


// Kommentar: Analog zu 'stellvertretend_nordost'.
void FarbenUpdate::stellvertretend_suedost(void)
    {
    f_schlange = f+1;
    g_schlange = g+1;

// SO(N)
    if(f!=0)
        {
        if((g!=m_minus1)&&(ep_ff[f-1][g+1]==1)&&(f!=n_minus1))
            {
            farbe_short_int = fgff[f+1][g];
            farbe_schlange = fgff[f_schlange][g_schlange];
            if(farbe_short_int != farbe_schlange)
                {
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f+1, g, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "9899");
                test_einfuegen(f+1, g, farbe_char, farbe_schlange, text);
#endif
                }
            // Auf der Ebene des Bezugspunktes auf dem Punkt und links daneben
            // Zusammen mit indirektes einfuegen nord wird der einzelne Feingitterpunkt mit Farbwerten umlaufen.
            // Das sind in diesem Falle 5 Punkte:  * *
            //                                     ? *
            //                                     * *
            }
        }

    if(g!=0)
        {
        if((f!=n_minus1)&&(ep_ff[f+1][g-1]==1)&&(g!=m_minus1))
// SO(W)
            {
            farbe_short_int = fgff[f+1][g];
            farbe_schlange = fgff[f_schlange][g_schlange];
            if(farbe_short_int != farbe_schlange)
                {
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f+1, g, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "9925");
                test_einfuegen(f+1, g, farbe_char, farbe_schlange, text);
#endif
                }
// Das (f+1,g) ist in beiden Fllen relevant, vgl. Heft 3.5.14.
            farbe_short_int = fgff[f][g+1];
            if(farbe_short_int != farbe_schlange)
                {
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f, g+1, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "9932");
                test_einfuegen(f, g+1, farbe_char, farbe_schlange, text);
#endif
                }
            }
        }
    }

// Kommentar: Hier wird nur die Variable erweitertes_snb_0 ergnzt - eine Variable, die nicht mehr verwendet wird,
// da die Funktion stellvertretendes_einfuegen nicht mehr benutzt wird. Daher: Keine Kommentierung.
void FarbenUpdate::indirektes_einfuegen0_suedost(void)
// Hier ist f_schlange nicht der Auenpunkt, sondern der indirekte Punkt. f_s sind die Auenpunkte!
    {
// Nchste Zeile: Das Folgende bezieht sich nur auf existente Punkte. Ein ep_ff - Feld mit 1 sen kann also gar nicht auftreten.
    if(ep_ff[f_schlange][g_schlange]!=0)
        {
        return;
        }
    if(fgff[f_schlange][g_schlange]==rang)
        {
        return;
        }
    if(fgff[f_schlange][g_schlange]<0)
        {
        return;
        }
    erweitertes_snb_0 |= SUED_OST;
    }

// Kommentar: Analog zu 'indirektes_einfuegen0_suedost'.
void FarbenUpdate::indirektes_einfuegen0_nordost(void)
// Hier ist f_schlange nicht der Auenpunkt, sondern der indirekte Punkt. f_s sind die Auenpunkte!
    {
// Nchste Zeile: Das Folgende bezieht sich nur auf existente Punkte. Ein ep_ff - Feld mit 1 sen kann also gar nicht auftreten.
    if(ep_ff[f_schlange][g_schlange]!=0)
        {
        return;
        }
    if(fgff[f_schlange][g_schlange]==rang)
        {
        return;
        }
    if(fgff[f_schlange][g_schlange]<0)
        {
        return;
        }
    erweitertes_snb_0 |= NORD_OST;
    }

// Kommentar: Analog zu 'stellvertretend_nordost'.
void FarbenUpdate::stellvertretend_suedwest(void)
    {
    f_schlange = f+1;
    g_schlange = g-1;

    if(f!=0)
        {
        if((g!=0)&&(ep_ff[f-1][g-1]==1)&&(f!=n_minus1))
// SW(N)
            {
            farbe_short_int = fgff[f+1][g];
            farbe_schlange = fgff[f_schlange][g_schlange];
            if(farbe_short_int != farbe_schlange)
                {
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f+1, g, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "9996");
                test_einfuegen(f+1, g, farbe_char, farbe_schlange, text);
#endif
                }
            }
        }
    if(f!=n_minus1)
        {
        if((g!=0)&&(ep_ff[f+1][g-1]==1))
// SW(O)
            {
            farbe_short_int = fgff[f+1][g];
            farbe_schlange = fgff[f_schlange][g_schlange];
            if(farbe_short_int != farbe_schlange)
                {
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f+1, g, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "10026");
                test_einfuegen(f+1, g, farbe_char, farbe_schlange, text);
#endif
                }
            farbe_short_int = fgff[f][g-1];
            if(farbe_short_int != farbe_schlange)
                {
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f, g-1, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "10034");
                test_einfuegen(f, g-1, farbe_char, farbe_schlange, text);
#endif
                }
            }
        }
    }

// Kommentar: Analog zu 'stellvertretend_nordost'.
void FarbenUpdate::stellvertretend_nordwest(void)
    {
    f_schlange = f-1;
    g_schlange = g-1;

    if(f!=0)
        {
        if((g!=m_minus1)&&(ep_ff[f-1][g+1]==1)&&(g!=0))
// NW(O)
            {
            farbe_short_int = fgff[f][g-1];
            farbe_schlange = fgff[f_schlange][g_schlange];
            if(farbe_short_int != farbe_schlange)
                {
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f, g-1, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "10057");
                test_einfuegen(f, g-1, farbe_char, farbe_schlange, text);
#endif
                }
            }
        }
    if(f!=n_minus1)
        {
        if((g!=0)&&(ep_ff[f+1][g-1]==1))
// NW(S)
            {
            farbe_short_int = fgff[f][g-1];
            farbe_schlange = fgff[f_schlange][g_schlange];
            if(farbe_short_int != farbe_schlange)
                {
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f, g-1, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "10072");
                test_einfuegen(f, g-1, farbe_char, farbe_schlange, text);
#endif
                }

            if(f!=0)
                {
                farbe_short_int = fgff[f-1][g];
                if(farbe_short_int != farbe_schlange)
                    {
                    farbe_char = (char) farbe_short_int;
                    kommunikation4er->einfuegen_fu(f-1, g, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                    strcpy(text, "10082");
                    test_einfuegen(f-1, g, farbe_char, farbe_schlange, text);
#endif
                    }
                }
            }
        }
    }

// Kommentar: Hier geschieht etwas vom selben Prinzip wie 'stellvertretend_nord', nur fr die Himmelsrichtung
// Nord - Ost.
void FarbenUpdate::stellvertretend_nordost(void)
    {
// Kommentar: Die Koordinaten fr den zu verstelltretenden Punkt werden bestimmt.
    f_schlange = f-1;
    g_schlange = g+1;

// Kommentar: Die im Folgenden zu stellenden Bedingung an f und g dienen nur dem Zweck, nicht zu arbeiten auf Punkten, die es
// auf dem Gitter nicht gibt.
    if(f!=n_minus1)
        {
// Kommentar: Mit (f,g) und (f-1,g+1) existiert auch (f,g+1). Da ep_ff(f+1,g+1) = 1 ost, mssen die Punkte nrdlich
// von (f,g) und (f,g+1) existieren. Da der Punkt (f-1,g+1) schon auf dem Kern mit ID_schlange = fgff(f_schlange,g_schlange) liegtm
// braucht nur noch der Punkt (f-1,g) - der ja existiert - auf den Kern mit Farbe ID_schlange bertragen zu werden.
// Und das nur dann, wenn sich die Farbwerte auf(f-1,g) und (f-1,g+1) unterscheiden. Genau das wird jetzt gemacht.
        if((g!=m_minus1)&&(ep_ff[f+1][g+1]==1)&&(f!=0))
// NO(S)
            {
            farbe_short_int = fgff[f-1][g];
            farbe_schlange = fgff[f_schlange][g_schlange];
            if(farbe_short_int != farbe_schlange)
                {
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f-1, g, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "10114");
                test_einfuegen(f-1, g, farbe_char, farbe_schlange, text);
#endif
                }
            }
        }
// Kommentar: Das folgende ist dasselbe, nur gespiegelt an der Sdwest - Nordost - Achse ((f+1,g-1) zu (f-1,g+1)). Der Eintrag fr
// (f-1,g) ist dabei aber zuviel => das wieder herausnehmen.
    if(f!=0)
        {
        if((g!=0)&&(ep_ff[f-1][g-1]==1))
// NW(S)
            {
// Kommentar: Diese Zeile bis zu kommunikation4er->einfuegen_fu wieder herausnehmen. 7.5.15. Nicht falsch, aber berflssig.
            farbe_short_int = fgff[f-1][g];
            farbe_schlange = fgff[f_schlange][g_schlange];
            if(farbe_short_int != farbe_schlange)
                {
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f-1, g, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "10132");
                test_einfuegen(f-1, g, farbe_char, farbe_schlange, text);
#endif
                }

            if(g!=m_minus1)
                {
                farbe_short_int = fgff[f][g+1];
                if(farbe_short_int != farbe_schlange)
                    {
                    farbe_char = (char) farbe_short_int;
                    kommunikation4er->einfuegen_fu(f, g+1, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                    strcpy(text, "10143");
                    test_einfuegen(f, g+1, farbe_char, farbe_schlange, text);
#endif
                    }
                }
            }
        }
    }

// Kommentar: Analog zu 'indirektes_einfuegen0_suedost'.
void FarbenUpdate::indirektes_einfuegen0_suedwest(void)
// Hier ist f_schlange nicht der Auenpunkt, sondern der indirekte Punkt. f_s sind die Auenpunkte!
    {
// Nchste Zeile: Das Folgende bezieht sich nur auf existente Punkte. Ein ep_ff - Feld mit 1 sen kann also gar nicht auftreten.
    if(ep_ff[f_schlange][g_schlange]!=0)
        {
        return;
        }
    if(fgff[f_schlange][g_schlange]==rang)
        {
        return;
        }
    if(fgff[f_schlange][g_schlange]<0)
        {
        return;
        }
    erweitertes_snb_0 |= SUED_WEST;
    }

// Kommentar: Analog zu 'indirektes_einfuegen0_suedost'.
void FarbenUpdate::indirektes_einfuegen0_nordwest(void)
// Hier ist f_schlange nicht der Auenpunkt, sondern der indirekte Punkt. f_s sind die Auenpunkte!
    {
// Nchste Zeile: Das Folgende bezieht sich nur auf existente Punkte. Ein ep_ff - Feld mit 1 sen kann also gar nicht auftreten.
    if(ep_ff[f_schlange][g_schlange]!=0)
        {
        return;
        }
    if(fgff[f_schlange][g_schlange]==rang)
        {
        return;
        }
    if(fgff[f_schlange][g_schlange]<0)
        {
        return;
        }
    erweitertes_snb_0 |= NORD_WEST;
    }

// Kommentar: Analog zu 'stellvertretend_nord'.
void FarbenUpdate::stellvertretend_ost(void)
    {
    f_schlange = f;
    g_schlange = g+1;

    if(f!=0)
        {
        if((g!=m_minus1)&&(ep_ff[f-1][g+1]==1)&&(f!=n_minus1))
            {
            farbe_short_int = fgff[f+1][g];
            farbe_schlange = fgff[f_schlange][g_schlange];
            if(farbe_short_int != farbe_schlange)
                {
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f+1, g, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "10206");
                test_einfuegen(f+1, g, farbe_char, farbe_schlange, text);
#endif
                }

            // Auf der Ebene des Bezugspunktes auf dem Punkt und links daneben
            // Zusammen mit indirektes einfuegen nord wird der einzelne Feingitterpunkt mit Farbwerten umlaufen.
            // Das sind in diesem Falle 5 Punkte:  * *
            //                                     ? *
            //                                     * *
            }
        }
    if(f!=n_minus1)
        {
        if((g!=m_minus1)&&(ep_ff[f+1][g+1]==1)&&(f!=0))
            {
            farbe_short_int = fgff[f-1][g];
            farbe_schlange = fgff[f_schlange][g_schlange];
            if(farbe_short_int != farbe_schlange)
                {
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f-1, g, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "10231");
                test_einfuegen(f-1, g, farbe_char, farbe_schlange, text);
#endif
                }
            }
        }
    }

// Kommentar: Diese Funktion wird von Punkten aufgerufen, die bereits exitieren. Man macht dafr ein Einfgen
// fr den stlichen Nachbarn. Das das ein zulssiger Punkt ist (g_schlange < m) wird im Vorfeld geklrt.
// Da das hier zu Kommunikationen fhrt, soll es so selten wie mglich erfolgen. Daher unterdrcken die ersten
// 3 if - Abfragen viele unntigen Kommunikationen.
void FarbenUpdate::indirektes_einfuegen0_ost(void)
// Hier ist f_schlange nicht der Auenpunkt, sondern der indirekte Punkt. f_s sind die Auenpunkte!
    {
#ifdef AAEE
    if((rang==2)&&(f==12)&(g==6))
        {
        cout << " AAEE rang 2 (12,6) ep - ff " << ep_ff[f_schlange][g_schlange] << " fgff ";
        cout << fgff[f_schlange][g_schlange] << " t " << t << endl;
        }
#endif
// Nchste Zeile: Das Folgende bezieht sich nur auf existente Punkte. Ein ep_ff - Feld mit 1 sen kann also gar nicht auftreten.

// Kommentar: In diesem Fall ist (f_schlange,g_schlange) ein direkter Punkt, fr den nichts indirektes zu tun bleibt.
    if(((f_schlange&1)==0)&&((g_schlange&1)==0))
        {
        return;
        }
// Kommentar: ep_ff = 1 => Der Punkt (f_schlange,g_schlange) existiert nicht mehr. Dafr ist natrlich nichts zu tun.
// ep_ff = 2 => das sind neue Punkte, und das sind immer wissende Punkte, weil sie dieselbe Farbe bekommen wie ihr
// zugehriger Grobgitterpunkt. Dafr ist auch nichts zu tun. Also verbleibt ep_ff = 0 als Wahl fr
// ein Einfgen an Stelle von.
    if(ep_ff[f_schlange][g_schlange]!=0)
        {
        return;
        }
// Kommentar: Wenn fgff = rang ist, liegt die Information vor, da beide Punkte, (f,g) und (f_schlange,g_schlange), auf demselben
// Kern liegen. Eine Kommunikation wre eine Kommunikation mit sich selber, und das ist sinnlos.
    if(fgff[f_schlange][g_schlange]==rang)
        {
        return;
        }
// Kommentar: Wenn fgff < 0 ist, dann hat dort vorher kein Punkt existiert. Entweder existiert dort immer noch kein Punkt,
// dann braucht der nicht behandelt zu werden. Oder der Punkt ist neu, dann ist er - vgl. die erste if - Abfrage - ein
// wissender Punkt, fr den indirekt nichts getan werden muss.
    if(fgff[f_schlange][g_schlange]<0)
        {
        return;
        }
// Kommentar: Diese Zeile ist eigentlich berflssig, weil die Funktion 'stellvertretendes_einfuegen', die diese
// Variable verwendet, nicht mehr aufgerufen wird.
#ifdef SV_herein2
    erweitertes_snb_0 |= OST;
#endif

// Kommentar: Der indirekte Punkt muss umgeben werden von Farbwerten - damit sein Farbenfeld exakt ist.
// Auf den gemeinsamen Punkten geschieht das durch direktes einfgen, vgl. direktes_einfuegen0.
// Damit verbleiben die Werte an den Seiten der Linie durch die gemeinsamen Punkte. Bei Ost nimmt
// man nur die 2 zwei Punkte nrdlich und sdlich des gemeisamen Punktes. Im Falle von Westen sind dies
// dann 4 Punkte. Fr diese 2 Punkte geschieht nun eine Kommunikation durch die Funktion indirekter_farbpunkt.
    if(f!=0)
        {
        indirekter_farbpunkt(f-1, g);
        // Diese Koordinaten verstehe ich nicht - im Falle von Osten.
        }
    if(f!=n_minus1)
        {
        indirekter_farbpunkt(f+1, g);
        }
//  Vgl. entfernen: Kein doppeltes Einfuegen, vorerst so realisiert.
    }

// Kommentar: Analog zu 'stellvertretend_nord'.
void FarbenUpdate::stellvertretend_west(void)
    {
    f_schlange = f;
    g_schlange = g-1;

    if(f!=0)
        {
        if((g!=0)&&(ep_ff[f-1][g-1]==1)&&(f!=n_minus1))
            {
            farbe_short_int = fgff[f+1][g];
            farbe_schlange = fgff[f_schlange][g_schlange];
            if(farbe_short_int != farbe_schlange)
                {
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f+1, g, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "10356");
                test_einfuegen(f+1, g, farbe_char, farbe_schlange, text);
#endif
                }

                {
                farbe_short_int = fgff[f+1][g-1];
                farbe_schlange = fgff[f_schlange][g_schlange];
                if(farbe_short_int != farbe_schlange)
                    {
                    farbe_char = (char) farbe_short_int;
                    kommunikation4er->einfuegen_fu(f+1, g-1, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                    strcpy(text, "10372");
                    test_einfuegen(f+1, g-1, farbe_char, farbe_schlange, text);
#endif
                    }
                }
            }
        }

    if(f!=n_minus1)
        {
        if((g!=0)&&(ep_ff[f+1][g-1]==1)&&(f!=0))
// Die stark einschrnkende Bedingung ep_ff == -1 scheint korrekt zu sein. Das Farbenfeld drum herum wird nmlich nur einmal
// upzudaten sein. Danach ist das Farbenfeld korrekt. Und das bleibt es dann auch. Daher korrekt.
// Ferner: Warum ep-ff == -1? Wenn dort alpha - Punkte existieren, dann braucht man kein stellvertretendes Einfgen.
// Also nur, wenn dort Punkte entfernt wurden. Sonst existieren die Farbwerte um (f+1,g) drum herum.
// Hier, diese Routine, kompensiert -1 sen im Farbenfeld, falls Punkte um einen nicht alpha - Punkt entfernt werden.
// Nur dann werden diese Farbwerte gebraucht - um die Nahe zu ep_ff == 1 entfernten Punkte zu kompensieren.
// Weit entfernt von ep_ff == 1 kann das nicht passieren.
            {
            farbe_short_int = fgff[f-1][g];
            farbe_schlange = fgff[f_schlange][g_schlange];
            if(farbe_short_int != farbe_schlange)
                {
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f-1, g, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "10395");
                test_einfuegen(f-1, g, farbe_char, farbe_schlange, text);
#endif
                }
                {
                farbe_short_int = fgff[f-1][g-1];
                    if(farbe_short_int != farbe_schlange)
                    {
                    farbe_char = (char) farbe_short_int;
                    kommunikation4er->einfuegen_fu(f-1, g-1, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                    strcpy(text, "10413");
                    test_einfuegen(f-1, g-1, farbe_char, farbe_schlange, text);
#endif
                    }
                }
            }
        }
    }

// Kommentar: Fast analog zu 'indirektes_einfuegen0_ost'. Der einzige Unterschied ist der Aufruf von indirekter Farbpunkt
// nicht nur nrdlich und sdlich neben dem gemeinsamen Punkt, sondern auch nrdlich und sdlich neben dem indirekten Punkt.
void FarbenUpdate::indirektes_einfuegen0_west(void)
    {
    if(((f_schlange&1)==0)&&((g_schlange&1)==0))
        {
        return;
        }
    if(ep_ff[f_schlange][g_schlange]!=0)
        {
        return;
        }
    if(fgff[f_schlange][g_schlange]==rang)
        {
        return;
        }
    if(fgff[f_schlange][g_schlange]<0)
        {
        return;
        }
#ifdef SV_herein2
    erweitertes_snb_0 |= WEST;
#endif
// Kommentar: Hier werden die indirekten Kommunikationen fr die 4 Punkte nrdlich und sdlich des direkten
// und indirekten Punktes - soweit mglich - durch Aufruf der Funktion indirekter_farbpunkt bewirkt.
    if(f!=0)
        {
        indirekter_farbpunkt(f-1, g);
        }
    if(f!=n_minus1)
        {
        indirekter_farbpunkt(f+1, g);
        }

    if((f!=0)&&(g!=0))
        {
        indirekter_farbpunkt(f-1, g-1);
        }
    if((f!=n_minus1)&&(g!=0))
        {
        indirekter_farbpunkt(f+1, g-1);
        }
    }

// Kommentar: Der Punkt (f-1,g) wird als Punkt angenommen, fr den ein stellvertretender Farbenupdate erfolgt.
// Bei dieser Funktion wird nur aktiv, wenn auf einer Seite - West oder Ost - die Punkte verschwinden.
void FarbenUpdate::stellvertretend_nord(void)
    {
// Kommentar: Das sind die Koordinaten fr den Punkt, fr den der stellvertretende Farbenupdate erfolgt.
    f_schlange = f-1;
    g_schlange = g;

// Kommentar: Fr die folgenden Schritte mssen f und g > 0 sein.
    if((g!=0)&&(f!=0))
        {
// Kommentar: Wenn hier und oben ein Punkt liegen, und links davon ein Punkt verschwindet, muss links
// nichts bertragen werden. Aber rechts davon, auf (f,g+1), mssen sich Punkte befinden, weil man
// eine 5 x 5 Breite hat. Und wenn weder links noch rechts Punkte liegen, htte man stattdessen eine Nord - Sd - Linie,
// was nicht sein kann.
        if(ep_ff[f-1][g-1]==1)
            {
// Kommentar: Falls der Punkt ganz im Osten liegt, ist in dieser Hinsicht nichts zuu tun.
            if(g!=m_minus1)
                {
// Kommentar: farbe_short_int ist die Farbe des Punktes, der nach farbe_schlange zu bermitteln ist.
                farbe_short_int = fgff[f][g+1];
                farbe_schlange = fgff[f_schlange][g_schlange];
// Kommentar: Wenn die Farben identisch sind, bleibt nichts zu tun.
                if(farbe_short_int != farbe_schlange)
                    {
                    farbe_char = (char) farbe_short_int;
                    //kommunikation4er->einfuegen_fu(f, g+1, farbe_char, fgff[f_schlange][g_schlange]);
// Kommentar: Der Kern der ID farbe_schlange soll auf der Stelle (f,g+1) die Farbe farbe_char eintragen - nach der
// Kommunikation.
                    kommunikation4er->einfuegen_fu(f, g+1, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                    strcpy(text, "10522");
                    test_einfuegen(f, g+1, farbe_char, farbe_schlange, text);
#endif
                    }

// Kommentar: Man macht dasselbe fr den stlichen Punkt darber. Auch die Farbe dieses Punktes soll
// der Kern mit der ID farbe_schlange bei sich eintragen.
                farbe_short_int = fgff[f-1][g+1];
                //farbe_schlange = fgff[f_schlange][g_schlange];
                if(farbe_short_int != farbe_schlange)
                    {
                    farbe_char = (char) farbe_short_int;
                    //kommunikation4er->einfuegen_fu(f-1, g+1, farbe_char, fgff[f_schlange][g_schlange]);
                    kommunikation4er->einfuegen_fu(f-1, g+1, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                    strcpy(text, "10542");
                    test_einfuegen(f-1, g+1, farbe_char, farbe_schlange, text);
#endif
                    }
// Kommentar: Fr noch einen Punkt darber trgt der direkte Punkt, weil gemeinsamer - oder alpha - Punkt, (f-2,g)
// seinen Farbwert ein.
                }
            }
        }
// Kommentar: Man macht dasselbe wie eben, nur an der Nord - Sd Linie durch (f,g)(und (f-1,g) gespiegelt, d.h.
// Ost und West vertauscht.
    if((g!=m_minus1)&&(f!=0))
        {
        if((ep_ff[f-1][g+1]==1))
// Die stark einschrnkende Bedingung ep_ff == 1 scheint korrekt zu sein. Das Farbenfeld drum herum wird nmlich nur einmal
// upzudaten sein. Danach ist das Farbenfeld korrekt. Und das bleibt es dann auch. Daher korrekt.
// Ferner: Warum ep-ff == 1? Wenn dort alpha - Punkte existieren, dann braucht man kein stellvertretendes Einfgen.
// Also nur, wenn dort Punkte entfernt wurden. Sonst existieren die Farbwerte um (f+1,g) drum herum.
// Hier, diese Routine, kompensiert -1 sen im Farbenfeld, falls Punkte um einen nicht alpha - Punkt entfernt werden.
// Nur dann werden diese Farbwerte gebraucht - um die Nahe zu ep_ff == 1 entfernten Punkte zu kompensieren.
// Weit entfernt von ep_ff == 1 kann das nicht passieren.
            {
            if(g!=0)
                {
                farbe_short_int = fgff[f][g-1];
                farbe_schlange = fgff[f_schlange][g_schlange];
                if(farbe_short_int != farbe_schlange)
                    {
                    farbe_char = (char) farbe_short_int;
                    kommunikation4er->einfuegen_fu(f, g-1, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                    strcpy(text, "10570");
                    test_einfuegen(f, g-1, farbe_char, farbe_schlange, text);
#endif
                    }
                farbe_short_int = fgff[f-1][g-1];
                if(farbe_short_int != farbe_schlange)
                    {
                    farbe_char = (char) farbe_short_int;
                    kommunikation4er->einfuegen_fu(f-1, g-1, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                    strcpy(text, "10589");
                    test_einfuegen(f-1, g-1, farbe_char, farbe_schlange, text);
#endif
                    }
                }
            }
        }
    }

// Kommentar: Analog zu 'indirektes_einfuegen0_west'.
void FarbenUpdate::indirektes_einfuegen0_nord(void)
    {
    if(((f_schlange&1)==0)&&((g_schlange&1)==0))
        {
        return;
        }

    if(ep_ff[f_schlange][g_schlange]!=0)
        {
        return;
        }
    if(fgff[f_schlange][g_schlange]==rang)
        {
        return;
        }
    if(fgff[f_schlange][g_schlange]<0)
        {
        return;
        }
#ifdef SV_herein2
    erweitertes_snb_0 |= NORD;
#endif
    if((f!=0)&&(g!=0))
        {
        indirekter_farbpunkt(f-1, g-1);
        }
    if((f!=0)&&(g!=m_minus1))
        {
        indirekter_farbpunkt(f-1, g+1);
        }
    if(g!=0)
        {
        indirekter_farbpunkt(f, g-1);
        }
    if(g!=m_minus1)
        {
        indirekter_farbpunkt(f, g+1);
        }
#ifdef ALTE_IMPLEMENTATION
    auf_alpha_eingefuegt = false;
    if(g!=0)
        {
        if(f!=0)
            {
            indirekter_farbpunkt(f-1, g-1);
            }
        indirekter_farbpunkt(f, g-1);
        if((ep_ff[f][g-1]==1)&&(g!=m_minus1))
            {
            farbe_short_int = fgff[f][g+1];
            farbe_char = (char) farbe_short_int;
            kommunikation4er->einfuegen_fu(f, g+1, farbe_char, fgff[f_schlange][g_schlange]);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "10678");
                test_einfuegen(f, g+1, farbe_char, fgff[f_schlange][g_schlange], text);
#endif
            farbe_short_int = fgff[f][g];
            farbe_char = (char) farbe_short_int;
            kommunikation4er->einfuegen_fu(f, g, farbe_char, fgff[f_schlange][g_schlange]);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "10681");
                test_einfuegen(f, g, farbe_char, fgff[f_schlange][g_schlange], text);
#endif
            auf_alpha_eingefuegt = true;
            if(f!=0)
                {
                farbe_short_int = fgff[f-1][g+1];
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f-1, g+1, farbe_char, fgff[f_schlange][g_schlange]);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "10687");
                test_einfuegen(f-1, g+1, farbe_char, fgff[f_schlange][g_schlange], text);
#endif
                }
            }
        }
    if(g!=m_minus1)
        {
        indirekter_farbpunkt(f, g+1);
        if((ep_ff[f][g+1]==1)&&(g!=0))
// Die stark einschrnkende Bedingung ep_ff == -1 scheint korrekt zu sein. Das Farbenfeld drum herum wird nmlich nur einmal
// upzudaten sein. Danach ist das Farbenfeld korrekt. Und das bleibt es dann auch. Daher korrekt.
// Ferner: Warum ep-ff == -1? Wenn dort alpha - Punkte existieren, dann braucht man kein stellvertretendes Einfgen.
// Also nur, wenn dort Punkte entfernt wurden. Sonst existieren die Farbwerte um (f+1,g) drum herum.
// Hier, diese Routine, kompensiert -1 sen im Farbenfeld, falls Punkte um einen nicht alpha - Punkt entfernt werden.
// Nur dann werden diese Farbwerte gebraucht - um die Nahe zu ep_ff == 1 entfernten Punkte zu kompensieren.
// Weit entfernt von ep_ff == 1 kann das nicht passieren.
            {
            farbe_short_int = fgff[f][g-1];
            farbe_char = (char) farbe_short_int;
            kommunikation4er->einfuegen_fu(f, g-1, farbe_char, fgff[f_schlange][g_schlange]);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "10705");
                test_einfuegen(f, g-1, farbe_char, fgff[f_schlange][g_schlange], text);
#endif
            if(auf_alpha_eingefuegt==false)
                {
                farbe_short_int = fgff[f][g];
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f, g, farbe_char, fgff[f_schlange][g_schlange]);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "10710");
                test_einfuegen(f, g, farbe_char, fgff[f_schlange][g_schlange], text);
#endif
                }
            if(f!=0)
                {
                farbe_short_int = fgff[f-1][g-1];
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f-1, g-1, farbe_char, fgff[f_schlange][g_schlange]);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "10716");
                test_einfuegen(f-1, g-1, farbe_char, fgff[f_schlange][g_schlange], text);
#endif
                }
            }
        }
#endif
    }

// Kommentar: Analog zu 'stellvertretend_nord'.
void FarbenUpdate::stellvertretend_sued(void)
    {
    f_schlange = f+1;
    g_schlange = g;

    if(g!=0)
        {
        if((f!=n_minus1)&&(ep_ff[f+1][g-1]==1)&&(g!=m_minus1))
            {
            farbe_short_int = fgff[f][g+1];
            farbe_schlange = fgff[f_schlange][g_schlange];
            if(farbe_short_int != farbe_schlange)
                {
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f, g+1, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "10740");
                test_einfuegen(f, g+1, farbe_char, farbe_schlange, text);
#endif
                }

            // Auf der Ebene des Bezugspunktes auf dem Punkt und links daneben
            // Zusammen mit indirektes einfuegen nord wird der einzelne Feingitterpunkt mit Farbwerten umlaufen.
            // Das sind in diesem Falle 5 Punkte:  * *
            //                                     ? *
            //                                     * *
            }
        }
    if(g!=m_minus1)
        {
        if((f!=n_minus1)&&(ep_ff[f+1][g+1]==1)&&(g!=0))
            {
            farbe_short_int = fgff[f][g-1];
            farbe_schlange = fgff[f_schlange][g_schlange];
            if(farbe_short_int != farbe_schlange)
                {
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f, g-1, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "10765");
                test_einfuegen(f, g, farbe_char, farbe_schlange, text);
#endif
                }
            }
        }
    }

// Kommentar: Analog zu 'indirektes_einfuegen0_ost'.
void FarbenUpdate::indirektes_einfuegen0_sued(void)
    {
    if(((f_schlange&1)==0)&&((g_schlange&1)==0))
        {
        return;
        }
    if(ep_ff[f_schlange][g_schlange]!=0)
        {
        return;
        }
    if(fgff[f_schlange][g_schlange]==rang)
        {
        return;
        }
    if(fgff[f_schlange][g_schlange]<0)
        {
        return;
        }
#ifdef SV_herein2
    erweitertes_snb_0 |= SUED;
#endif
    if(g!=0)
        {
        indirekter_farbpunkt(f, g-1);
        }
    if(g!=m_minus1)
        {
        indirekter_farbpunkt(f, g+1);
        }
#ifdef ALTE_IMPLEMENTATION
    auf_alpha_eingefuegt = false;
    if(g!=0)
        {
        indirekter_farbpunkt(f, g-1);
        if((ep_ff[f][g-1]==1)&&(g!=m_minus1))
            {
            farbe_short_int = fgff[f][g+1];
            farbe_char = (char) farbe_short_int;
            kommunikation4er->einfuegen_fu(f, g+1, farbe_char, fgff[f_schlange][g_schlange]);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
            test_einfuegen(f, g+1, farbe_char, fgff[f_schlange][g_schlange], "10834");
#endif
            farbe_short_int = fgff[f][g];
            farbe_char = (char) farbe_short_int;
            kommunikation4er->einfuegen_fu(f, g, farbe_char, fgff[f_schlange][g_schlange]);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
            test_einfuegen(f, g, farbe_char, fgff[f_schlange][g_schlange], "10837");
#endif
            // Auf der Ebene des Bezugspunktes auf dem Punkt und links daneben
            // Zusammen mit indirektes einfuegen nord wird der einzelne Feingitterpunkt mit Farbwerten umlaufen.
            // Das sind in diesem Falle 5 Punkte:  * *
            //                                     ? *
            //                                     * *
            auf_alpha_eingefuegt = true;
            }
        }
    if(g!=m_minus1)
        {
        indirekter_farbpunkt(f, g+1);
        if((ep_ff[f][g+1]==1)&&(g!=0))
            {
            farbe_short_int = fgff[f][g-1];
            farbe_char = (char) farbe_short_int;
            kommunikation4er->einfuegen_fu(f, g-1, farbe_char, fgff[f_schlange][g_schlange]);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                test_einfuegen(f, g-1, farbe_char, fgff[f_schlange][g_schlange], "10853");
#endif
            if(auf_alpha_eingefuegt==false)
                {
                farbe_short_int = fgff[f][g];
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f, g, farbe_char, fgff[f_schlange][g_schlange]);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                test_einfuegen(f, g, farbe_char, fgff[f_schlange][g_schlange], "10858");
#endif
                }
            }
        }
//  Kommentar: vgl. entfernen: kein doppeltes einfuegen.
#endif
    }

// Kommentar: Diese Funktion wird nur von stellvertretendes_einfuegen verwendet, und die ist aktuell deaktiviert,
// weil 'SV_herein' nicht definiert wird.
void FarbenUpdate::stellvertretend_sued_suedwest(void)
    {
    if(fgff[f+1][g]!=fgff[f+1][g-1])
        {
        stellvertretend_sued();
        stellvertretend_suedwest();
        }

    f_schlange = f+1;
    g_schlange = g;

    if(f!=n_minus1)
        {
        if((g!=m_minus1)&&(ep_ff[f+1][g+1]==1)&&(g!=0))
            {
            farbe_short_int = fgff[f][g-1];
            farbe_schlange = fgff[f_schlange][g_schlange];
            if(farbe_short_int != farbe_schlange)
                {

                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f, g-1, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "10890");
                test_einfuegen(f, g-1, farbe_char, farbe_schlange, text);
#endif
                }
            }
        }
    }

// Kommentar: Analog zu 'stellvertretend_sued_suedwest'.
void FarbenUpdate::stellvertretend_sued_suedost(void)
    {
    if(fgff[f+1][g]!=fgff[f+1][g+1])
        {
        stellvertretend_sued();
        stellvertretend_suedost();
        }

    f_schlange = f+1;
    g_schlange = g;

    if(f!=n_minus1)
        {
        if((g!=0)&&(ep_ff[f+1][g-1]==1)&&(g!=m_minus1))
            {
            farbe_short_int = fgff[f][g+1];
            farbe_schlange = fgff[f_schlange][g_schlange];
            if(farbe_short_int != farbe_schlange)
                {
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f, g+1, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "10918");
                test_einfuegen(f, g+1, farbe_char, farbe_schlange, text);
#endif
                }
            }
        }
    }

// Kommentar: Analog zu 'stellvertretend_sued_suedwest'.
void FarbenUpdate::stellvertretend_ost_suedost(void)
    {
    if(fgff[f][g+1]!=fgff[f+1][g+1])
        {
        stellvertretend_ost();
        stellvertretend_suedost();
        }

    f_schlange = f;
    g_schlange = g+1;

    if(f!=0)
        {
        if((g!=m_minus1)&&(ep_ff[f-1][g+1]==1)&&(f!=n_minus1))
            {
            farbe_short_int = fgff[f+1][g];
            farbe_schlange = fgff[f_schlange][g_schlange];
            if(farbe_short_int != farbe_schlange)
                {
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f+1, g, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "10946");
                test_einfuegen(f+1, g, farbe_char, farbe_schlange, text);
#endif
                }
            }
        }
    }

// Kommentar: Analog zu 'stellvertretend_sued_suedwest'.
void FarbenUpdate::stellvertretend_ost_nordost(void)
    {
    if(fgff[f][g+1]!=fgff[f-1][g+1])
        {
        stellvertretend_ost();
        stellvertretend_nordost();
        }

    f_schlange = f;
    g_schlange = g+1;

    if(f!=n_minus1)
        {
        if((g!=m_minus1)&&(ep_ff[f+1][g+1]==1)&&(f!=0))
            {
            farbe_short_int = fgff[f-1][g];
            farbe_schlange = fgff[f_schlange][g_schlange];
            if(farbe_short_int != farbe_schlange)
                {
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f-1, g, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "10974");
                test_einfuegen(f-1, g, farbe_char, farbe_schlange, text);
#endif
                }
            }
        }
    }

// Kommentar: Analog zu 'stellvertretend_sued_suedwest'.
void FarbenUpdate::stellvertretend_nord_nordost(void)
    {
    if(fgff[f-1][g]!=fgff[f-1][g+1])
        {
        stellvertretend_nord();
        stellvertretend_nordost();
        }

    f_schlange = f-1;
    g_schlange = g;

    if(f!=0)
        {
        if((g!=0)&&(ep_ff[f-1][g-1]==1)&&(g!=m_minus1))
            {
            farbe_short_int = fgff[f][g+1];
            farbe_schlange = fgff[f_schlange][g_schlange];
            if(farbe_short_int != farbe_schlange)
                {
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f, g+1, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "11002");
                test_einfuegen(f, g+1, farbe_char, farbe_schlange, text);
#endif
                }
            }
        }
    }

// Kommentar: Analog zu 'stellvertretend_sued_suedwest'.
void FarbenUpdate::stellvertretend_nord_nordwest(void)
    {
    if(fgff[f-1][g]!=fgff[f-1][g-1])
        {
        stellvertretend_nord();
        stellvertretend_nordwest();
        }

    f_schlange = f-1;
    g_schlange = g;

    if(f!=0)
        {
        if((g!=m_minus1)&&(ep_ff[f-1][g+1]==1)&&(g!=0))
            {
            farbe_short_int = fgff[f][g-1];
            farbe_schlange = fgff[f_schlange][g_schlange];
            if(farbe_short_int != farbe_schlange)
                {
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f, g-1, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "11028");
                test_einfuegen(f, g-1, farbe_char, farbe_schlange, text);
#endif
                }
            }
        }
    }

// Kommentar: Analog zu 'stellvertretend_sued_suedwest'.
void FarbenUpdate::stellvertretend_west_nordwest(void)
    {
    if(fgff[f][g-1]!=fgff[f-1][g-1])
        {
        stellvertretend_west();
        stellvertretend_nordwest();
        }

    f_schlange = f;
    g_schlange = g-1;

    if(f!=n_minus1)
        {
        if((g!=0)&&(ep_ff[f+1][g-1]==1)&&(f!=0))
            {
            farbe_short_int = fgff[f-1][g];
            farbe_schlange = fgff[f_schlange][g_schlange];
            if(farbe_short_int != farbe_schlange)
                {
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f-1, g, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "11058");
                test_einfuegen(f-1, g, farbe_char, farbe_schlange, text);
#endif
                }
            }
        }
    }

// Kommentar: Analog zu 'stellvertretend_sued_suedwest'.
void FarbenUpdate::stellvertretend_west_suedwest(void)
    {
    if(fgff[f][g-1]!=fgff[f+1][g-1])
        {
        stellvertretend_west();
        stellvertretend_nordwest();
        }

    f_schlange = f;
    g_schlange = g-1;

    if(f!=0)
        {
        if((g!=0)&&(ep_ff[f-1][g-1]==1)&&(f!=n_minus1))
            {
            farbe_short_int = fgff[f+1][g];
            farbe_schlange = fgff[f_schlange][g_schlange];
            if(farbe_short_int != farbe_schlange)
                {
                farbe_char = (char) farbe_short_int;
                kommunikation4er->einfuegen_fu(f+1, g, farbe_char, farbe_schlange);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                strcpy(text, "11086");
                test_einfuegen(f+1, g, farbe_char, farbe_schlange, text);
#endif
                }
            }
        }
    }

// Kommentar: Bei dieser Funktion hat man drei Punkte: (f,g) ist der Punkt, fr den Eigenschaftsinformation vorliegt,
// und der Kern S, der diesen Punkt hat, sendet. Auf (f_schlange,g_schlange) sitzt der empfangende Kern E, der
// die Nachricht empfngt und sein Farbenfeld beschreibt. Der Punkt (f_s,g_s) ist der Punkt, an dessen Stelle das
// Farbenfeld von S gelesen und das Farbenfeld von E gechrieben wird.
void FarbenUpdate::indirekter_farbpunkt(int f_s, int g_s)
    {
#ifdef AKTUELLE_KOMMENTARE
    if((rang==4)&&(f_s==33)&&(g_s==25))
        {
        cout << " KKLLMM (4,33,25) ep-ff " << ep_ff[f_s][g_s] << endl;
        }
#endif
// Was ist f_s, und was f_schlange? Und wie funktioniert diese Funktion?

// Kommentar: Es interessieren nur Farbwerte an Stellen, die neu sind. Wenn ep_ff[f_s][g_s] = 0 ist, ist,
// weil ep_ff[f_schlange][g_schlange] = 0 ist - siehe z.B. 'indirektes_einfuegen0_ost', dann ist die Farbinformation
// fgff[f_s][g_s] schon frher auf Kern(f_schlange,g_schlange) gewesen.
    if(ep_ff[f_s][g_s] == 2)
// Ich glaube, dass diese Abfrage nicht richtig ist. Denn dieses hier erfolgt nicht nur bei neuen Punkten, sondern
// auch beim Lasttransfer mit dadurch entstehenden indirekten Punkten.
// Beim Lasttransfer werden aber die Farbwerte der Punkte drum herum mit verschickt, so dass das hier wohl
// doch stimmt. 25.03.2014
// 16.04.2014: Das ist soweit richtig. Nur: Wenn eine Entfernung von Farbwerten von Gitterpunkten geschieht aufgrund weiter entfernter
// EP-SNB - Informationen, ist dies auch im Falle von ep-ff == 0 der Fall. Nur hat man dann wesentlich mehr
// Kommunikation. Das ist abzuwgen gegen eine Blockade der Entfernung von Gitterpunkten.
        {
/*
        kommunikation4er.fu_entfernen(f_schlange, g_schlange, fgff[f_schlange][g_schlange], ggff[f_s >> 1][g_s >> 1]);
            // Kommunikation von (f_schlange, g_schlange) nach Prozessor(f_s, g_s)
        kommunikation4er.fu_entfernen(f_s, g_s, ggff[f_s >> 1][g_s >> 1], fgff[f_schlange][g_schlange]);
            // Umgekehrte Kommunikation
*/

// Kommentar: Man bestimmt die Farbwerte auf (f_schlange,g_schlange) durch ablesen des Feingitterfarbenfeldes,
// und den auf (f_s,g_s) weil dieser Punkt neu ist durch ablesen des Grobgitterfarbenfeldes.
        farbe_short_int = fgff[f_schlange][g_schlange];
        farbe_char = (char) farbe_short_int;
        ggff_speichern = ggff[f_s >> 1][g_s >> 1];
// Kommentar: Wenn der neue Punkt nicht auf diesem Kern liegt.
        if(ggff_speichern!=rang)
            {
// Kommentar: Dann schicke an ihn die Informationen des Punktes (f_schlange,g_schlange), damit dort
// das Farbenfeld aktualisiert wird.
            kommunikation4er->einfuegen_fu(f_schlange, g_schlange, farbe_char, ggff_speichern);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
            strcpy(text, "11142");
            test_einfuegen(f_schlange, g_schlange, farbe_char, ggff_speichern, text);
#endif
            // Diese Verbindung kann schon existieren.
            // 19.2.15: bermittlung an den Auenrandpunkt fr den Kern des indirekten Punktes.
            }
#ifdef ZAEHLER_ALT
        else
            {
            ebene_ebenen_zaehler_h_minus1->inkrementieren(fgff[f_schlange][g_schlange]);
            // Diese Verbindung kann schon existieren.
            }
#endif
            // Kommunikation von (f_schlange, g_schlange) nach Prozessor(f_s, g_s)

// Kommentar: Auf jeden Fall bekommt der Kern mit dem Punkt (f_schlange,g_schlange) die Informationen auf der
// Koordinate (f_s,g_s) zugeschickt.
        kommunikation4er->einfuegen_fu(f_s, g_s, ggff_speichern, farbe_short_int);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
        strcpy(text, "11155");
        test_einfuegen(f_s, g_s, ggff_speichern, farbe_short_int, text);
#endif
            // Umgekehrte Kommunikation
            // 19.2.15: Fr den Auenrandpunkt (f_s,g_s) wird die Farbe an farbe_short_int bermittelt.
        }
    }

// Kommentar: Das direkte und indirekte einfgen sowie die Behandlung innerer Ecken wird fr schon vorhandene
// Gitterrandpunkte durchgefhrt. Dabei wird die Funktion 'fu_punkt_farbpunkt_einfuegen_0_nord' aufgerufen.
void FarbenUpdate::einfuegen_ein_punkt_ep_ff_0(int _f, int _g)
    {
// Kommentar: Kopie der Koordinaten des gemeinsamen Randpunktes.
    f = _f;
    g = _g;

// Kommentar: Fr die Randpunkte werden die fr die nicht auf dem Kern liegenden Nebenpunkte entsprechende
// Funktionen aufgerufen.
        switch(255 - feingitter_snb[f][g]) // Ich denke, dass der dynamische Rand auf dem feinen Gitter liegt.
            {
                {
                case 0:
                    break;
                case 1:
// Kommentar: Aufruf von 'fu_punkt_farbpunkt_einfuegen_0_nord'. Fr alle Flle darunter ein analoger Kommentar.
                    punkt_farbpunkt_einfuegen_0_nord;
                    break;
                case 2:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    break;
                case 3:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    break;
                case 4:
                    punkt_farbpunkt_einfuegen_0_ost;
                    break;
                case 5:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    break;
                case 6:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    break;
                case 7:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    break;
                case 8:
                    punkt_farbpunkt_einfuegen_0_suedost;
                    break;
                case 9:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    break;
                case 10:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    break;
                case 11:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    break;
                case 12:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    break;
                case 13:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    break;
                case 14:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    break;
                case 15:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    break;
                case 16:
                    punkt_farbpunkt_einfuegen_0_sued;
                    break;
                case 17:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_sued;

                    break;
                case 18:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    break;
                case 19:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    break;
                case 20:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    break;
                case 21:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    break;
                case 22:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    break;
                case 23:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;

                    punkt_farbpunkt_einfuegen_0_sued;

                    break;
                case 24:
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    break;
                case 25:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    break;
                case 26:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    break;
                case 27:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    break;
                case 28:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    break;
                case 29:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    break;
                case 30:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    break;
                case 31:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    break;
                case 32:
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    break;
                case 33:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    break;
                case 34:
                    punkt_farbpunkt_einfuegen_0_nordost;

                    punkt_farbpunkt_einfuegen_0_suedwest;
                    break;
                case 35:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    break;
                case 36:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    break;
                case 37:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    break;
                case 38:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    break;
                case 39:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    break;
                case 40:
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    break;
                case 41:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    break;
                case 42:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    break;
                case 43:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    break;
                case 44:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    break;
                case 45:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    break;
                case 46:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    break;
                case 47:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    break;
                case 48:
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    break;
                case 49:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    break;
                case 50:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    break;
                case 51:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    break;
                case 52:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    break;
                case 53:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    break;
                case 54:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    break;
                case 55:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    break;
                case 56:
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    break;
                case 57:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    break;
                case 58:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    break;
                case 59:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    break;
                case 60:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    break;
                case 61:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    break;
                case 62:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    break;
                case 63:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    break;

                case 64:
                    punkt_farbpunkt_einfuegen_0_west;
                    break;

                case 65:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 66:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 67:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 68:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 69:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 70:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 71:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 72:
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 73:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 74:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 75:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 76:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 77:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 78:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 79:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 80:
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 81:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 82:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 83:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 84:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 85:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 86:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 87:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;

                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 88:
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 89:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 90:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 91:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 92:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 93:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 94:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 95:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 96:
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;

                case 97:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 98:
                    punkt_farbpunkt_einfuegen_0_nordost;

                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 99:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 100:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 101:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 102:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 103:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 104:
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 105:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 106:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 107:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 108:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 109:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 110:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 111:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 112:
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 113:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 114:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 115:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 116:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 117:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 118:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 119:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 120:
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 121:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 122:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 123:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 124:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 125:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 126:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;
                case 127:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    break;

                case 128:
                    punkt_farbpunkt_einfuegen_0_nordwest;
                    break;
                case 129:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 130:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 131:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 132:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 133:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 134:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 135:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 136:
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 137:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 138:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 139:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 140:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 141:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 142:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 143:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 144:
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 145:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 146:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 147:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 148:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 149:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 150:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 151:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;

                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 152:
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 153:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 154:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 155:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 156:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 157:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 158:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 159:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 160:
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 161:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 162:
                    punkt_farbpunkt_einfuegen_0_nordost;

                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 163:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 164:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 165:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 166:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 167:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 168:
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 169:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 170:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 171:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 172:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 173:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 174:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 175:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 176:
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 177:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 178:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 179:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 180:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 181:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 182:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 183:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 184:
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 185:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 186:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 187:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 188:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 189:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 190:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 191:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_nordwest; break;

                case 192:
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest;
                    break;

                case 193:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 194:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 195:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 196:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 197:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 198:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 199:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 200:
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 201:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 202:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 203:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 204:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 205:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 206:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 207:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 208:
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 209:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 210:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 211:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 212:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 213:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 214:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 215:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;

                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 216:
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 217:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 218:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 219:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 220:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest;
                    break;
                case 221:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest;
                    break;
                case 222:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest;
                    break;
                case 223:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;

                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 224:
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;

                case 225:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 226:
                    punkt_farbpunkt_einfuegen_0_nordost;

                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 227:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 228:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 229:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 230:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 231:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 232:
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 233:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 234:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 235:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 236:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 237:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 238:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 239:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 240:
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 241:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 242:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 243:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 244:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 245:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 246:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 247:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 248:
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 249:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 250:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 251:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 252:
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;
                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 253:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 254:
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                case 255:
                    punkt_farbpunkt_einfuegen_0_nord;
                    punkt_farbpunkt_einfuegen_0_nordost;
                    punkt_farbpunkt_einfuegen_0_ost;
                    punkt_farbpunkt_einfuegen_0_suedost;
                    punkt_farbpunkt_einfuegen_0_sued;
                    punkt_farbpunkt_einfuegen_0_suedwest;

                    punkt_farbpunkt_einfuegen_0_west;
                    punkt_farbpunkt_einfuegen_0_nordwest; break;
                }
            }
    }

// Kommentar: Die folgende Funktion macht direktes Einfgen in Richtung Osten fr Punkte,
// die es noch nicht gab. Es werden auch die Inneren Ecken behandelt fr Osten, vgl. die Funktionen
// stellvertretung_fg_gleich_gg_HR  (HR fr Himmelsrichtung).
void FarbenUpdate::fu_punkt_farbpunkt_einfuegen_2_ost(void)
    {
// Kommentar: stlichste Punkte brauchen nicht behandelt zu werden.
    if(g==m_minus1)
        {
        return;
        }
// Kommentar: Der Paarpunkt wird fixiert.
    f_schlange = f;
    g_schlange = g+1;

// Kommentar: Direktes einfgen wird durchgefhrt.
    direktes_einfuegen2();

// Naechste Zeile neu 15.10.14
// Kommentar: Innere Ecken werden behandelt.
    stellvertretung_fg_gleich_gg_ost();
#ifdef AAHH
    test3fu("ost");
#endif
    }

// Kommentar: Fast analog zu 'fu_punkt_farbpunkt_einfuegen_2_ost': Nur die inneren Ecken werden hier nicht besonders
// behandelt. Sie treten nur fr Punkte auf, die im Norden, Osten, Sden oder Westen liegen. Bei diagonalen
// Nachbarschaften hat man keine innere Ecke.
void FarbenUpdate::fu_punkt_farbpunkt_einfuegen_2_nordost(void)
    {
    if(f==0)
        {
        return;
        }
    if(g==m_minus1)
        {
        return;
        }
    f_schlange = f-1;
    g_schlange = g+1;

    direktes_einfuegen2();
#ifdef AAHH
    test3fu("nordost");
#endif
    }

// Kommentar: Analog zu 'fu_punkt_farbpunkt_einfuegen_2_ost'.
void FarbenUpdate::fu_punkt_farbpunkt_einfuegen_2_nord(void)
    {
    if(f==0)
        {
        return;
        }
    f_schlange = f-1;
    g_schlange = g;

    direktes_einfuegen2();

// Naechste Zeile neu 15.10.14
    stellvertretung_fg_gleich_gg_nord();
#ifdef AAHH
    test3fu("nord");
#endif
    }

void FarbenUpdate::fu_punkt_farbpunkt_einfuegen_2_suedost(void)
// Kommentar: Analog zu 'fu_punkt_farbpunkt_einfuegen_2_nordost'.
    {
    if(g==m_minus1)
        {
        return;
        }
    if(f==n_minus1)
        {
        return;
        }
    f_schlange = f+1;
    g_schlange = g+1;

    direktes_einfuegen2();
#ifdef AAHH
    test3fu("suedost");
#endif
    }

// Kommentar: Analog zu 'fu_punkt_farbpunkt_einfuegen_2_ost'.
void FarbenUpdate::fu_punkt_farbpunkt_einfuegen_2_sued(void)
    {
    if(f==n_minus1)
        {
        return;
        }
    f_schlange = f+1;
    g_schlange = g;

    direktes_einfuegen2();
// Naechste Zeile neu 15.10.14
#ifdef AAHH
    test3fu("sued vor stellvertretung");
#endif
    stellvertretung_fg_gleich_gg_sued();
// Das wahrscheinlich doppelt gemoppelt.
#ifdef AAHH
    test3fu("sued");
#endif
    }

// Kommentar: Analog zu 'fu_punkt_farbpunkt_einfuegen_2_nordost'.
void FarbenUpdate::fu_punkt_farbpunkt_einfuegen_2_suedwest(void)
    {
    if(f==n_minus1)
        {
        return;
        }
    if(g==0)
        {
        return;
        }
    f_schlange = f+1;
    g_schlange = g-1;

    direktes_einfuegen2();
#ifdef AAHH
    test3fu("suedwest");
#endif
    }

// Kommentar: Analog zu 'fu_punkt_farbpunkt_einfuegen_2_ost'.
void FarbenUpdate::fu_punkt_farbpunkt_einfuegen_2_west(void)
    {
    if(g==0)
        {
        return;
        }
    f_schlange = f;
    g_schlange = g-1;

    direktes_einfuegen2();

// Naechste Zeile neu 15.10.14
    stellvertretung_fg_gleich_gg_west();
#ifdef AAHH
    test3fu("west");
#endif
    }

// Kommentar: Analog zu 'fu_punkt_farbpunkt_einfuegen_2_nordost'.
void FarbenUpdate::fu_punkt_farbpunkt_einfuegen_2_nordwest(void)
    {
    f_schlange = f-1;
    g_schlange = g-1;
    if(g==0)
        {
        return;
        }
    if(f==0)
        {
        return;
        }

    direktes_einfuegen2();
#ifdef AAHH
    test3fu("nordwest");
#endif
    }

// Kommentar: Das direkte und indirekte einfgen sowie die Behandlung innerer Ecken wird fr neue
// Gitterrandpunkte durchgefhrt. Dabei wird die Funktion 'fu_punkt_farbpunkt_einfuegen_2_nord' aufgerufen.
void FarbenUpdate::einfuegen_ein_punkt_ep_ff_2(int _f, int _g)
    {
// Kommentar: Kopie der Koordinaten des gemeinsamen Randpunktes.
    f = _f;
    g = _g;
// Kommentar: Fr die Randpunkte werden die fr die nicht auf dem Kern liegenden Nebenpunkte entsprechende
// Funktionen aufgerufen.
        switch(255 - feingitter_snb[f][g]) // Ich denke, dass der dynamische Rand auf dem feinen Gitter liegt.
            {
                {
               case 0:
                   break;
               case 1:
// Kommentar: Aufruf von 'fu_punkt_farbpunkt_einfuegen_0_nord'. Fr alle Flle darunter ein analoger Kommentar.
                   punkt_farbpunkt_einfuegen_2_nord;
                   break;
               case 2:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   break;
               case 3:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_nord;
                   break;
               case 4:
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 5:
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 6:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 7:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 8:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   break;
               case 9:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nord;
                   break;
               case 10:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   break;
               case 11:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nord;
                   break;
               case 12:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 13:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 14:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 15:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 16:
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 17:
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 18:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 19:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 20:
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 21:
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 22:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 23:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 24:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 25:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 26:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 27:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 28:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 29:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 30:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 31:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 32:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   break;
               case 33:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   break;
               case 34:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   break;
               case 35:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   break;
               case 36:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 37:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 38:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 39:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 40:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   break;
               case 41:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   break;
               case 42:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   break;
               case 43:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   break;
               case 44:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 45:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 46:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 47:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 48:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 49:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 50:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 51:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 52:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 53:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 54:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 55:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 56:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 57:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 58:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 59:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 60:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 61:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 62:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 63:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 64:
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 65:
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 66:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 67:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 68:
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 69:
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 70:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 71:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 72:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 73:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 74:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 75:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 76:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 77:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 78:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 79:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 80:
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 81:
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 82:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 83:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 84:
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 85:
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 86:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 87:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 88:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 89:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 90:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 91:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 92:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 93:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 94:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 95:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 96:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 97:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 98:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 99:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 100:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 101:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 102:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 103:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 104:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 105:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 106:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 107:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 108:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 109:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 110:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 111:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 112:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 113:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 114:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 115:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 116:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 117:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 118:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 119:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 120:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 121:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 122:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 123:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 124:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 125:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 126:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 127:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 128:
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   break;
               case 129:
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   break;
               case 130:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   break;
               case 131:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   break;
               case 132:
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 133:
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 134:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 135:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 136:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   break;
               case 137:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   break;
               case 138:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   break;
               case 139:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   break;
               case 140:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 141:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 142:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 143:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 144:
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 145:
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 146:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 147:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 148:
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 149:
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 150:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 151:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 152:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 153:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 154:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 155:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 156:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 157:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 158:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 159:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 160:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   break;
               case 161:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   break;
               case 162:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   break;
               case 163:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   break;
               case 164:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 165:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 166:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 167:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 168:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   break;
               case 169:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   break;
               case 170:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   break;
               case 171:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   break;
               case 172:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 173:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 174:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 175:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   break;
               case 176:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 177:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 178:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 179:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 180:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 181:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 182:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 183:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 184:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 185:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 186:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 187:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 188:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 189:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 190:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 191:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   break;
               case 192:
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 193:
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 194:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 195:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 196:
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 197:
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 198:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 199:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 200:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 201:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 202:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 203:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 204:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 205:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 206:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 207:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 208:
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 209:
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 210:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 211:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 212:
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 213:
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 214:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 215:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 216:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 217:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 218:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 219:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 220:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 221:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 222:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 223:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 224:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 225:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 226:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 227:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 228:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 229:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 230:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 231:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 232:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 233:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 234:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 235:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 236:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 237:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 238:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 239:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 240:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 241:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 242:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 243:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 244:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 245:
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 246:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 247:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 248:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 249:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 250:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 251:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 252:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 253:
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 254:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;
               case 255:
                   punkt_farbpunkt_einfuegen_2_nordost;
                   punkt_farbpunkt_einfuegen_2_suedost;
                   punkt_farbpunkt_einfuegen_2_suedwest;
                   punkt_farbpunkt_einfuegen_2_nordwest;
                   punkt_farbpunkt_einfuegen_2_nord;
                   punkt_farbpunkt_einfuegen_2_ost;
                   punkt_farbpunkt_einfuegen_2_sued;
                   punkt_farbpunkt_einfuegen_2_west;
                   break;

                }
            }
    }


// Kommentar: Die folgende Funktion macht direktes und indirektes Entfernen in Richtung Osten fr wissende Punkte.
void FarbenUpdate::fu_punkt_farbpunkt_entfernen_ost(void)
    {
// Kommentar: stlichste Punkte brauchen nicht behandelt zu werden.
    if(g==m_minus1)
        {
        return;
        }
// Kommentar: Der Paarpunkt wird fixiert.
    f_schlange = f;
    g_schlange = g+1;

#ifdef AUSGABEN_FARBENUPDATE
    cout << " Ost: Vor direktes entfernen rang " << rang << endl;
#endif
// Kommentar: Direktes und indirektes entfernen wird durchgefhrt.
    direktes_entfernen();
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Nach direktes entfernen rang " << rang << endl;
#endif
    indirektes_entfernen_ost();
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Nach indirektes entfernen rang " << rang << endl;
#endif
    }

// Kommentar: Fast analog zu 'fu_punkt_farbpunkt_entfernen_ost': Nur das indirekte_entfernen erfolgt nicht, weil
// die indirekten Punkte horizontal oder vertikal daneben liegen.
void FarbenUpdate::fu_punkt_farbpunkt_entfernen_nordost(void)
    {
    if(f==0)
        {
        return;
        }

    if(g==m_minus1)
        {
        return;
        }

    f_schlange = f-1;
    g_schlange = g+1;

#ifdef AUSGABEN_FARBENUPDATE
    cout << " NordOst: Vor direktes entfernen rang " << rang << endl;
#endif
    direktes_entfernen();
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Nach direktes entfernen rang " << rang << endl;
#endif
    }

// Kommentar: Analog zu 'fu_punkt_farbpunkt_entfernen_ost'.
void FarbenUpdate::fu_punkt_farbpunkt_entfernen_nord(void)
    {
    if(f==0)
        {
        return;
        }
    f_schlange = f-1;
    g_schlange = g;

#ifdef AUSGABEN_FARBENUPDATE
    cout << " Nord: Vor direktes entfernen rang " << rang << endl;
#endif
    direktes_entfernen();
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Nach direktes entfernen rang " << rang << endl;
#endif
    indirektes_entfernen_nord();
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Nach indirektes entfernen rang " << rang << endl;
#endif
    }

// Kommentar: Analog zu 'fu_punkt_farbpunkt_entfernen_nordost'.
void FarbenUpdate::fu_punkt_farbpunkt_entfernen_suedost(void)
    {
    if(f==n_minus1)
        {
        return;
        }
    if(g==m_minus1)
        {
        return;
        }
    f_schlange = f+1;
    g_schlange = g+1;

#ifdef AUSGABEN_FARBENUPDATE
    cout << " Suedost: Vor direktes entfernen rang " << rang << endl;
#endif
    direktes_entfernen();
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Nach direktes entfernen rang " << rang << endl;
#endif
    }

// Kommentar: Analog zu 'fu_punkt_farbpunkt_entfernen_ost'.
void FarbenUpdate::fu_punkt_farbpunkt_entfernen_sued(void)
    {
    if(f==n_minus1)
        {
        return;
        }
    f_schlange = f+1;
    g_schlange = g;

#ifdef AUSGABEN_FARBENUPDATE
    cout << " Sued: Vor direktes entfernen rang " << rang << endl;
#endif
    direktes_entfernen();
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Nach direktes entfernen rang " << rang << endl;
#endif
    indirektes_entfernen_sued();
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Nach indirektes entfernen rang " << rang << endl;
#endif
    }

// Kommentar: Analog zu 'fu_punkt_farbpunkt_entfernen_nordost'.
void FarbenUpdate::fu_punkt_farbpunkt_entfernen_suedwest(void)
    {
    if(f==n_minus1)
        {
        return;
        }
    if(g==0)
        {
        return;
        }
    f_schlange = f+1;
    g_schlange = g-1;

#ifdef AUSGABEN_FARBENUPDATE
    cout << " Suedwest: Vor direktes entfernen rang " << rang << endl;
#endif
    direktes_entfernen();
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Nach direktes entfernen rang " << rang << endl;
#endif
    }

// Kommentar: Analog zu 'fu_punkt_farbpunkt_entfernen_ost'.
void FarbenUpdate::fu_punkt_farbpunkt_entfernen_west(void)
    {
    if(g==0)
        {
        return;
        }

    f_schlange = f;
    g_schlange = g-1;

#ifdef AUSGABEN_FARBENUPDATE
    cout << " West: Vor direktes entfernen rang " << rang << endl;
#endif
    direktes_entfernen();
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Nach direktes entfernen rang " << rang << endl;
#endif
    indirektes_entfernen_west();
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Nach indirektes entfernen rang " << rang << endl;
#endif
    }

// Kommentar: Analog zu 'fu_punkt_farbpunkt_entfernen_nordost'.
void FarbenUpdate::fu_punkt_farbpunkt_entfernen_nordwest(void)
    {
    if(g==0)
        {
        return;
        }
    if(f==0)
        {
        return;
        }
    f_schlange = f-1;
    g_schlange = g-1;

#ifdef AUSGABEN_FARBENUPDATE
    cout << " Nordwest: Vor direktes entfernen rang " << rang << endl;
#endif
    direktes_entfernen();
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Nach direktes entfernen rang " << rang << endl;
#endif
//    indirektes_einfuegen0_ost();
    }

// Kommentar: Man macht ein indirektes_entfernen fr den stlichen Nachbarn.
// Das das ein zulssiger Punkt ist (g_schlange < m) wird im Vorfeld geklrt.
// Da das hier zu Kommunikationen fhrt, soll es so selten wie mglicg erfolgen. Daher unterdrcken die ersten
// 3 if - Abfragen viele unntigen Kommunikationen.
void FarbenUpdate::indirektes_entfernen_ost(void)
    {
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Ost 1 rang " << rang << " f_schlange " << f_schlange << " g_schlange " << g_schlange << endl;
#endif
// Kommentar: In unterem Fall ist (f_schlange,g_schlange) ein direkter Punkt.
    if(((f_schlange&1)==0)&&((g_schlange&1)==0))
        {
        return;
        }
// Kommentar: ep_ff = 1 => Der Punkt (f_schlange,g_schlange) existiert nicht mehr. Dafr ist natrlich nichts zu tun.
// ep_ff = 2 => das sind neue Punkte, und das sind immer wissende Punkte, weil sie dieselbe Farbe bekommen wie ihr
// zugehriger Grobgitterpunkt. Dafr ist auch nichts zu tun. Also verbleibt ep_ff = 0 als Wahl fr
// ein Einfgen an Stelle von.
    if(ep_ff[f_schlange][g_schlange]!=0)
        {
        return;
        }
// Kommentar: Wenn fgff = rang ist, liegt die Information vor, da beide Punkte, (f,g) und (f_schlange,g_schlange), auf demselben
// Kern liegen. Eine Kommunikation wre eine Kommunikation mit sich selber, und das ist sinnlos.
    if(fgff[f_schlange][g_schlange]==rang)
        {
        return;
        }
// Kommentar: Wenn fgff < 0 ist, dann hat dort vorher kein Punkt existiert. Entweder existiert dort immer noch kein Punkt,
// dann braucht der nicht behandelt zu werden. Oder der Punkt ist neu, dann ist er - vgl. die erste if - Abfrage - ein
// wissender Punkt, fr den indirekt nichts getan werden muss.
    if(fgff[f_schlange][g_schlange]<0)
        {
        return;
        }
//  Vorerst nicht einfuegen, damit nicht doppelt eingefgt wird => Zhler stimmen aonat nicht mehr.

// Kommentar: Der indirekte Punkt muss umgeben werden von Farbwerten - damit sein Farbenfeld exakt ist. Das gilt auch fr
// das entfernen. Auf den gemeinsamen Punkten geschieht das durch direktes entfernen, vgl. direktes_entfernen.
// Damit verbleiben die Werte an den Seiten der Linie durch die gemeinsamen Punkte. Bei Ost nimmt
// man nur die 2 zwei Punkte nrdlich und sdlich des gemeisamen Punktes. Im Falle von Westen sind dies
// dann 4 Punkte. Fr diese 2 Punkte geschieht nun eine Kommunikation durch die Funktion indirekter_farbpunkt_entfernen.
    if((f!=0)&&(g!=m_minus1))
        {
        indirekter_farbpunkt_entfernen(f-1, g+1);
        }
    if((f!=n_minus1)&&(g!=m_minus1))
        {
        indirekter_farbpunkt_entfernen(f+1, g+1);
        }
    }

// Kommentar: Das direkte Entfernen gilt fr den Punkt (f,g). Die Farbfeldnderung auf -1 erfolgt, wenn sie
// erfolgt, auf (f_schlange,g_schlange). Diese Variablen sind beim indirekten entfernen anders: Dort sind
// das die Punkte, fr die der Farbwechsel geschieht. Hier ist (f_schlange,g_schlange) das, was dort
// (f_s,g_s) ist.
// Diese Funktion fgt direkt - falls ntig - eine -1 ins Farbenfeld auf Position (f_schlange,g_schlange) ein.
void FarbenUpdate::direktes_entfernen(void)
    {
// Kommentar: Das Entfernen erfolgt nur fr Punkte, die entfernt werden; und die haben den ep_ff - Wert 1.
    if(ep_ff[f_schlange][g_schlange]!=1)
        {
        return;
        }
#ifdef AUSGABEN_DATEI
    da_eez->aus << " direktes entfernen (dek) Rang " << rang << " Ziel P " << fgff[f_schlange][g_schlange] << " f " << f << " g " << g << " m " << m << endl;
#endif
// Kommentar: Einfgen des Farbwertes -1 auf die Position mit ep_ff = 1.
    fgff[f_schlange][g_schlange] = -1;

#ifdef AADD
    if((rang==3)&&(t==5)&&(f_schlange==12)&&(g_schlange==6))
        {
        cout << " AADD 2 rang 3 fgff(12,6) " << fgff[12][6] << endl;
        }
#endif
    // Keine Kommunikation in diesem Fall. Weil der Prozessor P(f_schlange, g_schlange) im allgemeinen den Wert von direktes
    // einfuegen uebernimmt.

    }

// Kommentar: Fast analog zu 'indirektes_entfernen_ost'. Der einzige Unterschied ist der Aufruf von indirekter_farbpunkt_entfernen
// nicht nur nrdlich und sdlich neben dem gemeinsamen Punkt, sondern auch nrdlich und sdlich neben dem indirekten Punkt.
void FarbenUpdate::indirektes_entfernen_west(void)
    {
#ifdef AUSGABEN_FARBENUPDATE
    cout << " West 1 rang " << rang << " f_schlange " << f_schlange << " g_schlange " << g_schlange << endl;
#endif
    if(((f_schlange&1)==0)&&((g_schlange&1)==0))
        {
        return;
        }
    if(ep_ff[f_schlange][g_schlange]!=0)
        {
        return;
        }
    if(fgff[f_schlange][g_schlange]==rang)
        {
        return;
        }
    if(fgff[f_schlange][g_schlange]<0)
        {
        return;
        }
    if(f!=0)
        {
        indirekter_farbpunkt_entfernen(f-1, g);
        }
    if(f!=n_minus1)
        {
        indirekter_farbpunkt_entfernen(f+1, g);
        }
    if((f!=0)&&(g!=0))
        {
        indirekter_farbpunkt_entfernen(f-1, g-1);
        }
    if((g!=0)&&(f!=n_minus1))
        {
        indirekter_farbpunkt_entfernen(f+1, g-1);
        }
    }

// Kommentar: Analog zu 'indirektes_entfernen_west'.
void FarbenUpdate::indirektes_entfernen_nord(void)
    {
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Nord 1 rang " << rang << " f_schlange " << f_schlange << " g_schlange " << g_schlange << endl;
#endif
    if(((f_schlange&1)==0)&&((g_schlange&1)==0))
        {
        return;
        }
    if(ep_ff[f_schlange][g_schlange]!=0)
        {
        return;
        }
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Nord 2 rang " << rang << endl;
#endif
    if(fgff[f_schlange][g_schlange]==rang)
        {
        return;
        }
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Nord 3 rang " << rang << endl;
#endif
    if(fgff[f_schlange][g_schlange]<0)
        {
        return;
        }
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Nord 4 rang " << rang << " f " << f << " g " << g << endl;
#endif
    if((f!=0)&&(g!=0))
        {
        indirekter_farbpunkt_entfernen(f-1, g-1);
        }
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Nord 5 rang " << rang << endl;
#endif
    if((f!=0)&&(g!=m_minus1))
        {
        indirekter_farbpunkt_entfernen(f-1, g+1);
        }
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Nord 6 rang " << rang << endl;
#endif
    if(g!=0)
        {
        indirekter_farbpunkt_entfernen(f, g-1);
        }
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Nord 7 rang " << rang << endl;
#endif
    if(g!=m_minus1)
        {
        indirekter_farbpunkt_entfernen(f, g+1);
        }
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Nord 8 rang " << rang << endl;
#endif
    }

// Kommentar: Analog zu 'indirektes_entfernen_ost'.
void FarbenUpdate::indirektes_entfernen_sued(void)
    {
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Sued 1 rang " << rang << " f_schlange " << f_schlange << " g_schlange " << g_schlange << endl;
#endif
    if(((f_schlange&1)==0)&&((g_schlange&1)==0))
        {
        return;
        }
    if(ep_ff[f_schlange][g_schlange]!=0)
        {
        return;
        }
    if(fgff[f_schlange][g_schlange]==rang)
        {
        return;
        }
    if(fgff[f_schlange][g_schlange]<0)
        {
        return;
        }
#ifdef AUSGABEN_FARBENUPDATE
    cout << " ep_ff " << ep_ff[f_schlange][g_schlange] << " fgff " << fgff[f_schlange][g_schlange] << " rang " << rang << endl;
    if((f==15)&&(g==16)&&(f_schlange==16)&&(g_schlange==16))
       {
        cout << " In indirektes_entfernen Sued Zielfall gefunden " << endl;
       }
#endif
    if((f!=n_minus1)&&(g!=0))
        {
        indirekter_farbpunkt_entfernen(f+1, g-1);
        }
    if((f!=n_minus1)&&(g!=m_minus1))
        {
        indirekter_farbpunkt_entfernen(f+1, g+1);
        }
    }

// Kommentar: Auf dem Punkt (f,g) wird fr den Kern vom Nachbarpunkt (f_schlange,g_schlange) ggf. der
// Farbenfeldwert auf der Koordinate (f_s,g_s) auf -1 gesetzt. Dafr bedarf es einer Kommnikation.
void FarbenUpdate::indirekter_farbpunkt_entfernen(int f_s, int g_s)
    {
// Kommmentar: Es wird noch geprft, dass dort, wo die -1 eingetragen werden soll, wirklich auf dem Kern des Punktes (f,g)
// ep_ff = 1 ist.
    if(ep_ff[f_s][g_s] == 1)
        {
#ifdef ALT
// Kommentar: f_schlange und g_schlange werden zu short ints konvertiert. Die Befehle dazu sind aber auskommentiert.
        f_si = (short int) f_schlange;
        g_si = (short int) g_schlange;
#endif
// f_si und g_si werden berechnet, aber nicht gebraucht. Ein Fehler wird hier vermutet.
// Vielleicht sind diese Zeilen sowie ggff_speichern nur berflssig.
#ifdef ALT
// Diese Zeile beim kommentieren entfernt. Sie wird offenbar nicht mehr gebraucht.
        ggff_speichern = ggff[f_s >> 1][g_s >> 1];
#endif
// Kommentar: Der Kern mit dem Punkt (f_schlange,g_schlange) bekommt die Koordinaten (f_s,g_s) zugeschickt, um
// dort ins Farbenfeld eine -1 einzutragen.
        kommunikation4er->entfernen_fu(f_s, g_s, fgff[f_schlange][g_schlange]);
#ifdef AUSGABEN_FARBENUPDATE
        strcpy(text, "16392");
        test_entfernen(f_s, g_s, fgff[f_schlange][g_schlange], text);
        cout << " indireker_farbpunkt_entfernen 2 rang " << rang << " f_s " << f_s << " g_s " << g_s << " Ziel P " << fgff[f_schlange][g_schlange] << endl;
#endif
        }
    }

// Kommentar: Fr alle wissenden Punkte am Rand nacht dem P+ eingefgt und P- entfernt sind geschehen hier die
// Farbkorrekturen fr den Punkt (_f,_g).
void FarbenUpdate::entfernen_ein_punkt(int _f, int _g)
    {
// Kommentar: Die Koordinaten werden gesichert,
    f = _f;
    g = _g;

#ifdef AUSGABEN_FARBENUPDATE
    if((f==10)&&(g==4)&&(rang==3))
        {
        cout << " DDCCBB entfernen_ein_punkt f " << f << " g " << g << " rang " << rang << " 255-snb " << 255-feingitter_snb[f][g] << endl;
        }
#endif

// Kommentar: Zuerst wird eine 256 - Fallunterscheidung entfernen_korrigieren durchgefhrt, vgl. den Kommentar dort.
// Die Reihenfolge spielt eine Rolle - vgl. den Kommentar zu entfernen_korrigieren.
    entfernen_korrigieren();

#ifdef AUSGABEN_FARBENUPDATE
    cout << " 12900 rang " << rang << endl;
#endif
// Kommentar: Fr die Randpunkte werden die fr die nicht auf dem Kern liegenden Nebenpunkte entsprechende
// Funktionen aufgerufen.
        switch(255 - feingitter_snb[f][g]) // Ich denke, dass der dynamische Rand auf dem feinen Gitter liegt.
            {
                {
                case 0:
                    break;
                case 1:
// Kommentar: Die Funktion 'fu_punkt_farbpunkt_entfernen_nord' wird aufgerufen, vgl. oben.
// Die anderen Flle: Entsprechende 'fu_punkt_farbpunkt_entfernen_HR' mit HR eine Himmelsrichtung werden aufgerufen.
                    punkt_farbpunkt_entfernen_nord;
                    break;
                case 2:
                    punkt_farbpunkt_entfernen_nordost;
                    break;
                case 3:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    break;
                case 4:
                    punkt_farbpunkt_entfernen_ost;
                    break;
                case 5:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    break;
                case 6:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    break;
                case 7:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    break;
                case 8:
                    punkt_farbpunkt_entfernen_suedost;
                    break;
                case 9:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_suedost;
                    break;
                case 10:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    break;
                case 11:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    break;
                case 12:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    break;
                case 13:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    break;
                case 14:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    break;
                case 15:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    break;
                case 16:
                    punkt_farbpunkt_entfernen_sued;
                    break;
                case 17:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_sued;

                    break;
                case 18:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_sued;

                    break;
                case 19:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_sued;

                    break;
                case 20:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;

                    break;
                case 21:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;

                    break;
                case 22:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;

                    break;
                case 23:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;

                    punkt_farbpunkt_entfernen_sued;

                    break;
                case 24:
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;

                    break;
                case 25:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;

                    break;
                case 26:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;

                    break;
                case 27:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;

                    break;
                case 28:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    break;
                case 29:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;

                    break;
                case 30:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;

                    break;
                case 31:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;

                    break;
                case 32:
                    punkt_farbpunkt_entfernen_suedwest;
                    break;
                case 33:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_suedwest;
                    break;
                case 34:
                    punkt_farbpunkt_entfernen_nordost;

                    punkt_farbpunkt_entfernen_suedwest;
                    break;
                case 35:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedwest;
                    break;
                case 36:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedwest;
                    break;
                case 37:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedwest;
                    break;
                case 38:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedwest;
                    break;
                case 39:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedwest;
                    break;
                case 40:
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    break;
                case 41:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    break;
                case 42:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    break;
                case 43:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    break;
                case 44:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    break;
                case 45:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    break;
                case 46:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    break;
                case 47:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    break;
                case 48:
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;
                    break;
                case 49:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    break;
                case 50:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    break;
                case 51:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    break;
                case 52:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    break;
                case 53:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    break;
                case 54:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    break;
                case 55:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    break;
                case 56:
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    break;
                case 57:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    break;
                case 58:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    break;
                case 59:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    break;
                case 60:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;
                    break;
                case 61:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    break;
                case 62:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    break;
                case 63:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    break;

                case 64:
                    punkt_farbpunkt_entfernen_west;
                    break;

                case 65:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 66:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 67:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 68:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 69:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 70:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 71:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 72:
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 73:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 74:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 75:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 76:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 77:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 78:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 79:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 80:
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 81:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 82:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 83:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 84:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 85:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 86:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 87:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;

                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 88:
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 89:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 90:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 91:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 92:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 93:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 94:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 95:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 96:
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    break;

                case 97:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 98:
                    punkt_farbpunkt_entfernen_nordost;

                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 99:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 100:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 101:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 102:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 103:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 104:
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 105:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 106:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 107:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 108:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 109:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 110:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 111:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 112:
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 113:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 114:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 115:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 116:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 117:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 118:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 119:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 120:
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 121:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 122:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 123:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 124:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    break;
                case 125:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 126:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    break;
                case 127:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    break;

                case 128:
                    punkt_farbpunkt_entfernen_nordwest;
                    break;
                case 129:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 130:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 131:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 132:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 133:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 134:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 135:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 136:
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 137:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 138:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 139:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 140:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 141:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 142:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 143:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 144:
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 145:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 146:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 147:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 148:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 149:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 150:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 151:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;

                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 152:
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 153:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 154:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 155:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 156:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 157:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 158:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 159:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 160:
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 161:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 162:
                    punkt_farbpunkt_entfernen_nordost;

                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 163:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 164:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 165:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 166:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 167:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 168:
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 169:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 170:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 171:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 172:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 173:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 174:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 175:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 176:
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 177:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 178:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 179:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 180:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 181:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 182:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 183:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 184:
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 185:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 186:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 187:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 188:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 189:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 190:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_nordwest; break;
                case 191:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_nordwest; break;

                case 192:
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest;
                    break;

                case 193:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 194:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 195:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 196:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 197:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 198:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 199:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 200:
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 201:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 202:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 203:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 204:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 205:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 206:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 207:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 208:
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 209:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 210:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 211:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 212:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 213:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 214:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 215:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;

                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 216:
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 217:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 218:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 219:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 220:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest;
                    break;
                case 221:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest;
                    break;
                case 222:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest;
                    break;
                case 223:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;

                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 224:

                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest;
                    break;

                case 225:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 226:
                    punkt_farbpunkt_entfernen_nordost;

                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 227:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 228:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 229:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 230:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 231:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 232:
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 233:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 234:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 235:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 236:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 237:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 238:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 239:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 240:
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 241:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 242:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 243:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 244:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 245:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 246:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 247:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 248:
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 249:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 250:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 251:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 252:
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;
                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 253:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 254:
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                case 255:
                    punkt_farbpunkt_entfernen_nord;
                    punkt_farbpunkt_entfernen_nordost;
                    punkt_farbpunkt_entfernen_ost;
                    punkt_farbpunkt_entfernen_suedost;
                    punkt_farbpunkt_entfernen_sued;
                    punkt_farbpunkt_entfernen_suedwest;

                    punkt_farbpunkt_entfernen_west;
                    punkt_farbpunkt_entfernen_nordwest; break;
                }
            }
#ifdef AUSGABEN_FARBENUPDATE
    cout << " Ende 256er von 12900 rang " << rang << endl;
#endif
    }

// Kommentar: Da SV_herein nicht definiert ist, wird diese Funktion nicht aufgerufen, und daher wird sie nicht kommentiert.
void FarbenUpdate::stellvertretendes_einfuegen(void)
    {
    if(erweitertes_snb_0!=0)
        {
        farbe_short_int = fgff[f][g];
        farbe_char = (char) farbe_short_int;
        kommunikation4er->einfuegen_fu(f, g, farbe_char, fgff[f_schlange][g_schlange]);
        // Das sieht nicht richtig aus: f_schlange / g_schlange wird ja mehrfach beschrieben. 18.2.15
        }
        switch(erweitertes_snb_0) // Ich denke, dass der dynamische Rand auf dem feinen Gitter liegt.
            {

           case 0:
               break;
           case 1:
               stellvertretend_nord();
               break;
           case 2:
               stellvertretend_nordost();
               break;
           case 3:
               stellvertretend_nord_nordost();
               break;
           case 4:
               stellvertretend_ost();
               break;
           case 5:
               stellvertretend_nord();
               stellvertretend_ost();
               break;
           case 6:
               stellvertretend_ost_nordost();
               break;
           case 7:
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               break;
           case 8:
               stellvertretend_suedost();
               break;
           case 9:
               stellvertretend_nord();
               stellvertretend_suedost();
               break;
           case 10:
               stellvertretend_nordost();
               stellvertretend_suedost();
               break;
           case 11:
               stellvertretend_suedost();
               stellvertretend_nord_nordost();
               break;
           case 12:
               stellvertretend_ost_suedost();
               break;
           case 13:
               stellvertretend_nord();
               stellvertretend_ost_suedost();
               break;
           case 14:
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               break;
           case 15:
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               break;
           case 16:
               stellvertretend_sued();
               break;
           case 17:
               stellvertretend_nord();
               stellvertretend_sued();
               break;
           case 18:
               stellvertretend_nordost();
               stellvertretend_sued();
               break;
           case 19:
               stellvertretend_sued();
               stellvertretend_nord_nordost();
               break;
           case 20:
               stellvertretend_ost();
               stellvertretend_sued();
               break;
           case 21:
               stellvertretend_nord();
               stellvertretend_ost();
               stellvertretend_sued();
               break;
           case 22:
               stellvertretend_sued();
               stellvertretend_ost_nordost();
               break;
           case 23:
               stellvertretend_sued();
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               break;
           case 24:
               stellvertretend_sued_suedost();
               break;
           case 25:
               stellvertretend_nord();
               stellvertretend_sued_suedost();
               break;
           case 26:
               stellvertretend_nordost();
               stellvertretend_sued_suedost();
               break;
           case 27:
               stellvertretend_nord_nordost();
               stellvertretend_sued_suedost();
               break;
           case 28:
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               break;
           case 29:
               stellvertretend_nord();
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               break;
           case 30:
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               break;
           case 31:
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               break;
           case 32:
               stellvertretend_suedwest();
               break;
           case 33:
               stellvertretend_nord();
               stellvertretend_suedwest();
               break;
           case 34:
               stellvertretend_nordost();
               stellvertretend_suedwest();
               break;
           case 35:
               stellvertretend_suedwest();
               stellvertretend_nord_nordost();
               break;
           case 36:
               stellvertretend_ost();
               stellvertretend_suedwest();
               break;
           case 37:
               stellvertretend_nord();
               stellvertretend_ost();
               stellvertretend_suedwest();
               break;
           case 38:
               stellvertretend_suedwest();
               stellvertretend_ost_nordost();
               break;
           case 39:
               stellvertretend_suedwest();
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               break;
           case 40:
               stellvertretend_suedost();
               stellvertretend_suedwest();
               break;
           case 41:
               stellvertretend_nord();
               stellvertretend_suedost();
               stellvertretend_suedwest();
               break;
           case 42:
               stellvertretend_nordost();
               stellvertretend_suedost();
               stellvertretend_suedwest();
               break;
           case 43:
               stellvertretend_suedost();
               stellvertretend_suedwest();
               stellvertretend_nord_nordost();
               break;
           case 44:
               stellvertretend_suedwest();
               stellvertretend_ost_suedost();
               break;
           case 45:
               stellvertretend_nord();
               stellvertretend_suedwest();
               stellvertretend_ost_suedost();
               break;
           case 46:
               stellvertretend_suedwest();
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               break;
           case 47:
               stellvertretend_suedwest();
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               break;
           case 48:
               stellvertretend_sued_suedwest();
               break;
           case 49:
               stellvertretend_nord();
               stellvertretend_sued_suedwest();
               break;
           case 50:
               stellvertretend_nordost();
               stellvertretend_sued_suedwest();
               break;
           case 51:
               stellvertretend_nord_nordost();
               stellvertretend_sued_suedwest();
               break;
           case 52:
               stellvertretend_ost();
               stellvertretend_sued_suedwest();
               break;
           case 53:
               stellvertretend_nord();
               stellvertretend_ost();
               stellvertretend_sued_suedwest();
               break;
           case 54:
               stellvertretend_ost_nordost();
               stellvertretend_sued_suedwest();
               break;
           case 55:
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_sued_suedwest();
               break;
           case 56:
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               break;
           case 57:
               stellvertretend_nord();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               break;
           case 58:
               stellvertretend_nordost();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               break;
           case 59:
               stellvertretend_nord_nordost();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               break;
           case 60:
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               break;
           case 61:
               stellvertretend_nord();
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               break;
           case 62:
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               break;
           case 63:
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               break;
           case 64:
               stellvertretend_west();
               break;
           case 65:
               stellvertretend_nord();
               stellvertretend_west();
               break;
           case 66:
               stellvertretend_nordost();
               stellvertretend_west();
               break;
           case 67:
               stellvertretend_west();
               stellvertretend_nord_nordost();
               break;
           case 68:
               stellvertretend_ost();
               stellvertretend_west();
               break;
           case 69:
               stellvertretend_nord();
               stellvertretend_ost();
               stellvertretend_west();
               break;
           case 70:
               stellvertretend_west();
               stellvertretend_ost_nordost();
               break;
           case 71:
               stellvertretend_west();
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               break;
           case 72:
               stellvertretend_suedost();
               stellvertretend_west();
               break;
           case 73:
               stellvertretend_nord();
               stellvertretend_suedost();
               stellvertretend_west();
               break;
           case 74:
               stellvertretend_nordost();
               stellvertretend_suedost();
               stellvertretend_west();
               break;
           case 75:
               stellvertretend_suedost();
               stellvertretend_west();
               stellvertretend_nord_nordost();
               break;
           case 76:
               stellvertretend_west();
               stellvertretend_ost_suedost();
               break;
           case 77:
               stellvertretend_nord();
               stellvertretend_west();
               stellvertretend_ost_suedost();
               break;
           case 78:
               stellvertretend_west();
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               break;
           case 79:
               stellvertretend_west();
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               break;
           case 80:
               stellvertretend_sued();
               stellvertretend_west();
               break;
           case 81:
               stellvertretend_nord();
               stellvertretend_sued();
               stellvertretend_west();
               break;
           case 82:
               stellvertretend_nordost();
               stellvertretend_sued();
               stellvertretend_west();
               break;
           case 83:
               stellvertretend_sued();
               stellvertretend_west();
               stellvertretend_nord_nordost();
               break;
           case 84:
               stellvertretend_ost();
               stellvertretend_sued();
               stellvertretend_west();
               break;
           case 85:
               stellvertretend_nord();
               stellvertretend_ost();
               stellvertretend_sued();
               stellvertretend_west();
               break;
           case 86:
               stellvertretend_sued();
               stellvertretend_west();
               stellvertretend_ost_nordost();
               break;
           case 87:
               stellvertretend_sued();
               stellvertretend_west();
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               break;
           case 88:
               stellvertretend_west();
               stellvertretend_sued_suedost();
               break;
           case 89:
               stellvertretend_nord();
               stellvertretend_west();
               stellvertretend_sued_suedost();
               break;
           case 90:
               stellvertretend_nordost();
               stellvertretend_west();
               stellvertretend_sued_suedost();
               break;
           case 91:
               stellvertretend_west();
               stellvertretend_nord_nordost();
               stellvertretend_sued_suedost();
               break;
           case 92:
               stellvertretend_west();
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               break;
           case 93:
               stellvertretend_nord();
               stellvertretend_west();
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               break;
           case 94:
               stellvertretend_west();
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               break;
           case 95:
               stellvertretend_west();
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               break;
           case 96:
               stellvertretend_west_suedwest();
               break;
           case 97:
               stellvertretend_nord();
               stellvertretend_west_suedwest();
               break;
           case 98:
               stellvertretend_nordost();
               stellvertretend_west_suedwest();
               break;
           case 99:
               stellvertretend_nord_nordost();
               stellvertretend_west_suedwest();
               break;
           case 100:
               stellvertretend_ost();
               stellvertretend_west_suedwest();
               break;
           case 101:
               stellvertretend_nord();
               stellvertretend_ost();
               stellvertretend_west_suedwest();
               break;
           case 102:
               stellvertretend_ost_nordost();
               stellvertretend_west_suedwest();
               break;
           case 103:
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_west_suedwest();
               break;
           case 104:
               stellvertretend_suedost();
               stellvertretend_west_suedwest();
               break;
           case 105:
               stellvertretend_nord();
               stellvertretend_suedost();
               stellvertretend_west_suedwest();
               break;
           case 106:
               stellvertretend_nordost();
               stellvertretend_suedost();
               stellvertretend_west_suedwest();
               break;
           case 107:
               stellvertretend_suedost();
               stellvertretend_nord_nordost();
               stellvertretend_west_suedwest();
               break;
           case 108:
               stellvertretend_ost_suedost();
               stellvertretend_west_suedwest();
               break;
           case 109:
               stellvertretend_nord();
               stellvertretend_ost_suedost();
               stellvertretend_west_suedwest();
               break;
           case 110:
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               stellvertretend_west_suedwest();
               break;
           case 111:
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               stellvertretend_west_suedwest();
               break;
           case 112:
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               break;
           case 113:
               stellvertretend_nord();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               break;
           case 114:
               stellvertretend_nordost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               break;
           case 115:
               stellvertretend_nord_nordost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               break;
           case 116:
               stellvertretend_ost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               break;
           case 117:
               stellvertretend_nord();
               stellvertretend_ost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               break;
           case 118:
               stellvertretend_ost_nordost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               break;
           case 119:
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               break;
           case 120:
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               break;
           case 121:
               stellvertretend_nord();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               break;
           case 122:
               stellvertretend_nordost();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               break;
           case 123:
               stellvertretend_nord_nordost();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               break;
           case 124:
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               break;
           case 125:
               stellvertretend_nord();
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               break;
           case 126:
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               break;
           case 127:
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               break;
           case 128:
               stellvertretend_nordwest();
               break;
           case 129:
               stellvertretend_nord_nordwest();
               break;
           case 130:
               stellvertretend_nordost();
               stellvertretend_nordwest();
               break;
           case 131:
               stellvertretend_nord_nordost();
               stellvertretend_nord_nordwest();
               break;
           case 132:
               stellvertretend_ost();
               stellvertretend_nordwest();
               break;
           case 133:
               stellvertretend_ost();
               stellvertretend_nord_nordwest();
               break;
           case 134:
               stellvertretend_nordwest();
               stellvertretend_ost_nordost();
               break;
           case 135:
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_nord_nordwest();
               break;
           case 136:
               stellvertretend_suedost();
               stellvertretend_nordwest();
               break;
           case 137:
               stellvertretend_suedost();
               stellvertretend_nord_nordwest();
               break;
           case 138:
               stellvertretend_nordost();
               stellvertretend_suedost();
               stellvertretend_nordwest();
               break;
           case 139:
               stellvertretend_suedost();
               stellvertretend_nord_nordost();
               stellvertretend_nord_nordwest();
               break;
           case 140:
               stellvertretend_nordwest();
               stellvertretend_ost_suedost();
               break;
           case 141:
               stellvertretend_ost_suedost();
               stellvertretend_nord_nordwest();
               break;
           case 142:
               stellvertretend_nordwest();
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               break;
           case 143:
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               stellvertretend_nord_nordwest();
               break;
           case 144:
               stellvertretend_sued();
               stellvertretend_nordwest();
               break;
           case 145:
               stellvertretend_sued();
               stellvertretend_nord_nordwest();
               break;
           case 146:
               stellvertretend_nordost();
               stellvertretend_sued();
               stellvertretend_nordwest();
               break;
           case 147:
               stellvertretend_sued();
               stellvertretend_nord_nordost();
               stellvertretend_nord_nordwest();
               break;
           case 148:
               stellvertretend_ost();
               stellvertretend_sued();
               stellvertretend_nordwest();
               break;
           case 149:
               stellvertretend_ost();
               stellvertretend_sued();
               stellvertretend_nord_nordwest();
               break;
           case 150:
               stellvertretend_sued();
               stellvertretend_nordwest();
               stellvertretend_ost_nordost();
               break;
           case 151:
               stellvertretend_sued();
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_nord_nordwest();
               break;
           case 152:
               stellvertretend_nordwest();
               stellvertretend_sued_suedost();
               break;
           case 153:
               stellvertretend_sued_suedost();
               stellvertretend_nord_nordwest();
               break;
           case 154:
               stellvertretend_nordost();
               stellvertretend_nordwest();
               stellvertretend_sued_suedost();
               break;
           case 155:
               stellvertretend_nord_nordost();
               stellvertretend_sued_suedost();
               stellvertretend_nord_nordwest();
               break;
           case 156:
               stellvertretend_nordwest();
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               break;
           case 157:
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               stellvertretend_nord_nordwest();
               break;
           case 158:
               stellvertretend_nordwest();
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               break;
           case 159:
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               stellvertretend_nord_nordwest();
               break;
           case 160:
               stellvertretend_suedwest();
               stellvertretend_nordwest();
               break;
           case 161:
               stellvertretend_suedwest();
               stellvertretend_nord_nordwest();
               break;
           case 162:
               stellvertretend_nordost();
               stellvertretend_suedwest();
               stellvertretend_nordwest();
               break;
           case 163:
               stellvertretend_suedwest();
               stellvertretend_nord_nordost();
               stellvertretend_nord_nordwest();
               break;
           case 164:
               stellvertretend_ost();
               stellvertretend_suedwest();
               stellvertretend_nordwest();
               break;
           case 165:
               stellvertretend_ost();
               stellvertretend_suedwest();
               stellvertretend_nord_nordwest();
               break;
           case 166:
               stellvertretend_suedwest();
               stellvertretend_nordwest();
               stellvertretend_ost_nordost();
               break;
           case 167:
               stellvertretend_suedwest();
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_nord_nordwest();
               break;
           case 168:
               stellvertretend_suedost();
               stellvertretend_suedwest();
               stellvertretend_nordwest();
               break;
           case 169:
               stellvertretend_suedost();
               stellvertretend_suedwest();
               stellvertretend_nord_nordwest();
               break;
           case 170:
               stellvertretend_nordost();
               stellvertretend_suedost();
               stellvertretend_suedwest();
               stellvertretend_nordwest();
               break;
           case 171:
               stellvertretend_suedost();
               stellvertretend_suedwest();
               stellvertretend_nord_nordost();
               stellvertretend_nord_nordwest();
               break;
           case 172:
               stellvertretend_suedwest();
               stellvertretend_nordwest();
               stellvertretend_ost_suedost();
               break;
           case 173:
               stellvertretend_suedwest();
               stellvertretend_ost_suedost();
               stellvertretend_nord_nordwest();
               break;
           case 174:
               stellvertretend_suedwest();
               stellvertretend_nordwest();
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               break;
           case 175:
               stellvertretend_suedwest();
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               stellvertretend_nord_nordwest();
               break;
           case 176:
               stellvertretend_nordwest();
               stellvertretend_sued_suedwest();
               break;
           case 177:
               stellvertretend_sued_suedwest();
               stellvertretend_nord_nordwest();
               break;
           case 178:
               stellvertretend_nordost();
               stellvertretend_nordwest();
               stellvertretend_sued_suedwest();
               break;
           case 179:
               stellvertretend_nord_nordost();
               stellvertretend_sued_suedwest();
               stellvertretend_nord_nordwest();
               break;
           case 180:
               stellvertretend_ost();
               stellvertretend_nordwest();
               stellvertretend_sued_suedwest();
               break;
           case 181:
               stellvertretend_ost();
               stellvertretend_sued_suedwest();
               stellvertretend_nord_nordwest();
               break;
           case 182:
               stellvertretend_nordwest();
               stellvertretend_ost_nordost();
               stellvertretend_sued_suedwest();
               break;
           case 183:
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_sued_suedwest();
               stellvertretend_nord_nordwest();
               break;
           case 184:
               stellvertretend_nordwest();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               break;
           case 185:
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               stellvertretend_nord_nordwest();
               break;
           case 186:
               stellvertretend_nordost();
               stellvertretend_nordwest();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               break;
           case 187:
               stellvertretend_nord_nordost();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               stellvertretend_nord_nordwest();
               break;
           case 188:
               stellvertretend_nordwest();
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               break;
           case 189:
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               stellvertretend_nord_nordwest();
               break;
           case 190:
               stellvertretend_nordwest();
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               break;
           case 191:
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               stellvertretend_nord_nordwest();
               break;
           case 192:
               stellvertretend_west_nordwest();
               break;
           case 193:
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 194:
               stellvertretend_nordost();
               stellvertretend_west_nordwest();
               break;
           case 195:
               stellvertretend_nord_nordost();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 196:
               stellvertretend_ost();
               stellvertretend_west_nordwest();
               break;
           case 197:
               stellvertretend_ost();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 198:
               stellvertretend_ost_nordost();
               stellvertretend_west_nordwest();
               break;
           case 199:
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 200:
               stellvertretend_suedost();
               stellvertretend_west_nordwest();
               break;
           case 201:
               stellvertretend_suedost();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 202:
               stellvertretend_nordost();
               stellvertretend_suedost();
               stellvertretend_west_nordwest();
               break;
           case 203:
               stellvertretend_suedost();
               stellvertretend_nord_nordost();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 204:
               stellvertretend_ost_suedost();
               stellvertretend_west_nordwest();
               break;
           case 205:
               stellvertretend_ost_suedost();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 206:
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               stellvertretend_west_nordwest();
               break;
           case 207:
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 208:
               stellvertretend_sued();
               stellvertretend_west_nordwest();
               break;
           case 209:
               stellvertretend_sued();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 210:
               stellvertretend_nordost();
               stellvertretend_sued();
               stellvertretend_west_nordwest();
               break;
           case 211:
               stellvertretend_sued();
               stellvertretend_nord_nordost();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 212:
               stellvertretend_ost();
               stellvertretend_sued();
               stellvertretend_west_nordwest();
               break;
           case 213:
               stellvertretend_ost();
               stellvertretend_sued();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 214:
               stellvertretend_sued();
               stellvertretend_ost_nordost();
               stellvertretend_west_nordwest();
               break;
           case 215:
               stellvertretend_sued();
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 216:
               stellvertretend_sued_suedost();
               stellvertretend_west_nordwest();
               break;
           case 217:
               stellvertretend_sued_suedost();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 218:
               stellvertretend_nordost();
               stellvertretend_sued_suedost();
               stellvertretend_west_nordwest();
               break;
           case 219:
               stellvertretend_nord_nordost();
               stellvertretend_sued_suedost();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 220:
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               stellvertretend_west_nordwest();
               break;
           case 221:
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 222:
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               stellvertretend_west_nordwest();
               break;
           case 223:
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 224:
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               break;
           case 225:
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 226:
               stellvertretend_nordost();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               break;
           case 227:
               stellvertretend_nord_nordost();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 228:
               stellvertretend_ost();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               break;
           case 229:
               stellvertretend_ost();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 230:
               stellvertretend_ost_nordost();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               break;
           case 231:
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 232:
               stellvertretend_suedost();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               break;
           case 233:
               stellvertretend_suedost();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 234:
               stellvertretend_nordost();
               stellvertretend_suedost();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               break;
           case 235:
               stellvertretend_suedost();
               stellvertretend_nord_nordost();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 236:
               stellvertretend_ost_suedost();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               break;
           case 237:
               stellvertretend_ost_suedost();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 238:
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               break;
           case 239:
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 240:
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               break;
           case 241:
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 242:
               stellvertretend_nordost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               break;
           case 243:
               stellvertretend_nord_nordost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 244:
               stellvertretend_ost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               break;
           case 245:
               stellvertretend_ost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 246:
               stellvertretend_ost_nordost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               break;
           case 247:
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 248:
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               break;
           case 249:
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 250:
               stellvertretend_nordost();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               break;
           case 251:
               stellvertretend_nord_nordost();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 252:
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               break;
           case 253:
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;
           case 254:
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               break;
           case 255:
               stellvertretend_nord_nordost();
               stellvertretend_ost_nordost();
               stellvertretend_ost_suedost();
               stellvertretend_sued_suedost();
               stellvertretend_sued_suedwest();
               stellvertretend_west_suedwest();
               stellvertretend_west_nordwest();
               stellvertretend_nord_nordwest();
               break;


            }

    }



// Kommentar: Im Folgenden wird fr gemeinsame Punkte - auf diesen liegt die Eigenschaftsinformation vor -
// bei fehlender Eigenschaftsinformation das Farbenfeld auf -1 gesetzt. Es wird bewiesen, dass dieses
// Entfernen korrekt funktioniert.
void FarbenUpdate::entfernen_korrigieren(void)
    {
// Kommentar: Nur gemeinsame Punkte werden ausgewhlt. Die haben die Eigensschaftsinformation. Ferner muss
// die ep_ff[f][g] - Information so sein, dass ein neuer Punkt existiert. Bei ep_ff(HR[f][g]) = 2 hingegen wre
// diese Funktion berflssig: Das kann nmlich nicht sein. Diese Punkte liegen aber schon direkt
// nebeneinander.
    if(((f&1)==1)||((g&1)==1)||(ep_ff[f][g]!=2))
        {
        return;
        }

// Kommentar: Der zugeordnete Grobgitterpunkt wird bestimmt. Die Division geht wegen obiger Bedingung glatt auf.
    f_grob = f >> 1;
    g_grob = g >> 1;

#ifdef AUSGABEN_FARBENUPDATE
    cout << " 20286 rang " << rang << endl;
#endif

// Kommentar: Fr die Randpunkte werden die fr die nicht auf dem Kern liegenden Nebenpunkte entsprechende
// Funktionen aufgerufen.
    switch(255 - feingitter_snb[f][g]) // Ich denke, dass der dynamische Rand auf dem feinen Gitter liegt.
           {
           case 0:
               break;
           case 1:
// Kommentar: Hier wird die gleich angegebene Funktion aufgerufen, die jeweils dann, wenn dort eine nicht eigene Stelle ist,
// es ermglicht, einen echten Rand zu bilden. Die Zulssigkeit wird durch Informationen aus ep_snb_update getestet -
// vgl. diese Funktionen - wie entfernen_korrigieren_nord.
               entfernen_korrigieren_nord();
               break;
           case 2:
               entfernen_korrigieren_nordost();
               break;
           case 3:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               break;
           case 4:
               entfernen_korrigieren_ost();
               break;
           case 5:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               break;
           case 6:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               break;
           case 7:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               break;
           case 8:
               entfernen_korrigieren_suedost();
               break;
           case 9:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_suedost();
               break;
           case 10:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               break;
           case 11:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               break;
           case 12:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               break;
           case 13:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               break;
           case 14:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               break;
           case 15:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               break;
           case 16:
               entfernen_korrigieren_sued();
               break;
           case 17:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_sued();
               break;
           case 18:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_sued();
               break;
           case 19:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_sued();
               break;
           case 20:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               break;
           case 21:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               break;
           case 22:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               break;
           case 23:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               break;
           case 24:
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               break;
           case 25:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               break;
           case 26:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               break;
           case 27:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               break;
           case 28:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               break;
           case 29:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               break;
           case 30:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               break;
           case 31:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               break;
           case 32:
               entfernen_korrigieren_suedwest();
               break;
           case 33:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_suedwest();
               break;
           case 34:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedwest();
               break;
           case 35:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedwest();
               break;
           case 36:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedwest();
               break;
           case 37:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedwest();
               break;
           case 38:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedwest();
               break;
           case 39:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedwest();
               break;
           case 40:
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               break;
           case 41:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               break;
           case 42:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               break;
           case 43:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               break;
           case 44:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               break;
           case 45:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               break;
           case 46:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               break;
           case 47:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               break;
           case 48:
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               break;
           case 49:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               break;
           case 50:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               break;
           case 51:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               break;
           case 52:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               break;
           case 53:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               break;
           case 54:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               break;
           case 55:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               break;
           case 56:
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               break;
           case 57:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               break;
           case 58:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               break;
           case 59:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               break;
           case 60:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               break;
           case 61:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               break;
           case 62:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               break;
           case 63:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               break;
           case 64:
               entfernen_korrigieren_west();
               break;
           case 65:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_west();
               break;
           case 66:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_west();
               break;
           case 67:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_west();
               break;
           case 68:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_west();
               break;
           case 69:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_west();
               break;
           case 70:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_west();
               break;
           case 71:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_west();
               break;
           case 72:
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_west();
               break;
           case 73:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_west();
               break;
           case 74:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_west();
               break;
           case 75:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_west();
               break;
           case 76:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_west();
               break;
           case 77:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_west();
               break;
           case 78:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_west();
               break;
           case 79:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_west();
               break;
           case 80:
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               break;
           case 81:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               break;
           case 82:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               break;
           case 83:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               break;
           case 84:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               break;
           case 85:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               break;
           case 86:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               break;
           case 87:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               break;
           case 88:
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               break;
           case 89:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               break;
           case 90:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               break;
           case 91:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               break;
           case 92:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               break;
           case 93:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               break;
           case 94:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               break;
           case 95:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               break;
           case 96:
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 97:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 98:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 99:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 100:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 101:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 102:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 103:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 104:
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 105:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 106:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 107:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 108:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 109:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 110:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 111:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 112:
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 113:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 114:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 115:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 116:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 117:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 118:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 119:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 120:
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 121:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 122:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 123:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 124:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 125:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 126:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 127:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               break;
           case 128:
               entfernen_korrigieren_nordwest();
               break;
           case 129:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordwest();
               break;
           case 130:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_nordwest();
               break;
           case 131:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_nordwest();
               break;
           case 132:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_nordwest();
               break;
           case 133:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_nordwest();
               break;
           case 134:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_nordwest();
               break;
           case 135:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_nordwest();
               break;
           case 136:
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_nordwest();
               break;
           case 137:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_nordwest();
               break;
           case 138:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_nordwest();
               break;
           case 139:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_nordwest();
               break;
           case 140:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_nordwest();
               break;
           case 141:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_nordwest();
               break;
           case 142:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_nordwest();
               break;
           case 143:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_nordwest();
               break;
           case 144:
               entfernen_korrigieren_sued();
               entfernen_korrigieren_nordwest();
               break;
           case 145:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_nordwest();
               break;
           case 146:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_nordwest();
               break;
           case 147:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_nordwest();
               break;
           case 148:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_nordwest();
               break;
           case 149:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_nordwest();
               break;
           case 150:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_nordwest();
               break;
           case 151:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_nordwest();
               break;
           case 152:
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_nordwest();
               break;
           case 153:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_nordwest();
               break;
           case 154:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_nordwest();
               break;
           case 155:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_nordwest();
               break;
           case 156:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_nordwest();
               break;
           case 157:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_nordwest();
               break;
           case 158:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_nordwest();
               break;
           case 159:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_nordwest();
               break;
           case 160:
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 161:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 162:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 163:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 164:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 165:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 166:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 167:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 168:
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 169:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 170:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 171:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 172:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 173:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 174:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 175:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 176:
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 177:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 178:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 179:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 180:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 181:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 182:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 183:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 184:
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 185:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 186:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 187:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 188:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 189:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 190:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 191:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_nordwest();
               break;
           case 192:
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 193:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 194:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 195:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 196:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 197:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 198:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 199:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 200:
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 201:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 202:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 203:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 204:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 205:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 206:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 207:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 208:
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 209:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 210:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 211:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 212:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 213:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 214:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 215:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 216:
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 217:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 218:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 219:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 220:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 221:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 222:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 223:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 224:
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 225:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 226:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 227:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 228:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 229:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 230:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 231:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 232:
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 233:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 234:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 235:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 236:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 237:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 238:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 239:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 240:
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 241:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 242:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 243:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 244:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 245:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 246:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 247:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 248:
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 249:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 250:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 251:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 252:
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 253:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 254:
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           case 255:
               entfernen_korrigieren_nord();
               entfernen_korrigieren_nordost();
               entfernen_korrigieren_ost();
               entfernen_korrigieren_suedost();
               entfernen_korrigieren_sued();
               entfernen_korrigieren_suedwest();
               entfernen_korrigieren_west();
               entfernen_korrigieren_nordwest();
               break;
           }
    }


// NORD 1 NORD-OST 2 OST 4 SUED - OST 8 SUED 16 SUED - WEST 32 WEST 64 NORD - WEST 128

// Kommentar: Vergleiche den Kommentar zu 'entfernen_korrigieren_nord'. Man kann hier genauso argumentieren.
void FarbenUpdate::entfernen_korrigieren_sued(void)
    {
    if(f==n_minus1)
        {
        return;
        }
    if((((ep_snb_update->SNB[f_grob+1][g_grob])&199)==0)&&(ep_snb_update->feld_aktuell[f_grob+1][g_grob]==false))
    // N + NW + NO + W + O = 1 + 128 + 2 + 64 + 4 = 7 + 192 = 199
        {
        fgff[f+1][g] = -1;
        }
    }

// Kommentar: Vergleiche den Kommentar zu 'entfernen_korrigieren_nordost'. Man kann hier genauso argumentieren.
void FarbenUpdate::entfernen_korrigieren_suedost(void)
    {
    if((f==n_minus1)||(g==m_minus1))
        {
        return;
        }
    if((((ep_snb_update->SNB[f_grob+1][g_grob])&7)==0)&&(ep_snb_update->feld_aktuell[f_grob][g_grob]==false))
    // N + NO + O = 1 + 2 + 4 = 7
        {
        fgff[f+1][g+1] = -1;
        }
    }

// Kommentar: Vergleiche den Kommentar zu 'entfernen_korrigieren_nordost'. Man kann hier genauso argumentieren.
void FarbenUpdate::entfernen_korrigieren_suedwest(void)
    {
    if((f==n_minus1)||(g==0))
        {
        return;
        }
    if((((ep_snb_update->SNB[f_grob+1][g_grob-1])&7)==0)&&(ep_snb_update->feld_aktuell[f_grob+1][g_grob-1]==false))
    // N + NO + O = 1 + 2 + 4 = 7
        {
        fgff[f+1][g-1] = -1;
        }
    }

// Kommentar: Vergleiche den Kommentar zu 'entfernen_korrigieren_nord'. Man kann hier genauso argumentieren.
void FarbenUpdate::entfernen_korrigieren_west(void)
    {
    if(g==0)
        {
        return;
        }
    if((((ep_snb_update->SNB[f_grob][g_grob-1])&31)==0)&&(ep_snb_update->feld_aktuell[f_grob][g_grob-1]==false))
    // N + NO + O + SO + S = 1 + 2 + 4 + 8 + 16 = 31
        {
        fgff[f][g-1] = -1;
        }
    }

// Kommentar: Vergleiche den Kommentar zu 'entfernen_korrigieren_nordost'. Man kann hier genauso argumentieren.
void FarbenUpdate::entfernen_korrigieren_nordwest(void)
    {
    if((f==0)||(g==0))
        {
        return;
        }
    if((((ep_snb_update->SNB[f_grob][g_grob-1])&7)==0)&&(ep_snb_update->feld_aktuell[f_grob][g_grob-1]==false))
    // N + NO + O = 1 + 2 + 4 = 7
        {
        fgff[f-1][g-1] = -1;
        }
    }

// Kommentar: Wenn (f,g) - entsprechend der Voraussetzungen an den Aufruf der Funktion - ein gemeinsamer Punkt ist,
// und (f-1,g) nicht auf dem eigenen Kern liegt, dann kann der Punkt als nicht existent erkannt werden.
// Die if - Abfrage gewhrleistet das.
void FarbenUpdate::entfernen_korrigieren_nord(void)
// Kommentar: Ist der Punkt ein nrdlicher Punkt, dann gibt es den Punkt (f-1,g) nicht.
    {
    if(f==0)
        {
        return;
        }
// Kommentar: Die 6 Grobgitterpunkte, die um (f-1,g) drum herum liegen, zeigen keine Verfeinerung an.
// Daraus lt sich schlieen, dass an der Stelle (f-1,g) kein Punkt existieren kann:
// * * *
// * * *
// Diese 6 Sternchen zeigen die Punkte an, wo feld_aktuell = false ist. Und zwischen den beiden mittleren
// Sternchen liegt (f-1,g). Damit schirmen dieses 6 Punkte eine wirksame Verfeinerungsentscheidung fr diesen Punkt
// sicher ab.
// Wie man auf die 6 Sterne kommt? feld_aktuell(f_grob-1,g_grob] liegt oben in der Mitte. Und wegen
// O + SO + S + SW + W = 64 + 32 + 16 + 8 + 4 = 124 liegen die anderen Punkte in Richtung von O + SO + S + SW + W
// in Bezug auf diesen Punkt, die durch ep_snb_update->SNB[f_grob-1][g_grob])&124)==0 auf false getestet werden.
    if((((ep_snb_update->SNB[f_grob-1][g_grob])&124)==0)&&(ep_snb_update->feld_aktuell[f_grob-1][g_grob]==false))
    // O + SO + S + SW + W = 64 + 32 + 16 + 8 + 4 = 124
        {
        fgff[f-1][g] = -1;
        }
    }

// Kommentar: Wenn (f,g) - entsprechend der Voraussetzungen an den Aufruf der Funktion - ein gemeinsamer Punkt ist,
// und (f-1,g+1) nicht auf dem eigenen Kern liegt, dann kann der Punkt als nicht existent erkannt werden.
// Die if - Abfrage gewhrleistet das.
void FarbenUpdate::entfernen_korrigieren_nordost(void)
    {
// Kommentar: Liegt (f-1,g+1) nicht auf dem Gitter, beendet das hier den Aufruf der Funktion.
    if((f==0)||(g==m_minus1))
        {
        return;
        }

// Kommentar: Die 4 Grobgitterpunkte, die um (f-1,g) drum herum liegen, zeigen keine Verfeinerung an.
// Daraus lt sich schlieen, dass an der Stelle (f-1,g) kein Punkt existieren kann:
// * *
// * *
// Diese 4 Sternchen zeigen die Punkte an, wo feld_aktuell = false ist. Und zwischen den beiden mittleren
// Sternchen liegt (f-1,g). Damit schirmen dieses 4 Punkte eine wirksame Verfeinerungsentscheidung fr diesen Punkt
// sicher ab.
    if((((ep_snb_update->SNB[f_grob-1][g_grob])&28)==0)&&(ep_snb_update->feld_aktuell[f_grob-1][g_grob]==false))
    // O + SO + S = 16 + 8 + 4 = 28
        {
        fgff[f-1][g+1] = -1;
        }
    }

// Kommentar: Vergleiche den Kommentar zu 'entfernen_korrigieren_nord'. Man kann hier genauso argumentieren.
void FarbenUpdate::entfernen_korrigieren_ost(void)
    {
    if(g==m_minus1)
        {
        return;
        }

    if((((ep_snb_update->SNB[f_grob][g_grob+1])&241)==0)&&(ep_snb_update->feld_aktuell[f_grob][g_grob+1]==false))
    // N + NW + W + SW + S = 128 + 64 + 32 + 16 + 1 = 241
        {
        fgff[f][g+1] = -1;
        }
    }

///////////////////////////////////////////////////////////////////
/////////////////  Neuer Code fr innere Ecken   //////////////////
/////////////////  22.10.2014                    //////////////////
///////////////////////////////////////////////////////////////////

// Kommentar: Bei inneren Ecken - vgl. die Dissertation - hat man z.B. so eine Konstellation:
// . *
// * *
// Wo der Punkt liegt, gibt es keinen Punkt. Der Kern, der den dazu diagonalen Punkt hat (ein Sternchen),
// bermittelt jeweils den einen Sternwert zum anderen. Auf diese Weise wird diese Information gesichert.
// Und das tritt exakt nur bei inneren Ecken auf.
// Bemerkung: Hier hat man die Flle
// . *  sowie  * .
// * *         * *
// wobei jeweils vom Entscheidungspunkt (diagonal zu '.') fr den nrdlichen Punkt die stliche oder westlichen
// Information bermittelt wird. Durch Aufruf von ...nord ...west werden aber bei
// . *
// * *
// beide nicht diagonalen Informationen bermittelt.
void FarbenUpdate::stellvertretung_fg_gleich_gg_nord(void)
    {
#ifdef AKTUELLE_KOMMENTARE_NEU
    if((f==12)&&(g==20)&&(rang==0))
        {
        cout << " ERTZ epff-1,0 " << ep_ff[f-1][g] << " fgf-1,0 " << feingitter_feld[f-1][g] << " t " << t << endl;
        }
#endif
#ifdef AAHH
    if((rang==2)&&(f==20)&&(g==8))
        {
        cout << " AAHH stellvertretung nord rang 1 f 20 g 8 t " << t << " epff(f-1)(g) " << ep_ff[f-1][g];
        cout << " fgff[f-1][g] " << fgff[f-1][g] << " ep_ff[f-1][g-1] " << ep_ff[f-1][g-1];
        cout << " fgff[f-1][g-1] " << fgff[f-1][g-1] << " ep_ff[f][g-1] " << ep_ff[f][g-1];
        cout << " fgff[f][g-1] " << fgff[f][g-1] << endl;
        }
#endif

// Kommentar: Nur auf gemeinsamen -  oder alpha - Punkten knnen diese Entscheidungen getroffen werden.
    if(((f&1)==1)||((g&1)==1))
        {
        return;
        }
    // 13.3.15

// Kommentar: Zuerst wird die ep_ff - Information des nrdlichen Punktes gelesen.
    switch(ep_ff[f-1][g])
        {
// Kommentar: Im Norden ist ein neuer Punkt entstanden.
        case 2:
// Kommentar: Die Farbenfeld - Information, die dort eingetragen wurde, wird jetzt gelesen.
            ggff_hilfe = ggff[(f-1)>>1][g>>1];
#ifdef AUSGABE_LMNO
            if((f==12)&&(g==20)&&(rang==0))
                {
                cout << " ggff_hilfe " << ggff_hilfe << " epff0,-1 " << ep_ff[f][g-1] << " ep_ff-1-1 " << ep_ff[f-1][g-1] << " fgf-1-1 " << feingitter_feld[f-1][g-1] << " fgff0-1 " << fgff[f][g-1] << endl;
                }
#endif
// Kommentar: Wenn dort ein eigener Punkt liegt, braucht nichts kommuniziert zu werden.
            if(ggff_hilfe!=rang)
                {
#ifdef AUSGABE_LMNO
                if((f==12)&&(g==20)&&(rang==0))
                    {
                    cout << " LMNO Nord 2 " << endl;
                    }
#endif
// Kommentar: Jetzt soll die Information nach Westen, nmlich (f,g-1), verschickt werden. Also wird g > 0 getestet.
                if(g>0)
                    {

// Kommentar: Man kann Einschrnkungen machen, um die Kommunikation sehr zu begrenzen.
// Diese Bedingung gilt, wenn kein Punkt an der Stelle von '.' im Bild
// . *
// * *
// existiert (ep_ff[f-1][g-1]==0)&&(fgff[f-1][g-1]<0) oder nicht mehr existiert (ep_ff[f-1][g-1]==1).
                    if(((ep_ff[f-1][g-1]==0)&&(fgff[f-1][g-1]<0))||(ep_ff[f-1][g-1]==1))
                        {
// Kommentar: Jetzt wird angeschaut, was fr eine Information verschickt werden kann, d.h. was an der Stelle (f,g-1)
// vorliegt.
                        switch(ep_ff[f][g-1]) // Bei ep_ff == 0 ist fgff[f][g-1]>=0; 20.10.14
                            {
// Kommentar: Erster Fall: Dort ist im Bezug auf Hinzufgen und Entfernen nichts passiert.
                            case 0:
// Kommentar: Dann ist nur dann was zu tun, wenn dort schon ein Punkt existiert hat.
                                if(fgff[f][g-1]>=0)
                                    {
#ifdef AUSGABE_LMNO
                                    if((f==12)&&(g==20)&&(rang==0))
                                        {
                                        cout << " LMNO Nord 3 " << endl;
                                        }
#endif
// Kommentar: Bestimme dann dessen Farbwert.
                                        farbe_short_int = fgff[f][g-1];
                                        farbe_char = (char) farbe_short_int;
                                        //kommunikation4er->einfuegen_fu(f-1, g, farbe_char, fgff[f_schlange][g_schlange]);
#ifdef AUSGABE_LMNO
                                        cout << " LMNO Nord 4 f " << f << " g-1 " << g-1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                        cout << " f-1/2 " << ((f-1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
// Kommentar: Und verschicke die Information (f,g-1,farbe(f,g-1)) an den Kern des nrdlichen Punktes.
                                        kommunikation4er->einfuegen_fu(f, g-1, farbe_char, ggff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                        strcpy(text, "24685");
                                        test_einfuegen(f, g-1, farbe_char, ggff_hilfe, text);
#endif
                                    }
                                break;
                            case 1: // 19.9.2015
                                kommunikation4er->einfuegen_fu(f, g-1, farbe_char_m1, ggff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "24689");
                                test_einfuegen(f, g-1, farbe_char_m1, ggff_hilfe, text);
#endif
                                break;
                            case 2:
// Kommentar: Der Punkt links ist neu. Dann besorgt man sich den Farbwert ber das Grobgitterfarbenfeld, und
// verschickt diese Farbfeldinformation mit (f,g-1) an den Kern des nrdlichen Punktes.
                                farbe_short_int = ggff[f>>1][(g-1)>>1];
                                farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                cout << " LMNO Nord 4 f " << f << " g-1 " << g-1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                cout << " f-1/2 " << ((f-1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                kommunikation4er->einfuegen_fu(f, g-1, farbe_char, ggff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "24700");
                                test_einfuegen(f, g-1, farbe_char, ggff_hilfe, text);
#endif
                                break;
                            }
                        }
                    }
// Kommentar: Jetzt soll die Information nach Osten, nmlich (f,g+1), verschickt werden. Also wird g < m_minus1 getestet.
// Alles weitere erfolgt wie fr den Fall nach Westen - siehe den Fall 'if(g>0) usw.'.
                if(g<m_minus1)
                    {
                    if(((ep_ff[f-1][g+1]==0)&&(fgff[f-1][g+1]<0))||(ep_ff[f-1][g+1]==1))
                        {
                        switch(ep_ff[f][g+1])
                            {
                            case 0:
                                if(fgff[f][g+1]>=0) // Ein Punkt ist da, 23.10.14 => Innere Ecke
                                    {
                                    farbe_short_int = fgff[f][g+1];
                                    farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                    cout << " LMNO Nord 4 f " << f << " g+1 " << g+1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                    cout << " f-1/2 " << ((f-1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                    kommunikation4er->einfuegen_fu(f, g+1, farbe_char, ggff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                    strcpy(text, "24725");
                                    test_einfuegen(f, g+1, farbe_char, ggff_hilfe, text);
#endif
                                    }
                                break;
                            case 1: // 19.9.2015
                                kommunikation4er->einfuegen_fu(f, g+1, farbe_char_m1, ggff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "24729");
                                test_einfuegen(f, g+1, farbe_char_m1, ggff_hilfe, text);
#endif
                                break;
                            case 2:
                                farbe_short_int = ggff[f>>1][(g+1)>>1]; // nderung 2.11.15
                                farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                cout << " LMNO Nord 4 f " << f << " g+1 " << g+1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                cout << " f-1/2 " << ((f-1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                kommunikation4er->einfuegen_fu(f, g+1, farbe_char, ggff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "24739");
                                test_einfuegen(f, g+1, farbe_char, ggff_hilfe, text);
#endif
                                break;
                            }
                        }
                    }
                }
            break;
// Kommentar: Dieser Fall: Der Punkt oben existierte schon, oder er existierte weder vorher noch nachher.
        case 0:

// Kommentar: Dann kann man das Farbenfeld oben ablesen.
            fgff_hilfe = fgff[f-1][g];
#ifdef AUSGABE_LMNO
            if((f==12)&&(g==20)&&(rang==0))
                {
                cout << " ggff_hilfe " << ggff_hilfe << " epff0,-1 " << ep_ff[f][g-1] << " ep_ff-1-1 " << ep_ff[f-1][g-1] << " fgf-1-1 " << feingitter_feld[f-1][g-1] << " fgff0-1 " << fgff[f][g-1] << endl;
                }
#endif

// Kommentar: Diese if - Abfrage checkt, dass der Punkt schon existierte. (Weder vorher noch nachher nicht kann man ausschlieen.)
            if(fgff_hilfe>=0)
                {
#ifdef AUSGABE_LMNO
                if((f==12)&&(g==20)&&(rang==0))
                    {
                    cout << " LMNO Nord 2 " << endl;
                    }
#endif
// Kommentar: Fast alles was jetzt kommt, kam schon oben im Fall case = 2 der Anweisung switch(ep_ff[f-1][g]).
// Denn es geht nur noch darum, die Information nach Westen oder Osten nach Norden zu bertragen.
// Und das geht fast genauso wie oben nach der Anweisung 'if(g>0) usw.'.
// Der einzige Unterschied besteht darin, dass bei dem Befehl 'kommunikation4er->einfuegen_fu' der letzte Parameter
// hier jeweils fgff_hilfe ist, und dort ggff_hilfe. Das leitet sich aber von dem in beiden Fllen ermittelten
// Farbwertes an der Stelle (f-1,g) ab.
                if(g>0)
                    {
                    if(((ep_ff[f-1][g-1]==0)&&(fgff[f-1][g-1]<0))||(ep_ff[f-1][g-1]==1))
                        {
                        switch(ep_ff[f][g-1]) // Bei ep_ff == 0 ist fgff[f][g-1]>=0; 20.10.14
                            {
                            case 0:
                                if(fgff[f][g-1]>=0)
                                   {
#ifdef AUSGABE_LMNO
                                    if((f==12)&&(g==20)&&(rang==0))
                                        {
                                        cout << " LMNO Nord 3 " << endl;
                                        }
#endif
                                    farbe_short_int = fgff[f][g-1];
                                    farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                    cout << " LMNO Nord 4 f " << f << " g-1 " << g-1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                    cout << " f-1/2 " << ((f-1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                    kommunikation4er->einfuegen_fu(f, g-1, farbe_char, fgff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "24802");
                                test_einfuegen(f, g-1, farbe_char, fgff_hilfe, text);
#endif
                                   }
                                break;
                            case 1: // 19.9.2015
                                kommunikation4er->einfuegen_fu(f, g-1, farbe_char_m1, fgff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "24806");
                                test_einfuegen(f, g-1, farbe_char, fgff_hilfe, text);
#endif
                                break;
                            case 2:
                                farbe_short_int = ggff[f>>1][(g-1)>>1];
                                farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                cout << " LMNO Nord 4 f " << f << " g-1 " << g-1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                cout << " f-1/2 " << ((f-1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                kommunikation4er->einfuegen_fu(f, g-1, farbe_char, fgff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "24815");
                                test_einfuegen(f, g-1, farbe_char, fgff_hilfe, text);
#endif
                                break;
                            }
                        }
                    }
                if(g<m_minus1)
                    {
                    if(((ep_ff[f-1][g+1]==0)&&(fgff[f-1][g+1]<0))||(ep_ff[f-1][g+1]==1))
                        {
                        switch(ep_ff[f][g+1])
                            {
                            case 0:
                                if(fgff[f][g+1]>=0)
                                    {
                                    farbe_short_int = fgff[f][g+1];
                                    farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                    cout << " LMNO Nord 4 f " << f << " g+1 " << g+1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                    cout << " f-1/2 " << ((f-1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                    kommunikation4er->einfuegen_fu(f, g+1, farbe_char, fgff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "24838");
                                test_einfuegen(f, g+1, farbe_char, fgff_hilfe, text);
#endif
                                    }
                                break;
                            case 1: // 19.9.2015
                                kommunikation4er->einfuegen_fu(f, g+1, farbe_char_m1, fgff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "24841");
                                test_einfuegen(f, g+1, farbe_char_m1, fgff_hilfe, text);
#endif
                                break;
                            case 2:
                                farbe_short_int = ggff[f>>1][(g+1)>>1];
//                                farbe_short_int = ggff[f>>1][(g+2)>>1]; nderung 5.2.16
                                farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                cout << " LMNO Nord 4 f " << f << " g+1 " << g+1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                cout << " f-1/2 " << ((f-1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                kommunikation4er->einfuegen_fu(f, g+1, farbe_char, fgff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "24852");
                                test_einfuegen(f, g+1, farbe_char, fgff_hilfe, text);
#endif
                                break;
                            }
                        }
                    }
                }
            break;
        }
    }

// Kommentar: Analog zu 'stellvertretung_fg_gleich_gg_sued'.
void FarbenUpdate::stellvertretung_fg_gleich_gg_sued(void)
    {
    if(((f&1)==1)||((g&1)==1))
        {
        return;
        }
    // 13.3.15
#ifdef AUSGABE_LMNO
    if((f==12)&&(g==20)&&(rang==0))
        {
        cout << " ERTZ epff-1,0 " << ep_ff[f-1][g] << " fgf-1,0 " << feingitter_feld[f-1][g] << " t " << t << endl;
        }
#endif

#ifdef AAHH
    if((rang==3)&&(f==19)&&(g==8))
        {
        cout << " AAHH stellvertretung sued rang 1 f 19 g 8 t " << t << " epff(f+1)(g) " << ep_ff[f+1][g];
        cout << " ep_ff[f+1][g-1] " << ep_ff[f+1][g-1];
        cout << " fgff[f+1][g-1] " << fgff[f+1][g-1] << " ep_ff[f][g-1] " << ep_ff[f][g-1];
        cout << " fgff[f][g-1] " << fgff[f][g-1] << endl;
        }
#endif

    switch(ep_ff[f+1][g])
        {
        case 2:
            ggff_hilfe = ggff[(f+1)>>1][g>>1];
#ifdef AUSGABE_LMNO
            if((f==12)&&(g==20)&&(rang==0))
                {
                cout << " ggff_hilfe " << ggff_hilfe << " epff0,-1 " << ep_ff[f][g-1] << " ep_ff+1-1 " << ep_ff[f+1][g-1] << " fgf+1-1 " << feingitter_feld[f+1][g-1] << " fgff0-1 " << fgff[f][g-1] << endl;
                }
#endif
            if(ggff_hilfe!=rang)
                {
#ifdef AUSGABE_LMNO
                if((f==12)&&(g==20)&&(rang==0))
                    {
                    cout << " LMNO Sued 2 " << endl;
                    }
#endif
                if(g>0)
                    {
                    if(((ep_ff[f+1][g-1]==0)&&(fgff[f+1][g-1]<0))||(ep_ff[f+1][g-1]==1))
                        {
                        switch(ep_ff[f][g-1]) // Bei ep_ff == 0 ist fgff[f][g-1]>=0; 20.10.14
                            {
                            case 0:
                                if(fgff[f][g-1]>=0)
                                    {
#ifdef AUSGABE_LMNO
                                    if((f==12)&&(g==20)&&(rang==0))
                                        {
                                        cout << " LMNO Sued 3 " << endl;
                                        }
#endif
                                    farbe_short_int = fgff[f][g-1];
                                    farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                    cout << " LMNO Sued 4 f " << f << " g-1 " << g-1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                    cout << " f+1/2 " << ((f+1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                    kommunikation4er->einfuegen_fu(f, g-1, farbe_char, ggff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                    strcpy(text, "25121");
                                    test_einfuegen(f, g-1, farbe_char, ggff_hilfe, text);
#endif
                                    }
                                break;
                            case 1: // 19.9.2015
                                kommunikation4er->einfuegen_fu(f, g-1, farbe_char_m1, ggff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25125");
                                test_einfuegen(f, g-1, farbe_char, ggff_hilfe, text);
#endif
                                break;
                            case 2:
                                farbe_short_int = ggff[f>>1][(g-1)>>1];
                                farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                cout << " LMNO Sued 4 f " << f << " g-1 " << g-1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                cout << " f+1/2 " << ((f+1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                kommunikation4er->einfuegen_fu(f, g-1, farbe_char, ggff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25134");
                                test_einfuegen(f, g-1, farbe_char, ggff_hilfe, text);
#endif
                                break;
                            }
                        }
                    }
                if(g<m_minus1)
                    {
                    if(((ep_ff[f+1][g+1]==0)&&(fgff[f+1][g+1]<0))||(ep_ff[f+1][g+1]==1))
                        {
                        switch(ep_ff[f][g+1])
                            {
                            case 0:
                                if(fgff[f][g+1]>=0) // Ein Punkt ist da, 23.10.14 => Innere Ecke
                                    {
                                    farbe_short_int = fgff[f][g+1];
                                    farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                    cout << " LMNO Sued 4 f " << f << " g+1 " << g+1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                    cout << " f+1/2 " << ((f+1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                    kommunikation4er->einfuegen_fu(f, g+1, farbe_char, ggff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25157");
                                test_einfuegen(f, g+1, farbe_char, ggff_hilfe, text);
#endif
                                    }
                                break;
                            case 1: // 19.9.2015
                                kommunikation4er->einfuegen_fu(f, g+1, farbe_char_m1, ggff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25160");
                                test_einfuegen(f, g+1, farbe_char, ggff_hilfe, text);
#endif
                                break;
                            case 2:
                              farbe_short_int = ggff[f>>1][(g+1)>>1]; // nderung 5.2.16
//                                farbe_short_int = ggff[f>>1][(g+2)>>1];
                                farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                cout << " LMNO Sued 4 f " << f << " g+1 " << g+1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                cout << " f+1/2 " << ((f+1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                kommunikation4er->einfuegen_fu(f, g+1, farbe_char, ggff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25171");
                                test_einfuegen(f, g+1, farbe_char, ggff_hilfe, text);
#endif
                                break;
                            }
                        }
                    }
                }
            break;
        case 0:
            fgff_hilfe = fgff[f+1][g];
#ifdef AUSGABE_LMNO
            if((f==12)&&(g==20)&&(rang==0))
                {
                cout << " ggff_hilfe " << ggff_hilfe << " epff0,-1 " << ep_ff[f][g-1] << " ep_ff+1-1 " << ep_ff[f+1][g-1] << " fgf+1-1 " << feingitter_feld[f+1][g-1] << " fgff0-1 " << fgff[f][g-1] << endl;
                }
#endif
            if(fgff_hilfe>=0)
                {
#ifdef AUSGABE_LMNO
                if((f==12)&&(g==20)&&(rang==0))
                    {
                    cout << " LMNO Sued 2 " << endl;
                    }
#endif
                if(g>0)
                    {
                    if(((ep_ff[f+1][g-1]==0)&&(fgff[f+1][g-1]<0))||(ep_ff[f+1][g-1]==1))
                        {
                        switch(ep_ff[f][g-1]) // Bei ep_ff == 0 ist fgff[f][g-1]>=0; 20.10.14
                            {
                            case 0:
                                if(fgff[f][g-1]>=0)
                                   {
#ifdef AUSGABE_LMNO
                                    if((f==12)&&(g==20)&&(rang==0))
                                        {
                                        cout << " LMNO Sued 3 " << endl;
                                        }
#endif
                                    farbe_short_int = fgff[f][g-1];
                                    farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                    cout << " LMNO Sued 4 f " << f << " g-1 " << g-1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                    cout << " f+1/2 " << ((f+1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                    kommunikation4er->einfuegen_fu(f, g-1, farbe_char, fgff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25223");
                                test_einfuegen(f, g-1, farbe_char, fgff_hilfe, text);
#endif
                                   }
                                break;
                            case 1: // 19.9.2015
                                kommunikation4er->einfuegen_fu(f, g-1, farbe_char_m1, fgff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25227");
                                test_einfuegen(f, g-1, farbe_char_m1, fgff_hilfe, text);
#endif
                                break;
                            case 2:
                                farbe_short_int = ggff[f>>1][(g-1)>>1];
                                farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                cout << " LMNO Sued 4 f " << f << " g-1 " << g-1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                cout << " f+1/2 " << ((f-1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                kommunikation4er->einfuegen_fu(f, g-1, farbe_char, fgff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25236");
                                test_einfuegen(f, g-1, farbe_char, fgff_hilfe, text);
#endif
                                break;
                            }
                        }
                    }
                if(g<m_minus1)
                    {
                    if(((ep_ff[f+1][g+1]==0)&&(fgff[f+1][g+1]<0))||(ep_ff[f+1][g+1]==1))
                        {
                        switch(ep_ff[f][g+1])
                            {
                            case 0:
                                if(fgff[f][g+1]>=0)
                                    {
                                    farbe_short_int = fgff[f][g+1];
                                    farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                    cout << " LMNO Sued 4 f " << f << " g+1 " << g+1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                    cout << " f+1/2 " << ((f+1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                    kommunikation4er->einfuegen_fu(f, g+1, farbe_char, fgff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25259");
                                test_einfuegen(f, g+1, farbe_char, fgff_hilfe, text);
#endif
                                    }
                                break;
                            case 1: // 19.9.2015
                                kommunikation4er->einfuegen_fu(f, g+1, farbe_char_m1, fgff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25263");
                                test_einfuegen(f, g+1, farbe_char_m1, fgff_hilfe, text);
#endif
                                break;
                            case 2:
                                farbe_short_int = ggff[f>>1][(g+1)>>1];
//                                farbe_short_int = ggff[f>>1][(g+2)>>1]; nderung 5.2.16
                                farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                cout << " LMNO Sued 4 f " << f << " g+1 " << g+1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                cout << " f+1/2 " << ((f+1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                kommunikation4er->einfuegen_fu(f, g+1, farbe_char, fgff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25273");
                                test_einfuegen(f, g+1, farbe_char, fgff_hilfe, text);
#endif
                                break;
                            }
                        }
                    }
                }
            break;
        }
    }

// Kommentar: Analog zu 'stellvertretung_fg_gleich_gg_sued'.
void FarbenUpdate::stellvertretung_fg_gleich_gg_west(void)
    {
    if(((f&1)==1)||((g&1)==1))
        {
        return;
        }
    // 13.3.15
#ifdef AAHH
    if((rang==2)&&(f==20)&&(g==8))
        {
        cout << " AAHH stellvertretung rang 1 f 20 g 8 t " << t << " epff(f)(g-1) " << ep_ff[f][g-1];
        cout << " fgff[f][g-1] " << fgff[f][g-1] << " ep_ff[f-1][g-1] " << ep_ff[f-1][g-1];
        cout << " fgff[f-1][g-1] " << fgff[f-1][g-1] << " ep_ff[f-1][g] " << ep_ff[f-1][g];
        cout << " fgff[f-1][g] " << fgff[f-1][g] << endl;
        }
#endif
#ifdef AUSGABE_LMNO
    if((f==12)&&(g==20)&&(rang==0))
        {
        cout << " ERTZ epff-1,0 " << ep_ff[f-1][g] << " fgf-1,0 " << feingitter_feld[f-1][g] << " t " << t << endl;
        }
#endif

    switch(ep_ff[f][g-1])
        {
        case 2:
            ggff_hilfe = ggff[f>>1][(g-1)>>1];
#ifdef AUSGABE_LMNO
            if((f==12)&&(g==20)&&(rang==0))
                {
                cout << " ggff_hilfe " << ggff_hilfe << " epff0,-1 " << ep_ff[f][g-1] << " ep_ff+1-1 " << ep_ff[f+1][g-1] << " fgf+1-1 " << feingitter_feld[f+1][g-1] << " fgff0-1 " << fgff[f][g-1] << endl;
                }
#endif
            if(ggff_hilfe!=rang)
                {
#ifdef AUSGABE_LMNO
                if((f==12)&&(g==20)&&(rang==0))
                    {
                    cout << " LMNO West 2 " << endl;
                    }
#endif
                if(f>0)
                    {
                    if(((ep_ff[f-1][g-1]==0)&&(fgff[f-1][g-1]<0))||(ep_ff[f-1][g-1]==1))
                        {
                        switch(ep_ff[f-1][g]) // Bei ep_ff == 0 ist fgff[f][g-1]>=0; 20.10.14
                            {
                            case 0:
                                if(fgff[f-1][g]>=0)
                                    {
#ifdef AUSGABE_LMNO
                                    if((f==12)&&(g==20)&&(rang==0))
                                        {
                                        cout << " LMNO West 3 " << endl;
                                        }
#endif
                                    farbe_short_int = fgff[f-1][g];
                                    farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                    cout << " LMNO West 4 f-1 " << f-1 << " g " << g << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                    cout << " f/2 " << (f >> 1) << " (g-1)/2 " << ((g-1)>>1) << endl;
#endif
                                    kommunikation4er->einfuegen_fu(f-1, g, farbe_char, ggff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                    strcpy(text, "25359");
                                    test_einfuegen(f-1, g, farbe_char, ggff_hilfe, text);
#endif
                                    }
                                break;
                            case 1: // 19.9.2015
                                kommunikation4er->einfuegen_fu(f-1, g, farbe_char_m1, ggff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25363");
                                test_einfuegen(f-1, g, farbe_char_m1, ggff_hilfe, text);
#endif
                                break;
                            case 2:
                                farbe_short_int = ggff[(f-1)>>1][g>>1];
                                farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                cout << " LMNO West 4 f-1 " << f-1 << " g " << g << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                cout << " f/2 " << (f >> 1) << " (g-1)/2 " << ((g-1)>>1) << endl;
#endif
                                kommunikation4er->einfuegen_fu(f-1, g, farbe_char, ggff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25372");
                                test_einfuegen(f-1, g, farbe_char, ggff_hilfe, text);
#endif
                                break;
                            }
                        }
                    }
                if(f<n_minus1)
                    {
                    if(((ep_ff[f+1][g-1]==0)&&(fgff[f+1][g-1]<0))||(ep_ff[f+1][g-1]==1))
                        {
                        switch(ep_ff[f+1][g])
                            {
                            case 0:
                                if(fgff[f+1][g]>=0) // Ein Punkt ist da, 23.10.14 => Innere Ecke
                                    {
                                    farbe_short_int = fgff[f+1][g];
                                    farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                    cout << " LMNO West 4 f " << f << " g+1 " << g+1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                    cout << " f+1/2 " << ((f+1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                    kommunikation4er->einfuegen_fu(f+1, g, farbe_char, ggff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                    strcpy(text, "25395");
                                    test_einfuegen(f+1, g, farbe_char, ggff_hilfe, text);
#endif
                                    }
                                break;
                            case 1: // 19.9.2015
                                kommunikation4er->einfuegen_fu(f+1, g, farbe_char_m1, ggff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25399");
                                test_einfuegen(f+1, g, farbe_char, ggff_hilfe, text);
#endif
                                break;
                            case 2:
//                                farbe_short_int = ggff[(f+2)>>1][g>>1];
                                farbe_short_int = ggff[(f+1)>>1][g>>1]; // nderung 5.2.16
                                farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                cout << " LMNO West 4 f " << f << " g+1 " << g+1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                cout << " f+1/2 " << ((f+1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                kommunikation4er->einfuegen_fu(f+1, g, farbe_char, ggff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25409");
                                test_einfuegen(f+1, g, farbe_char, ggff_hilfe, text);
#endif
                                break;
                            }
                        }
                    }
                }
            break;
        case 0:
            fgff_hilfe = fgff[f][g-1];
#ifdef AUSGABE_LMNO
            if((f==12)&&(g==20)&&(rang==0))
                {
                cout << " ggff_hilfe " << ggff_hilfe << " epff0,-1 " << ep_ff[f][g-1] << " ep_ff+1-1 " << ep_ff[f+1][g-1] << " fgf+1-1 " << feingitter_feld[f+1][g-1] << " fgff0-1 " << fgff[f][g-1] << endl;
                }
#endif
            if(fgff_hilfe>=0)
                {
#ifdef AUSGABE_LMNO
                if((f==12)&&(g==20)&&(rang==0))
                    {
                    cout << " LMNO West 2 " << endl;
                    }
#endif
                if(f>0)
                    {
                    if(((ep_ff[f-1][g-1]==0)&&(fgff[f-1][g-1]<0))||(ep_ff[f-1][g-1]==1))
                        {
                        switch(ep_ff[f-1][g]) // Bei ep_ff == 0 ist fgff[f][g-1]>=0; 20.10.14
                            {
                            case 0:
                                if(fgff[f-1][g]>=0)
                                   {
#ifdef AUSGABE_LMNO
                                    if((f==12)&&(g==20)&&(rang==0))
                                        {
                                        cout << " LMNO West 3 " << endl;
                                        }
#endif
                                    farbe_short_int = fgff[f-1][g];
                                    farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                    cout << " LMNO Sued 4 f " << f << " g-1 " << g-1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                    cout << " f+1/2 " << ((f+1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                    kommunikation4er->einfuegen_fu(f-1, g, farbe_char, fgff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                    strcpy(text, "25461");
                                    test_einfuegen(f-1, g, farbe_char, fgff_hilfe, text);
#endif
                                   }
                                break;
                            case 1: // 19.9.2015
                                kommunikation4er->einfuegen_fu(f-1, g, farbe_char_m1, fgff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25465");
                                test_einfuegen(f-1, g, farbe_char_m1, fgff_hilfe, text);
#endif
                                break;
                            case 2:
                                farbe_short_int = ggff[(f-1)>>1][g>>1];
                                farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                cout << " LMNO Sued 4 f " << f << " g-1 " << g-1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                cout << " f+1/2 " << ((f-1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                kommunikation4er->einfuegen_fu(f-1, g, farbe_char, fgff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25474");
                                test_einfuegen(f-1, g, farbe_char, fgff_hilfe, text);
#endif
                                break;
                            }
                        }
                    }
                if(f<n_minus1)
                    {
                    if(((ep_ff[f+1][g-1]==0)&&(fgff[f+1][g-1]<0))||(ep_ff[f+1][g-1]==1))
                        {
                        switch(ep_ff[f+1][g])
                            {
                            case 0:
                                if(fgff[f+1][g]>=0)
                                    {
                                    farbe_short_int = fgff[f+1][g];
                                    farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                    cout << " LMNO West 4 f " << f << " g+1 " << g+1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                    cout << " f+1/2 " << ((f+1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                    kommunikation4er->einfuegen_fu(f+1, g, farbe_char, fgff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                    strcpy(text, "25497");
                                    test_einfuegen(f+1, g, farbe_char, fgff_hilfe, text);
#endif
                                    }
                                break;
                            case 1: // 19.9.2015
                                kommunikation4er->einfuegen_fu(f+1, g, farbe_char_m1, fgff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25501");
                                test_einfuegen(f+1, g, farbe_char_m1, fgff_hilfe, text);
#endif
                                break;
                            case 2:
//                                farbe_short_int = ggff[(f+2)>>1][g>>1];
                                farbe_short_int = ggff[(f+1)>>1][g>>1]; // nderung 5.2.16
                                farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                cout << " LMNO Sued 4 f " << f << " g+1 " << g+1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                cout << " f+1/2 " << ((f+1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                kommunikation4er->einfuegen_fu(f+1, g, farbe_char, fgff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25511");
                                test_einfuegen(f+1, g, farbe_char, fgff_hilfe, text);
#endif
                                break;
                            }
                        }
                    }
                }
            break;
        }
    }

// Kommentar: Analog zu 'stellvertretung_fg_gleich_gg_sued'.
void FarbenUpdate::stellvertretung_fg_gleich_gg_ost(void)
    {
#ifdef AUSGABE_LMNO
    if((f==12)&&(g==20)&&(rang==0))
        {
        cout << " ERTZ epff-1,0 " << ep_ff[f-1][g] << " fgf-1,0 " << feingitter_feld[f-1][g] << " t " << t << endl;
        }
#endif
    if(((f&1)==1)||((g&1)==1))
        {
        return;
        }
    // 13.3.15
    switch(ep_ff[f][g+1])
        {
        case 2:
            ggff_hilfe = ggff[f>>1][(g+1)>>1];
#ifdef AUSGABE_LMNO
            if((f==12)&&(g==20)&&(rang==0))
                {
                cout << " ggff_hilfe " << ggff_hilfe << " epff0,-1 " << ep_ff[f][g-1] << " ep_ff+1-1 " << ep_ff[f+1][g-1] << " fgf+1-1 " << feingitter_feld[f+1][g-1] << " fgff0-1 " << fgff[f][g-1] << endl;
                }
#endif
            if(ggff_hilfe!=rang)
                {
#ifdef AUSGABE_LMNO
                if((f==12)&&(g==20)&&(rang==0))
                    {
                    cout << " LMNO Ost 2 " << endl;
                    }
#endif
                if(f>0)
                    {
                    if(((ep_ff[f-1][g+1]==0)&&(fgff[f-1][g+1]<0))||(ep_ff[f-1][g+1]==1))
                        {
                        switch(ep_ff[f-1][g]) // Bei ep_ff == 0 ist fgff[f][g-1]>=0; 20.10.14
                            {
                            case 0:
                                if(fgff[f-1][g]>=0)
                                    {
#ifdef AUSGABE_LMNO
                                    if((f==12)&&(g==20)&&(rang==0))
                                        {
                                        cout << " LMNO Ost 3 " << endl;
                                        }
#endif
                                    farbe_short_int = fgff[f-1][g];
                                    farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                    cout << " LMNO West 4 f-1 " << f-1 << " g " << g << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                    cout << " f/2 " << (f >> 1) << " (g-1)/2 " << ((g-1)>>1) << endl;
#endif
                                    kommunikation4er->einfuegen_fu(f-1, g, farbe_char, ggff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                    strcpy(text, "25587");
                                    test_einfuegen(f-1, g, farbe_char, ggff_hilfe, text);
#endif
                                    }
                                break;
                            case 1: // 19.9.2015
                                kommunikation4er->einfuegen_fu(f-1, g, farbe_char_m1, ggff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25591");
                                test_einfuegen(f-1, g, farbe_char_m1, ggff_hilfe, text);
#endif
                                break;
                            case 2:
                                farbe_short_int = ggff[(f-1)>>1][g>>1];
                                farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                cout << " LMNO West 4 f-1 " << f-1 << " g " << g << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                cout << " f/2 " << (f >> 1) << " (g-1)/2 " << ((g-1)>>1) << endl;
#endif
                                kommunikation4er->einfuegen_fu(f-1, g, farbe_char, ggff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25600");
                                test_einfuegen(f-1, g, farbe_char, ggff_hilfe, text);
#endif
                                break;
                            }
                        }
                    }
                if(f<n_minus1)
                    {
                    if(((ep_ff[f+1][g+1]==0)&&(fgff[f+1][g+1]<0))||(ep_ff[f+1][g+1]==1))
                        {
                        switch(ep_ff[f+1][g])
                            {
                            case 0:
                                if(fgff[f+1][g]>=0) // Ein Punkt ist da, 23.10.14 => Innere Ecke
                                    {
                                    farbe_short_int = fgff[f+1][g];
                                    farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                    cout << " LMNO Ost 4 f " << f << " g+1 " << g+1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                    cout << " f+1/2 " << ((f+1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                    kommunikation4er->einfuegen_fu(f+1, g, farbe_char, ggff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                    strcpy(text, "25623");
                                    test_einfuegen(f+1, g, farbe_char, ggff_hilfe, text);
#endif
                                    }
                                break;
                            case 1: // 19.9.2015
                                kommunikation4er->einfuegen_fu(f+1, g, farbe_char_m1, ggff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25627");
                                test_einfuegen(f+1, g, farbe_char_m1, ggff_hilfe, text);
#endif
                                break;
                            case 2:
//                                farbe_short_int = ggff[(f+2)>>1][g>>1];
                                farbe_short_int = ggff[(f+1)>>1][g>>1]; // nderung 5.2.16
                                farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                cout << " LMNO West 4 f " << f << " g+1 " << g+1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                cout << " f+1/2 " << ((f+1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                kommunikation4er->einfuegen_fu(f+1, g, farbe_char, ggff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25638");
                                test_einfuegen(f+1, g, farbe_char, ggff_hilfe, text);
#endif
                                break;
                            }
                        }
                    }
                }
            break;
        case 0:
            fgff_hilfe = fgff[f][g+1];
#ifdef AUSGABE_LMNO
            if((f==12)&&(g==20)&&(rang==0))
                {
                cout << " ggff_hilfe " << ggff_hilfe << " epff0,-1 " << ep_ff[f][g-1] << " ep_ff+1-1 " << ep_ff[f+1][g-1] << " fgf+1-1 " << feingitter_feld[f+1][g-1] << " fgff0-1 " << fgff[f][g-1] << endl;
                }
#endif
            if(fgff_hilfe>=0)
                {
#ifdef AUSGABE_LMNO
                if((f==12)&&(g==20)&&(rang==0))
                    {
                    cout << " LMNO West 2 " << endl;
                    }
#endif
                if(f>0)
                    {
                    if(((ep_ff[f-1][g+1]==0)&&(fgff[f-1][g+1]<0))||(ep_ff[f-1][g+1]==1))
                        {
                        switch(ep_ff[f-1][g]) // Bei ep_ff == 0 ist fgff[f][g-1]>=0; 20.10.14
                            {
                            case 0:
                                if(fgff[f-1][g]>=0)
                                   {
#ifdef AUSGABE_LMNO
                                    if((f==12)&&(g==20)&&(rang==0))
                                        {
                                        cout << " LMNO West 3 " << endl;
                                        }
#endif
                                    farbe_short_int = fgff[f-1][g];
                                    farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                    cout << " LMNO Sued 4 f " << f << " g-1 " << g-1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                    cout << " f+1/2 " << ((f+1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                    kommunikation4er->einfuegen_fu(f-1, g, farbe_char, fgff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                    strcpy(text, "25689");
                                    test_einfuegen(f-1, g, farbe_char, fgff_hilfe, text);
#endif
                                   }
                                break;
                            case 1: // 19.9.2015
                                kommunikation4er->einfuegen_fu(f-1, g, farbe_char_m1, fgff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25693");
                                test_einfuegen(f-1, g, farbe_char_m1, fgff_hilfe, text);
#endif
                                break;
                            case 2:
                                farbe_short_int = ggff[(f-1)>>1][g>>1];
                                farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                cout << " LMNO Sued 4 f " << f << " g-1 " << g-1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                cout << " f+1/2 " << ((f-1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                kommunikation4er->einfuegen_fu(f-1, g, farbe_char, fgff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25702");
                                test_einfuegen(f-1, g, farbe_char, fgff_hilfe, text);
#endif
                                break;
                            }
                        }
                    }
                if(f<n_minus1)
                    {
                    if(((ep_ff[f+1][g+1]==0)&&(fgff[f+1][g+1]<0))||(ep_ff[f+1][g+1]==1))
                        {
                        switch(ep_ff[f+1][g])
                            {
                            case 0:
                                if(fgff[f+1][g]>=0)
                                    {
                                    farbe_short_int = fgff[f+1][g];
                                    farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                    cout << " LMNO West 4 f " << f << " g+1 " << g+1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                    cout << " f+1/2 " << ((f+1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                    kommunikation4er->einfuegen_fu(f+1, g, farbe_char, fgff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                    strcpy(text, "25725");
                                    test_einfuegen(f+1, g, farbe_char, fgff_hilfe, text);
#endif
                                    }
                                break;
                            case 1: // 19.9.2015
                                kommunikation4er->einfuegen_fu(f+1, g, farbe_char_m1, fgff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25729");
                                test_einfuegen(f+1, g, farbe_char_m1, fgff_hilfe, text);
#endif
                                break;
                            case 2:
//                                farbe_short_int = ggff[(f+2)>>1][g>>1];
                                farbe_short_int = ggff[(f+1)>>1][g>>1]; // nderung 5.2.16
                                farbe_char = (char) farbe_short_int;
#ifdef AUSGABE_LMNO
                                cout << " LMNO Sued 4 f " << f << " g+1 " << g+1 << " fsi " << farbe_short_int << " ggff_hilfe " << ggff_hilfe;
                                cout << " f+1/2 " << ((f+1) >> 1) << " g/2 " << (g>>1) << endl;
#endif
                                kommunikation4er->einfuegen_fu(f+1, g, farbe_char, fgff_hilfe);
#ifdef AUSGABEN_FARBENUPDATE_EINFUEGEN
                                strcpy(text, "25738");
                                test_einfuegen(f+1, g, farbe_char, fgff_hilfe, text);
#endif
                                break;
                            }
                        }
                    }
                }
            break;
        }
    }

// Kommentar: Die beiden Hilfsausgaben optional.

void FarbenUpdate::test2fu(char * wo)
    {
    }

void FarbenUpdate::test3fu(char * wo)
    {
    }

void FarbenUpdate::test_entfernen(int ff, int gg, int farbe_ff_gg, char * stelle)
    {
    if((ff==16)&&(gg==15)&&(farbe_ff_gg==0))
        {
        cout << " Farbenfeld von (16,15) reset an Stelle " << stelle;
        cout << " f_schlange " << f_schlange << " g_schlange " << g_schlange << " f " << f << " g " << g << endl;
        }
    if((ff==26)&&(gg==10))
        {
        cout << " Farbenfeld von (26,10) reset an Stelle " << stelle << " Farbe " << farbe_ff_gg;
        cout << " f_schlange " << f_schlange << " g_schlange " << g_schlange << " f " << f << " g " << g << endl;
        }
    }

void FarbenUpdate::test_einfuegen(int f_p, int g_p, char farbe_char, int farbe_zielprozessor, char * stelle)
    {
    if((f_p==4)&&(g_p==61))
        {
        int farbe_int;
        farbe_int = (int) farbe_char;
        cout << "  FU test_einfuegen rang " << rang << " Farbe_char "  << farbe_int << " Farbe ZP " << farbe_zielprozessor << " Stelle " << stelle << endl;
        }
    }
